////////////////////////////////////////////////////////////////////////////////
// MODULE DrawMaster:   Pairing interface to GTour        G.Kaniuk 06/06/04
//
#include <list>
#include <algorithm>
#include <vector>
#include <string>

#include "stdlib.h"
#include <values.h>

#include "Pairsort.h"
#include "PairTool.h"
#include "Pair.h"

#pragma hdrstop
//---------------------------------------------------------------------------

#include "Format.h"
#include "Sort.h"
#include "Model.h"

#include "PrefsMaster.h"
#include "TourMaster.h"
#include "RegMaster.h"
#include "DrawMaster.h"
//---------------------------------------------------------------------------

namespace DrawMaster										// REGISTER SUPPORT
{
  	using Tours::GTour;

	using TourMaster::GoTour;

	void ChangeGrade
   (
   	const GTour& Tour,		// Grade change may cross a bar
      ints  oGrd, ints nGrd,	// Old and new grades
      ints  oMMi, ints oMMs,	// Old initial and current scores
      ints& nMMi, ints& nMMs	// New initial and current scores
   )
   {
   	// A change in grade does not affect the stored mms per round, because
      // the stored mms is history, like colour and parity

      if(nGrd == oGrd) return;

      if(nGrd > Tour.Rules.MacBar)
      	nMMi = Tour.Rules.MacBar;												// Demoted to bar level

      if(nGrd < Tour.Rules.MinGrade)
      	nMMi = Tour.Rules.MinGrade;											// Promoted to novice level

      if(Tour.Rules.MinGrade <= nGrd && nGrd <= Tour.Rules.MacBar)	// Continuous range
      	nMMi = nGrd;

      nMMs = oMMs + nMMi - oMMi;													// nMMi - oMMi is the      																					// McMahon increment
   }

   int BoardCount(void)
   {
      int Boards = GoTour.Pair.size() - 1;

      return Boards>0 ? Boards : 0;
   }

   int FindBoard(const size_t pid)
   {
   	const int boardcount = BoardCount();
      const int players = (int)GoTour.Players;

      for(int b=1; b<=boardcount; b++)
      {
      	const GBoard& pair = GoTour.Pair[b];

         if(1<=pair.Black && pair.Black<=players)
         {
         	const GPlayerPtr black = GoTour.Listed[pair.Black];

            if(black)
         		if(black->Pid == pid)
            		return b;
         }

         if(1<=pair.White && pair.White<=players)
         {
         	const GPlayerPtr white = GoTour.Listed[pair.White];

            if(white)
               if(white->Pid == pid)
                  return b;
         }
      }

      return -1;
   }
}
//------------------------------------------------------------------------------
//
//
// CONVERTORS
namespace DrawMaster
{
	using std::string;
   using Tools::to_string;
	using Tours::GBoard;

	string points_string(const GPlayerPtr P)
   {
   	if(!P) return string();

   	Tools::ints points = GoTour.PointsValue(true, GoTour.Roundl, P);

      return Tours::wins_ustring(points);
   }

   string wins_string(const GPlayerPtr P)
   {
   	if(!P) return string();

   	Tools::ints wins = GoTour.WinsValue(P);

      return Tours::wins_ustring(wins);
   }

   string homegames_string(const GPlayerPtr P)
   {
   	if(!P) return string();

   	Tools::ints homes = GoTour.HomeGamesValue(P);

      return Tools::to_string(homes);
   }

   string ForgeKomi(const GPlayerPtr Black, const GPlayerPtr White)
   {
   	// DISPLAY HANDICAP AND KOMI POINTS
   	int Komi, Stone;

   	PairMaster::KomiStones(Black, White, Komi, Stone);

      Stone /= Tours::mc_scale;	 Komi  /= Tours::mc_scale;

      if(!Stone) return string();

      // HANDICAP																		// GK 6.4.0 corrected H++K
      const string plus     = string(Tools::sPlus);
      const string strStone = to_string(Stone);
      const string rawKomi  = !Komi ? string() : to_string(Komi);
      const string strKomi	 = Komi>0 ? plus + rawKomi : rawKomi;

      string Handicap  = string();

      switch(GoTour.Rules.Style)
      {
      	case Tours::stSwissHcap:
            if(Black->Grade < White->Grade) Handicap = strStone + strKomi;
         break;

         case Tours::stAllPlayHcap:
            if(Black->Grade < White->Grade) Handicap = strStone + strKomi;
         break;

         case Tours::stMcMahon:
         	if(Black->McScore < White->McScore) Handicap = strStone + strKomi;
         break;

         default:
				Handicap = string();
      }
		return Handicap;
   }


	void DrawEditInfo(int black, int white, bool& Repeat, bool& BadOrder)
   {
   	const GPlayerPtr Black = GoTour.Listed[black];
      const GPlayerPtr White = GoTour.Listed[white];

      Repeat = BadOrder = false;

      if(!Black || !White) return;

   	// CHECK FOR REPEAT GAME IN PREVIOUS ROUNDS
      Repeat = false;

      for(size_t r=0; r<GoTour.Round-1; r++)
      	if(Black->Games[r].Oppid == (size_t)white)
         	{ Repeat = true;  break; }

      // CHECK FOR BAD BOARD ORDER
      {
         int komipt=0,stones=0;

         PairMaster::KomiStones(Black, White, komipt, stones);

         BadOrder = stones>0 ? Black->Grade > White->Grade : false;
      }
   }
}
//---------------------------------------------------------------------------

namespace DrawMaster											// BAR AND ROUND SUPPORT
{
	using Tools::ints;

	using Tours::GTour;

   const int MinBarPopUK[Tours::max_rounds] =		// Minumum UK pop
   {
   	2, 3, 4,  5,  6,  7,  8,  9, 10, 11, 12, 13
   };

   const int MaxBarPopUK[Tours::max_rounds] =		// Maximum UK pop
   {
   	4, 6, 8, 10, 12, 15, 18, 22, 26, 30, 35, 40
   };

   const int MinBarPopEU[Tours::max_rounds] =		// Minumum EU pop
   {
   	2, 3, 4,  8, 12, 16, 20, 24, 28, 32, 36, 40
   };

   const int MaxBarPopEU[Tours::max_rounds] =		// Maximum EU pop
   {
   	4, 5, 6, 12, 18, 24, 30, 36, 42, 48, 54, 60
   };

   bool GetBarPops(GTour* Tour, int& LoPop, int& HiPop)
   {
   	const int Rounds = (int)Tour->Rules.Rounds;

   	LoPop = HiPop = 0;

   	switch (Tour->Rules.RuleSet)
      {
      	case Tours::rtBGA:
         {
         	LoPop = MinBarPopUK[Rounds-1];  HiPop = MaxBarPopUK[Rounds-1];
         }
         break;

         case Tours::rtEGF:
         {
         	LoPop = MinBarPopEU[Rounds-1];  HiPop = MaxBarPopEU[Rounds-1];
         }
         break;

         default: return false;
      }
      return true;
   }

   //---------------------------------------------------------------------------
   // GK BARDEPTH 19/11/2015
   // Major modification to include limits to bardepth = maxgrade-bargrade.
   //
   // A. Require LoPop<=MacPop<=HiPop AND bardepth<=LimBarDepth.
   //
   // B. We accept bardepth=MaxBardepth when MacPop changes from
   //    MacPop<LoPop to LoPop<=MacPop<=HiPop and
   //    bardepth changes from bardepth<=LimBardepth to bardepth=MaxBarDepth
   //
   // C. Defaults:  LimBardepth = 2, MaxBarDepth = 3
   //---------------------------------------------------------------------------

   int SetBar(GTour* Tour, int& MacPop)
   {
   	using std::list;

   	using Tools::ints;

      using Tours::stMcMahon;
      using Tours::stBarLow;

      // IDENTIFY POPULATION RANGE IN BAR
      int LoPop(0), HiPop(0);

      if(!GetBarPops(Tour, LoPop, HiPop))
      	return brNoBar;

      // SORT IN GRADE ORDER - ALL PLAYERS ARE ALREADY REGISTERED
      RegMaster::RegSort(Tours::skGrade, 1);
   	int Players = Tour->Players, ActivePlayers = 0;

      MacPop = Players;	if(!Players) return brSleep;

      // GET GRADE GROUPS
      std:list<DGroup>Grades;  Grades.clear();

      ints GroupScore, Grade;
      ints LowestGrade = Tour->Rules.MinGrade, HighestGrade;

      int GroupSize = 1;

      Tour->Rules.MacBar = LowestGrade;

      for(int p=1; p <= Players; p++)										// Sorted in descending grade order
      {
      	GPlayerPtr Player = Tour->Listed[Tour->Sorted[p]];

         Grade = Player->Grade;

         if(Player->PlayingAll())											// GK 6.3.9 BGA RULES SAME AS EGF
         {
         	if(!ActivePlayers)												// The first one!
            {
            	HighestGrade = Grade;

            	Tour->Rules.MacBar = Grade;

               GroupScore = Grade;	GroupSize = 0;
            }

            if(Grade<GroupScore)												// Grade change starts a new group
            {
               Grades.push_back(DGroup(GroupSize,GroupScore));		// Completed group on list

               GroupScore = Grade;  GroupSize = 0;						// Start a new group
            }

            ++GroupSize;	++ActivePlayers;
         }

         if(p==Players)													// Last player terminates group
         {
            Grades.push_back(DGroup(GroupSize,Grade));      // May be a new group of 1 or more

            LowestGrade = Grade;

            break;
         }
      }

      const Tours::style_t& Style =  Tour->Rules.Style;

      if(stMcMahon < Style && Style < stBarLow)	// Everybody equal
      {
      	MacPop = ActivePlayers;

         Tour->Rules.MacBar = 0;

         return brSet;
      }

      // MARCH THROUGH GRADES LIST REDUCING BAR TO MAXIMISE POPULATION
      // WHILE KEEPING BAR-DEPTH BELOW MaxBarDepth
      const ints LimBarDepth=2*Tours::mc_scale, MaxBarDepth=3*Tours::mc_scale;

      list<DGroup>::iterator sGroup = Grades.begin();
      list<DGroup>::iterator eGroup = Grades.end();
      list<DGroup>::iterator iGroup = sGroup;

      ints& MacBar(Tour->Rules.MacBar); 		// Current bar
      ints& MacDepth(Tour->Rules.MacDepth);

      MacBar   = sGroup->Score;				   // Start at the highest grade
      MacDepth = HighestGrade-MacBar;
      MacPop   = sGroup->Size;

      if(MacPop > HiPop)return brHiBar;		// First group is already too big
      if(MacPop == HiPop)return brSet;			// First group max possible

      ints NewDepth = MacDepth;
      ints NewBar   = MacBar;

      int NewPop=0;

      do                                 	 	// Scan the groups list
      {
      	NewPop  += iGroup->Size;				// We have a new group
      	NewBar   = iGroup->Score;
         NewDepth = HighestGrade-NewBar;

      	if(NewDepth>LimBarDepth) break;		// Accept previous bar
			if(NewPop>HiPop) break;

         MacPop = NewPop;							// We have a new bar
      	MacBar = NewBar;
         MacDepth = NewDepth;
      }
      while(iGroup++ != eGroup);

      if(iGroup==eGroup)							// Entire population within limits
      {
       	if(MacPop<LoPop) return brNoMcmon;	// Not viable McMahon
         if(MacDepth==0) return brSet;			// Entire pop all same grade

         return brManual;							// All bars may be possible
      }

      if(NewDepth==MaxBarDepth)					// Extreme case
      	if(LoPop<=NewPop && NewPop<=HiPop)
      		if(MacPop<LoPop)						// There was a gap
            	MacBar = NewBar;

      	return brSet;
   }

   int SetInitialScores(ints NovBar, ints KomBar, ints MacBar, ints& BarDepth, GTour* Tour)
   {
   	BarDepth = 0;

   	for(size_t p = 1; p<=Tour->Players; p++)
      {
      	GPlayerPtr Player = Tour->Listed[p];

         // STANDARD INITIAL SCORES
         switch(Tour->Rules.Style)
         {
         	case Tours::stMcMahon: case Tours::stBarLow:
            {
            	const ints& Grade(Player->Grade);

               if(Grade <= NovBar)
                  Player->MiScore = NovBar;

               if(NovBar < Grade && Grade< MacBar)
                  Player->MiScore = Grade;

               if(Grade >= MacBar)											// GK 6.3.9 	NEW BGA RULES SAME AS EGF RULES
               {
                 const ints bdepth = Grade-MacBar;						// Find the maximum bar depth

                  if(Player->PlayingAll())								// Above the bar play all rounds
                  {
                  	Player->MiScore =  MacBar;

                     if(bdepth > BarDepth) BarDepth = bdepth;
                  }
                  else
                  	Player->MiScore =  MacBar - Tours::mc_scale;	// The group just below the bar
               }
            }
            break;

            case Tours::stSwiss: case Tours::stSwissHcap:
            	Player->MiScore = 0;
            break;

            case Tours::stAllPlay: case Tours::stAllPlayHcap:
            	Player->MiScore = 0;
            break;

            default:
               Player->MiScore = 0;
         }

         // SUPERGROUP SCORE
         const string& GroupCode = Player->Group;

         if(GroupCode[1] == Tours::gpTop)
         {
         	int iSuper = (int)(GroupCode[0] - Tools::cZero);

            if(iSuper>0)
            	Player->MiScore = MacBar + iSuper*GoTour.Rules.SuperDiff;
         }

         Player->McScore = Player->MiScore;
      }
      return brSet;
   }

   void SetMcMahonScores(void)
   {
   	for(size_t p=1; p<=GoTour.Players; p++)
      {
      	GPlayerPtr Player = GoTour.Listed[p];

      	Player->McScore = Player->McMahonScore(GoTour.Round);
      }
   }

   void SetPlayerStates(void)						// CALLED AT START OF DRAW OR MANUAL PAIRING
   {
   	using Tours::psRegistered;
      using Tours::psFinished;
      using Tours::psSleep;

   	const size_t Round = GoTour.Round;			if(Round<1) return;

      const intsi psFree = Round==1 ? psRegistered : psFinished;

   	for(size_t p=1; p<=GoTour.Players; p++)
      {
         GPlayerPtr Player = GoTour.Listed[p];

         if(GoTour.PastRound())
         	Player->State = Player->Awake(Round) ? psFinished : psSleep;
         else
            Player->State = Player->Awake(Round) ? psFree : psSleep;
      }
	}

   void SetNewRound(void)
   {
      switch(GoTour.State)
      {
         case Tours::tsRegClosed: { ++GoTour.Round;  GoTour.Roundl = GoTour.Round; }
            break;

         case Tours::tsRoundDone: { ++GoTour.Round;  GoTour.Roundl = GoTour.Round; }
            break;

         case Tours::tsForcedPair: { GoTour.Roundl = GoTour.Round; }
            break;
      }
      GoTour.PopulationCount();
   }

   void ResetRound(void)
   {
   	--GoTour.Round;  GoTour.Roundl = GoTour.Round;

      GoTour.State = GoTour.Round>0 ? Tours::tsRoundDone : Tours::tsRegClosed;

      GoTour.PopulationCount();
   }
}
//---------------------------------------------------------------------------

namespace DrawMaster											// THE DRAW
{
	using std::vector;

   using Tours::GGame;

   using PairMaster::PairCount;
   using PairMaster::ClearOpponents;

   void StartManualDraw(PairCount& ModelCount)
   {
      ClearOpponents();										// Start from clean slate
      SetPlayerStates();									// Defines who is playing this round
   	SetMcMahonScores();									// Sleepers get their points now

      Model::UnevenHomeScan(ModelCount);				// Ideal draw
   }

   void CancelManualDraw(void)
   {
   	if(GoTour.Round>0)--GoTour.Round;			   // Roundl was never changed

      SetPlayerStates();									// Original round states

      GoTour.PopulationCount();

      SetMcMahonScores();									// Original round scores
   }

   void EndFullManualDraw(PairCount& ActualCount)			// ALL PLAYERS ALREADY PAIRED MANUALLY
   {
   	const int Round = GoTour.Round;							// Round guaranteed > 0

   	const int BoardCount = GoTour.PlayNowCount/2;		// PlayNowCount guaranteed even

   	GoTour.Pair.resize(BoardCount+1);						// Board numbers start at 1

      for(size_t p=1; p<=GoTour.Players; p++)
      {
      	GPlayerPtr Player = GoTour.Listed[p];

         GGame& Game = Player->Games[Round-1];

         int pBlack=0, pWhite=0;

         if(Game.Board)												// We preserve the manual board order
         {
            if(Game.Colour==Tours::gcBlack)
            {
               pBlack = p;  pWhite = Game.Oppid;
            }

            if(Game.Colour==Tours::gcWhite)
            {
               pWhite = p;  pBlack = Game.Oppid;
            }

            GoTour.Pair[Game.Board] = GBoard(pBlack, pWhite);
         }
      }
   	AnalyseDraw(ActualCount);
   }


   bool StartDraw(PairCount& ModelCount)
   {
      DrawMaster::SetNewRound();                	// This round ...

      if(GoTour.PlayNowCount%2)							// ... MUST be even
      {
         ResetRound();

         return false;
      }

      PairMaster::ClearOpponents();						// Start from clean slate
      SetPlayerStates();									// Registered or Finished
      SetMcMahonScores();									// Sleepers get points here
      PairTool::SetUpDnCounts();							// Play up and lose or down and win

      Model::UnevenHomeScan(ModelCount);

      return true;
   }

   bool CycleDraw
   (
   	const Byte PairMode, const int PairedCountRemaining,
      PairCount& Actual, vector<PairCount>& Actuals
   )
   {
   	using Tours::stAllPlay;
      using Tours::stAllPlayHcap;
      using Tours::stMcMahon;
      using Tours::stBarLow;

      const Tours::style_t& Style = GoTour.Rules.Style;

   	int TrialCount = PairMaster::GetTrials(GoTour.PairParam);

      if(TrialCount<=0 || TrialCount>PairMaster::size_monitor) return false;

      bool(*Pair)(int&) = PairMode==PairMaster::pmModel
      						? Model::Pairing
                        : Pair::Pairing;

      // SPECIFIC SEEDING ENABLES EXACT RECOVER OF BEST DRAW
   	vector<int>Seed(TrialCount);

      for(int s=0; s<TrialCount; s++)
      	Seed[s] = GoTour.PlayNowCount + s*GoTour.Round;

      // ALL PLAY ALL
      if(Style==stAllPlay || Style==stAllPlayHcap)
      	if(Pair::AllPlayAll())
         {
         	AnalyseDraw(Actual);

         	return true;
         }
         else
         	return false;

      // ANY 'MUST PLAY' PAIRED BEFORE THE REST
      int MustPlayCount = 0;
      {
     		if(!Pair::MustPlay(MustPlayCount))
         	return false;

         if(MustPlayCount == GoTour.PlayNowCount)
         {
         	AnalyseDraw(Actual);

         	return true;
         }
      }

      // THE DRAW CYCLE
      {
      	bool Paired = false;	 int PairedCount=0;

      	PairSort::StartSort(GoTour.Players+1);

         for(int s=0; s<TrialCount; s++)
         {
            if(TrialCount>1)srand(Seed[s]);					// Single shot trial does not reseed -  V6.5.0

            Actuals[s].Clear();									// Set all to -1

            Paired = Pair(PairedCount);

            PairedCount += MustPlayCount;

            if(Paired)
            {
               if(PairedCount == PairedCountRemaining)// All paired
               	AnalyseDraw(Actuals[s]);				// Set all to non-negative

               ClearDraw(Tours::psChosen);				// Any pre-draw untouched
            }
            else
            	break;
         }

      	PairSort::EndSort();

         if(!Paired)
         	return false; 										// Pairing failed due to software bug
      }

      // SEARCH FOR BEST DRAW
      int Best = 0;												//  BUGFIX in V6.2.1 WAS Best = -1;
       																// If all attempts were stuck,
      																// we choose the first

      int MinUneven, MinLand,  MinClub;
      int uneven,    sameland, sameclub;

      switch(Style)
      {
      	case stMcMahon: case stBarLow:
         	MinUneven = MinLand = MinClub = GoTour.Players;
         break;

         case Tours::stSwiss:
         {
         	MinUneven = GoTour.Players;

            MinLand = MinClub = 0;
         }
         break;

         case Tours::stSwissHcap:
         {
         	MinUneven = GoTour.Players;

            MinLand = MinClub = 0;
         }
         break;

         default:
            MinUneven = MinLand = MinClub = 0;
      }

      for(int s=0; s<TrialCount; s++)
      {
      	Actual = Actuals[s];

      	uneven   = Actual.Uneven;
         sameland = Actual.SameLand;
         sameclub = Actual.SameClub;

         if(GoTour.Rules.Style != Tours::stMcMahon)
         	sameland = sameclub = 0;

         if(uneven >= 0 && sameland >= 0 && sameclub >= 0)
            if(uneven <= MinUneven && sameland <= MinLand && sameclub <= MinClub)
            {
               MinUneven = uneven;  MinLand = sameland;  MinClub = sameclub;

               Best = s;
            }
      }
      																// BUGFIX in V6.2.1 WAS if(Best < 0) return false;
      Actual = Actuals[Best];                         // Best is now guaranteed >=0

      // RECREATE BEST PAIRING
      {
      	int PairedCount = 0;
         
      	PairSort::StartSort(GoTour.Players+1);

         if(TrialCount>1) srand(Seed[Best]);				// V6.5.0

         bool PairDone = Pair(PairedCount);

         // SOME PLAYERS NOT PAIRED IN TRAINING MODE
         if(PairDone)
            if(PairMode==PairMaster::pmModel)
               Model::DisturbPairing();

         PairSort::EndSort();
         
         return PairDone;
      }
   }
}
//------------------------------------------------------------------------------

namespace DrawMaster										// POST-DRAW SUPPORT
{
	using std::string;

   using Tours::GBoard;
   using Tours::GPlayerPtr;
   using Tours::GGame;
   using Tours::GBoard;

  	using TourMaster::GoTour;

   void GameComment(GBoard Board, std::vector<bool>& Note)
   {
      using Tours::mc_scale;
      using Tours::Score;
      using Tours::NoClub;
      
      using DrawMaster::coCount;

      const Tours::style_t Style = GoTour.Rules.Style;

      for(size_t n=0; n<Note.size(); n++)Note[n] = false;

      if(Board.Black<1 || Board.Black>(int)GoTour.Players)return;
      if(Board.White<1 || Board.White>(int)GoTour.Players)return;

      const int Round = GoTour.Round;  if(Round<1) return;

      GPlayerPtr Black = GoTour.Listed[Board.Black];
      GPlayerPtr White = GoTour.Listed[Board.White];

      GGame BlackG = Black->Games[Round-1];
      GGame WhiteG = White->Games[Round-1];

      // SAME CLUB/COUNTRY
      const bool SameClub = Black->Club != NoClub && White->Club != NoClub && Black->Club==White->Club;

      Note[coClub] = PrefsMaster::PrefsTube.ScreenFlagClub && SameClub;
      Note[coLand] = PrefsMaster::PrefsTube.ScreenFlagLand && Black->Land==White->Land;


      // UNEVEN/HANDICAP GAMES IN McMAHON TOURNAMENTS
      Note[coUneven] = Note[coHcap] = false;

      if(Style == Tours::stMcMahon || Style == Tours::stBarLow)
      {
	      const int Stones    = Black->Games[Round-1].Stones;
	      const int Komi      = Black->Games[Round-1].Komi;

	      const bool Uneven   = BlackG.Parity != Tours::gdEven;
	      const bool Handicap = Stones > Tours::mc_scale || Komi;

         if(Handicap)
            Note[coHcap]   = true;
         else
            Note[coUneven] = Uneven;
      }

      // BOARD ORDER CONTRADICTS MMS SCORE DIFFERENCE AT START OF ROUND
      if(Style == Tours::stMcMahon || Style == Tours::stBarLow)
      {
      	int RawWhite = White->McScore -  Score(WhiteG.Result);
      	int RawBlack = Black->McScore -  Score(BlackG.Result);

     	 	Note[coOrder] = RawBlack - RawWhite > abs(GoTour.Rules.KomiDiff);
      }

      // BOARD ORDER DETERMINED BY GRADE DIFF FOR HANDICAP TOURNAMENTS
      if(Style == Tours::stSwissHcap || Style == Tours::stAllPlayHcap)
      	Note[coOrder] = Black->Grade-White->Grade > abs(GoTour.Rules.KomiDiff);

      // REPEAT GAMES
      for(int r=0; r<Round-1; r++)
      	if(Black->Games[r].Oppid == (size_t)Board.White)
				Note[coRepeat] = true;
   }

 	int AllocateBoards(int& PairedCount)
   {
   	using Tours::psChosen;
      using Tours::psPlaying;

      PairedCount = 0;

      // ADJUST ALREADY PAIRED PLAYER'S STATE
      for(size_t p=1; p<=GoTour.Players; p++)
         if(GoTour.Listed[p]->State == psPlaying)			// Must play or ForcedPair
            GoTour.Listed[p]->State = psChosen;

      // BOARDS ALLOCATED IN RANK ORDER
      int Active     = Tours::RankSort(&GoTour);			if(!Active) return -1;
      int Round      = GoTour.Round;							if(Round<1) return -2;
      int BoardCount = GoTour.PlayNowCount/2;
      int Board      = 1;

      GoTour.Pair.resize(BoardCount+1);						// Board numbers start at 1

      for(int b=1; b<=BoardCount; b++)							// Boards may be empty
      	GoTour.Pair[b] = Tours::GBoard(0,0);            // if draw incomplete

      // ALLOCATE CHOSEN PLAYERS
      for(int p=1; p<=Active; p++)								// Inactives at back of Listed
      {
      	int one = GoTour.Sorted[p];

      	GPlayerPtr One = GoTour.Listed[one];	

         if(One->State == psChosen)
         {
         	int two = One->Games[Round-1].Oppid;

         	GPlayerPtr Two = GoTour.Listed[two];
            			if(Two->State != psChosen)
                     	return -3;

            if(One->Games[Round-1].Colour == Tours::gcBlack)
            	GoTour.Pair[Board] = Tours::GBoard(one, two);
            else
               GoTour.Pair[Board] = Tours::GBoard(two, one);

            One->Games[Round-1].Board = Board;
            Two->Games[Round-1].Board = Board;

            One->State = Two->State = psPlaying;	PairedCount += 2;

            ++Board;		if(Board>BoardCount)break;
         }
      }

      RegMaster::RegSort(GoTour.SortOrder, GoTour.SortSense);

      return 0;
   }

   void ListBoards(void)
   {
   	using TourMaster::GoTour;

      const int Round = GoTour.Round;

   	if(GoTour.State <= Tours::tsRegClosed || Round<1)
      {
      	GoTour.Pair.resize(0);

         return;
      }
      
      // ENSURE ENOUGH BOARDS FOR PLAYERS, JOINERS AND LEAVERS
   	size_t Boards = 0, MaxBoard = 0, LimBoard = (GoTour.Players+1)/2;

   	for(size_t p=1; p<=GoTour.Players; p++)
      {
      	Tours::GPlayerPtr P = GoTour.Listed[p];

         const size_t Board = P->Games[Round-1].Board;

         if(Board > MaxBoard) MaxBoard = Board;

         bool Awake = P->Awake(Round);
         bool Join  = P->Joining(Round);
         bool Leave = P->Leaving(Round);

         if(Awake || Join || Leave)
        		++Boards;
      }
      Boards = (Boards+1)/2;									// Players of course should be even

      if(MaxBoard > Boards)Boards = MaxBoard;			// So some are empty
      if(Boards > LimBoard)Boards = LimBoard;

      GoTour.Pair.resize(Boards+1);

      for(size_t b=1; b<=Boards; b++)						// Clear all boards
      	GoTour.Pair[b] = Tours::GBoard(0,0);

      // POPULATE BOARDS WITH PLAYERS AND LEAVERS
      for(size_t p=1; p<=GoTour.Players; p++)
      {
         Tours::GPlayerPtr P = GoTour.Listed[p];

         Tours::GGame& Game = P->Games[Round-1];

         if(1<=Game.Board && Game.Board <= Boards)
            switch(Game.Colour)
            {
               case Tours::gcBlack:  GoTour.Pair[Game.Board].Black = p; break;
               case Tours::gcWhite:  GoTour.Pair[Game.Board].White = p; break;

               default:	break;
            }
      }
   }

   void LayoutBoards(void)											// BOARD ALLOCATION
   {
   	const size_t Bneeded = BoardCount();					// Boards required

      size_t Bsupplied = 0, LastRoom = 0;			   		// Specified in RulesForm

      for(size_t room=0; room<Tours::size_rooms; room++)	// Check each room
      {
      	const size_t& Binroom(GoTour.Boards[room]);

         if(Binroom)													// Accumulate boards
      	{
         	Bsupplied += Binroom;

            LastRoom = room;										// Highest room used
         }
      }

      const int Bextra = (int)Bneeded - (int)Bsupplied;

      if(Bsupplied==0)												// No rooms set up yet
      {
      	GoTour.Boards[0] = Bneeded;	return;
      }

      if(Bextra > 0)													// All missing boards in last room  V6.5.0
      	GoTour.Boards[LastRoom] += Bextra;					// Unused rooms have 0 Boards
    }

   void AnalyseDraw(PairMaster::PairCount& Actual)
   {
   	using Tools::modulus;
   	using Tours::psChosen;
      using Tours::psPlaying;

   	// TODO
   	// This routine will be enhanced to also calculate the
      // pairing 'merit', and will then have an additional parameter

   	Actual.Uneven = Actual.SameClub = Actual.SameLand = 0;

      for(size_t p=1; p<=GoTour.Players; p++)						// Clear pairing marks
      	GoTour.Listed[p]->Spare[0] = 0;

   	for(size_t p=1; p<=GoTour.Players; p++)
      {
      	GPlayerPtr One = GoTour.Listed[p];

         Byte& OneMark = One->Spare[0];

         if(!OneMark)														// Marked players are already counted
         if(One->State == psChosen || One->State == psPlaying)	// A new pair
         {
         	int two = One->Games[GoTour.Round-1].Oppid;

            if(two)															// Must have opponent
            {
               GPlayerPtr Two = GoTour.Listed[two];

               Byte& TwoMark = Two->Spare[0];

               if(modulus(One->McScore - Two->McScore) >= Tours::gsOne)
                  ++Actual.Uneven;

               if(One->Land == Two->Land)
               {
                  if(One->Club == Two->Club)
                     ++Actual.SameClub;

                  ++Actual.SameLand;
               }

               OneMark = 1;	TwoMark = 1;         					// Mark pair as done
         	}
         }
		}
   }

   bool ClearDraw(void)														// USER cancels
   {
   	using PairMaster::UnpairPlayer;
      using PairMaster::ClearMonitor;

   	if(GoTour.Round<1)return false;

   	for(size_t p=1; p<=GoTour.Players; p++)
      {
      	const int Oppid = GoTour.Listed[p]->Games[GoTour.Round-1].Oppid;

         UnpairPlayer(p);
         UnpairPlayer(Oppid);
      }

      ResetRound();
      	SetPlayerStates();
      		SetMcMahonScores();

		ClearMonitor(GoTour.PairParam);

      return true;
   }

   bool ClearDraw(Tours::PlayerState State)							// PAIRING support
   {
   	using PairMaster::UnpairPlayers;

   	if(GoTour.Round<1) return false;

   	for(size_t p=1; p<=GoTour.Players; p++)
      {
      	Tours::GPlayerPtr Player = GoTour.Listed[p];

         if(Player->State==State)
         	UnpairPlayers(p, Player->Games[GoTour.Round-1].Oppid);
      }
      return true;
   }

}
//---------------------------------------------------------------------------

namespace DrawMaster										// EDITING THE DRAW
{
	GHandicap::GHandicap(int board, string handicap)
   {
   	Board = board;

      ParseKomi(handicap, Stones, Komi);
   }

   int ParseKomi(string komi, int& Stones, int& Komi)
   {
   	using Tools::cPlus;
      using Tools::cMinus;
      using Tools::next_posint;

      const int MaxStones = 32;   								// 128/mc_scale;
      const int MaxKomi   = 1000;								// Display limited

      string::iterator skomi = komi.begin();
      string::iterator ekomi = komi.end();

      Stones = Komi = 0;

      if(komi.empty())												// Clear the handicap
      {
      	Stones = Komi = 0;

         return 0;
      }

      if(!isdigit(*skomi))											// Must start with digit
      	return -1;

      if(next_posint(skomi, ekomi, Stones))
      {
      	if(Stones >= MaxStones)
         	return 5;

      	Stones *= Tours::mc_scale;


         if(skomi == ekomi) return 0;							// Only stones, no komi

         char op = *skomi;

         if(op != cPlus && op != cMinus)						// Invalid character
            return 1;

         ++skomi;

         if(skomi==ekomi) 											// Missing komi
            return 2;

         if(!isdigit(*skomi))										// Komi must start with digit
         	return -1;

         if(next_posint(skomi, ekomi, Komi))			      // Komi points ...
            if(skomi == ekomi)			   					// ... properly terminated
            {
            	if(Komi >= MaxKomi)
               	return 6;

            	Komi *= Tours::mc_scale;

            	if(op == cMinus) Komi = -Komi;

               return 0;
            }
         return 3;													// Komi garbage
      }
      else
         return 4;
   }

   void MergeKomiLists
   (
   	list<GHandicap>& OldKomis,
      list<GHandicap>& NewKomis
   )
   {
   	if(OldKomis.empty() && NewKomis.empty()) return;

      list<GHandicap>::iterator sOldKomi = OldKomis.begin();
      list<GHandicap>::iterator eOldKomi = OldKomis.end();

      list<GHandicap>::iterator sNewKomi = NewKomis.begin();
      list<GHandicap>::iterator eNewKomi = NewKomis.end();

      list<GHandicap>::iterator fNewKomi;

      while(sOldKomi != eOldKomi)
		{
      	fNewKomi = find(sNewKomi, eNewKomi, *sOldKomi);

         if(fNewKomi == eNewKomi)
         {
            NewKomis.push_back(GHandicap(sOldKomi->Board, 0, 0));		// The komi was deleted
                                                                     // so we still need to set it to zero
            eNewKomi = NewKomis.end();
         }
         ++sOldKomi;
      }

      if(!NewKomis.empty())														// In increasing board order
      	NewKomis.sort();
   }

   
	bool ChangeDraw													// NEW PAIRS AND HANDICAPS FROM DRAW EDIT
   (
   	vector<GBoard>& NewPairs,									// New pairing after edit
      list<GHandicap>OldKomis,									// Original komi list
      list<GHandicap>NewKomis										// New komi list
   )
   {
   	using DrawMaster::GHandicap;

   	using PairMaster::UnpairPlayers;
      using PairMaster::UnpairPlayer;
   	using PairMaster::ManualPairing;
      using PairMaster::ManualHandicap;

      // WE KNOW THAT PlayNowCount IS EVEN
      // NOTE THAT NEWPAIRS MIGHT BE TOO BIG
      // IF PLAYERS WERE REMOVED
      GoTour.Pair.resize(GoTour.PlayNowCount/2+1);

      int Round = GoTour.Round; 		if(Round<1) return false;

      MergeKomiLists(OldKomis, NewKomis);

      list<GHandicap>::iterator nKomi = NewKomis.begin();
      list<GHandicap>::iterator eKomi = NewKomis.end();

      // THERE MAY BE SOME EMPTY BOARDS
      int board = 0;

      for(size_t b=1; b<NewPairs.size(); b++)
      {
      	GBoard& Board = NewPairs[b];

      	size_t black = Board.Black, white = Board.White;

         if(black && white)											// Ignore empty boards
         {                                                  // It is guaranteed that b=0 implies w=0 and vica versa
            GPlayerPtr Black = GoTour.Listed[black];
            GPlayerPtr White = GoTour.Listed[white];

            GGame& BlackG = Black->Games[Round-1];
            GGame& WhiteG = White->Games[Round-1];

            bool NewPair = false;

            if (BlackG.Oppid != white)								// Black has a new opponent
               NewPair = true;
            else
               if(BlackG.Colour != Tours::gcBlack)				// Black is now white
                  NewPair = true;

            size_t nBoard = nKomi->Board;							// Next board with a change in handicaps

            int Stones=0, Komipt=0;									// No change in handicaps

            if(nKomi != eKomi)
            if(b == nBoard)											// There is a change in handicaps
            {
               Stones = nKomi->Stones;
               Komipt = nKomi->Komi;

               ++nKomi;
            }

            // MAKE THE CHANGE
            if(NewPair)
            {
            	UnpairPlayer(BlackG.Oppid);		// Unpair opponent first
            	UnpairPlayer(black);             // because player's opponent is zeroed

               UnpairPlayer(WhiteG.Oppid);
               UnpairPlayer(white);

               ManualPairing(b, black, white, Stones, Komipt);
            }
            else
               if(b==nBoard)											// The only change is in the handicap
                  ManualHandicap(black, white, Stones,Komipt);

            ++board;														// The next board number

            BlackG.Board = board;  WhiteG.Board = board;		// The board number can change

            GoTour.Pair[board] = Board;
         }
      }

      // MOP UP PLAYERS ON THE SAME BOARD LEAVING THE DRAW
      for(size_t p=1; p<=GoTour.Players; p++)
      {
      	GPlayerPtr Player = GoTour.Listed[p];

      	if(Player->Leaving(Round))
         	UnpairPlayer(p);			// State != Sleep, but Result is Sleep
      }

      SetMcMahonScores();
      
      return true;
   }

	int ManualDraw(const int b, const GBoard& Board, string& Handicap)
  	{
   	int Stones, Komi;

      int Report = ParseKomi(Handicap, Stones,Komi);

      if(Report) return Report;

      PairMaster::ManualPairing(b, Board.Black, Board.White, Stones, Komi);

      return 0;
  	}
}
//---------------------------------------------------------------------------

namespace DrawMaster										// RESULTS SUPPORT
{
	using Tools::intsi;
   using Tours::GPlayerPtr;
   using Tours::GGame;
   using TourMaster::GoTour;

   static void ClearResult(GGame& Game, GPlayerPtr Player)
   {
      Player->McScore -= Tours::Score(Game.Result);
      Player->State   = Tours::psPlaying;

      Game.Result   = Tours::grNone;

      GoTour.MissingCount += 1;
   }

   static void EditResult(intsi Result, GGame& Game, GPlayerPtr Player)
   {
   	using TourMaster::GoTour;

      Player->McScore -= Tours::Score(Game.Result);	// Excise the old result
      Player->McScore += Tours::Score(Result);        // Add the new result

      Game.Result = Result;

      if(Game.Result == Tours::grNone)
      	Player->State = Tours::psPlaying;
      else
      	Player->State = Tours::psFinished;

      if(Player->MissingResult(GoTour.Round))
      	++GoTour.MissingCount;
   }

	bool EditResult(int board, intsi Bres, intsi Wres)
   {
   	const int Round = GoTour.Round;		if(Round<1) return false;

      const Tours::TourState OldState = GoTour.State;

   	Tours::GBoard& Board = GoTour.Pair[board];

      GPlayerPtr Black = GoTour.Listed[Board.Black];
      GPlayerPtr White = GoTour.Listed[Board.White];

      GGame& Bgame = Black->Games[Round-1];
      GGame& Wgame = White->Games[Round-1];

      // NO CHANGE
      if(Bres==Bgame.Result && Wres==Wgame.Result)
      	return true;

      while(true)
      {
      	// NEW RESULT
         if(Bgame.Result==Tours::grNone || Wgame.Result==Tours::grNone)
         {
				if(GoTour.MissingCount)
            {
               if(Black->MissingResult(Round))
                  GoTour.MissingCount -= 1;

               if(White->MissingResult(Round))
                  GoTour.MissingCount -= 1;
            }
            else
            	return false;

            EditResult(Bres, Bgame, Black);  EditResult(Wres, Wgame, White);

            break;
         }

         // CLEAR EXISTING RESULT
         if(Bres==Tours::grNone)			// The parser has set Bres=Wres to grNone
         {
            if(!Black->MissingResult(Round))
            	ClearResult(Bgame, Black);

            if(!White->MissingResult(Round))
            	ClearResult(Wgame, White);

            break;
         }

         // CHANGE EXISTING RESULT
         EditResult(Bres, Bgame, Black);
         EditResult(Wres, Wgame, White);

         break;
      }

      GoTour.SetState();

      if(GoTour.State != OldState)
      	if(!GoTour.SaveHeader())
         	return false;

      if(!GoTour.SavePlayer(Board.Black)) return false;
      if(!GoTour.SavePlayer(Board.White)) return false;

   	return true;
   }

   void RestoreResult(GPlayerPtr Player)
   {
      using Tours::grLose; using Tours::grWin;  using Tours::grJigo;
      using Tours::grDlos; using Tours::grFwin;
      using Tours::grVoid; using Tours::grUsolv;
      using Tours::grNone;

      using TourMaster::GoTour;

      const size_t Round = GoTour.Round;

      if(Round<1 || Round>GoTour.Rules.Rounds) return;

      size_t Opp = Player->Games[Round-1].Oppid;

      GGame& PGame = Player->Games[Round-1];
      GGame& QGame = GoTour.Listed[Opp]->Games[Round-1];

      intsi &PRes = PGame.Result, &QRes = QGame.Result;

      switch(QRes)
      {
      	case grWin:   PRes = grLose;  break;
         case grLose:  PRes = grWin;   break;

         case grFwin:  PRes = grDlos;  break;
         case grDlos:  PRes = grFwin;  break;

         case grJigo:  PRes = grJigo;  break;
         case grUsolv: PRes = grUsolv; break;
         case grVoid:  PRes = grVoid;  break;

         default:		  PRes = grNone;
      }
   }

   int Restore(GPlayerPtr Player)						// Removes State/Game result contradiction
   {
   	using Tours::grSleep;

   	const size_t Round = GoTour.Round;

      if(Round<1 || Round>GoTour.Rules.Rounds)
      	return 0;

   	if(Player->Joining(Round))							// Player State is still psSleep
      {
      	Player->Games[Round-1].Result = grSleep;

         return -1;
      }

      if(Player->Leaving(Round))							// Player State is still psPlaying or pFinished
      {
         RestoreResult(Player);							// Can reset if opponent has result

         return  1;
      }

      return 0;												// Player not joining or leaving
   }

   intsi GetCurrentResult(GPlayerPtr Player)
   {
   	size_t Round = GoTour.Round;

   	if(Round<1 || Round>Player->Games.size()) return 0;

      return Player->Games[Round-1].Result;
   }

   
}
//---------------------------------------------------------------------------

namespace DrawMaster										// RULES SUPPORT
{
	void ChangeStyle(void)								// CALLED ONLY WHEN STYLE HAS CHANGED
   {
   	using Tours::GPlayer;
      using Tours::GGame;

   	const string BarGroup = string("0:"), NoGroup = string("  ");

      // RE-COMPUTE THE BAR
   	// GK REGSORT 2015  const int OldOrder = GoTour.SortOrder, OldSense = GoTour.SortSense;
      int BarPop;	 SetBar(&GoTour, BarPop);
      // GK REGSORT 2015  RegMaster::RegSort(OldOrder, OldSense);

      // RE-SET PLAYER GROUP
      for(size_t p=1; p<=GoTour.Players; p++)
      {
      	GPlayerPtr P = GoTour.Listed[p];

         if(GoTour.Rules.Style == Tours::stMcMahon || GoTour.Rules.Style == Tours::stBarLow)
         	if(P->Grade >= GoTour.Rules.MacBar)
            	P->Group = BarGroup;		// Above bar in McMahon
            else
            	P->Group = NoGroup;		// Below bar
         else
         	P->Group = NoGroup;			// Non-McMahon, so there are no groups
      }

      // RE-SET INITIAL AND CURRENT SCORES
      const ints NovBar = GoTour.Rules.MinGrade;
      const ints KomBar = GoTour.Rules.KomiGroup;
      const ints MacBar = GoTour.Rules.MacBar;

      ints BarDepth;	SetInitialScores(NovBar, KomBar, MacBar, BarDepth, &GoTour);

      SetMcMahonScores();

      // RE-SET HANDICAPS
      if(GoTour.Round>0)
      {
         const int round = GoTour.Round-1;

         for(size_t b=1; b<GoTour.Pair.size(); b++)
         {
            const Tours::GBoard Board = GoTour.Pair[b];

            GPlayerPtr Black = GoTour.Listed[Board.Black], White = GoTour.Listed[Board.White];

            int Komi, Stones;	PairMaster::KomiStones(Black, White, Komi, Stones);

            GGame &BlackG = Black->Games[round], &WhiteG = White->Games[round];

            BlackG.Stones =  Stones;	BlackG.Komi =  Komi;
            WhiteG.Stones = -Stones;	WhiteG.Komi = -Komi;
         }
      }
   }
}
//---------------------------------------------------------------------------

