////////////////////////////////////////////////////////////////////////////////
// MODULE Model: Model Pairing and Simulated results 					07/11/04
////////////////////////////////////////////////////////////////////////////////

#include <string>
#include <list>
#include <algorithm>

#include <stdlib.h>

#pragma hdrstop

#include "TourMaster.h"
#include "DrawMaster.h"
#include "PairMaster.h"
#include "PairSort.h"
#include "PairTool.h"
#include "Model.h"
#include "Statistics.h"

namespace Model												// PAIRING
{
   static const int HomeScanCount = 3;

   bool Pairing(int& PairedCount)
   {
   	using TourMaster::GoTour;

      using PairSort::Sorted;

      using PairTool::FormMMGroups;
      using PairTool::Choose;
      using PairMaster::PairPlayers;

      int mmstart = 1, mmend = GoTour.Players;

      PairedCount = 0;
      
      PairSort::McMahonSort(mmstart, mmend);

      int mfold = mmend/2;

      if(mmend %2)
      	return false;

      for(int m=1; m<=mfold; m++)
      	PairMaster::PairPlayers(Sorted[m], Sorted[mfold+m]);

      PairedCount = 2*mfold;

      return true;
   }

   void DisturbPairing(void)
   {
   	using TourMaster::GoTour;

      using PairTool::Choose;
      using PairMaster::GetPartDraw;
      using PairMaster::UnpairPlayers;

      const int PartDraw = GetPartDraw(GoTour.PairParam);

      if(!PartDraw) return;
      
      for(size_t p = 1; p <= GoTour.Players; p++)
      {
         Tours::GPlayerPtr Black = GoTour.Listed[p];

         const int pOpp = Black->Games[GoTour.Round-1].Oppid;

         if(pOpp)
         {
            Tours::GPlayerPtr White = GoTour.Listed[pOpp];

            const bool BlackMust = Black->Group[1] == Tours::gpAlways;
            const bool WhiteMust = White->Group[1] == Tours::gpAlways;

            if(!BlackMust && !WhiteMust)
            	if(Black->State==Tours::psChosen)
               	if(!Choose(PartDraw))
                     UnpairPlayers(p, pOpp);
         }
      }
   }

}
//------------------------------------------------------------------------------

namespace Model												// UNEVEN/HOME COUNT
{
	using PairTool::MMGroup;
   using PairTool::MMGroups;

   using PairTool::ARLands;
   using PairTool::ARLand;
   using PairTool::ARClub;

   using PairTool::FormAreas;

   // The minimum number of uneven games in the pairing is counted exactly.
   // The minimum number of home games in an even group is counted exactly.
   // There is variation in the home game count for the uneven games.
   // The minimum would be reached after many simulated pairings.
   //
   // The down player is chosen with probability proportional to the largest
   // area in the group.  The up player is chosen with probability proportional
   // to the largest area excluding the down area.

   static void CountEvenHome(MMGroup& McGroup, int& LandCount, int& ClubCount)
   {
   	if(!McGroup.Pool) return;

   	int a = McGroup.Start;
      int b = McGroup.Start + McGroup.Size - 1;			// Full group range

		ARLands Areas;  FormAreas(a, b, Areas);			// The range [a,b] is Free

      ARLand& BigLand = Areas.Lands[0];					// A representative largest area

      const int HalfSize = (b-a+1)/2;						// Unavailable are at back of group

      const int SameLand = BigLand.Size > HalfSize ?	// The largest land folds
      				 			BigLand.Size - HalfSize :
                   			0;
      LandCount += SameLand;

      if(SameLand)												// There may be same club games
      {
      	ARClub& BigClub = BigLand.Clubs[0];				// A representative largest club

         if(BigClub.Club==Tours::NoClub)
         	ClubCount += 0;
         else
         	ClubCount += BigClub.Size > HalfSize ?		// The largest club folds
      					 	 BigClub.Size - HalfSize :
                  	    0;
      }

      McGroup.Pool = 0;
   }


   static void CountOddHome
   (
   	MMGroup& McFirst, MMGroup& McSecond, int& LandCount, int& ClubCount
   )
   {
   	using PairSort::SortedPlayer;

   	if(!McFirst.Size || !McSecond.Size) return;

      // CHOOSE FIRST PLAYER AT RANDOM
   	int aF = McFirst.Start;
      int bF = McFirst.Start + McFirst.Size - 1;

      int One = PairTool::Choose(aF, bF);

      // CHOOSE SECOND PLAYER AT RANDOM
      int aS = McSecond.Start;
      int bS = McSecond.Start + McSecond.Size - 1;

      int Two = PairTool::Choose(aS, bS);

      // MARK PLAYERS AS PAIRED
      PairSort::SortedPlayer(One)->State = Tours::psChosen;
      PairSort::SortedPlayer(Two)->State = Tours::psChosen;

      // NOW COUNT
      if(SortedPlayer(One)->Land == SortedPlayer(Two)->Land)
      {
         if(SortedPlayer(One)->Club == SortedPlayer(Two)->Club)
         	++ClubCount;

         ++LandCount;
      }
      --McFirst.Pool;  --McSecond.Pool;
   }

   void UnevenHomeScan(PairMaster::PairCount& Model)
   {
      // THIS PROCESS LEAVES THE PLAYER STATE INVARIANT
      // It is called just BEFORE the draw to produce optimal estimates
      // of uneven and home games.

   	using Tours::psRegistered;
      using Tours::psFinished;

   	using TourMaster::GoTour;

      using PairTool::FormMMGroups;

      const int Round = GoTour.Round, Players = GoTour.Players;

      if(PairMaster::GetTrials(GoTour.PairParam)>1)				// Allow for random single shot V6.5.0
      	srand(Round*Players);

      PairSort::StartSort(GoTour.Players+1);

      Model.Uneven = Model.SameLand = Model.SameClub = 0;

      for(int Trial=0; Trial<HomeScanCount; Trial++)
      {
      	// COUNT UNEVEN GAMES ARISING FROM ODD GROUPS
      	MMGroups McGroups;	FormMMGroups(McGroups);

         int uneven = 0, land = 0, club = 0;

         for(size_t m=0; m<McGroups.size(); m++)
         	if(McGroups[m].Pool%2 != 0)
            {
               ++uneven;

               CountOddHome(McGroups[m],McGroups[m+1], land, club);
            }

         // ALL GROUPS EVEN - COUNT HOME GAMES
         for(size_t m=0; m<McGroups.size(); m++)
         	if(McGroups[m].Pool)
            	CountEvenHome(McGroups[m], land, club);

         if(Trial==0)
         {
         	Model.Uneven = uneven; 	Model.SameLand = land;	Model.SameClub = club;
         }
         else
         {
         	bool Ubest = uneven <= Model.Uneven;
            bool Lbest = land   <= Model.SameLand;
            bool Cbest = club   <= Model.SameClub;

         	if(Ubest && Lbest && Cbest)
            {
               Model.Uneven = uneven;
            	Model.SameLand = land;
            	Model.SameClub = club;
            }
         }

         // CLEAN UP THE PLAYER STATE
         for(int p=1; p<=Players; p++)
         {
         	Tours::GPlayerPtr Player = PairSort::SortedPlayer(p);

            if(Player->State == Tours::psChosen)
            	Player->State = Round>1 ? psFinished : psRegistered;
         }
      }
      PairSort::EndSort();
   }
}
//------------------------------------------------------------------------------

namespace Model												// RESULT SIMULATION
{
	using Tools::real;
   using Tools::intsi;
   using Tools::erf;

   // FITTING CONSTANTS FROM pwin0009
   //
	static const real pwin_u[4] = { 0.057186500, 0.126337000, 0.0, 0.0139085 };
	static const real pwin_v[4] = { 0.002713730, 0.002390260, 0.0, 0.0000040 };
	static const real pwin_K    =   0.204062;

	static real ProbWin(int r, int s)  // Probability that r beats s
	{
		real q, qq;

		q = s-r; qq = q*q;

		real lambda=0.0;

		for(int n=0; n<4; n++)
		{
			const real cq = (pwin_u[n] +pwin_v[n]*qq)*q;

			const real nkrs = n*pwin_K *(q>=0 ? r : s);

			lambda +=  cq*exp(nkrs);
		}

		return 0.5*(1 - erf(lambda));
	}

   static const int  pW_Range  = 1000;	  			// Pwin to 3 dec places
   static const int  pJ_Range  = pW_Range;  	  // Jigo
   static const int  pF_Range  = pJ_Range+10;  // Free win
   static const int  pD_Range  = pF_Range+8;   // Default loss
   static const int  pV_Range  = pD_Range+6;   // Void
   static const int  pU_Range  = pV_Range+4;   // Unresolved
   static const int  pT_Range  = pU_Range+2;   // Total range

   static void SimResults( int Bmms, int Wmms, intsi& Bres, intsi& Wres)
   {
      int Q = PairTool::Choose(pT_Range);

      if(Q > pU_Range) { Bres = Wres = Tours::grUsolv; 					 return; }
      if(Q > pV_Range) { Bres = Wres = Tours::grVoid;   					 return; }
      if(Q > pD_Range) { Bres = Tours::grDlos;  Wres = Tours::grFwin; return; }
      if(Q > pF_Range) { Bres = Tours::grFwin;  Wres = Tours::grDlos; return; }
      if(Q > pJ_Range) { Bres = Tours::grJigo;  Wres = Tours::grJigo; return; }

      real P = ProbWin(Bmms, Wmms)*pW_Range;

      if((real)Q < P)
         { Bres = Tours::grWin;  Wres = Tours::grLose; }
      else
         { Bres = Tours::grLose; Wres = Tours::grWin; }
   }

   bool SimulateGame(int board)
   {
   	using TourMaster::GoTour;
      using Tours::GPlayerPtr;

      Tours::GBoard Board = GoTour.Pair[board];	// board is guaranteed within range

      GPlayerPtr Pblack = GoTour.Listed[Board.Black];
      GPlayerPtr Pwhite = GoTour.Listed[Board.White];

      const int Bmms = (7*Pblack->Grade + 3*Pblack->McScore)/10;
      const int Wmms = (7*Pwhite->Grade + 3*Pwhite->McScore)/10;

      intsi Bres,Wres;

      SimResults(Bmms,Wmms, Bres, Wres);

      const Byte PairMode = PairMaster::GetPairMode(GoTour.PairParam);

      if(PairMode == PairMaster::pmTest)
      	if(Bres == Tours::grJigo || Bres == Tours::grUsolv)
         {
         	Bres = PairTool::Choose(1);
            Wres = !Bres ? Tours::grWin : Tours::grLose;
         }

      return DrawMaster::EditResult(board, Bres, Wres);
   }
}
//------------------------------------------------------------------------------







