////////////////////////////////////////////////////////////////////////////////
// MODULE Pair: The GoDraw Pairing Algorithms 				 G.Kaniuk 17/04/05
////////////////////////////////////////////////////////////////////////////////

#include <string>
#include <vector>
#include <algorithms>
#pragma hdrstop

#include "PalgRES.h"
#include "PairRES.h"

#include "TourMaster.h"
#include "PairMaster.h"
#include "PairSort.h"
#include "PairTool.h"
#include "PairMcMahon.h"
#include "Pair.h"

////////////////////////////////////////////////////////////////////////////////
// 'MUST PLAY' PAIRING ALGORITHM
//
namespace Pair
{
	using std::vector;
   using std::string;

   using Tours::GPlayerPtr;
   using Tours::GGame;

   using TourMaster::GoTour;
   using PairSort::Sorted;
   using PairSort::SortedPlayer;
   
   typedef vector<int> Sequence;

   class PRoundTable
   {
   	private:														// PRIVATE DATA
      	vector< vector<int> > Graph;						// Square matrix of rounds
         int aGroup;												// Start index of group

      private:														// PRIVATE TOOLS
      	int MinimumRound(const int i)const;
         bool PairingHasRepeats(const int period, const int r);
      	bool Cycle(const int period, const int r);

      public:
      	bool AllPaired;										// Pairing is free from repeat games

      	PRoundTable(const int a, const int b);			// Does it all

         void SeeTable(const int a);
   };

   int PRoundTable::MinimumRound(const int i)const
   {
   	int MinRound = GoTour.Round;							// Minimum entry>0 in row i

      for(size_t j = 0; j<Graph.size(); j++)
      {
         const int r = Graph[i][j];

         if(r)														// So played in round r
            if(r < MinRound)
               MinRound = r;
      }
      return MinRound;									  		// Player joined group in MinRound
   }

   bool PRoundTable::PairingHasRepeats(const int period, const int r)
   {
   	if(r==0) return false;									// No repeats in round 1

   	for(int g=0; g<period/2; g++)							// period is even
      {
      	const int One = Sorted[aGroup + g];				// Pairing
         const int Two = Sorted[aGroup + period - 1 - g];

         if(PairMaster::PlayedBefore(One, Two, r))		// We check UPTO round r
         	return true;
      }
   	return false;
   }

   bool PRoundTable::Cycle(const int period, const int r)
   {
   	AllPaired = false;

      for(int k=0; k<period; k++)
         if(PairingHasRepeats(period, r))
         {
            const int CarryRound = Sorted[aGroup + period-1];

            for(int i=period-1; i>=1; i--)
               Sorted[aGroup+i] = Sorted[aGroup+i-1];

            Sorted[aGroup+1] = CarryRound;
         }
         else
         {
         	AllPaired = true;

            break;
         }

      return AllPaired;
   }

   PRoundTable::PRoundTable(const int a, const int b)
   {
   	// COUNT THE NUMBER OF PLAYERS IN THE GROUP
   	const int GroupSize = b - a + 1;

      aGroup = a;

      // CONSTRUCT A CLEAN TABLE
      Graph.resize(GroupSize);

      for(int p = 0; p<GroupSize; p++)
         Graph[p] = vector<int>(GroupSize, 0);

      // LOAD THE TABLE: Table[i][j] = round number
      Sequence::iterator sPlayer = Sorted.begin() + a;
      Sequence::iterator ePlayer = Sorted.begin() + b + 1;
      Sequence::iterator jPlayer;

      for(int i=0; i<GroupSize; i++)
      {
         const GPlayerPtr Player = PairSort::SortedPlayer(a+i);

         for(size_t r=0; r<GoTour.Round; r++)
         {
            jPlayer = std::find(sPlayer, ePlayer, (int)Player->Games[r].Oppid);

            if(jPlayer != ePlayer)
            {
					const int j = jPlayer - sPlayer;

            	Graph[i][j] = r+1;
            }
         }
      }

      // COUNT PLAYERS JOINING EACH ROUND
      const int Round = GoTour. Round;
      vector<int> JoinedCount(Round,0);					// Number of players joining at round r

      for(int i=0; i<GroupSize; i++)  						// March though all rows in the Graph
      {
			int MinRound = MinimumRound(i)-1;  				// Minimum round number entry in Graph

         ++JoinedCount[MinRound];
      }

      // IDENTIFY PLAYERS JOINING EACH ROUND
      Sequence Players(GroupSize);							// Packed in order of joining at round r

      Sequence iJoined(Round,0);							  // Next available slot for round r in Players

      for(int r=1; r<Round; r++)								// Start the joined index
      	iJoined[r] = iJoined[r-1] + JoinedCount[r-1];

      for(int i=0; i<GroupSize; i++)
      {
			int MinRound = MinimumRound(i)-1;  				// Player joins group at MinRound

         Players[iJoined[MinRound]] = Sorted[a+i];		// Pack joining player into next slot

         ++iJoined[MinRound];
      }

      for(int i=0; i<GroupSize; i++)						// Move packed players back to Sorted
      	Sorted[a+i] = Players[i];

      // SORT EACH JOINING SUBGROUP INTO PID ORDER
      int pa, pb = 0;

      for(int r=0; r<Round; r++)
      {
      	pa = pb + 1;  pb = pa + JoinedCount[r] - 1;

      	if(pa<=pb)
         	PairSort::PlayerIdSort(pa, pb);
      }

      // NOW CYCLE, WITH NEW INTAKE AT BACK OF PLAYERS
      int JoinedSize = 0;

      for(int r=0; r<Round; r++)
      {
      	JoinedSize += JoinedCount[r];

         if(JoinedSize)
         	if(!Cycle(JoinedSize, r))
            	break;
      }
   }

   void PRoundTable::SeeTable(const int a)
   {
   	for(size_t i=0; i<Graph.size(); i++)
      for(size_t j=0; j<Graph.size(); j++)
      {
      	int Pi = Sorted[a+i], Pj = Sorted[a+j];

         int r = Graph[i][j];

         int s = r;

         r = s;   Pj=Pi; s =Pi+Pj;
      }
   }

	bool MustPlay(int &PairedCount)
   {
   	PairSort::StartSort(GoTour.Players+1);

   	int pa = 1, pb = GoTour.Players, a,b;

   	PairSort::GroupCodeSort(pa, pb);

      string GroupCode = string();

      a = b = 0;

      int Report = 0;

      PairedCount = 0;

      while(PairSort::NextGroupCode(pa,pb, GroupCode, a, b))
      {
      	if(GroupCode[1] != Tours::gpAlways) break;

         const int GroupSize = b-a+1;

      	if(GroupSize % 2)
         {
         	Report = IDQ_OddMustPlay;

            break;
         }

      	PRoundTable RoundTable(a, b);

         if(RoundTable.AllPaired)
            for(int g=0; g<GroupSize/2; g++)
            {
               int One = a + g, Two = b - g;

               PairMaster::PairPlayers(Sorted[One], Sorted[Two]);

               PairSort::SortedPlayer(One)->State = Tours::psPlaying;
               PairSort::SortedPlayer(Two)->State = Tours::psPlaying;

               PairedCount += 2;
            }
         else
         	Report = IDQ_MustPlayRepeat;
      }

      if(Report)
      {
      	// GK 6.3.9 ToolsGUI::OperatorError(Report);

         GoTour.PostTourInfo(Report);
      }

      PairSort::EndSort();


      return !Report;
   }

   bool AllPlayAll(void)
   {
   	PairSort::StartSort(GoTour.Players+1);

   	PairSort::PlayerIdSort(1, GoTour.Players);

      int PairedCount = 0, Report = 0;

      for(size_t p=1; p<=GoTour.Players; p++)
      	if(!PairSort::SortedPlayer(p)->Playing(GoTour.Round))
         {
            ToolsGUI::OperatorError(IDQ_NoAllPlayAll);

            PairSort::EndSort();

            return false;
      	}

      PRoundTable RoundTable(1, GoTour.Players);

      const int Round = GoTour.Round;

      if(RoundTable.AllPaired)
         for(size_t g=0; g<GoTour.Players/2; g++)
         {
            const int one = 1 + g, two = GoTour.Players - g;

            const int One = Sorted[one], Two = Sorted[two];

            if(GoTour.Rules.Style==Tours::stAllPlay)	// ROBIN ON BOARD 1 ALTERNATES COLOUR EACH ROUND
            {                                         // REMAINING SEATS HAVE FIXED COLOUR
               const int Board = g+1;

               int Black, White, Stones = 0, Komi = 0;

               if(Board == 1)
                  if(Round%2)
                     { Black = Two; White = One; }		// Odd round
                  else
                     { Black = One; White = Two; }		// Even round
               else
                  if(!(Board%2))
                     { Black = Two; White = One; }		// Even board
                  else
                     { Black = One; White = Two; }		// Odd board

               PairMaster::ManualPairing(Board, Black, White, Stones, Komi);
            }
            else													// NORMAL HANDICAP RULES
            	PairMaster::PairPlayers(One, Two);

            PairSort::SortedPlayer(one)->State = Tours::psPlaying;
            PairSort::SortedPlayer(two)->State = Tours::psPlaying;

            PairedCount += 2;
         }
      else
         Report = IDQ_AllPlayAllRepeat;

      if(Report)
      {
         // GK 6.3.9 ToolsGUI::OperatorError(Report);

         GoTour.PostTourInfo(Report);
      }

      PairSort::EndSort();

      return !Report;
   }
}

////////////////////////////////////////////////////////////////////////////////
// 'SPLIT AND MATCH' PAIRING ALGORITHM
//
namespace Pair
{
	static void SplitAndMatch(void)
   {
   	int a=1, b=GoTour.Players;

   	PairSort::McMahonSort(a,b);

      const int Fold = (b-a+1)/2;

      for(int f=0; f<Fold; f++)
      	PairMaster::PairPlayers(Sorted[a+f], Sorted[a+f+Fold]);
   }
}

////////////////////////////////////////////////////////////////////////////////
// SWISS TOURNAMENTS
//
// 1. Initial score zero
// 2. No club bias
// 3. McMahon draw up/down rules apply
//
namespace Pair
{
	static bool SwissDraw(int& PairedCount)
   {
   	// INITIAL SCORES SET 0 AT CLOSE REG
      return McMahon::Draw(PairedCount);
   }

   static bool SwissHandicap(int& PairedCount)
   {
      // HANDICAPS ALLOCATED BY GRADE IN PairPlayers();
      return McMahon::Draw(PairedCount);
   }
}


////////////////////////////////////////////////////////////////////////////////
// GENERAL PAIRING CONTROL
//
namespace Pair
{
   bool Pairing(int& PairedCount)
   {
   	for(size_t p = 1; p<=GoTour.Players; p++)
      	PairSort::Sorted[p] = p;

      PairedCount = 0;

   	switch(GoTour.Rules.Style)
      {
      	case Tours::stMcMahon:
         case Tours::stBarLow:		return McMahon::Draw(PairedCount);

         case Tours::stSwiss:	  		return SwissDraw(PairedCount);
         case Tours::stSwissHcap:	return SwissHandicap(PairedCount);

         default: 				  		return false;
      }
   }
}
//
////////////////////////////////////////////////////////////////////////////////


