////////////////////////////////////////////////////////////////////////////////
// MODULE PairMaster: GTour pairing support 							G.Kaniuk 01/06/05
////////////////////////////////////////////////////////////////////////////////

#include <vector>
#include <algorithm>
#include <stdlib.h>

#pragma hdrstop

#include "PalgRES.h"
#include "TourMaster.h"
#include "PairMaster.h"
#include "PairSort.h"
#include "PairTool.h"

////////////////////////////////////////////////////////////////////////////////
// PAIRING PARAMETERS
//
namespace PairMaster
{
   using Tools::real;

   // THE RAW BYTE STRUCTURE OF THE Params ARRAY IS
   //
   // 4*int						Parity, Trials Area , Colour
   // 4*real   				Alpha, Beta, GammaM, GammaP
   // 2*Byte   				PairMode, PartDraw
   // 3*size_monitor*ints  Monitor

   PairParams::PairParams()
   {
   	Trials = def_trials;
      Parity = def_parity;
      Area   = def_area;
      Colour = def_colour;
      Alpha  = def_alpha;
      Beta   = def_beta;
      GammaP = def_gamma;
      GammaM = def_gamma;

      PairMode = pmNormal;
      PartDraw = 10;

      for(int m=0; m<3*size_monitor; m++)
      	Monitor[m] = 0;
   }

   PairParams::PairParams(Byte* TourStream)
   {
   	GetStream(TourStream);
   }

   void PairParams::PutStream(Byte* Stream)
   {
   	using Tools::ints;

      if(!Stream) return;

   	int* sTrials = (int*)Stream + 0;
      int* sParity = (int*)Stream + 1;
      int* sArea   = (int*)Stream + 2;
      int* sColour = (int*)Stream + 3;
      int* sNext   = (int*)Stream + 4;

      real* sAlpha = (real*)sNext + 0;
      real* sBeta  = (real*)sNext + 1;
      real* sGamaP = (real*)sNext + 2;
      real* sGamaM = (real*)sNext + 3;
      real* sEnd   = (real*)sNext + 4;

      Byte* sModes = (Byte*)sEnd;
      Byte* sPmode = sModes + 0;
      Byte* sRmode = sModes + 1;
      Byte* sEmode = sModes + 2;

      ints* sMonit = (ints*)sEmode;

      *sTrials = Trials<=size_monitor ? Trials : size_monitor;
      *sParity = Parity;
      *sArea   = Area;
      *sColour = Colour;

      *sAlpha  = Alpha;
      *sBeta   = Beta;
      *sGamaP  = GammaP;
      *sGamaM  = GammaM;

      *sPmode  = PairMode;
      *sRmode  = PartDraw;

      for(size_t m = 0; m<3*size_monitor; m++)
      	sMonit[m] = Monitor[m];
   }

   void PairParams::GetStream(Byte* Stream)
   {
      if(!Stream) return;

   	int* sTrials = (int*)Stream + 0;
      int* sParity = (int*)Stream + 1;
      int* sArea   = (int*)Stream + 2;
      int* sColour = (int*)Stream + 3;
      int* sNext   = (int*)Stream + 4;

      real* sAlpha = (real*)sNext + 0;
      real* sBeta  = (real*)sNext + 1;
      real* sGamaP = (real*)sNext + 2;
      real* sGamaM = (real*)sNext + 3;
      real* sEnd   = (real*)sNext + 4;

      Byte* sModes = (Byte*)sEnd;
      Byte* sPmode = sModes + 0;
      Byte* sRmode = sModes + 1;
		Byte* sEmode = sModes + 2;

      ints* sMonit = (ints*)sEmode;

      Trials = *sTrials;
      Parity = *sParity;
      Area   = *sArea;
      Colour = *sColour;

      Alpha  = *sAlpha;
      Beta   = *sBeta;
      GammaP = *sGamaP;
      GammaM = *sGamaM;

      PairMode = *sPmode;
      PartDraw = *sRmode;

      for(size_t m = 0; m<3*size_monitor; m++)
      	Monitor[m] = sMonit[m];
   }

	void SetDefaultParameters(Byte* DefPars)
   {
      if(!DefPars) return;

      PairParams Def;

      Def.PutStream(DefPars);
   }

   void PutTrials(int trials, Byte* Stream)
   {
   	if(trials<=0 || trials > size_monitor-2) return;

   	*((int*)Stream) =  trials;
   }

   int GetTrials(Byte* Stream)
   {
   	int trials = *((int*)Stream);

      if(trials<=0 || trials > size_monitor-2)
      	return 0;
      else
      	return trials;
   }

   void PutPairMode(Byte pairmode, Byte* Stream)
   {
   	Byte* sPmode = Stream + 4*sizeof(int) + 4*sizeof(real);

      *sPmode = pairmode;
   }

   Byte GetPairMode(Byte* Stream)
   {
   	Byte* sPmode = Stream + 4*sizeof(int) + 4*sizeof(real);

      return *sPmode;
   }


   void PutPartDraw(Byte partdraw, Byte* Stream)
   {
   	Byte* sPartDraw = Stream + 4*sizeof(int) + 4*sizeof(real) + sizeof(Byte);

      *sPartDraw = partdraw;
   }

   Byte GetPartDraw(Byte* Stream)
   {
   	Byte* sPartDraw = Stream + 4*sizeof(int) + 4*sizeof(real) + sizeof(Byte);

      return *sPartDraw;
   }

   void PutMonitor(PairCount& Model, PairCount& Acual, PairCounts& Trials, Byte* Stream)
   {
   	ints* sMonit = (ints*)(Stream +  4*sizeof(int) + 4*sizeof(real) + 2*sizeof(Byte));

      sMonit[0] = Model.Uneven;  sMonit[1] = Model.SameClub;	sMonit[2] = Model.SameLand;
      sMonit[3] = Acual.Uneven;  sMonit[4] = Acual.SameClub;	sMonit[5] = Acual.SameLand;

      if(Trials.size() > size_monitor) return;

      for(size_t r = 0; r<Trials.size(); r++)
      {
      	sMonit[6 + 3*r + 0] = Trials[r].Uneven;
         sMonit[6 + 3*r + 1] = Trials[r].SameClub;
         sMonit[6 + 3*r + 2] = Trials[r].SameLand;
      }
   }

   void GetMonitor(Byte* Stream, PairCount& Model, PairCount& Acual, PairCounts& Trials)
   {
   	ints* sMonit = (ints*)(Stream +  4*sizeof(int) + 4*sizeof(real) + 2*sizeof(Byte));

      int iMonit[6];

      for(int i=0; i<6; i++)
      {
      	iMonit[i] = (int)sMonit[i];

         // if(iMonit[i] == 255) iMonit[i] = -1;
      }

      Model.Uneven = iMonit[0];  Model.SameClub = iMonit[1];	Model.SameLand = iMonit[2];
      Acual.Uneven = iMonit[3];  Acual.SameClub = iMonit[4];	Acual.SameLand = iMonit[5];

      if(Trials.size() > size_monitor) return;

      for(size_t r = 0; r<Trials.size(); r++)
      {
      	Trials[r].Uneven   = sMonit[6 + 3*r + 0];
         Trials[r].SameClub = sMonit[6 + 3*r + 1];
         Trials[r].SameLand = sMonit[6 + 3*r + 2];
      }
   }

   void ClearMonitor(Byte* Stream)
   {
   	ints* sMonit = (ints*)(Stream +  4*sizeof(int) + 4*sizeof(real) + 2*sizeof(Byte));

      for(int m=0; m<6; m++) sMonit[m] = -1;
   }
}

////////////////////////////////////////////////////////////////////////////////
// HOUSEKEEPING
//
namespace PairMaster
{
	using TourMaster::GoTour;

	void ClearOpponents(void)
   {
   	const int Round = GoTour.Round;	if(Round<1) return;

      for(size_t p = 1; p<=GoTour.Players; p++)
      {
      	Tours::GGame& Game =  GoTour.Listed[p]->Games[Round-1];

         Game.Oppid = 0;

      	Game.Stones = 0;
         Game.Komi   = 0;
      }
   }
}


////////////////////////////////////////////////////////////////////////////////
// PAIRING SUPPORT
//

using PairSort::Throw;

namespace PairMaster
{
	using Tours::GPlayerPtr;
   using Tours::GGame;

   static void komi_stones(const int bscore, const int wscore, int& komi, int& stones)
   {
   	using TourMaster::GoTour;

      const ints scorediff  = abs(bscore - wscore);
      const ints komidiff  = GoTour.Rules.KomiDiff;
      const ints largestone  = 9*Tours::mc_scale;

      // WE FIRST DECIDE IF A HANDICAP IS POTENTIALLY TO BE AWARDED
      // THIS DECISION IS BASED SOLELY ON AN INEQUALITY IN THE SCORE DIFFERENCE
      // THEN THE HANDICAP AWARDED IS MODIFIED BY komidiff.
      // if(scorediff > abs(komidiff))  Prior 6.3.4

      if(scorediff == 0)					// Even game
      {
      	komi = stones = 0;	return;
      }

      stones = scorediff + komidiff;	// Potential number of stones

      if(stones <= 0)						// So 0<scorediff<=-komidiff (and komidiff <0)
      {                                // Therefore scorediff too small for hcap
         stones = komi = 0;  return;
      }

      if(GoTour.Rules.KomiRate == 0)	// We accept whatever stones have been awarded
      {
         komi = 0;	return;
      }

      if(stones > largestone)				// The komi is awarded if too many stones
      {
         komi = (stones - largestone)/Tours::mc_scale*GoTour.Rules.KomiRate;

         stones = largestone;
      }
      else
         komi = 0;							// Normal handicap, so no komi
   }

   static void komi_stones_colour(const int oneS, const int twoS, GGame& oneG, GGame& twoG)
   {
   	int komi, stones; 	komi_stones(oneS, twoS, komi, stones);

      if(oneS > twoS)
      {
         oneG.Stones = -stones;	oneG.Komi = -komi;  oneG.Colour = Tours::gcWhite;
         twoG.Stones =  stones;	twoG.Komi =  komi;  twoG.Colour = Tours::gcBlack;
      }

      if(oneS < twoS)
      {
         oneG.Stones =  stones;	oneG.Komi =  komi;  oneG.Colour = Tours::gcBlack;
         twoG.Stones = -stones;	twoG.Komi = -komi;  twoG.Colour = Tours::gcWhite;
      }
   }

   static void barlow_komi_stones
   (
   	GPlayerPtr black, GPlayerPtr white,
      int& komipt, int& stones
   )
   {
   	const bool barblack = black->MiScore >= GoTour.Rules.MacBar;
      const bool barwhite = white->MiScore >= GoTour.Rules.MacBar;

      if(barblack && barwhite)
         komipt = stones = 0;

      if(!barblack && barwhite)
         komi_stones(black->Grade, white->Grade, komipt, stones);

      if(barblack && !barwhite)
         komi_stones(black->Grade, white->Grade, komipt, stones);

      if(!barblack && !barwhite)
         komi_stones(black->McScore, white->McScore, komipt, stones);
   }

   static void barlow_komi_stones_colour
   (
   	GPlayerPtr one, GPlayerPtr two,
      GGame& oneG, GGame& twoG
   )
   {
   	// PROCESS PLAYERS WHERE EITHER IS BELOW BAR
   	const bool barone = one->MiScore >= GoTour.Rules.MacBar;
      const bool bartwo = two->MiScore >= GoTour.Rules.MacBar;

      if(!barone && bartwo)
         komi_stones_colour(one->Grade, two->Grade, oneG, twoG);

      if(barone && !bartwo)
         komi_stones_colour(one->Grade, two->Grade, oneG, twoG);

      if(!barone && !bartwo)
         komi_stones_colour(one->McScore, two->McScore, oneG, twoG);
   }

	void KomiStones
   (
   	const GPlayerPtr Black, const GPlayerPtr White,
      int& Komipt, int& Stones
   )
   {
      switch(GoTour.Rules.Style)
      {
      	case Tours::stMcMahon: // Both players below the bar get handicap
         {
         	const bool BarBlack = Black->MiScore >= GoTour.Rules.MacBar;
         	const bool BarWhite = White->MiScore >= GoTour.Rules.MacBar;

         	if(!BarBlack && !BarWhite)
            	komi_stones(Black->McScore, White->McScore, Komipt, Stones);
            else
            	Komipt = Stones = 0;
         }
         break;

         case Tours::stBarLow:	// Both players above the bar - no handicap
         	barlow_komi_stones(Black, White, Komipt, Stones);
         break;

         case Tours::stSwissHcap:
         	komi_stones(Black->Grade, White->Grade, Komipt, Stones);
         break;

         case Tours::stAllPlayHcap:
         	komi_stones(Black->Grade, White->Grade, Komipt, Stones);
         break;

         default: Komipt = Stones = 0;
      }
   }


   void PairPlayers(int one, int two)
   {
   	using Tools::intsi;

   	using Tours::mc_scale;

    	using Tours::gdDown;		using Tours::gdEven;		using Tours::gdUp;
      using Tours::gcBlack;	using Tours::gcWhite; 	using Tours::gcNone;

   	using Tours::psChosen;

      using Tours::GPlayerPtr;
      using Tours::GGame;

      using PairSort::Player;
      
      using PairTool::Choose;

   	if(one<1 || one>(int)GoTour.Players)Throw(IDA_BadPairPlayerId);
      if(two<1 || two>(int)GoTour.Players)Throw(IDA_BadPairPlayerId);

      int r = GoTour.Round - 1;	  if(r<0)Throw(IDA_BadPairRound);

      GPlayerPtr One = Player(one), 	   Two = Player(two);
      GGame    &OneG = One->Games[r],   &TwoG = Two->Games[r];

      // OPPONENTS
      One->State = psChosen;	OneG.Oppid = (size_t)two;
      Two->State = psChosen;	TwoG.Oppid = (size_t)one;

      // PARITY
      if(One->McScore > Two->McScore) { OneG.Parity = gdDown;  TwoG.Parity = gdUp;   }
      if(One->McScore == Two->McScore){ OneG.Parity = gdEven;  TwoG.Parity = gdEven; }
      if(One->McScore <  Two->McScore){ OneG.Parity = gdUp;    TwoG.Parity = gdDown; }

      One->Mbal += OneG.Parity;			Two->Mbal += TwoG.Parity;
      One->Mout += abs(OneG.Parity); 	Two->Mout += abs(TwoG.Parity);

      // COLOUR
      if(One->Cbal < Two->Cbal){ OneG.Colour = gcWhite; TwoG.Colour = gcBlack; }
      if(One->Cbal > Two->Cbal){ OneG.Colour = gcBlack; TwoG.Colour = gcWhite; }

   	if(One->Cbal == Two->Cbal)								// Try for alternating colours
      {
      	bool alternate = r>0;								// Random for round 1

         while(alternate)
         {
         	const GGame& OnePrev = One->Games[r-1];	// Previous games
            const GGame& TwoPrev = Two->Games[r-1];

            alternate = false;

            if(OnePrev.Oppid && TwoPrev.Oppid)			// They both played
            	if(OnePrev.Colour != TwoPrev.Colour)	// So we can swap colours
               {
                  OneG.Colour = -OnePrev.Colour;
                  TwoG.Colour = -TwoPrev.Colour;

                  alternate = true;
            	}
            break;
         }

         if(!alternate)											// Give up - choose at random
         {
         	OneG.Colour = Choose(2) ? gcBlack : gcWhite;
         	TwoG.Colour = -OneG.Colour;
         }
      }

      // HANDICAPS DEPEND ON STYLE
      const bool BarOne = One->MiScore >= GoTour.Rules.MacBar;
      const bool BarTwo = Two->MiScore >= GoTour.Rules.MacBar;

      OneG.Stones = TwoG.Stones = 0;
      OneG.Komi   = TwoG.Komi   = 0;

      switch(GoTour.Rules.Style)
      {
      	case Tours::stMcMahon:	// Normal handicap for both below bar
         	if(!BarOne && !BarTwo)
         		komi_stones_colour(One->McScore, Two->McScore, OneG, TwoG);
         break;

         case Tours::stBarLow:	// Grade handicap if one below bar, other above
         	if(!BarOne || !BarTwo)
            	barlow_komi_stones_colour(One, Two, OneG, TwoG);
         break;

         case Tours::stSwissHcap:
            komi_stones_colour(One->Grade, Two->Grade, OneG, TwoG);
         break;

         case Tours::stAllPlayHcap:
            komi_stones_colour(One->Grade, Two->Grade, OneG, TwoG);
         break;

         default:
            OneG.Stones = TwoG.Stones = 0;
            OneG.Komi   = TwoG.Komi   = 0;
      }


      // EPILOGUE
      One->Cbal += OneG.Colour;			Two->Cbal += TwoG.Colour;
   }


   void ManualPairing
   (
   	size_t board,
   	size_t black, size_t white,
      int Stones, int Komi
   )
   {
    	using Tours::gdDown;		using Tours::gdEven;		using Tours::gdUp;
      using Tours::gcBlack;	using Tours::gcWhite; 	using Tours::gcNone;

   	using Tours::psPlaying;
      using Tours::GGame;

   	if(black<1 || black>GoTour.Players)Throw(IDA_BadManualPlayerId);
      if(white<1 || white>GoTour.Players)Throw(IDA_BadManualPlayerId);

      int r = GoTour.Round - 1;	  if(r<0)Throw(IDA_BadManualRound);

      GPlayerPtr Black = GoTour.Listed[black];
      GPlayerPtr White = GoTour.Listed[white];

      GGame    &BlackG = Black->Games[r],   &WhiteG = White->Games[r];

      // OPPONENTS
      Black->State = psPlaying;  			  BlackG.Oppid = white;		BlackG.Board = board;
      White->State = psPlaying;				  WhiteG.Oppid = black;    WhiteG.Board = board;

      // PARITY
      if(Black->McScore > White->McScore) { BlackG.Parity = gdDown;  WhiteG.Parity = gdUp;   }
      if(Black->McScore == White->McScore){ BlackG.Parity = gdEven;  WhiteG.Parity = gdEven; }
      if(Black->McScore <  White->McScore){ BlackG.Parity = gdUp;    WhiteG.Parity = gdDown; }

      Black->Mbal += BlackG.Parity;			White->Mbal += WhiteG.Parity;
      Black->Mout += abs(BlackG.Parity); 	White->Mout += abs(WhiteG.Parity);
      Black->Cbal += gcBlack;	  				White->Cbal += gcWhite;

      // COLOUR
      BlackG.Colour = gcBlack;  				WhiteG.Colour = gcWhite;

      // HANDICAPS
      BlackG.Stones =  Stones;  				BlackG.Komi =  Komi;
      WhiteG.Stones = -Stones;  				WhiteG.Komi = -Komi;
   }

   void ManualHandicap(size_t black, size_t white, int Stones, int Komi)
   {
   	using Tours::GGame;

      if(black<1 || black>GoTour.Players)Throw(IDA_BadHcapId);
      if(white<1 || white>GoTour.Players)Throw(IDA_BadHcapId);

      int r = GoTour.Round - 1;	   if(r<0)Throw(IDA_BadHcapRound);

      GGame &BlackG = GoTour.Listed[black]->Games[r];	if(BlackG.Oppid != white) return;
      GGame &WhiteG = GoTour.Listed[white]->Games[r]; if(WhiteG.Oppid != black) return;

      BlackG.Stones =  Stones;		BlackG.Komi =  Komi;
      WhiteG.Stones = -Stones;  		WhiteG.Komi = -Komi;
   }

   void UnpairPlayer(size_t one)
   {
      using Tours::psRegistered;	using Tours:: psFinished;
      using Tours::psPlaying;		using Tours:: psSleep;

      using Tours::gdEven; using Tours::gcNone; using Tours::grNone;

      const int Round = GoTour.Round;		 if(GoTour.Round<1) return;

      if(one<1 || one>GoTour.Players) return;

      Tours::GPlayerPtr One = GoTour.Listed[one];
      Tours::GGame&    OneG = One->Games[Round-1];

      // STATE AND RESULT
      while(true)
      {
      	if(One->Joining(Round))						// State is Sleep, but result is not Sleep
         {
         	GoTour.RestoreRegStatus = true;		// Forces update of reg grid

         	One->State = psPlaying;

            break;
         }

         if(One->Leaving(Round))						// State is Playing, but result is Sleep
         {
         	GoTour.RestoreRegStatus = true;

          	One->State = psSleep;

            break;
         }

         if(One->Playing(Round))
         {
         	// State and Result are reset to values at start of round
         	One->State  = Round==1 ? psRegistered : psFinished;
         	OneG.Result = grNone;

            break;
         }

         if(One->Sleeping(Round))
         	break;

         break;
      }

      // OPPONENT
      OneG.Oppid = OneG.Board = 0;

      // PARITY
      One->Mbal -= OneG.Parity; One->Mout -= abs(OneG.Parity);
      OneG.Parity = gdEven;

      // COLOUR
      One->Cbal -= OneG.Colour;
      OneG.Colour = gcNone;

      // HANDICAPS
      OneG.Stones = 0;	OneG.Komi = 0;
   }

   void UnpairPlayers(int one, int two)
   {
      if(one && two)
      {
      	UnpairPlayer(one);	UnpairPlayer(two);
      }
   }

   bool PlayedBefore(const int One, const int Two, const int round)
   {
   	using TourMaster::GoTour;

      if(One<1 || One>(int)GoTour.Players)Throw(IDA_BadRepeatId);
      if(Two<1 || Two>(int)GoTour.Players)Throw(IDA_BadRepeatId);

      const PairTool::PGames& OneGames = GoTour.Listed[One]->Games;

      for(int r=0; r<round; r++)
      	if(OneGames[r].Oppid == (size_t)Two)
       		return true;

      return false;
   }
}
//
////////////////////////////////////////////////////////////////////////////////








