////////////////////////////////////////////////////////////////////////////////
// MODULE PairMcMahon: The McMahon Pairing Algorithm 			  G.Kaniuk 13/06/05
////////////////////////////////////////////////////////////////////////////////

#include <vector>
#include <string>
#include <algorithms>                      
#pragma hdrstop

#include "PalgRES.h"
#include "PairRES.h"

#include "TourMaster.h"
#include "PairMaster.h"
#include "PairSort.h"
#include "PairTool.h"
#include "Pair.h"


////////////////////////////////////////////////////////////////////////////////
// 'McMAHON' PAIRING ALGORITHM
//
namespace McMahon
{
	using std::string;
   using std::vector;
   
   using Tools::ints;
   using Tools::modulus;

   using Tours::mc_scale;
   using Tours::GPlayerPtr;
   using Tours::GGame;

   using TourMaster::GoTour;

   using PairSort::Throw;
   using PairSort::Sorted;
   using PairSort::SortedPlayer;

   using PairSort::MibSort;
   using PairSort::CibSort;
   using PairSort::ClassSort;
   using PairSort::PoolSort;
   using PairSort::KomiSort;
   using PairSort::GroupCodeSort;
   using PairSort::NextGroupCode;

	using PairTool::MMGroups;
   using PairTool::MMGroup;

	class PMcMahon
   {
   	private:
      	enum	// DRAWSTATUS
         { dsFreshStart, dsInprogress, dsStuck, dsDone, dsCount };

         enum	// PLAYER CLASS
         { pcNone,  pcUp, pcDown, pcOut };

         enum  // GROUP PARITY
         {
         	gpNoParity = -1, gpEvenParity = 0, gpOddParity = 1
         };

         enum	//DRAW CLASS
         {
         	dcNoClass = -1, dcFragile = 0, dcNormal = 1
         };

         enum	//GROUP CLASS
         {
         	gcNoClass = -1, gcFragile = 0, gcNormal = 1
         };

         enum // GROUP START POINT
         {
         	gsNoGroup = -1
         };

   		int DrawStatus;				// Draw progress - one of dsThing
         int DrawClassNow;          // Overall class - one of dcThing
         int DrawClassLast;			// Previous DrawClass
         int ParityNow;					// Parity of an MMGroup is even or odd or none

         int gOne, gTwo;				// Groups for players One and Two
         int gLast;						// Previous gOne
         int gStep;						// Search direction for fragile groups

         int pOne, pTwo;				// Players One and Two - index to Listed;

         int aPlayer, bPlayer;		// The pairing range in the current pass

         MMGroups McGroups;			// McMahon groups at this pairing pass

         int FreePlayerCount;			// Number of unpaired players
         int FragileCount;				// No players with no opponent in own group>1
         int NormalCount;			   // No players with an opponent in own group or singleton

      public:
      	PMcMahon();
         bool Draw(int& PairedCount);

      private:
      	void AnalyseClass(bool& ClassEvent);
         void ClassifyPlayer(const int p);
         void ClassifyGroup(const int g);

         void PruneGroups(void);
         bool PruneTail(void);

         void CheckNearestWorkGroup(const int dir, int& grp, int& mms);
         void FindNextWorkGroup(const int dir, int& grp, int& mms);
         
         void FindGroupOne(void);
         void FindGroupTwo(void);
         void FindKomiGroupTwo(void);

         void FindPlayerOne(void);
         void FindPlayerTwo(void);
   		void FindUnevenVolunteers(int& a, int& b);

         void PurgeSmallPools(int& a, int& b);
         void BookPlayerOne(const int one);
   };

   PMcMahon::PMcMahon()
   {
   	DrawStatus = dsFreshStart;

      FreePlayerCount = 0;
      FragileCount    = 0;
      NormalCount     = 0;

      gLast  =  0;  										// Group index increases from highest MMS to lowest
      gOne   =  0;    gTwo = 0;						// First group is 0
      pOne   =  0;    pTwo = 0;						// First player is 1

      gStep  =  1;
      
      for(size_t p=1; p<=GoTour.Players; p++)	// Clear pairing marks
      	PairSort::Player(p)->Spare[0] = 0;
   }

   bool PMcMahon::Draw(int& PairedCount)
   {
      // THIS IS A NEW PAIRING
      DrawStatus    = dsFreshStart;
      DrawClassNow  = dcNoClass;
      DrawClassLast = dcNoClass;
      ParityNow     = gpNoParity;

      FreePlayerCount = -1;

      PairedCount = 0;

      // START A NEW PAIRING PASS
   	while(DrawStatus == dsFreshStart)
      {
      	DrawClassLast = dcNoClass;
         		  gLast = gsNoGroup;
      	DrawStatus    = dsInprogress;

         // PAIR UNTIL THERE IS A CHANGE IN STATUS
         while(DrawStatus == dsInprogress)
         {
         	PairTool::FormMMGroups(McGroups);

            const int McCount = McGroups.size();

            aPlayer = McGroups[0].Start;
            bPlayer = McGroups[McCount-1].Start + McGroups[McCount-1].Size - 1;

            FreePlayerCount = bPlayer - aPlayer + 1;	// All non-free players were removed in the McMahon sort

            // ANALYSE THE POOL FOR EACH PLAYER
            bool ClassEvent = false;

            AnalyseClass(ClassEvent);

            if(ClassEvent)										// We have a change in PoolsClass
            {
            	DrawStatus = dsFreshStart;

               break;
            }

            // MAKE ONE PAIRING AND PAIR THE PLAYERS
            PruneGroups();										// Make one pairing and maybe modify DrawStatus

            if(DrawStatus != dsStuck)
            {
               PairMaster::PairPlayers(pOne, pTwo);

               if(gTwo==gOne)									// Pool for gOne updated in BookPlayerOne
               	--McGroups[gTwo].Pool;

               FreePlayerCount -= 2;

               PairedCount += 2;
            }
            else													// Unhook the players
            {
               PairMaster::UnpairPlayer(pOne);
               PairMaster::UnpairPlayer(pTwo);

               return true;									// Post draw analysis lists unpaired players
            }

            // WE ARE DONE WHEN FREE PLAYERS EXHAUSTED
            if(!FreePlayerCount)
            {
            	DrawStatus = dsDone;

               return true;
            }

            // EPILOGUE: MAINTAIN THE SEARCH START POINT FOR GROUP ONE
            if(!McGroups[gOne].Pool)						// We have lost a group
            	--gLast;

            if(gTwo<gOne)										// pOne is playing up
            	if(!McGroups[gTwo].Pool)					// We have lost another pool
               	--gLast;                            // Earlier than gOne

            if(!McGroups[McCount-1].Pool)
            	gLast = -1;

         }// while DrawStatus==dsInprogress

         if(DrawClassNow > DrawClassLast) ParityNow = gpOddParity;
         if(DrawClassNow < DrawClassLast) ParityNow = gpNoParity;

      }// while DrawStatus==dsFreshStart
      return false;
   }

   bool Draw(int& PairedCount)
   {
   	PMcMahon McMahon;

      bool DrawDone = McMahon.Draw(PairedCount);

      return DrawDone;
   }
}

/////////////////////////////////////////////////////////////////////////////
// CLASS ANALYSIS
//
namespace McMahon
{
   void PMcMahon::ClassifyPlayer(const int p)
   {
      GPlayerPtr Player = SortedPlayer(p);

      const int go = Player->McGroup;

      const int gsize = McGroups.size();

      // PLAYER'S POOL COUNTS
      int po = McGroups[go].Pool - 1;    		// Own pool
      int pu=0, pd=0;								// Up and Down pools

      // NEAREST UP AND DOWN POOLS
      const int gu =  go-1, gd = go+1;
      if(gu>=0)    pu += McGroups[gu].Pool;
      if(gd<gsize) pd += McGroups[gd].Pool;

      // NEXT NEAREST UP AND DOWN POOLS
      const int guu = go-2, gdd = go+2;
      if(guu>=0)    pu += McGroups[guu].Pool;
      if(gdd<gsize) pd += McGroups[gdd].Pool;

      // REMOVE OPPONENTS FROM ALL THE POOLS IN PREVIOUS ROUNDS
      bool nopools = false;

      for(size_t r=0; r<GoTour.Round-1; r++)
      {
      	const GGame& Game = Player->Games[r];

         if(Game.Oppid)
         {
         	const GPlayerPtr Opponent = GoTour.Listed[Game.Oppid];

         	if(Opponent->Free())
            {
		         const int gopr = Opponent->McGroup;

         		if(gopr==gu || gopr==guu) --pu;	// Exclude opponents in up groups
         		if(gopr==gd || gopr==gdd) --pd;  // Exclude opponents in down groups

         		if(gopr==go) --po;					// Exclude opponents in same group

               if(pu<0) pu = 0;						// No opponents left in up group
               if(pd<0) pd = 0;                 // No opponents left in down group
               if(po<0) po = 0;                 // No opponents left in same group

               nopools = !pu && !pd && !po;
            }
         }
      }

      // REMOVE ANY 'MUST NOT' PLAY OPPONENTS
      const string& pCode = Player->Group;

      if(pCode[1] == Tours::gpNever)
      	for(int g=guu; g<=gdd; g++)
      		if(0<=g && g<gsize)
      		{
            	const PairTool::MMGroup& mg = McGroups[g];
               const int a = mg.Start, b = mg.Start + mg.Size - 1;

               for(int q = a;  q <= b;  q++)
               	if(q != p && SortedPlayer(q)->Group == pCode)
                  {
                  	if(g==guu || g==gu) --pu;
                     if(g==gdd || g==gd) --pd;

                     if(g==go) --po;

                     if(pu<0) pu = 0;						// No opponents left in up group
                     if(pd<0) pd = 0;                 // No opponents left in down group
                     if(po<0) po = 0;                 // No opponents left in same group

                     nopools = !pu && !pd && !po;
                  }
      		}

      // THE PLAYER'S CLASS IS DECIDED BY THE POOL DISTRIBUTION
      if(pu==0 && pd==0) Player->Class = pcNone;		// We will have to search far for opponents
      if(pu>0  && pd==0) Player->Class = pcUp;			// Opponents upstream only
      if(pd>0  && pu==0) Player->Class = pcDown;      // Opponents downstream only
      if(pu>0  && pd>0)  Player->Class = pcOut;       // Opponents in both directions

      // THE PLAYER'S POOL
      Player->Pool = po;

      nopools = nopools ||  (!pu && !pd && !po);		// Test point 
   }
   //
   /////////////////////////////////////////////////////////////////////////////
   //
   void PMcMahon::ClassifyGroup(const int g)
   {
   	if(g<0 || g>=(int)McGroups.size())Throw(IDA_BadClassifyGroup);

      PairTool::MMGroup& mg = McGroups[g];

      if(mg.Size == 0)Throw(IDA_BadGroupSize);

      // MOST GROUPS ARE NORMAL
      mg.Class = gcNormal;

      if(mg.Size == 1) return;

      // A GROUP OF MORE THAN 1 IS FRAGILE
      // IF THERE EXISTS A PLAYER IN THE GROUP
      // WITH NO OPPONENTS
      const int a = mg.Start, b = mg.Start+mg.Size-1;

      for(int p=a; p<=b; p++)
      	if(!SortedPlayer(p)->Pool)
         	{ mg.Class = gcFragile;		return; }
   }
   //
   /////////////////////////////////////////////////////////////////////////////
   //
   void PMcMahon::AnalyseClass(bool& ClassEvent)
   {
      // WE COUNT NUMBER OF GROUPS IN EACH CLASS
      const int McCount = McGroups.size();

      // CLASSIFY ALL THE PLAYERS
      for(int p=aPlayer; p<=bPlayer; p++)
      	ClassifyPlayer(p);

      // CLASSIFY ALL THE GROUPS, COUNTING NUMBER IN EACH CLASS
      FragileCount = NormalCount = 0;

      for(int g=0; g<McCount; g++)
      {
      	PairTool::MMGroup& McGroup = McGroups[g];

      	ClassifyGroup(g);

         if(McGroup.Class == gcFragile) ++FragileCount;
         if(McGroup.Class == gcNormal)  ++NormalCount;
		}

      if(!FragileCount && !NormalCount)
			PairSort::Throw(IDA_BadGroupClass);

      // THE CURRENT DRAW CLASS AND PARITY.
      // NOTE THAT gcFragile AND gcNormal ARE MUTUALLY EXCLUSIVE
      DrawClassNow = FragileCount ? dcFragile : dcNormal;

      if(DrawClassLast == dcNoClass)
      {
      	DrawClassLast = DrawClassNow;

         ParityNow = DrawClassNow == dcFragile ? gpNoParity : gpOddParity;
      }

      // DETECT CHANGE OF CLASS EVENT
      ClassEvent = DrawClassNow !=  DrawClassLast;
   }
}

/////////////////////////////////////////////////////////////////////////////
// PRUNING
//
namespace McMahon
{
   void PMcMahon::PruneGroups(void)
   {
   	// START A NEW SEARCH FOR ONE PAIR
      pOne = pTwo = 0;

   	// WE FIRST DRAW FROM THE THIN TAIL
    	if(PruneTail())
      	return;

      // SEARCH FOR GROUP ONE BY CLASS AND PARITY
   	FindGroupOne();

      // CHOOSE PlayerOne BY CLASS, MIB
      FindPlayerOne();

      // SEARCH FOR GROUP TWO WHEN FRAGILE OR ODD
      if(DrawClassNow == dcFragile || ParityNow == gpOddParity)
      {
      	FindGroupTwo();		if(DrawStatus == dsStuck) return;
      }
      else
      	gTwo = gOne;

      if(DrawClassNow == dcNoClass && ParityNow == gpNoParity)
      	Throw(IDA_BadClassParity);

      FindPlayerTwo();
   }
   //
   /////////////////////////////////////////////////////////////////////////////
   //
   bool PMcMahon::PruneTail(void)
   {
   	// FRAGILE GROUP MANAGED BY FindGroupOne
   	if(DrawClassNow == dcFragile)return false;

   	// CHECK FOR THIN TAIL - LAST GROUPS HAVE JUST ONE PLAYER EACH
      const int McCount = McGroups.size();

      MMGroup TailGroup = McGroups[McCount-1];

      if(McCount==1)																// There is only one group
      {
         if(TailGroup.Size>1) return false;								// It is not a thin tail

         DrawStatus = dsStuck; return true;     						// Stuck if there is only one player in the group
      }

      if(TailGroup.Size==1)													// Thin tail
      {
      	gOne = McCount-1;    gLast = gOne;

      	BookPlayerOne(bPlayer);
      }
      else																			// The last group is bushy
      {
         if(McGroups[McCount-2].Size>1)									// The last group does not shadow a thin tail
         	return false;

         if(TailGroup.Komip==0)return false;								// No komi-players players here - treat as normal

         gOne = McCount-1;		gLast = gOne;								// We need to trim the bush

         FindPlayerOne();
      }

      if(PairSort::Player(pOne)->Grade > GoTour.Rules.KomiGroup)  // PlayerOne is booked
      	FindGroupTwo();
      else
      	FindKomiGroupTwo();

      if(DrawStatus == dsStuck) return true;

      FindPlayerTwo();

      return true;
   }
}


/////////////////////////////////////////////////////////////////////////////
// FINDING GROUPS
//
namespace McMahon
{
   void PMcMahon::FindGroupOne(void)
   {
   	using PairTool::ChooseExcluding;

      const int McCount = McGroups.size();

		// LINEAR SEARCH FOR FRAGILE GROUP
		int gbegin, gend, g;											// The group search range

      if(DrawClassNow == dcFragile)
      {
      	gLast = gStep>0 ? gsNoGroup : McCount;

         gbegin = gLast + gStep;
         gend   = gbegin + gStep*McCount;

         for(g=gbegin; g != gend; g += gStep)
         	if(McGroups[g].Class == gcFragile)				// Guaranteed for some g
            {
            	gOne = g;											// Group one found

               gLast = gOne;										// Start for next circular search

               gStep = -gStep;									// Reverse for next time

               return;
            }
      }

      // FIRST FIND THE START POINT OF THE CIRCULAR SEARCH
      if(gLast < -2 || gLast > McCount)
      	Throw(IDA_BadGroupLast);

      gStep = 1;	gbegin = gLast + gStep;

      if(gbegin<0 ||gbegin >= McCount)
      	gbegin = 0;

      gend   = gbegin + McCount;									// Unwrapped range

      // NOW FIND GROUP MATCHING PARITY
      for(g = gbegin; g != gend; g++)
      {
      	const int f = g>=McCount ? g - McCount : g;		// Circular index

         if(f<0 || f>=McCount)
         	PairSort::Throw(IDA_BadGroupIndex);

         if(McGroups[f].Pool%2 == ParityNow)					// Odd or Even guaranteed
         {
         	gOne = f;

            break;
         }
      }

      if(g == gend)													// SEARCH FAILS
      {
      	gOne = 0;

         switch(ParityNow)
         {
         	case gpOddParity:	 ParityNow=gpEvenParity; break;	// All odd groups done, switch  to even

            case gpEvenParity: Throw(IDA_BadEvenParity);			// Cannot happen as all groups even
            case gpNoParity: 	 Throw(IDA_BadNoParity);			// Only Fragile groups have NoParity
            default:           Throw(IDA_BadParityNow);
         }
      }
      gLast = gOne;
   }
   //
   /////////////////////////////////////////////////////////////////////////////
   //
   void PMcMahon::CheckNearestWorkGroup(const int dir, int& grp, int& mms)
   {
   	// NOTE THAT THE WORK POOLS HAVE ALREADY BEEN PURGED OF
      // OLD OPPS AND 'MUST NOT PLAY'
      const int McCount = McGroups.size();

      grp = gOne + dir;

      // BOUNDARY CHECK
		if((dir>0 && grp >= McCount) || (dir<0 && grp<0) )
      {
         grp = gsNoGroup;  return;
      }

      // POPULATION CHECK
      if(!McGroups[grp].Work)
      {
         grp = gsNoGroup;  return;
      }

      // WANT SCORE DIFFERENCE
      mms = modulus(McGroups[grp].Score - McGroups[gOne].Score);
   }
   //
   /////////////////////////////////////////////////////////////////////////////
   //
   void PMcMahon::FindNextWorkGroup(const int dir, int& grp, int& mms)
   {
   	// NOTE THAT THE WORK POOLS HAVE ALREADY BEEN PURGED OF OLD OPPS AND 'MUST NOT PLAY'
      const int McCount = McGroups.size();

      int popgrp = 0, gre;

      if(dir == 1)
      	gre = McCount;
      else
      	gre = -1;

      grp = gOne + 2*dir;

      if(grp<0 || grp>=McCount)
      {
      	grp = gsNoGroup;	return;
      }

      // SEARCH FOR FIRST GROUP CONTAINING AT LEAST ONE OPPONENT
      while(grp != gre && !popgrp)
      {
      	popgrp = McGroups[grp].Work;	// Population of free players in upstream group

         if(!popgrp) grp += dir;
      }

      // WE WANT THE SCORE DIFFERENCE
      if(popgrp)
      	mms = modulus(McGroups[grp].Score - McGroups[gOne].Score);
      else
      	grp = gsNoGroup;
   }
   //
   /////////////////////////////////////////////////////////////////////////////
   //
   void PMcMahon::FindGroupTwo(void)
   {
      // ELIMINATE UNPLAYABLE OPPONENTS
      PairTool::PurgeOpponents(pOne, McGroups);

      // ANALYSE NEAREST GROUPS FOR OPPONENTS
      int gu, mu;  CheckNearestWorkGroup(-1, gu, mu);
      int gd, md;  CheckNearestWorkGroup( 1, gd, md);

		// SEARCH FOR NEXT NEAREST GROUPS CONTAINING AT LEAST ONE OPPONENT
		int guu, muu;	FindNextWorkGroup(-1, guu, muu);
      int gdd, mdd;	FindNextWorkGroup( 1, gdd, mdd);

      // SEARCH FOR A GROUP HAS FAILED
      if(gu == gsNoGroup && guu == gsNoGroup)
      if(gd == gsNoGroup && gdd == gsNoGroup)
      {
      	gTwo = gsNoGroup;	DrawStatus = dsStuck;

         return;
      }

      // PLAYOUT BECAUSE gOne IS ODD
      if(DrawClassNow == dcNormal)
      {
      	if(gd != gsNoGroup)  { gTwo = gd; return; }
         if(gdd != gsNoGroup) { gTwo = gdd; return; }

       	if(gu != gsNoGroup)  { gTwo = gu; return; }
         if(guu != gsNoGroup) { gTwo = guu; return; }
      }

      // FRAGILE gOne - MINIMISE THE McMAHON SCORE DIFFERENCE
      const GPlayerPtr Player = PairSort::Player(pOne);

      int mumin,mdmin, gumin,gdmin,gomin;

      mumin = mdmin = 100*Tours::mc_scale;
      gumin = gdmin = gomin = gsNoGroup;

      // NOTE THAT SINCE ONE OF guu, gu, gd, gdd IS SET,
      // AND SINCE EVERY mxx < momin IT FOLLOWS THAT
      // AT LEAST ONE OF THE FOLLOWING FOUR INEQUALITIES IS TRUE
      if(guu != gsNoGroup && muu < mumin) { mumin = muu; gumin = guu; }
      if(gu  != gsNoGroup &&  mu < mumin) { mumin = mu;  gumin = gu;  }
      if(gdd != gsNoGroup && mdd < mdmin) { mdmin = mdd; gdmin = gdd; }
      if(gd  != gsNoGroup &&  md < mdmin) { mdmin = md;  gdmin = gd;  }

      if(mumin  < mdmin) gomin = gumin;
      if(mdmin  < mumin) gomin = gdmin;
      if(mdmin == mumin) gomin = gsNoGroup;

      // PLAYOUT BECAUSE gOne IS FRAGILE
      if(gomin != gsNoGroup)
      	gTwo = gomin;
      else
         gTwo = Player->Mbal>= 0 ? gdmin : gumin;
   }
   void PMcMahon::FindKomiGroupTwo(void)
   {
      // ELIMINATE UNPLAYABLE OPPONENTS
      PairTool::PurgeOpponents(pOne, McGroups);

      // FIND UPSTREAM WORKING KOMI GROUP
      for(int g = gOne; g>=0; g--)						// Search upstream
      {
      	const MMGroup& McGroup = McGroups[g];

			if(ExistsKomiOpponent(pOne, McGroup))		// McGroup has at least 1 komi opponent
			{
         	gTwo = g; 	return;
         }
      }

      FindGroupTwo();										// Find nearest group
   }
}

////////////////////////////////////////////////////////////////////////////////
// FINDING PLAYERS
//
namespace McMahon
{
   using PairSort::mrOut;
   using PairSort::mrUp;
   using PairSort::mrDown;

   using PairSort::crGrey;
   using PairSort::crBlack;
   using PairSort::crWhite;

   void PMcMahon::FindPlayerOne(void)	// GUARANTEED TO FIND A PLAYER
   {
   	// WE ALREADY KNOW THE GROUP - FIND A PLAYER
   	PairTool::MMGroup& GroupOne = McGroups[gOne];

      int ta = GroupOne.Start, tb = GroupOne.Start+GroupOne.Size-1;  // Backtrack store
      int  a = ta, 			     b = tb;								  		// Dynamic range

      bool evengame = true;

      // FOR A SINGLETON GROUP, NO SEARCH NEEDED
      if(a == b)
		{
      	BookPlayerOne(GroupOne.Start);

         return;
      }

      // PLAYERS WITH FEWEST OPPONENTS GET PRIORITY
      // SUPPOSE A PLAYER HAS THE SMALLEST POOL IN THE GROUP,
      // BUT TWO OTHER POTENTIAL OPPONENTS ARE PAIRED
      // THEN NEXT TIME ROUND< THE PLAYER'S POOL IS REDUCED BY 2
      PoolSort(a, b);

      if(DrawClassNow == dcNormal)										// We can expect problems if
      {                                                        // the smallest pool is
      	const int aPool = SortedPlayer(a)->Pool;              // smaller than the rounds remaining

         const int RoundsLeft = GoTour.Rules.Rounds - GoTour.Round + 1;

         if(aPool-2 > RoundsLeft)										// BUGFIX V6.2.1  WAS aPool > RoundsLeft
         {                     											// No problem, so we will choose from
         	a = ta;  b = tb;												// all players in group
         }

         evengame = true; 													// Optimist
      }

      // FOR ODD GROUPS FIND VOLUNTEERS TO PLAY DOWN
      int MibRule = PairSort::mrNone, Mib = 0;

      if(DrawClassNow == dcNormal && ParityNow == gpOddParity)
      {
      	MibRule = PairSort::mrDown;									// Select played up and lost a lot

         ta = a;  tb = b;

         int Mib = MibSort(a, b, MibRule);							// MibRule reset to mrNone if no volunteers

         if(Mib<=0) MibRule = PairSort::mrNone;

         evengame = false;
      }

      // FOR FRAGILE GROUPS FIND VOLUNTEERS TO PLAY OUT
      if(DrawClassNow==dcFragile)
      {
      	const int   McCount = McGroups.size();
         const ints OneScore = McGroups[gOne].Score;

         const int gu = gOne-1,  gd = gOne+1;  						// Neighbours
         bool    UpOK = false, DnOK = false;							// Desired neighbours

         if(gu>=0) 		UpOK = McGroups[gu].Score - OneScore <= mc_scale;
         if(gd<McCount) DnOK = OneScore - McGroups[gd].Score <= mc_scale;

         if(UpOK==DnOK)    MibRule = mrOut;							// Volunteer to play out
         if(UpOK && !DnOK) MibRule = mrUp;         				// Choose from high played down and won
         if(DnOK && !UpOK) MibRule = mrDown;							// Choose from high played up and lost

         ta = a; tb = b;

         Mib = MibSort(a, b, MibRule);

         if(Mib==0) MibRule = mrNone;                          // No Voulunteer for out
         if(Mib < 0) MibRule = mrUp;									// Wants to play up
         if(Mib > 0) MibRule = mrDown;									// Wants to play down

         evengame = false;
      }

      // FOR ODD OR FRAGILE GROUPS TRY TO MATCH THE DESIRED DRAW DIRECTION
      // WITH THE ACHIEVABLE DRAW DIRECTION AS DEFINED BY THE PLAYER'S CLASS
      if(!evengame)
      {
      	if(MibRule != mrNone)							// Try to match desired direction
         	if(!ClassSort(a,b, MibRule))				// Cannot match
            	MibRule = mrNone;

         if(MibRule == mrNone)							// No desired/achieved match possible
         {
         	a = ta;	b = tb;								// Backtrack to maximal useful pool

            (void)MibSort(a,b, mrOut);					// Find those not yet played up or down

            if(!ClassSort(a,b, mrDown))				// Of these, look for those ABLE to play down
            {
            	a = ta; b = tb; 							// Backtrack if still no one
            }
         }
      }

      const bool AreaStyle = GoTour.Rules.Style == Tours::stMcMahon
                                                ||
      							  GoTour.Rules.Style == Tours::stBarLow;

      if(AreaStyle)		// WE IGNORE AREA FOR SWISS TOURNAMENTS
      {
         // WHATEVER THE CLASS, WE CHOOSE FROM THE LARGEST AREA
         //PairTool::FormMaximumArea(a,b);

         // WITHIN THE LARGEST AREA WE GIVE PRIORITY TO PLAYERS
         // WITH GROUP CODES MustNot, PreferNot, AND Prefer
         // BECAUSE THESE HAVE REDUCED POOLS
         // NOTE THAT THERE ARE NO MustPlay AS THESE HAVE ALREADY BEEN PAIRED
         GroupCodeSort(a,b);
         {
            string GroupCode;									// Identified by NextGroupCode

            const int ga = a, gb = b;					   // Group range

            a = 0;                                    // First call to NextGroupCode

            // IDENTIFY PLAYERS WITH SAME GROUP CODE CODE
            NextGroupCode(ga,gb, GroupCode, a,b);
         }

         // BELOW THE BAR GIVE PRIORITY TO PLAYERS WITH HIGH NUMBER OF SAME-AREA GAMES
         PairTool::FormSameAreaCount(PairSort::saForward, a,b);
		}

      // CHOOSE A PLAYER WITH THE WORST COLOUR IMBALANCE
      CibSort(a,b, PairSort::crGrey);

      // FINALLY - CHOOSE THE PLAYER AT RANDOM
      const int One = PairTool::Choose(a,b);

      BookPlayerOne(One);
   }
   //
   /////////////////////////////////////////////////////////////////////////////
   //
   void PMcMahon::FindPlayerTwo(void)
   {
   	// IDENTIFY RANGE OF PLAYERS IN GROUP TWO
      const GPlayerPtr PlayerOne = PairSort::Player(pOne);

   	MMGroup& GroupTwo = McGroups[gTwo];

      int a = GroupTwo.Start, b = GroupTwo.Start + GroupTwo.Size - 1;

      if(a==b)
      {
      	pTwo = Sorted[a];	return;
      }

      // THERE MAY BE SOME PLAYERS IN GROUP TWO WHICH CANNOT PLAY pOne.
      // REMOVE THEM AND GIVE PRIORITY TO SMALLEST POOLS
		PurgeSmallPools(a,b);

      // FIND VOLUNTEERS ABOVE KOMIBAR IF GAME IS NOT EVEN
      if(gOne != gTwo)
      	if(GroupTwo.Komip==0)								// GK 6.4.0 There are no komi players
         	FindUnevenVolunteers(a, b);

      // IF TAGGED 'PREFER', FIND COMPATIBLE OPPONENTS
      PairSort::CompatibleSort(a, b, Tours::gpPrefer, PlayerOne->Group);

      // FIND PLAYERS FROM DIFFERENT REGION IF REQUIRED
      const bool AreaStyle = GoTour.Rules.Style == Tours::stMcMahon
                                                ||
      							  GoTour.Rules.Style == Tours::stBarLow;

      if(AreaStyle)
      {
         const int BarScore = GoTour.Rules.MacBar;   	// GK 6.4.0 was GoTour.Rules.MacBar+(GoTour.Round-1)*Tours::mc_scale;

         bool AlienGame =  true;

         if(McGroups[gOne].Score >= BarScore)			// If BOTH players below Bar
            AlienGame = GoTour.Rules.BarSeg;          // we try to segregate

         if(McGroups[gTwo].Score >= BarScore)
            AlienGame = GoTour.Rules.BarSeg;

         if(AlienGame)
            PairTool::FormAlienArea(PlayerOne->Land, PlayerOne->Club, a, b);

         // IF TAGGED 'AVOID', EXCLUDE INCOMPATIBLE OPPONENTS
         PairSort::CompatibleSort(a, b, Tours::gpAvoid, PlayerOne->Group);
      }

      // TRY TO MINIMISE THE COLOUR IMBALANCE OF THE PLAYERS
      {
      	int CibOne = PlayerOne->Cbal, rule;

         if(CibOne >  0) rule = crBlack;	// One played much white
         if(CibOne == 0) rule = crGrey;   // One doesnt care
         if(CibOne <  0) rule = crWhite;  // One played much black

         CibSort(a,b, rule);
      }

      // CHOOSE TWO AT RANDOM
      const int Two = PairTool::Choose(a,b);

      pTwo = Sorted[Two];
   }
  	//
  	//////////////////////////////////////////////////////////////////////////////
  	//
  	void PMcMahon::BookPlayerOne(const int one)
  	{
   	// PLAYER ONE IS SELECTED
      GPlayerPtr PlayerOne = SortedPlayer(one);

   	pOne = Sorted[one];

      PlayerOne->State = Tours::psChosen;

      // THE POOL OF THE PLAYER'S GROUP IS THEREFORE REDUCED
      MMGroup& GroupOne = McGroups[PlayerOne->McGroup];

      --GroupOne.Pool;

      if(!GroupOne.Pool)
      	GroupOne.Class = gcNoClass;
  	}
   //
   /////////////////////////////////////////////////////////////////////////////
   //
   void PMcMahon::FindUnevenVolunteers(int& a, int& b)
   {
   	using PairSort::MibSort;

      if(a==b) return;

   	int ta = a, tb = b;								// Backtrack

      if(gOne < gTwo) 									// mmsTwo is weaker, so gTwo playing up
      {
      	int Mib = MibSort(a,b, mrUp);				// Volunteers have been playing down, have Mib<0

         if(Mib >= 0)									// There are NO volunteers to play up
         {
         	a = ta;	b = tb;							// Backtrack

         	Mib = MibSort(a,b, mrOut);				// Choose a virgin
         }
      }

      if(gTwo < gOne)									// mmsTwo is stronger, so playing down
      {
      	int Mib = MibSort(a,b, mrDown);			// Volunteers have been playing up, so have Mib>0

         if(Mib<=0)										// There are NO volunteers to play down
         {
            a = ta;  b = tb;							// ... so backtrack

            Mib = MibSort(a,b, mrOut);				// Choose a virgin
         }
      }
   }
   //
   /////////////////////////////////////////////////////////////////////////////
   //
   void PMcMahon::PurgeSmallPools(int& a, int& b)
   {
   	// NUMBER OF ROUNDS LEFT INCLUDING THIS ONE
   	const int RoundsLeft = GoTour.Rules.Rounds - GoTour.Round + 1;

      if(a==b) return;

   	// THERE MAY BE SOME PLAYERS IN GROUP TWO WHICH CANNOT PLAY ONE - REMOVE THEM
      PairTool::MarkOpponents(pOne, gTwo, McGroups);

      PairSort::BusySort(a, b);

      // GIVE PRIORITY TO PLAYERS WITH SMALLEST POOLS
   	const int ta = a, tb = b;

      PairSort::PoolSort(a,b);							// We can expect problems if
      {                                            // the smallest pool is
      	const int aPool = SortedPlayer(a)->Pool;  // smaller than the rounds remaining

         if(aPool > RoundsLeft)							// No problem
         {                     							// so we will choose from
         	a = ta;  b = tb;								// all players in group
         }
      }
   }
}
//
////////////////////////////////////////////////////////////////////////////////



