////////////////////////////////////////////////////////////////////////////////
// MODULE PairSort: All sort functions for pairing  				G.Kaniuk 16/10/04
////////////////////////////////////////////////////////////////////////////////
#include <functional>
#include <string>
#include <vector>
#include <ctype.h>
#include <stdio.h>
#include <vcl.h>
#pragma hdrstop

#include "PalgRES.h"
#include "ETour.h"
#include "Maths.h"
#include "PairSort.h"
#include "TourMaster.h"

////////////////////////////////////////////////////////////////////////////////
// ENVIRON
//
// IN ALL SORT ROUTINES a AND b DEFINE AN INDEX RANGE TO THE VECTOR Sorted
// SUCH THAT ALL PLAYERS i IN a<=i<=b ARE SORTED
//
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// TOOLS
//
namespace PairSort
{
	using std::vector;
	using std::sort;
   using std::string;
   using std::binary_function;

   using Tools::ints;
   
   using Tours::GPlayerPtr;
   
   using TourMaster::GoTour;

	std::vector<int> Sorted;		  // Index to PlayerList

   void Throw(int ID_message)
   {
   	using ToolsGUI::ThrowError;

      if(LoadStr(ID_message).IsEmpty())
      	throw Exception("Unknown pairing message");

   	ThrowError(ToolsGUI::erBug, IDA_Pairing, IDA_PairingProc, ID_message);
   }

   void StartSort(int Size)
   {
   	if(Size<1) Throw(IDA_BadStartSort);

   	try
      {
         Sorted.resize(Size);

         for(int i=0; i<Size; i++)Sorted[i] = i;

         return;
      }
      catch(...)
      {
      	Throw(IDA_MemStartSort);
      }
   }

   void EndSort(void)
   {
   	Sorted.clear();
   }

   bool ValidSortRange(const int a, const int b)
   {
   	if(a<1 || b<a || b > (int)Sorted.size())
      	return false;
      else
      	return true;
   }

   GPlayerPtr SortedPlayer(const int p)
   {
   	if(p<1 || p > (int)GoTour.Players)Throw(IDA_BadSortedPlayerIndex);

   	return GoTour.Listed[Sorted[p]];
   }

   GPlayerPtr Player(const int p)
   {
   	if(p<1 || p > (int)GoTour.Players)Throw(IDA_BadPlayerIndex);

   	return GoTour.Listed[p];
   }
}

////////////////////////////////////////////////////////////////////////////////
// McMAHON SORT
//
namespace PairSort
{
   struct OrderMcMahon : binary_function<size_t, size_t, bool>
   {
      bool operator()(const size_t p,  const size_t q)
      {
         GPlayerPtr P = GoTour.Listed[p], Q = GoTour.Listed[q];

         if(GoTour.Round < 1) return p<q;					// Only used from Round 1 up

         bool Pfree = P->Free();								// Waiting for draw if Registered or Finished
         bool Qfree = Q->Free();

         if(Pfree && Qfree)									// Both playing
            return P->McScore > Q->McScore;				// Descending MMorder

         if(Pfree && !Qfree)									// Sleepers or paired to back
            return true;
         else
            return false;
      }
   };

	void McMahonSort(const int a, int& b)
   {
   	if(!ValidSortRange(a,b))Throw(IDA_BadMcMahonRange);

      // SINGLETON EPILOGUE
      if(b==a)
      	if(SortedPlayer(a)->Free())
         	return;
         else
         	Throw(IDA_BadMcMahonSingle);

      	Iterator bSort = Sorted.begin() + a;
      	Iterator eSort = Sorted.begin() + b + 1;

     		sort(bSort, eSort, OrderMcMahon());						// Firstly McMahon Order

      // TRIM TO AVAILABLE
      int e = b+1;														// Exclude unavailable players

      while(e != a)
      {
      	--e;

      	if(SortedPlayer(e)->Free())								// Last player waiting to be paired
         	break;
      }
      b=e;																	// b is Last Awake player, or it is a

      if(!SortedPlayer(b)->Free()) 								// false if ALL players asleep
      	Throw(IDA_BadMcMahonEnd);
   }
   //
   /////////////////////////////////////////////////////////////////////////////
   //
   static ints McScore;

   struct OrderModMcDiff : binary_function<size_t, size_t, bool>
   {
      bool operator()(const size_t p,  const size_t q)
      {
         const int pScore = Tools::modulus(GoTour.Listed[p]->McScore - McScore);
         const int qScore = Tools::modulus(GoTour.Listed[q]->McScore - McScore);

         return pScore < qScore;
      }
   };

   void ModMcDiffSort(const int a, int& b, const ints mcscore)
   {
   	using Tools::modulus;

   	McScore = mcscore;

   	if(!ValidSortRange(a,b))Throw(IDA_BadMcDiffRange);

      // NO SINGLETON EPILOGUE
      if(b==a)return;

      	Iterator bSort = Sorted.begin() + a;
      	Iterator eSort = Sorted.begin() + b + 1;

     		sort(bSort, eSort, OrderModMcDiff());

      // TRIM TO LOWEST DIFFERENCE
      const int aMcDiff = modulus(SortedPlayer(a)->McScore - McScore);

      for(int p=a+1; p<=b; p++)
      {
      	const int pMcDiff = modulus(SortedPlayer(p)->McScore - McScore);

         if(pMcDiff != aMcDiff){ b = p-1; break; }
      }
   }
}


////////////////////////////////////////////////////////////////////////////////
// KOMI SORT
//
// Ordering is dependent on parameter KomiPlayer:
//
// KomiPlayer true:  Free Komi players,then NonKomi players
// KomiPlayer true:  FreeNonKomi players,then Komi players
//
namespace PairSort
{
	static bool KomiPlayer;

   struct OrderKomi : binary_function<size_t, size_t, bool>
   {
      bool operator()(const size_t p,  const size_t q)
      {
         GPlayerPtr P = GoTour.Listed[p], Q = GoTour.Listed[q];

         bool Pfree = P->Free();
         bool Qfree = Q->Free();

         bool Pkomi = P->Grade <= GoTour.Rules.KomiGroup;
         bool Qkomi = Q->Grade <= GoTour.Rules.KomiGroup;

         if(KomiPlayer)
         	return (Pfree && Pkomi) && !(Qfree && Qkomi);
         else
         	return (Pfree && !Pkomi) && !(Qfree && !Qkomi);
      }
   };

	void KomiSort(const int a, int& b, const bool komiplayer)
   {
   	KomiPlayer = komiplayer;

   	if(!ValidSortRange(a,b))Throw(IDA_BadKomiRange);

      // NO SINGLETON EPILOGUE
      if(b==a)return;

      	Iterator bSort = Sorted.begin() + a;
      	Iterator eSort = Sorted.begin() + b + 1;

     		sort(bSort, eSort, OrderKomi());

      // TRIM TO THE SET WANTED - IF ANY
      int p;

      for(p=a+1; p<=b; p++)
      {
      	GPlayerPtr P = SortedPlayer(p);

         if(!P->Free())break;

         if(KomiPlayer)
         	if(P->Grade > GoTour.Rules.KomiGroup)break;

         if(!KomiPlayer)
         	if(P->Grade <= GoTour.Rules.KomiGroup)break;
      }

      // EPILOGUE - IF NONE EXIST LEAVE b ALONE
      int q = p-1;																	// b is Last Awake player, or it is a

      if(!SortedPlayer(q)->Free())return;

      if(KomiPlayer)
      	if(SortedPlayer(q)->Grade > GoTour.Rules.KomiGroup)return;

      if(!KomiPlayer)
      	if(SortedPlayer(q)->Grade <= GoTour.Rules.KomiGroup)return;

      b = q;
   }
}

////////////////////////////////////////////////////////////////////////////////
// AREA SORT
//
// AreaSort organises in lexicographic ordering of the area code = Land+Club
//
// AreaSizeSort assumes that the players have been marked with the size of their
// their areas.
//
// AlienAreaSort works with free players only and excludes a specific AreaCode
//
namespace PairSort
{
   struct OrderArea :  binary_function<size_t, size_t, bool>
   {
      bool operator()(const size_t p,  const size_t q)
      {
         GPlayerPtr P = GoTour.Listed[p], Q = GoTour.Listed[q];

         std::string Parea = P->AreaCode();
         std::string Qarea = Q->AreaCode();

         bool Pfree = P->Free(), Qfree = Q->Free();				// Players free to be paired

         if(Pfree && Qfree)
            return Parea < Qarea;										// Lexicographic order separates areas

         if(Pfree && !Qfree)												// Unavailables to back
            return true;
         else
            return false;
      };
   };

   void AreaSort(const int a, int& b)
   {
   	if(!ValidSortRange(a,b))Throw(IDA_BadAreaSortRange);

       // SINGLETON EPILOGUE
   	if(b==a)
      	if(SortedPlayer(a)->Free())
         	return;
         else
         	Throw(IDA_BadAreaSingle);

   		Iterator bSort = Sorted.begin() + a;
      	Iterator eSort = Sorted.begin() + b + 1;

      	sort(bSort, eSort, OrderArea());

      // TRIM TO AVAILABLES
      int e = b+1;														// Exclude unavailable players

      while(e != a)
      {
      	--e;  if(SortedPlayer(e)->Free()) break;
      }
      b=e;																	// b is Last available player, or it is a

      if(!SortedPlayer(e)->Free())
      	Throw(IDA_BadAreaSortEnd);
   }
   //
   /////////////////////////////////////////////////////////////////////////////
   //
   struct OrderAreaSize:  binary_function<size_t, size_t, bool>
   {
      bool operator()(const size_t p,  const size_t q)
      {
         using Tools::ints;

         GPlayerPtr P = GoTour.Listed[p], Q = GoTour.Listed[q];

         ints PLandSize = *((ints*)(P->Spare));
         ints PClubSize = *((ints*)(P->Spare + sizeof(ints)));

         ints QLandSize = *((ints*)(Q->Spare));
         ints QClubSize = *((ints*)(Q->Spare + sizeof(ints)));


         if(PLandSize == QLandSize)
            if(PClubSize == QClubSize)
               return P->AreaCode() < Q->AreaCode();
            else
               return PClubSize > QClubSize;
         else
            return PLandSize > QLandSize;
      };
   };

   void AreaSizeSort(const int a, const int b)
   {
   	if(!ValidSortRange(a,b))Throw(IDA_BadAreaSizeRange);

      // NO SINGLETON EPILOGUE
   	if(b==a) return;

   		Iterator bSort = Sorted.begin() + a;
      	Iterator eSort = Sorted.begin() + b + 1;

      	sort(bSort, eSort, OrderAreaSize());
   }
   //
   /////////////////////////////////////////////////////////////////////////////
   // GK 6.4.0
   // A players with club code "NOC" can play any other player
   //
   const string noc=string("NOC");

   static string AlienLand, AlienClub;

   struct OrderAlienArea :  binary_function<size_t, size_t, bool>
   {
      bool operator()(const size_t p,  const size_t q)
      {
         GPlayerPtr P = GoTour.Listed[p], Q = GoTour.Listed[q];

         const bool Palienland = P->Land != AlienLand;		// So P<Q if Palienland is true and Qalienland is false
         const bool Qalienland = Q->Land != AlienLand;

         if(Palienland != Qalienland)
         	return Palienland && !Qalienland;

         const bool Pnoclub = P->Club==noc;
         const bool Qnoclub = Q->Club==noc;

         const bool Palienclub = Pnoclub || P->Club != AlienClub;
         const bool Qalienclub = Qnoclub || Q->Club != AlienClub;

         return Palienclub && !Qalienclub;
      };
   };

   bool AlienAreaSort
   (
   	const int a, int& b,
      const string alienland, const string alienclub
   )
   {
   	if(!ValidSortRange(a,b))Throw(IDA_BadAlienRange);

   	AlienLand = alienland;    AlienClub = alienclub;

      if(AlienClub==noc) return true;

      // SINGLETON
   	if(b==a)
      {
      	const GPlayerPtr aPlayer = SortedPlayer(a);
         const string aLand = aPlayer->Land, aClub = aPlayer->Club;

      	bool SameArea = aLand==alienland && aClub==alienclub;

         return !SameArea;
      }

      // SORT TO PLACE NON-ALIENS AT THE BACK
      Iterator bSort = Sorted.begin() + a;
      Iterator eSort = Sorted.begin() + b + 1;

      sort(bSort, eSort, OrderAlienArea());

      // TRIM TO ALIENS
      const GPlayerPtr bPlayer = SortedPlayer(b);

      const string bLand = bPlayer->Land, bClub = bPlayer->Club;		// Identifies same-area group

      for(int i=b; i>=a; --i)
      {
         const GPlayerPtr iPlayer = SortedPlayer(i);
      	const string iLand = iPlayer->Land, iClub = iPlayer->Club;

         if(iLand != bLand){ b = i; return true; }
         if(iClub != bClub){ b = i; return true; }
      }

      // THERE ARE NO ALIENS
      return false;
   }

   //
   /////////////////////////////////////////////////////////////////////////////
   //	GK 6.4.0 Give players with large number of same area games priority for
   // 			alien pairing.
   // We order players by decreasing number of same-area games for saForward
   // 					  by increasing for saReverse
   //
   static int sa_sortdir;

   struct OrderSameArea:  binary_function<size_t, size_t, bool>
   {
      bool operator()(const size_t p,  const size_t q)
      {
         using Tools::ints;

         GPlayerPtr P = GoTour.Listed[p], Q = GoTour.Listed[q];

         bool sameland=P->Spare[0]==Q->Spare[0], orderland, orderclub;

         switch(sa_sortdir)
         {
         	case saForward:
            {
            	orderland = P->Spare[0]>Q->Spare[0];
               orderclub = P->Spare[1]>Q->Spare[1];
            }
            break;

            case saReverse:
            {
            	orderland = P->Spare[0]<Q->Spare[0];
               orderclub = P->Spare[1]<Q->Spare[1];
            }
            break;

            default: return true;
         }

         if(orderland) return true;					// Same-land count

         if(sameland) return orderclub;			// Same-club count

         return false;
   	};
   };

   void SameAreaSort(const int sortdir, const int a, int& b)
   {
   	if(sortdir!=saForward &&  sortdir!=saReverse) return;
   	if(!ValidSortRange(a,b))Throw(IDA_BadSameAreaRange);

       // SINGLETON
   	if(b==a)
      	if(SortedPlayer(a)->Free())
         	return;
         else
         	Throw(IDA_BadSameAreaSingle);

      // SORT IN SAME-AREA ORDER
      sa_sortdir = sortdir;

      Iterator bSort = Sorted.begin() + a;
      Iterator eSort = Sorted.begin() + b + 1;

      sort(bSort, eSort, OrderSameArea());

      // TRIM TO FIRST SAME-AREA COUNT GROUP
      Byte nland = SortedPlayer(a)->Spare[0], nclub = SortedPlayer(a)->Spare[1];
      int e = a;
      while(e != b)
      {
      	++e;

         if(SortedPlayer(e)->Spare[0] != nland) break;
         if(SortedPlayer(e)->Spare[1] != nclub) break;
      }
      b=e;

      if(!SortedPlayer(e)->Free())
      	Throw(IDA_BadSameAreaEnd);
   }
}

////////////////////////////////////////////////////////////////////////////////
// GROUP CODES
//
// The Group command code sort order is  * ~ - + : space
// If two commands are equal, we sort in order of Group ID:
//     alpha, digit, none
//
namespace PairSort
{
   enum { grpAlpha, grpDigit, grpNone, grpUnknown };	// GPlayer.Group = Letter + Attribute

   struct OrderGroupCode:  binary_function<size_t, size_t, bool>
   {
      bool operator()(const size_t p,  const size_t q)
      {
         using Tools::cSpace;

         GPlayerPtr P = GoTour.Listed[p], Q = GoTour.Listed[q];

         const string &Pcode = P->Group, &Qcode = Q->Group;

         const int  Pcmd = Tours::GroupCodes.find(Pcode[1], 0);
         const int  Qcmd = Tours::GroupCodes.find(Qcode[1], 0);

         const bool Pfree = P->Free(), Qfree = Q->Free();

         if(!Pfree)return false;   			// Sleepers to back: P<Q is false
         if(!Qfree)return true;				// 						P<Q is true here

         if(Pcmd < Qcmd)return true;   	// Rcmd ordering is * ~ - + : space
         if(Pcmd > Qcmd)return false;

         // QUANTIFY THE GROUP ID CODE IN ORDER ALPHA, DIGIT, SPACE
         const char Pgrp = Pcode[0],				Qgrp = Qcode[0];

         int 					Pcat = grpUnknown, 						   Qcat = grpUnknown;
         if(isalpha(Pgrp)) Pcat = grpAlpha;		if(isalpha(Qgrp)) Qcat = grpAlpha;
         if(isdigit(Pgrp)) Pcat = grpDigit;     if(isdigit(Qgrp)) Qcat = grpDigit;
         if(Pgrp==cSpace)	Pcat = grpNone;      if(Qgrp==cSpace)	Qcat = grpNone;

         if(Pcat < Qcat)return true;
         if(Pcat > Qcat)return false;

         switch(Pcat)
         {
            case grpAlpha:
            {
               if(Pgrp<Qgrp) return true;		// Alphabetic order
               if(Pgrp>Qgrp) return false;
            }
            break;

            case grpDigit:
            {
               if(Pgrp>Qgrp) return true;		// Descending supergroup order
               if(Pgrp<Qgrp) return false;
            }
            break;

            default: ;								// No group
         }
         return P->Pid < Q->Pid;					// Registration order
      };
   };

   void GroupCodeSort(const int a, int& b)
   {
   	if(!ValidSortRange(a,b))Throw(IDA_BadGroupCodeRange);

      // NO SINGLETON EPILOGUE
      if(b==a)
      	if(SortedPlayer(a)->Free())     	// Must be free
         	return;
         else
      		Throw(IDA_BadGroupCodeSingle);

         Iterator bSort = Sorted.begin() + a;
         Iterator eSort = Sorted.begin() + b + 1;

         sort(bSort, eSort, OrderGroupCode());

		// TRIM - EXCISE ANY SLEEPERS AT END
		while(b > a)
      {
      	if(SortedPlayer(b)->Free()) break;

         --b;
      }
      if(!SortedPlayer(b)->Free())     	// The last Free player
      	Throw(IDA_BadGroupCodeEnd);
   }

   bool NextGroupCode
   (
   	const int rs, const int re,			// Range of players
      string& GroupCode,						// The code found
      int& a, int& b								// The group range:
   )                                      // rs <= a<=b <= re
   {
   	GroupCode = string();

   	if(a<0) return false;

      // SINGLETON
      if(rs==re)
      {
      	a=rs; b=re;

         GroupCode = SortedPlayer(a)->Group;

         return false;
      }

      // PREVIOUS GROUP MARKED FROM a TO b INCLUSIVE UNLESS a=0
   	if(!a) 																// Identifies first call
      	a = rs;		// Start of player range
      else
      	a = b+1;		// Start of next group

      b = a;

      if(a > re)															// No more groups within the given range
      	return false;

      // SLEEPING PLAYERS AT END TERMINATE GROUPS
      const GPlayerPtr aPlayer = SortedPlayer(a);
      	if(!aPlayer->Free())
         	return false;

      // THE ID AND CODE IS CONSTANT ACROSS THE GROUP
      GroupCode = aPlayer->Group;

      const char pgroup = GroupCode[0];
      const char pcode  = GroupCode[1];

      // FIND END OF GROUP
      int e = a + 1;

      while(e <= re)
      {
      	const GPlayerPtr ePlayer = SortedPlayer(e);

      	const string eGroup = ePlayer->Group;

         if(!ePlayer->Free()) break;

         if(eGroup[0] != pgroup)break;
         if(eGroup[1] != pcode)break;

         ++e;
      }

      b = e - 1;

      return true;
   }
   //
   /////////////////////////////////////////////////////////////////////////////
   //
   static string CompatCode;

   struct OrderCompatible:  binary_function<size_t, size_t, bool>
   {
      bool operator()(const size_t p,  const size_t q)
      {
         const bool Pcompat = GoTour.Listed[p]->Group == CompatCode;
         const bool Qcompat = GoTour.Listed[q]->Group == CompatCode;

         const bool PreferCode = CompatCode[1] == Tours::gpPrefer;

         if(PreferCode)
         	return Pcompat && !Qcompat;   // Prefer to play
         else
         	return !Pcompat && Qcompat;   // Avoid playing
      }
   };

   void CompatibleSort(const int a, int& b, const char tag, string compatcode)
   {
   	CompatCode = compatcode;

   	if(!ValidSortRange(a,b))Throw(IDA_BadCompatibleRange);

      // WE SORT IF THE GROUP ATTRIBUTE MATCHES ONE OF THE PREFER/AVOID TAGS
      const char GroupCmd = compatcode[1];

      if(GroupCmd != tag) return;

      if(tag != Tours::gpPrefer && tag != Tours::gpAvoid)
      	return;

      // NO SINGLETON EPILOGUE
      if(b==a) return;

         Iterator bSort = Sorted.begin() + a;
         Iterator eSort = Sorted.begin() + b + 1;

         sort(bSort, eSort, OrderCompatible());

      // TRIM TO REQUIRED COMPATIBLE GROUP
      for(int p=a; p<=b; p++)
      {
      	if(GroupCmd==Tours::gpPrefer)
         	if(SortedPlayer(p)->Group[1] != Tours::gpPrefer)
            	if(p>a)
               	{ b = p-1;	return; }
               else
               	return;

         if(GroupCmd==Tours::gpAvoid)
         	if(SortedPlayer(p)->Group[1] == Tours::gpAvoid)
            	if(p>a)
               	{ b = p-1;	return; }
               else
               	return;
      }
   }
}

////////////////////////////////////////////////////////////////////////////////
// POOL SORT
//
// Increasing order of Player.Pool.  Previous opponents and 'Must Not' play
// have already been removed from the players Pool.
//
// We want the set of players with the lowest pool
//
namespace PairSort
{
   struct OrderPool:  binary_function<size_t, size_t, bool>
   {
      bool operator()(const size_t p,  const size_t q)
      {
         return GoTour.Listed[p]->Pool <  GoTour.Listed[q]->Pool;
      }
   };

   void PoolSort(const int a, int& b)
   {
   	if(!ValidSortRange(a,b))Throw(IDA_BadPoolRange);

      // NO SINGLETON EPILOGUE
      if(b==a) return;

         Iterator bSort = Sorted.begin() + a;
         Iterator eSort = Sorted.begin() + b + 1;

         sort(bSort, eSort, OrderPool());

      // TRIM TO SMALLEST POOL
      const int aPool = SortedPlayer(a)->Pool;

      for(int p=a+1; p<=b; p++)
         if(SortedPlayer(p)->Pool != aPool)
         {
            b = p-1;	break;
         }
   }
}

///////////////////////////////////////////////////////////////////////////////
// MIB SORT
//
// The sort depends on the 'rule' parameter.  It identifies the player's
// DESIRED direction of draw.
//
// mrOut:	increasing .Mout
//          If played even a lot, VOLUNTEER to play out
//
// mrUp:		decreasing .DnWin and increasing .Mbal
//   			If played down and won a lot, but not playing up to compensate,
//				you get a large negative Mbal, and so you WISH to play up
//
// mrDown:	decreasing .UpLose and decreasing .Mbal
//				If played up and lost a lot, but not playing down to compensate,
//				you get a large positive Mbal, so you WISH to play down
//
// COMMENT
// If you play down and win then play up and lose, your Mbal will be zero
// so you have less chance of volunteering to play either up or down
//
namespace PairSort
{
	static int MibRule;

   struct OrderMib:  binary_function<size_t, size_t, bool>
   {
      bool operator()(const size_t p,  const size_t q)
      {
      	GPlayerPtr P = GoTour.Listed[p], Q = GoTour.Listed[q];

      	switch(MibRule)
         {
         	case mrOut:  return P->Mout < Q->Mout;

            case mrUp:   return P->DnWin  > Q->DnWin  && P->Mbal < Q->Mbal;
            case mrDown: return P->UpLose > Q->UpLose && P->Mbal > Q->Mbal;

            default:		 return P->Mbal < Q->Mbal;
         }
      }
   };

   int MibSort(const int a, int& b, const int rule)
   {
   	if(!ValidSortRange(a,b))Throw(IDA_BadMibRange);

      if(rule<mrNone || rule>mrOut) Throw(IDA_BadMibRule);

      if(rule==mrNone) return 0;
      
      // SINGLETON
      if(b==a)
      	return SortedPlayer(a)->Mbal;

      MibRule = rule;

         Iterator bSort = Sorted.begin() + a;
         Iterator eSort = Sorted.begin() + b + 1;

         sort(bSort, eSort, OrderMib());

      // FIND THE FIRST p ST a < p UNDER MibOrder
      // IF THERE WAS NO SUCH p, THEN b IS LEFT INVARIANT
      OrderMib MibOrder;

      for(int p=a+1; p<=b; p++)
      	if(MibOrder(Sorted[a], Sorted[p]))
         {
         	b = p-1; 	break;
         }

      return SortedPlayer(a)->Mbal;
   }
}

///////////////////////////////////////////////////////////////////////////////
// CLASS SORT
//
// The sort depends on the 'rule' parameter.  It identifies the player's
// ABILITY to play in a particular direction.
//
// mrUp:		Class pcUp or pcOut first, then pcDown or pcNone
// mrDown:  Class pcDown or pcOut first, then pcUp or pcNone
//
namespace PairSort
{
	static int ClassRule;

   struct OrderClass:  binary_function<size_t, size_t, bool>
   {
      bool operator()(const size_t p,  const size_t q)
      {
         int pClass = GoTour.Listed[p]->Class, qClass = GoTour.Listed[q]->Class;

         // IF A PLAYER CLASS IS mrOUT THEN THE RULE IS SATISFIED
         if(pClass == mrOut) pClass = ClassRule;
         if(qClass == mrOut) qClass = ClassRule;

         return pClass==ClassRule && qClass!=ClassRule;
      }
   };

   bool ClassSort(const int a, int& b, int rule)
   {
   	if(!ValidSortRange(a,b))Throw(IDA_BadClassRange);
      if(rule<mrUp || rule >mrDown) Throw(IDA_BadClassRule);

      // SINGLETON EPILOGUE
      int aClass = SortedPlayer(a)->Class;
      if(aClass == mrOut) aClass = rule;

      if(b==a)
      	return aClass == rule;

      ClassRule = rule;

         Iterator bSort = Sorted.begin() + a;
         Iterator eSort = Sorted.begin() + b + 1;

         sort(bSort, eSort, OrderClass());

   	// TRIM TO FIRST CLASS, LEAVING b INVARIANT IF THE RANGE IS ALL ONE CLASS
      OrderClass ClassOrder;

      aClass = SortedPlayer(a)->Class;
      if(aClass == mrOut) aClass = rule;

      for(int p=a+1; p<=b; p++)
      	if(ClassOrder(Sorted[a], Sorted[p]))
         {
         	b = p-1;		break;
         }

      // EPILOGUE
      return aClass == rule;
	}
}

///////////////////////////////////////////////////////////////////////////////
// CIB SORT
//
// The sort depends on the 'rule' parameter.  It identifies the players
// who HAVE to played a particular colour a lot.  Colour imbalance is defined as
// White games - Black games
//
// crBlack:	increasing Cib
// crWhite:	decreasing Cib
// crGrey:	increasing |Cib| but with priority to Cib <0
//
namespace PairSort
{
	using Tools::modulus;
   
	static int CibRule;

   struct OrderCib:  binary_function<size_t, size_t, bool>
   {
      bool operator()(const size_t p,  const size_t q)
      {
         int pCib = GoTour.Listed[p]->Cbal, qCib = GoTour.Listed[q]->Cbal;

         switch(CibRule)
         {
         	case crBlack: return pCib < qCib;
            case crWhite: return pCib > qCib;

            default:
            {
            	int mpCib = modulus(pCib), mqCib = modulus(qCib);

               if(mpCib == mqCib)
               	return pCib<qCib;			// Increasing Cib to give Black priority
               else
               	return mpCib > mqCib;	// Decreasing |Cib| to reduce imbalance
            }
         }
      }
   };


   void CibSort(const int a, int& b, const int rule)
   {
   	if(!ValidSortRange(a,b))Throw(IDA_BadCibRange);

      // NO SINGLETON EPILOGUE
      if(b==a) return;

      CibRule = rule;		if(rule<crGrey || rule >crWhite) Throw(IDA_BadCibRule);

         Iterator bSort = Sorted.begin() + a;
         Iterator eSort = Sorted.begin() + b + 1;

         sort(bSort, eSort, OrderCib());

   	// TRIM TO THE FIRST SET SATISFYING THE RULE
      OrderCib CibOrder;

      for(int p=a+1; p<=b; p++)
      	if(CibOrder(Sorted[a],Sorted[b]))
         {
         	b = p-1;		break;
         }
   }
}


////////////////////////////////////////////////////////////////////////////////
// BUSY SORT
//
// The ordering is:
// Potential opponents of player, then unplayables
// The field Spare[0] is set to 0 if player available for pairing ,
//                and is set to 1 if player is not available in
// the routine MarkOpponents
//
namespace PairSort
{
	struct OrderBusy:  binary_function<size_t, size_t, bool>
   {
      bool operator()(const size_t p,  const size_t q)
   	{
      	const bool pBusy = GoTour.Listed[p]->Spare[0];
       	const bool qBusy = GoTour.Listed[q]->Spare[0];

      	return !pBusy && qBusy;
   	}
   };

	void BusySort(const int a, int& b)
   {
   	if(!ValidSortRange(a,b))Throw(IDA_BadBusyRange);

      // NO SINGLETON EPILOGUE
      if(b==a) return;

      	Iterator bSort = Sorted.begin() + a;
      	Iterator eSort = Sorted.begin() + b + 1;

     		sort(bSort, eSort, OrderBusy());

      // TRIM TO FREE GROUP
      for(int p=a+1; p<=b; p++)
      	if(SortedPlayer(p)->Spare[0])
         	{ b = p-1;  break; }
   }
}


////////////////////////////////////////////////////////////////////////////////
// PURE SORT INTERFACE - NO POST PROCESSING
//
namespace PairSort
{
	struct OrderPid:  binary_function<size_t, size_t, bool>
   {
      bool operator()(const size_t p,  const size_t q)
   	{
      	return GoTour.Listed[p]->Pid < GoTour.Listed[q]->Pid;
   	}
   };

	void PlayerIdSort(const int a, const int b)
   {
   	if(!ValidSortRange(a,b))Throw(IDA_BadPlayerIdRange);

      if(b==a) return;

      	Iterator bSort = Sorted.begin() + a;
      	Iterator eSort = Sorted.begin() + b + 1;

     		sort(bSort, eSort, OrderPid());
   }
}

////////////////////////////////////////////////////////////////////////////////
// DEVELOPMENT
//
namespace Tools
{
	extern std::string to_string(const int value);
}

namespace PairSort
{
	using Tools::to_string;
   
   void ScanSorted(const int a, const int b)
   {
   	int Round = GoTour.Round;

   	for(int p=a; p<=b; p++)
      {
      	int Pause=0, McScore, State,Pid,McGroup, Pool;
         int Mbal,Cbal, DnWin,UpLose,Mout, Class;
         int Opp;

         std::string Name, Club, Land, Group;
         std::string Grade,Spare0,Spare1;

      	Tours::GPlayerPtr Player = SortedPlayer(p);

         Name = Player->Name;  Club = Player->Club;  Land = Player->Land;
         Group = Player->Group;

         Grade = to_string(Player->Grade);
         Spare0 = to_string(Player->Spare[0]);
         Spare1 = to_string(Player->Spare[1]);

         Pid = Player->Pid;
         McScore = Player->McScore;
         State = (int)Player->State;
         McGroup = Player->McGroup;
         Pool = Player->Pool;
         Class = Player->Class;
         Mout = Player->Mout;  DnWin = Player->DnWin;  UpLose = Player->UpLose;
         Mbal = Player->Mbal;		Cbal = Player->Cbal;

         string Line = Name+Grade+Club+Spare0+Spare1;

         //har Line[80];
         //sprintf(Line, "%s  %s  %s  [%2d] [%2d] ...", Name.c_str(), Club.c_str(),Spare0, Spare1);

         Opp = Player->Games[Round-1].Oppid;

         ++Pause;
      }
   }
}
//
////////////////////////////////////////////////////////////////////////////////





