////////////////////////////////////////////////////////////////////////////////
// MODULE PairTool: Pure library functions for the pairing G.Kaniuk 10/10/04
////////////////////////////////////////////////////////////////////////////////

#include <vector>
#include <algorithm>
#include <stdlib.h>
#pragma hdrstop

#include "PalgRES.h"

#include "TourMaster.h"
#include "PairMaster.h"
#include "PairSort.h"
#include "Sort.h"
#include "PairTool.h"
#include "Pmatch.h"

////////////////////////////////////////////////////////////////////////////////
// CHOOSE
//
namespace PairTool
{
	using Tours::GPlayerPtr;

   using TourMaster::GoTour;

   using PairSort::SortedPlayer;
   
   int Choose(int Size)
   {
   	if(Size <=1 ) return 0;

      return (int)(_lrand()%(long) Size );	// Guaranteed in [0, Size-1]
   }

   int Choose(int a, int b)
   {
   	if(b <= a) return a;

      long Size = b - a + 1;

      return a + (int)(_lrand()%Size);			// Guaranteed in [a, b]
   }

   int ChooseExcluding(int One, int a, int b)
   {
   	if(a==b && a==One) return a;				// Give up

   	if(One<a || One>b) return Choose(a,b);	// One is outside [a, b]

      if(One == a) return Choose(a+1,b);
      if(One == b) return Choose(a,b-1);

   	int Chose = Choose(b-a);					// Choose from 1 fewer

      if(a+Chose < One)
      	return a + Chose;							// Left of One
      else
      	return a + Chose + 1;					// Right of One
   }

   int ChooseExcluding(int a, int b, int e, int f)
   {
   	// We choose from [a,b] excluding [e,f]
   	// We assume [e,f] is a subinterval of [a,b]

      if(e==a && f==b) return Choose(a,b);	// Give up

      if(e==a && f<b) return Choose(f+1,b);	// Excluded starts at a
      if(f==b && a<e) return Choose(a,e-1);	// Excluded ends at f

      int L = e-a, R = b-f;						// Sizes of remaining sub intervals

      int S = Choose(L+R);							// Choose 0<=S<L+R from union

      if(S<L)											// Chose in left
      	return a+S;
      else												// Chose right
      	return f+1+S-L;
   }

   int ChooseExcluding(string& land, string& club, int a, int b)
   {
   	// Assumes [a,b] sorted by land,club

      if(a==b) return a;							// No choice

      int e = a-1, f = a-1;

      for(int  p=a; p<=b; p++)
      {
      	Tours::GPlayerPtr Player = SortedPlayer(p);

      	if(Player->Land==land && Player->Club==club)
         	if(e<a)
            	e = f = p;							// Given Land,Club detected
            else
            	f = p;								// Inflate end of area range
      }

      if(e<a)											// There is no exclusion
      	return Choose(a,b);
      else												// Land,Club occupies range [e,f]
      	return ChooseExcluding(a,b, e,f);
   }
}


////////////////////////////////////////////////////////////////////////////////
// HOUSEKEEPING
//
namespace PairTool
{
   void SetUpDnCounts(void)
   {
      for(size_t p = 1; p<=GoTour.Players; p++)
      {
      	GPlayerPtr Player = GoTour.Listed[p];

         Player->DnWin = 0;	Player->UpLose = 0;

         for(size_t r=0; r<GoTour.Round; r++)
         {
         	const Tours::GGame& Game = Player->Games[r];

            if(Game.Parity==Tours::gdDown && Game.Result==Tours::grWin)
            	++Player->DnWin;

            if(Game.Parity==Tours::gdUp && Game.Result==Tours::grLose)
            	++Player->UpLose;
         }
      }
   }
}

////////////////////////////////////////////////////////////////////////////////
// McMAHON GROUPS
//
namespace PairTool
{
	using std::vector;
   using std::list;
   using Tours::GTour;
   using PairSort::Iterator;

   void ScanGroups(MMGroups& Groups)
   {
   	int Pause = 0;

      int Score, Start, Size, Pool, Class, Work;

   	for(size_t g=0; g<Groups.size(); g++)
      {
      	PairTool::MMGroup& Group = Groups[g];

         Score = Group.Score;
         Start = Group.Start;
         Size  = Group.Size;
         Pool  = Group.Pool;
         Work  = Group.Work;
         Class = Group.Class;

         ++Pause;
      }
      
      Pause = Score+Start+Size+Pool+Class+Work;
   }

	void FormMMGroups(MMGroups& McGroups)
   {
      McGroups.clear();

      int Players = GoTour.Players;

      // SORT FROM 1 TO  PLAYERS IN McCMAHON ORDER
      if(PairMaster::GetTrials(GoTour.PairParam) == 1)
      {
      	Iterator bSorted = PairSort::Sorted.begin() + 1;
      	Iterator eSorted = PairSort::Sorted.end();

         std::random_shuffle(bSorted, eSorted);
      }

		PairSort::McMahonSort(1, Players);					// Unavailable players go to back


      // FORM GROUPS ON A LIST
      int p = 1;													// Start of top group

      std::list<MMGroup>McList;

      MMGroup McGroup(SortedPlayer(1)->McScore, 1, 0);

      while(p <= Players)
      {
         ints McScore = SortedPlayer(p)->McScore;

      	if( McScore < McGroup.Score)						// Triggers change of group
         {
         	McGroup.Pool = McGroup.Size;					// All players available

         	McList.push_back(McGroup);						// Save the old

            McGroup = MMGroup(McScore, p, 0);			// Start the new
         }

         ++McGroup.Size;

         ++p;

         bool Complete = false;

         if(p > Players)
         	Complete = true;									// Last player completes the last group
         else
         	Complete = !SortedPlayer(p)->Free();		// Last Free player completes the last group

         if(Complete)
			{
         	McGroup.Pool = McGroup.Size;					// All players available

         	McList.push_back(McGroup);
         }
      }

      // ATTACH EACH PLAYER TO THEIR GROUP
      VectorLoad<MMGroup>(McList, McGroups);

      for(size_t g=0; g<McGroups.size(); g++)
      {
      	MMGroup& Group = McGroups[g];
         
         const int a = Group.Start, b = Group.Start + Group.Size - 1;

         Group.Komip = 0;

         for(int p=a; p<=b; p++)
         {
         	GPlayerPtr Player = SortedPlayer(p);

         	Player->McGroup = g;

            if(Player->Grade <= GoTour.Rules.KomiGroup)
            	++Group.Komip;
         }
      }
   }
}

////////////////////////////////////////////////////////////////////////////////
// AREA GROUPS
//
namespace PairTool
{
	static void MarkAreaSize(ARLand Group)
   {
   	for(int p = Group.Start; p <  Group.Start + Group.Size; p++)
      {
      	GPlayerPtr P = SortedPlayer(p);

         *((ints*)(P->Spare)) = (ints)Group.Size;						// Bytes 0 and 1
      }
   }

   static void MarkAreaSize(ARClub Group)
   {
   	for(int p = Group.Start; p <  Group.Start + Group.Size; p++)
      {
      	GPlayerPtr P = SortedPlayer(p);

         *((ints*)(P->Spare + sizeof(ints))) = (ints)Group.Size;	// Bytes 2 and 3
      }
   }
   void FormAreas(int a, int& b, ARLands& Areas)
   {
   	if(a==b) return;
      
   	Areas.Lands.clear();

   	// SORT BY COUNTRY/CLUB TRIMMING b TO FREE PLAYERS
   	PairSort::AreaSort(a, b);

      // MARK EACH PLAYER WITH THEIR LAND/CLUB GROUP SIZES
      int p=a;

      string LandCode = SortedPlayer(a)->Land;
      string ClubCode = SortedPlayer(a)->Club;

      ARLand LandArea(LandCode, a);
      ARClub ClubArea(ClubCode, a);

      while(p <= b)
      {
      	LandCode = SortedPlayer(p)->Land;
         ClubCode = SortedPlayer(p)->Club;

         if(LandCode != LandArea.Land)				// Current LandArea now complete
         {
         	MarkAreaSize(LandArea);
            MarkAreaSize(ClubArea);

            LandArea = ARLand(LandCode, p);		// Start new LandArea
            ClubArea = ARClub(ClubCode, p);  	// Start first club in new LandArea
         }
         ++LandArea.Size;

         if(ClubCode != ClubArea.Club)				// Current ClubArea now complete
         {
         	MarkAreaSize(ClubArea);

            ClubArea = ARClub(ClubCode, p);		// Start Next club in LandArea
         }
         ++ClubArea.Size;


         ++p;

         if(p>b)
         {
           MarkAreaSize(LandArea);
           MarkAreaSize(ClubArea);
         }
      }

      // SORT THE GROUP IN DECREASING ORDER OF AREA SIZE
      PairSort::AreaSizeSort(a,b);

      // FORM CLUB/COUNTRY GROUPS ON A LIST
      std::list<ARLand>LandList;
      std::list<ARClub>ClubList;

      LandCode = SortedPlayer(a)->Land;
      ClubCode = SortedPlayer(a)->Club;

      LandArea = ARLand(LandCode, a);
      ClubArea = ARClub(ClubCode, a);

      p=a;

      while(p <= b)
      {
      	LandCode = SortedPlayer(p)->Land;
      	ClubCode = SortedPlayer(p)->Club;

         if(LandCode != LandArea.Land)				// This completes LandArea
         {
         	ClubList.push_back(ClubArea);			// Complete the ClubList
         	LandArea.Load(ClubList);				// Load the Clubs vector from ClubList
         	LandList.push_back(LandArea);		   // List completed land with clubs

            LandArea = ARLand(LandCode, p);  	// Start new LandArea
            ClubArea = ARClub(ClubCode, p);  	// Start new ClubArea
            ClubList.clear();							// Start new ClubList
         }
         ++LandArea.Size;

         if(ClubCode != ClubArea.Club)
         {
         	ClubList.push_back(ClubArea);

            ClubArea = ARClub(ClubCode, p);
         }
         ++ClubArea.Size;

         ++p;

         if(p>b)
         {
         	ClubList.push_back(ClubArea);			// Complete the ClubList
         	LandArea.Load(ClubList);				// Vectorise
         	LandList.push_back(LandArea);		   // List completed land with clubs
         }
      }
      Areas.Load(LandList);							// Copy LandList to Lands vector
   }

   void FormMaximumArea(int& a, int& b)
   {
   	if(a==b) return;

   	ARLands Region;	FormAreas(a, b, Region);

      if(!Region.SizeMax || !Region.Lands.size())
      	PairSort::Throw(IDA_BadFormMaximumArea);

      if(Region.Lands.size()>1)						// All the equally largest countries
      {
      	a = Region.Start;
      	b = Region.Start + Region.SizeMax - 1;

      	return;
      }
      else													// All the equally largest clubs in the one country
      {
      	a = Region.Start;
      	b = Region.Start + Region.Lands[0].SizeMax - 1;

         return;
      }
   }

   void FormAlienArea(const string Land, const string Club, int& a, int& b)
   {
   	if(a==b) return;

      if(PairSort::AlienAreaSort(a,b, Land, Club)) return;					// Each group here has no players in (Land,Club)

      FormSameAreaCount(PairSort::saReverse, a, b);							// No aliens
   }

   void FormSameAreaCount(const int dir, const int a, int& b)			  // GK 6.4.0 Count same area games
   {
      if(GoTour.Roundl==1) return;

      const int MacBar = GoTour.Rules.MacBar;

      int maxnland=0, maxnclub=0;

   	for(int p=a; p<=b; p++)														// We first count the number of same-area games
      {
      	Tours::GPlayerPtr Player = SortedPlayer(p);

         const string &pLand(Player->Land), &pClub(Player->Club);
         Byte &nland(Player->Spare[0]), &nclub(Player->Spare[1]);

         nland = nclub = 0;

         if(Player->MiScore>=MacBar)											// Player is above the bar
         	if(!GoTour.Rules.BarSeg)											// But we do not segregate above the bar
            	continue;

         if(pClub==PairSort::noc)												// Players with no club play everybody
         	continue;

         for(size_t r=0; r<GoTour.Roundl; r++)
         {
            const size_t pOpp = Player->Games[r].Oppid;

            if(pOpp)
            {
               const GPlayerPtr Opp = PairSort::Player(pOpp);

               const string &oLand(Opp->Land), &oClub(Opp->Club);

               if(oLand==pLand)
               {
                	nland += nland<255 ? 1 : 0;								// Same-land game

                  if(oClub==pClub) nclub += nclub<255 ? 1 : 0;			// Same-club game

                  if(nclub>maxnclub) maxnclub=nclub;
               }

               if(nland>maxnland) maxnland=nland;
            }
         }
      }

   	if(maxnland==0) return;														// All games were between different countries
      if(maxnclub>0)																	// Do nothing if there are no same-club games
      	PairSort::SameAreaSort(dir, a,b);
   }
}

////////////////////////////////////////////////////////////////////////////////
// FILTERS
//
namespace PairTool
{
   void PurgeOpponents(const int p, MMGroups& McGroups)
   {
   	// INITIALISE THE WORK POOLS TO THE NUMBER OF FREE PLAYERS IN EACH GROUP
      const int McCount = McGroups.size();

      for(int g=0; g<McCount; g++)
      	McGroups[g].Work = McGroups[g].Pool;

      // WE PURGE OLD OPPONENTS WHO ARE FREE
      const GPlayerPtr Player = PairSort::Player(p);

      for(size_t r=0; r<GoTour.Round; r++)
      {
      	const size_t pOpp = Player->Games[r].Oppid;

         if(pOpp)
         {
         	const GPlayerPtr Opp = PairSort::Player(pOpp);

            if(Opp->Free())
            	--McGroups[Opp->McGroup].Work;
         }
      }

      // PURGE PLAYERS IN THE SAME 'MUST NOT PLAY' SET
      const std::string& pGroupCode = Player->Group;

      if(pGroupCode[1] == Tours::gpNever)
      	for(size_t q=1; q<=GoTour.Players; q++)
      	{
      		const GPlayerPtr Fam = PairSort::Player(q);

            if(Fam->Group == pGroupCode)
            	if(McGroups[Fam->McGroup].Work>0)
            		--McGroups[Fam->McGroup].Work;
         }
   }
	//
   /////////////////////////////////////////////////////////////////////////////
   //
   void MarkOpponents(const int pOne, const int gTwo, const MMGroups& McGroups)
   {
   	// THE PAIRING MARKS
   	const Byte pmFree = 0, pmBusy = 1;

   	// THE CHOSEN PLAYER
   	const GPlayerPtr PlayerOne = PairSort::Player(pOne);

      const PGames& Games = PlayerOne->Games;

      // THE CHOSEN GROUP
   	const MMGroup& GroupTwo = McGroups[gTwo];

      const int pa =  GroupTwo.Start, pb = GroupTwo.Start + GroupTwo.Size - 1;

      // ANY PAIRED PLAYERS IN GROUP TWO ARE BUSY, THE REST ARE FREE
      for(int p=pa; p<=pb; p++)
      {
      	GPlayerPtr Player = PairSort::SortedPlayer(p);

         if(Player->State == Tours::psChosen)
         	Player->Spare[0] = pmBusy;
         else
         	Player->Spare[0] = pmFree;
      }

      // PREVIOUS OPPONENTS OF ONE IN GROUP TWO ARE NOT FREE
      for(size_t r=0; r<GoTour.Round; r++)
      {
      	const size_t pOpp = Games[r].Oppid;

         if(pOpp)
         {
         	GPlayerPtr OppOne = PairSort::Player(pOpp);

            if(OppOne->McGroup == gTwo && OppOne->Free())
            	OppOne->Spare[0] = pmBusy;
         }
      }

      // PLAYERS IN THE SAME 'MUST NOT PLAY' SET ARE NOT FREE
      const std::string& OneGroupCode = PlayerOne->Group;

      if(OneGroupCode[1] == Tours::gpNever)
         for(int p = pa; p<=pb; p++)
            if(SortedPlayer(p)->Group == OneGroupCode)
               SortedPlayer(p)->Spare[0] = pmBusy;
   }
	//
   /////////////////////////////////////////////////////////////////////////////
   //
   bool ExistsKomiOpponent(const int pkomi, const MMGroup& McGroup)
   {
   	if(McGroup.Work==0) return false;	// No opponents for pkomi in this group
      if(McGroup.Komip==0) return false;	// No komi opponents for pkomi in this group

   	// WE HAVE A CHOSEN KOMI PLAYER
   	GPlayerPtr KomiPlayer = PairSort::Player(pkomi);

      const string& KomiGroupCode = KomiPlayer->Group;

      const bool KomiMustNot = KomiGroupCode[1] == Tours::gpNever;

      const string &KomiClub = KomiPlayer->Club, &KomiLand = KomiPlayer->Land;

      // FIND A FREE OPPONENT WHO IS ALSO A KOMI PLAYER
      const int a = McGroup.Start, b = McGroup.Start + McGroup.Size - 1;

      for(int p=a; p<=b; p++)
      	if(PairSort::Sorted[p] != pkomi)
         {
         	const GPlayerPtr Player = PairSort::SortedPlayer(p);

            // EXCLUDE ANY 'MUST NOT PLAY'
            bool SeePlayer = true;

            if(KomiMustNot && Player->Group == KomiGroupCode)
            	SeePlayer = false;

            // PLAYER MUST BE BELOW KOMI BAR
            if(SeePlayer && Player->Grade <= GoTour.Rules.KomiGroup)
            {
            	// EXCLUDE PREVIOUS OPPONENTS
               bool PlayedBefore = false;

            	for(size_t r=0; r<GoTour.Round; r++)
            		if(Player->Games[r].Oppid == (size_t)pkomi)
                  {  PlayedBefore = true;  break; }

               // FOUND A PLAYER - WANT ALIEN REGION
               if(!PlayedBefore)
               {
               	const string &Club = Player->Club, &Land = Player->Land;

                  if(Land != KomiLand) return true;
                  if(Club != KomiClub) return true;
               }
            }
         }
		return false;
   }
}

////////////////////////////////////////////////////////////////////////////////
// POLISH
//
// The draw is polished just before AllocateBoards by examining each group of 3
// boards to try to minimise the square mms defference.  Manually paired players
// are left untouched as are any must plays.  Players below the lower bar are not
// polished.
//
namespace PairTool
{
   using Tours::psChosen;
   using Tours::psPlaying;
   using Tours::psSleep;

   GPlayerPtr ranked_player(const int r)
   {
   	return GoTour.Listed[GoTour.Sorted[r]];
   }

   GPlayerPtr listed_player(const int p)
   {
   	return GoTour.Listed[p];
   }

   bool fixed_pair(const int one, const int two)
   {
   	if(one<=0 || two<=0)
      	return true;

      const GPlayerPtr One = listed_player(one);	// GK 6.3.9 , Two = listed_player(two);

      // MANUALLY PAIRED PLAYERS ARE LEFT ALONE
      if(One->State==Tours::psPlaying)					  // Two has the same State
      	return true;

      // DO NOT VIOLATE must play RESTRICTION
      if(One->Group[1]==Tours::gpAlways)				  // Two has same group code
      	return true;

      return false;
    }

   bool eligible_pair(const int one, const int two)
   {
      using Tours::gpNever;
      using PairMaster::PlayedBefore;

   	if(one<=0 || two<=0)
      	return false;

      const GPlayerPtr One = listed_player(one), Two = listed_player(two);

      // DO NOT VIOLATE musn't play RESTRICTION
      const char onecmd = One->Group[1], twocmd = Two->Group[1];

      if(onecmd==gpNever && twocmd==gpNever)
      	return One->Group[0]!= Two->Group[0];

      // PLAYERS HAVE NOT PLAYED IN A ROUND PRECEEDING CURRENT
      return !PlayedBefore(one, two, GoTour.Roundl-1);
   }
   
   int sqsum_dmms(const vector<int>& match)
   {
   	const int nmatch = match.size()/2;  if(nmatch<2) return -1;

   	int sqsum = 0;

      for(int e=0; e<nmatch; e++)
      {
      	const int one=match[2*e], two=match[2*e+1];

         if(eligible_pair(one, two))
         {
            const GPlayerPtr One=listed_player(one), Two=listed_player(two);

            const ints mmsdiff = One->McScore - Two->McScore;

            sqsum += mmsdiff*mmsdiff;
         }
         else
         	return -1;
      }
      return sqsum;
   }

   void unhook_player(const int p)
   {
   	// THIS IS A CUT DOWN VERSION OF UnpairPlayer
      // THERE IS NO CHANGE OF PLAYER State
      using Tours::gdEven; using Tours::gcNone; using Tours::grNone;

      const int Round = GoTour.Round;

      Tours::GPlayerPtr One = listed_player(p);
      Tours::GGame&    OneG = One->Games[Round-1];

      // OPPONENT
      OneG.Oppid = OneG.Board = 0;

      // PARITY
      One->Mbal -= OneG.Parity; One->Mout -= abs(OneG.Parity);
      OneG.Parity = gdEven;

      // COLOUR
      One->Cbal -= OneG.Colour;
      OneG.Colour = gcNone;

      // HANDICAPS
      OneG.Stones = 0;	OneG.Komi = 0;
   }

   int ImprovePairing(void)
   {
      using Tours::GBoard;
      using Tours::mc_scale;

      using PairMaster::PairPlayers;
      using Pmatch::PMatch;

      enum { poFree, poPaired, poSleep };

      // PASS 1 - WE LAY PLAYERS ON BOARDS IN RANK ORDER
      const int Pcount = GoTour.PlayNowCount;		if(Pcount<4) return 0;
      const int Active = Tours::RankSort(&GoTour); if(Active<Pcount) return 0;
      const int Bcount = Pcount/2;

      GoTour.Pair.resize(Bcount);

      int p=1, b=0;									// Index players, boards

      for(p=1; p<=Active; p++)
      {
      	GPlayerPtr Player = ranked_player(p);

         Player->Spare[0] = poFree;				// All active players are free
      }

      // LAY ALL FREE AWAKE PLAYERS ON BOARDS WITH
      // OPPONENTS IN RANK ORDER
      p=1;	b=0;

      do
      {
      	// SEARCH FOR NEXT FREE PLAYER
      	while(ranked_player(p)->Spare[0] && p<=Active)
         	++p;

         // GET THE PLAYER PAIR
         const int  one = GoTour.Sorted[p];
         GPlayerPtr One = listed_player(one);

         if(One->State==psSleep)
         {
         	One->Spare[0] = poSleep;

            ++p;	continue;
         }

         const int  two = One->Games[GoTour.Round-1].Oppid;
         GPlayerPtr Two = listed_player(two);

         // LAY PLAYERS ON THE BOARD
         GoTour.Pair[b] = GBoard(one,two);

         One->Spare[0] = poPaired; 					// Mark as paired
         Two->Spare[0] = poPaired;

         ++b;  ++p;
      }
      while(b<Bcount);

      if(b != Bcount)
      	return 1;

      // PASS 2 - TRY TO MINIMISE THE SUM OF SQUARE MMS DIFFERENCES
      //          FOR EACH GROUP OF THREE BOARDS NEAR THE BAR
      const int bargroup  = GoTour.Rules.MacBar - GoTour.Rules.Rounds*mc_scale;
      const int komigroup = GoTour.Rules.KomiGroup;
      const int topgroup  = komigroup<bargroup ? bargroup : komigroup;

      b=0;

      do
      {
      	// SET UP THE CURRENT MATCHING FOR 6 PLAYERS
         int barcount = 0;

         vector<int>matcho(0);

      	for(int e=0; e<3; e++)
            if(b+e<=Bcount-1)
            {
               const GBoard& eboard = GoTour.Pair[b+e];

               const int black = eboard.Black, white = eboard.White;

               const GPlayerPtr Black = listed_player(black);
               const GPlayerPtr White = listed_player(white);

               if(Black->Grade>=topgroup) ++barcount;
               if(White->Grade>=topgroup) ++barcount;

               if(!fixed_pair(black,white))
               {
               	matcho.push_back(black);  matcho.push_back(white);
               }
            }
         if(barcount<2){ b+=3; continue; }			// Too few above the bar

         const int sqsummso = sqsum_dmms(matcho);
         if(sqsummso <= 0) { b+=3; continue; }		// There is no improvement

         // SCAN ALL PERFECT MATCHINGS
         PMatch perfect(matcho.size());

         vector<int>match(matcho), matchi(matcho);

         int sqsummsi = sqsummso;

         bool improved = false;

         while(perfect.next_matching(match))
         {
         	const int sqsumms = sqsum_dmms(match);

            if(-1<sqsumms && sqsumms<sqsummsi)
            {
            	sqsummsi = sqsumms;	matchi = match;

               improved = true;
            }
         }

         // RE-PAIR
         if(improved)
         {
         	// UNHOOK ALL THE PLAYERS IN THE OLD MATCHING
            const int nmatch = matcho.size()/2;

            for(int e=0; e<nmatch && b+e<Bcount-1; e++)
            {
               const int one = matcho[2*e], two = matcho[2*e+1];

               if(eligible_pair(one,two))
               {
               	unhook_player(one);
                  unhook_player(two);
               }
               else
               	return 2;
            }

            // PAIR ALL THE PLAYERS IN THE IMPROVED MATCHING
            for(int e=0; e<nmatch && b+e<Bcount-1; e++)
            {
            	const int one = matchi[2*e], two = matchi[2*e+1];

               if(eligible_pair(one,two))
            		PairPlayers(one,two);
               else
               	return 3;
            }
         }
         b += 3;												// Next 3 boards
      }
      while(b<Bcount-1);									// Need at least 2 boards
      
      return 0;
   }
}


////////////////////////////////////////////////////////////////////////////////
// DEVELOPMENT
//
namespace PairTool
{
	using PairSort::ScanSorted;
   using PairSort::mrNone;
   using PairSort::mrUp;
   using PairSort::mrDown;
   using PairSort::mrOut;

   using PairSort::crGrey;
   using PairSort::crBlack;
   using PairSort::crWhite;

   using PairSort::ClassSort;
   using PairSort::MibSort;
   using PairSort::CibSort;
   using PairSort::KomiSort;
   using PairSort::CompatibleSort;
   using PairSort::AlienAreaSort;
   using PairSort::GroupCodeSort;
   using PairSort::PoolSort;
   using PairSort::BusySort;

   void ApproveSort(const int ao,const int bo)
   {
   	int a = ao, b = bo;

   	int rule = mrNone;

      string sUK("UK"), sXXX("XXX"), sCam("CAM");

      a = ao;  b= bo; rule = mrUp;   ClassSort(a,b, rule);  ScanSorted(ao,bo);
      a = ao;  b= bo; rule = mrDown; ClassSort(a,b, rule);  ScanSorted(ao,bo);
 //   a = ao;  b= bo; rule = mrOut;  ClassSort(a,b, rule);  ScanSorted(ao,bo);

      a = ao;  b= bo; rule = mrUp;   MibSort(a,b, rule);  ScanSorted(ao,bo);
      a = ao;  b= bo; rule = mrDown; MibSort(a,b, rule);  ScanSorted(ao,bo);
      a = ao;  b= bo; rule = mrOut;  MibSort(a,b, rule);	ScanSorted(ao,bo);

      a = ao;  b= bo; rule = crBlack; CibSort(a,b, rule); ScanSorted(ao,bo);
      a = ao;  b= bo; rule = crWhite; CibSort(a,b, rule); ScanSorted(ao,bo);
      a = ao;  b= bo; rule = crGrey;  CibSort(a,b, rule);	ScanSorted(ao,bo);

      a = ao;  b= bo; KomiSort(a,b, true);     				ScanSorted(ao,bo);
      a = ao;  b= bo; KomiSort(a,b, false);    				ScanSorted(ao,bo);

      a = ao;  b = bo; PoolSort(a,b);							ScanSorted(ao,bo);

      a = ao;  b = bo; GroupCodeSort(a,b);					ScanSorted(ao,bo);


      a = ao;  b= bo; CompatibleSort(a, b, '*', "A*");	ScanSorted(ao,bo);
      a = ao;  b= bo; CompatibleSort(a, b, '~', "B~");	ScanSorted(ao,bo);
      a = ao;  b= bo; CompatibleSort(a, b, '-', "C-");	ScanSorted(ao,bo);
      a = ao;  b= bo; CompatibleSort(a, b, '+', "D+");	ScanSorted(ao,bo);

      a = ao;  b = bo; FormMaximumArea(a,b);					ScanSorted(ao,bo);
      a = ao;  b = bo; FormAlienArea(sUK, sXXX, a,b);		ScanSorted(ao,bo);
      a = ao;  b = bo; FormAlienArea(sUK, sCam, a,b);		ScanSorted(ao,bo);


      a = ao;  b = bo;  for(int p=a; p<=b; p++) SortedPlayer(p)->Spare[0] = p%2;

      BusySort(a,b);							ScanSorted(ao,bo);

      a = a;
   }
}
//
////////////////////////////////////////////////////////////////////////////////








