////////////////////////////////////////////////////////////////////////////////
// MODULE PairTool: Pure library functions for the pairing G.Kaniuk 10/10/04
////////////////////////////////////////////////////////////////////////////////
#include <list>
#include <vector>
#include <string>

#include "Maths.h"
#include "Tour.h"

#ifndef PairToolH
#define PairToolH

////////////////////////////////////////////////////////////////////////////////
// VECTORISE A LIST
//
namespace PairTool
{
	using std::string;
   using std::list;
   using std::vector;

	template <class Group>				// VECTORISE A LIST
   void VectorLoad
   (
   	list<Group>& List,				// Given list
   	vector<Group>& Groups			// Fill with list elements
   )
   {
   	Groups.resize(List.size());

      list<Group>::iterator sList = List.begin();
      list<Group>::iterator eList = List.end();

      int g = 0;

      while(sList != eList)
      {
      	Groups[g] = *sList;

         ++sList;  ++g;
      }
   }

	template <class Group>
   void VectorLoad
   (
   	list<Group>& List,				// Given list of size sorted groups
   	vector<Group>& Groups,			// Fill with list elements

      int& Size,							// Total size of all groups
      int& SizeMax						// Total size of equally largest groups
   )
   {
   	Groups.resize(List.size());

      list<Group>::iterator sList = List.begin();
      list<Group>::iterator eList = List.end();

      int bigsize = sList->Size;		 // Largest group size

      int g = 0;

      Size = SizeMax = 0;

      while(sList != eList)
      {
      	Groups[g] = *sList;

         if(sList->Size == bigsize)			// All these groups have the same largest size
         	SizeMax += bigsize;				// Accumulate the sum

         Size += sList->Size;					// Total size all groups

         ++sList;  ++g;
      }
   }

   typedef vector<Tours::GGame> PGames;	// The Games supplied in GPlayer
}

////////////////////////////////////////////////////////////////////////////////
// CHOOSE
//
namespace PairTool
{
   int Choose(int Size);

   int Choose(int a, int b);

   int ChooseExcluding(int One, int a, int b);

   int ChooseExcluding(int a, int b, int e, int f);

   int ChooseExcluding(string& Land, string& Club, int a, int b);
}

///////////////////////////////////////////////////////////////////////////////
// HOUSEKEEPING
//
namespace PairTool
{
	void SetUpDnCounts(void);

   void ApproveSort(const int ao,const int bo);

   int ImprovePairing(void);
}

////////////////////////////////////////////////////////////////////////////////
// MCMAHON GROUPS
//
namespace PairTool
{
	using Tools::ints;

	struct MMGroup
   {
      ints Score;		// The group MMS

      int Start;		// The start index to Sorted
      int Size;		// Number of players in group
      int Pool;		// Number of unpaired players in group
      int Work;		// Working pool for group searching
      int Class; 		// One of mcClass set after analysis
      int Komip;		// Number of Komi players in group

      MMGroup()
      {
      	Score = 0; Start = 0;
         Size  = 0; Pool  = 0;
         Work  = 0; Komip = 0;
         Class = 0;
      }

      MMGroup(ints score, int start, int size)
      {
      	Score = score;  Start = start;
         Size  = size;   Pool  = size;
         Work  = 0;
         Class = 0;
      }
   };

   typedef std::vector<MMGroup> MMGroups;

   void ScanGroups(MMGroups& Groups);

   void FormMMGroups(MMGroups& McGroups);

   void PurgeOpponents(int pOne, MMGroups& McGroups);

   void MarkOpponents(const int pOne, const int gTwo, const MMGroups& McGroups);

   bool ExistsKomiOpponent(const int p, const MMGroup& McGroup);
}

////////////////////////////////////////////////////////////////////////////////
// AREA GROUPS
//
namespace PairTool
{
	using std::string;
   using std::vector;
   using std::list;

   struct ARClub									// CLUB GROUP
   {
      string Club;								// Club code

      int Start;									// Index to Sorted
      int Size;									// Number of players in group

      ARClub()
      {
      	Club = string();

         Start = Size = 0;
      }

      ARClub(string club, int start)
      {
      	Club = club;  Start = start;  Size = 0;
      }
   };

   struct ARLand									// LAND GROUP
   {
   	string Land;								// Country code

      int Start;									// Index to Sorted
      int Size;									// Number of players in group

      int SizeMax;								// Number of players in all equally largest clubs

      vector<ARClub> Clubs;

      ARLand()
      {
      	Land = string();

         Start = Size = 0;

         Clubs.clear();
      }

      ARLand(string land, int start)
      {
      	Land = land;

         Start = start;	Size  = 0;

         Clubs.clear();
      }

      void Load(list<ARClub> ClubList)
      {
      	VectorLoad<ARClub>(ClubList, Clubs, Size, SizeMax);

         Start = Clubs[0].Start;
      }
   };

   struct ARLands									// LANDS SEQUENCE
   {
   	int Start;									// Index to first player in the sequence of Lands
      int Size;									// Total number of players in all Lands

      int SizeMax;								// Number of players in all equally largest Lands

      vector<ARLand> Lands;					// The sequence of Lands

      ARLands()
      {
      	Start = Size = 0;

         SizeMax = 0;

         Lands.clear();
      }

      void Load(list<ARLand> LandList)
      {
      	VectorLoad<ARLand>(LandList, Lands, Size, SizeMax);

         Start = Lands[0].Start;

      }
   };

   void FormAreas(int a, int& b, ARLands& Areas);

   void FormMaximumArea(int& a, int& b);

   void FormAlienArea(const string Land, const string Club, int& a, int& b);

   void FormSameAreaCount(const int dir, const int a, int& b);
}
#endif
//
////////////////////////////////////////////////////////////////////////////////
