////////////////////////////////////////////////////////////////////////////////
// MODULE RegMaster:  RegGrid interface to GTour        G.Kaniuk 13/06/04
//
#include <algorithm>
#include <vcl.h>
#pragma hdrstop
#pragma package(smart_init)

#include "GoErrMsgRES.h"
#include "GoInfoRES.h"

#include "Format.h"
#include "Strings.h"

#include "Sort.h"
#include "TourStream.h"
#include "TourMaster.h"
#include "DrawMaster.h"
#include "RegMaster.h"
//---------------------------------------------------------------------------

namespace RegMaster											// ERRORS
{
   static void Throw(ToolsGUI::EType etype, short idrout, short idmsg)
   {
      ToolsGUI::ThrowError(etype, IDE_RegMaster, idrout, idmsg);
   }
}

namespace RegMaster											// LOCAL PARSERS
{
	using Tools::sComma;
   using Tools::sSpace;

   enum
   {
   	chSpace = ' ', chHyphen = '-', chQuote = '\'',
      chKyu   = 'k', chDan    = 'd', chPro   = 'p'
   };

   int parse_pid(string& text, int& pid)
   {
      using Tools::to_int;
      using Tools::to_string;

      using TourMaster::GoTour;

      for(size_t i=0; i<text.size(); i++)
      	if(!isdigit(text[i]))
         	return pidBad;

      if(!to_int(text.begin(), text.end(), pid)) return pidBad;

      text = to_string("0000", pid);

      if(pid==0) return pidZero;

      const int nextpid = (int)GoTour.NewPid;

      if(pid>nextpid)
      {
         pid = nextpid;

         return pidBig;
      }

      return pidLegal;
   }

   bool legal_region_syntax(string text)
   {
   	using Tools::is_alphabetic;

   	string::iterator ftext = text.begin(), etext = text.end();

      if(!is_alphabetic(*ftext)) return false;	++ftext;	// First char is alpha

      if(!is_alphabetic(*(etext-1))) return false;       // Last char must be alpha

      while(ftext != etext)
      {
      	bool puncton = false;									// Punctuation mark '-' or ' '

      	switch(*ftext)
         {
         	case chHyphen: puncton = true; break;
            case chSpace:  puncton = true; break;

            default: if(!is_alphabetic(*ftext)) return false;
         }

         if(puncton)                                     // Two in a row not allowed
         {
         	if(*(ftext-1) == Tools::cMinus) return false;
            if(*(ftext-1) == Tools::cSpace) return false;
         }

         ++ftext;
      }
      return true;
   }

   void capitalise(string& text)				// Assume legal name syntax, and may
   {													// contain cSpace, cQuote, or cMinus
   	if(text.empty()) return;

      string::iterator stext = text.begin(), etext = text.end();

      *stext = Tools::uppercase(*stext); ++stext;

      while(stext != etext)
      {
         switch(*(stext-1))
         {
            case chSpace:
               *stext = Tools::uppercase(*stext);
            break;

            case chHyphen:
               *stext = Tools::uppercase(*stext);
            break;

            case chQuote:
               *stext = Tools::uppercase(*stext);
            break;

            default: ;
         }
         ++stext;
      }
   }


   void rounds_string(const vector<bool>& playing,  string& text)
   {
      string items = string();

      for(size_t r=0; r<playing.size(); r++)
      {
         if(playing[r])
            items.append(Tools::to_string(r+1));
         else
            items.append(string(r<9 ? 1 : 2, Tools::cNone));

         if(r<playing.size() - 1)
            items.append(string(1,Tools::cSpace));
      }

      text = items;
   }

   void string_rounds(string text, vector<bool>& playing)
   {
      int rounds = playing.size(), round;

      for(int r=0; r<rounds; r++)playing[r] = false;

      if(text.empty()) return;

      string::iterator stext = text.begin(), etext = text.end();

      int rcount = 0;

      while(Tools::next_posint(stext, etext, round))
      {
         if(1 <= round && round <= rounds)
            playing[round-1] = true;

         ++rcount;  if(rcount>Tours::max_rounds)break;
      }
   }

   string name_family(const string fname)
   {
      string given,middle,family, familygiven, fullname = fname;

      Tools::parse_extended_family(fullname, given,middle,family);

      if(given.empty()) return family;

      familygiven = family + string(sComma) + given;

      if(middle.empty()) return familygiven;

      return familygiven + string(sSpace) + middle;
   }

   string family_comma_name(int form, const string sname)
   {
   	using Tools::sSpace;
      using Tools::sComma;

      string given, middle, family, name = sname;

      given = string();  middle = string();  family = string();

      switch(form)
      {
         case TourMaster::nfGivenFamily:
            if(!Tools::parse_name(name, given,middle, family)) return string();
         break;

         case TourMaster::nfFamilyGiven:
            if(!Tools::parse_extended_family(name, given,middle, family)) return string();
         break;

         case TourMaster::nfFamily:
         	if(!Tools::parse_name(name, family)) return string();
         break;
         
         default: return string();
      }

      capitalise(family);	if(given.empty()) return family;

      capitalise(given);

      string familygiven = family + string(sComma) + given;

      if(middle.empty()) return familygiven;

      capitalise(middle);

      return familygiven +  string(sSpace) + middle;
   }

   
   bool parse_grade(string& sgrade, ints& vgrade)
   {
      // sgrade is np, nd, nk or nnk where n is a digit

      using Tools::lexical_class;
      using Tools::lxDigit;
      using Tools::to_posint;
      using Tools::cZero;

      using Tours::mc_scale;
      using Tours::pro_grade;

      using TourMaster::GoTour;
      
      vgrade = 0;   if(sgrade.empty()) return false;


      sgrade = Tools::lowercase(sgrade);

      int gbodysize = sgrade.size() - 1;

      if(gbodysize<1 || gbodysize>2) return false;

      if(lexical_class(sgrade[0])!= Tools::lxDigit)
         return false;

      char suffix = sgrade[gbodysize];

      switch(suffix)
      {
         case chKyu:
         {
            if(gbodysize==2)
            if(lexical_class(sgrade[1])!= Tools::lxDigit)
                  return false;

            vgrade = -to_posint(sgrade.begin(), sgrade.begin()+gbodysize)*mc_scale;

            if(vgrade == 0)
            	if(GoTour.State > Tours::tsRegClosed)
               {
               	vgrade = -99*Tours::mc_scale;

               	return false;
               }
               else
            		vgrade = Tours::zero_kyu;

            return true;
         }

         case chDan:
         {
            vgrade = ((ints)(sgrade[0] - Tools::cZero) - 1)*mc_scale;

            if(vgrade<0)
            	if(GoTour.State > Tours::tsRegClosed)
               {
               	vgrade = Tours::min_grade;

               	return false;
               }
               else
            		vgrade = Tours::zero_dan;

            return gbodysize==1;
         }

         case chPro:
         {
            vgrade = ((ints)(sgrade[0] - Tools::cZero) - 1)*mc_scale + pro_grade;

            if(vgrade<pro_grade)
            	if(GoTour.State > Tours::tsRegClosed)
               {
               	vgrade = Tours::min_grade;

                  return false;
               }
               else
            		vgrade = Tours::zero_pro;

            return gbodysize==1;
         }

         default: return false;
      }
   }


   bool parse_rounds(string& srounds, vector<bool>& vrounds)
   {
   	using Tours::stAllPlay;
      using Tours::stAllPlayHcap;

   	using TourMaster::GoTour;

      // srounds is a sequence of POSINT separated by space
      // POSINT::positive integer
      // As this is meant for register import we retain a flexible
      // syntax eg POSINT could be in random order

      int rounds = TourMaster::GoTour.Rules.Rounds, r;

      if(vrounds.size() != size_t(rounds))
      	return false;

      if(srounds.empty())
      {
         for(r=0; r<rounds; r++)vrounds[r] = true;

         return true;
      }

      if(GoTour.Rules.Style== stAllPlay || GoTour.Rules.Style== stAllPlayHcap)
      {
         for(r=0; r<rounds; r++)vrounds[r] = true;

         rounds_string(vrounds,  srounds);

         return true;
      }

      string::iterator rs = srounds.begin();
      string::iterator re = srounds.end();

      for(r=0; r<rounds; r++)vrounds[r] = false;

      while(Tools::next_posint(rs, re, r))
         if(1 <= r && r <= rounds)
            vrounds[r-1] = true;
         else
            return false;

      return true;
   }

   bool parse_group(string& group, char& id, int& type)				// GK correction for 6.4.2
   {                                                              // On any error, type is set to -peGroupXXX
   	using Tools::cSpace;
      using Tours::GroupCodes;

      type = 0;

   	if(group.empty())
      {
      	group = string(2, cSpace);  return true;
      }

      if																				// Only McMahon or Barlow can have groups
      (
      	TourMaster::GoTour.Rules.Style != Tours::stMcMahon
         										 &&
         TourMaster::GoTour.Rules.Style != Tours::stBarLow
      )
      {
      	group = string(2, cSpace);

         type = -rgGroupForbid;  return false;
      }

      if(group.size() == 1)													// User types id or code
      {
      	id = Tools::uppercase(group[0]);

         switch(Tools::lexical_class(id))
      	{
         	case Tools::lxUpper: type = -rgGroupSpecial; break;	// Trying to create special group
            case Tools::lxDigit:	type = -rgGroupSuper; break;		// Trying to create super group

            default:					type = 0;  break;						// Not sure
      	}

      	if(type<0) return false;

         while(true)																// It might be a code
         {
         	if(id==Tours::gpTop){ type = -rgGroupSuper; break; }	// Maybe trying supergroup

            const int igt = GroupCodes.find(id,0);

            if(0<=igt && igt<=3)												// Maybe trying specialgroup
            {
            	type = -rgGroupSpecial; break;
            }

            break;
         }

         if(type==0) type = -rgGroupOther;								// No idea

         group = string(2, cSpace);	 group[0]=id;

         return false;
      }

      if(group.size()>2)group.resize(2);									// Theoretically impossible!

      id = Tools::uppercase(group[0]);    								// 0 - 9 or A - Z is group id

      switch(Tools::lexical_class(id))
      {
      	case Tools::lxDigit: 												// Super group identifier
         {
         	if (group[1] == Tours::gpTop) return true;

            type = -rgGroupSuper;											// Bad super group code
         }
         break;

         case Tools::lxUpper: 												// Special group identifier
         {
           	const char gt = group[1];

            const int igt = GroupCodes.find(gt,0);

            if(0<=igt && igt<=3) 							 				// See Tour.h GroupCodes
            	return true;
            else
            	type = -rgGroupSpecial;
        	}
         break;

         default: type = -rgGroupOther;
      }
      return false;
   }

   static bool parse_status(string& status)
   {
   	if(status.empty()) return true;

      for(int i=1; i<4; i++)
      	if(status==string(TourMaster::RegCode[i])) return true;

      return false;
   }
}
namespace RegMaster											// FULL PLAYER PARSER
{
	bool make_region(const int sizecode, string& area, string& areacode)
   {
      using TourMaster::GoTour;
      using Tours::size_club;
      using Tours::size_land;
      using ToolsGUI::BugReport;

      if(area.empty())
      {
      	areacode = string();

         return false;
      }

      if(!legal_region_syntax(area))return false;

      capitalise(area);

      switch(sizecode)
      {
      	case size_club: areacode = GoTour.ClubCode(area);  break;
         case size_land: areacode = GoTour.LandCode(area);  break;

         default:	areacode = string();
      }

      if(areacode.empty())
      {
      	BugReport(IDE_RegMaster, IDE_make_region, IDE_BadCodeSize);

         return false;
      }

      return true;
   }

	bool make_family(string& full)
   {
   	using Tools::parse_extended_family;

      if(full.empty()) return false;
      
   	string given, middle, family, familyname;

      string fullname = Tools::compactify_name(full);

      if
      (
      	!parse_extended_family
         (
         	fullname, given, middle, family
         )
      )
      return false;

      capitalise(given);  capitalise(middle);  capitalise(family);

      familyname = family;

      if(!given.empty())
      {
      	familyname += string(Tools::sComma) + given;

         if(!middle.empty())
         	familyname += string(Tools::sSpace) + middle;
      }

      full = familyname;

      return true;
   }

   bool make_rated_grade(string sgor, string& strength, string& tgrade)
   {
      // A default value for tgrade is supplied by caller

   	using Tools::real;	using Tools::to_stringr;
      using Tools::round;	using Tools::to_stringi;

      const string kstring = string("k"),  dstring = string("d") , pstring = string("p");

   	int gor;

      if(!Tools::to_int(sgor.begin(), sgor.end(), gor))
      {
      	strength = string("??.?z");

         return false;
      }

      if(gor<100)
      {
      	strength = string("??.?n");									// Novice - n

       	return false;
      }

      if(gor<=2050)
      {
      	const real   krated  = -(gor-2100)/100.0;
         const int    kgrade  = round(krated);

         strength = to_stringr(1, krated) + kstring;
         tgrade   = to_stringi(kgrade) + kstring;

         return true;
      }

      if(gor<2700)
      {
      	const real drated = (gor-2000)/100.0;
         const int  dgrade  = round(drated);

         strength = to_stringr(1, drated) + dstring;
         tgrade   = to_stringi(dgrade) + dstring;

         return true;
      }

      if(gor<2970)
      {
      	const real prated = (gor-2670)/30.0;
         const int pgrade	= round(prated);

      	strength = to_stringr(1, prated) + pstring;
         tgrade   = to_stringi(pgrade) + pstring;

         return true;
      }

      strength = string("??.?x");

      return false;
   }
}
//------------------------------------------------------------------------------

namespace RegMaster											// FULL PLAYER PARSER
{
	using std::string;
   using std::vector;
   using std::list;

   using Tours::GPlayer;
	using TourMaster::GoTour;

   enum
   {
      rfNumber, rfStatus,
      rfName,   rfGrade, rfClub, rfLand, rfRounds,
      rfGroup,  rfNote,
      rfFields
   };

   void player_round_change(bool playing, Tools::intsi& result)
   {
   	using Tours::grSleep;
      using Tours::grNone;

   	switch(result)
      {
      	case grNone:  if(!playing) result = grSleep; break;	// Put to sleep
         case grSleep: if(playing)  result = grNone;  break;   // Wake up

         default:		  if(!playing) result = grSleep; break;   // Dont touch result if playing
      }
   }


   int parse_player(vector<string>& fields, GPlayer& player)
   {
   	using Tools::legal_name_syntax;
      using ToolsGUI::ErrorReport;

   	// CONVERT PLAYER NUMBER
      fields[rfNumber] = Tours::pid_string(player.Pid) + Tools::sSpace;

      // CHECK REGISTRATION STATUS
      string regstatus = fields[rfStatus];

      if(!parse_status(regstatus))
      	return rfStatus;

      if(GoTour.State < Tours::tsRegClosed)
			player.State = !regstatus.empty() ? Tours::psRegistered : Tours::psEntered;
      else
      	if(regstatus.empty()) return rfStatus;


      // CHECK NAME
      string given, middle, family, familyname;

      if(!Tools::parse_extended_family(fields[rfName], given, middle, family))
      	return rfName;

      capitalise(given);  capitalise(middle);  capitalise(family);

      familyname = family;

      if(!given.empty())
      {
      	familyname += string(Tools::sComma) + given;

         if(!middle.empty())
         	familyname += string(Tools::sSpace) + middle;
      }
      fields[rfName] = player.Name = familyname;


      // CHECK GRADE
      if(!parse_grade(fields[rfGrade], player.Grade))
      	return rfGrade;


      // CHECK CLUB
      string club = fields[rfClub];

      if(!legal_region_syntax(club))
      	return rfClub;

      capitalise(club);	fields[rfClub] = club;

      string clubcode = TourMaster::GoTour.ClubCode(club);

      if(clubcode.empty())
      {
      	ErrorReport(IDE_RegMaster, IDE_parse_player, IDE_BadClubCodes);

         return rfClub;
      }
      else
      	player.Club =  clubcode;


      // CHECK COUNTRY
      string land = fields[rfLand];

      if(!legal_region_syntax(land))
      	return rfLand;

      capitalise(land);		fields[rfLand] = land;

      string landcode =  TourMaster::GoTour.LandCode(land);

      if(landcode.empty())
      {
      	ErrorReport(IDE_RegMaster, IDE_parse_player, IDE_BadClubCodes);

         return rfLand;
      }
      else
      	player.Land = landcode;


      // CHECK ROUNDS
      // A game result is changed only if the player is put to sleep or woken up

      vector<bool>rounds(GoTour.Rules.Rounds);

      if(!parse_rounds(fields[rfRounds], rounds))
      	return rfRounds;

      for(size_t r=0; r<GoTour.Rules.Rounds; r++)
      	player_round_change(rounds[r], player.Games[r].Result);


      // CHECK GROUP
      char gid;  int gtype;

      string group = Tools::trim(fields[rfGroup]);

      if(!parse_group(group, gid, gtype))
      	return rfGroup;

      player.Group = group;							// Require 2 spaces for no group

      fields[rfGroup] = Tools::trim(group);   	// Require empty string for no group


      // CHECK NOTE
      string note = fields[rfNote];

      if(!Tools::is_ascii(note))
      	return rfNote;

      player.Note = note;

      return rfFields;
   }

   void start_new_player(GPlayer& player)
   {
      // ASSIGN INITIAL MCMAHON SCORE - DEPENDS ON GRADE AND GROUP
      initial_score(player);

      // SET MCMAHON SCORE AND ACCOUNT FOR ROUNDS MISSED
      player.McScore = player.MiScore;

      for(size_t r=0; r<GoTour.Roundl; r++)
         if(r && r%2==0)
            player.McScore += Tours::gsOne;

      // BY DEFAULT A PLAYER ADDED AFTER CLOSE OF REG DOES NOT PLAY THIS ROUND
      // A PLAYER JOINING THIS ROUND IS ANALYSED BY player_rounds_changed()
      if(GoTour.State > Tours::tsRegClosed)
         player.State = Tours::psSleep;
   }

   void end_player_edit
   (
   	GPlayer& playerold, GPlayer& playernow,
      const int iplayer,
      int& playrep
   )
   {
      // DETECT RESTORED JOINING/LEAVING
   	const int OldForcedEdCount = GoTour.ForcedEdCount;

      // DETECT ANY NEW CLUBS,LANDS
   	const bool newcodes = GoTour.UpdatePopulationCount(playerold, playernow);

      if(OldForcedEdCount && !GoTour.ForcedEdCount)
      {
      	TourMaster::GoTour.PopulationCount();	// MissingCount needed

      	GoTour.SetState();

         DrawMaster::ListBoards();

         playrep = prFullyRestored;
      }

      *GoTour.GetPlayerPtr(iplayer) = playernow;

      if(save_player(newcodes, GoTour.Sorted[iplayer])<0)
      	playrep = prCorrupt;
   }

   void player_details_changed(GPlayer& pold, GPlayer& pnow, int& playrep )
   {
   	using TourMaster::GoTour;

      // WHILE VIEWING PAST ROUND YOU CAN CHANGE THE RESULT OR THE DRAW
      // BUT NOT ANY PLAYER DETAILS

      initial_score(pnow);					// Recalc MiScore according to currrent bars

      pnow.McScore
      =
      pnow.McMahonScore(GoTour.Round);	// Recalc McScore from new MiScore

      if(!playrep)
      	if(pold.MiScore != pnow.MiScore)
      		playrep = prNewMMS;
   }

   /////////////////////////////////////////////////////////////////////////////
   //  player_rounds_changed
   //  Reports if player Joining or Leaving Draw and if so, sets
   //  GoTour.State to tsEditForced
   //  
   /////////////////////////////////////////////////////////////////////////////

   void player_rounds_changed(GPlayer& pold, GPlayer& pnow, int& playrep)
   {
   	using Tools::intsi;
      using Tours::tsEditForced;

      const int Round = GoTour.Round;		// The current round count

      playrep = prNone;

      // CHANGE IN CURRENT ROUND MAY TRIGGER A STATE CHANGE
      intsi rold = pold.Games[Round-1].Result, &rnow = pnow.Games[Round-1].Result;

      if(rnow == rold) return;

      if(!pold.Awake(Round) && pnow.Joining(Round))	// User edit to add player to draw
      {
      	 GoTour.State = tsEditForced;

          playrep = prJoin;
      }

      if( pold.Awake(Round) && pnow.Leaving(Round))	// User edit to remove player from draw
      {
      	GoTour.State = tsEditForced;

         playrep = prLeave;
      }

      if(pold.Joining(Round) && !pnow.Awake(Round))	// User edits to undo Joining
      	playrep = prRestored;

      if(pold.Leaving(Round) &&  pnow.Awake(Round))	// User edits to undo Leaving
      {
         DrawMaster::RestoreResult(&pnow);				// Can restore if opponent still has result

      	playrep = prRestored;
      }

      if(GoTour.ForcedEdCount==1 && playrep==prRestored)		// Restoring the last of the Joiners or Leavers
         playrep = prFullyRestored;

   }
}
//-----------------------------------------------------------------------------

namespace RegMaster								  			// TourStream INTERFACE
{
	using Tours::ftGoDraw;
   using Tours::faOpen;
	using TourMaster::GoTour;

	int save_player(bool newcodes, int plist)
   {
    	try
      {
   		Tours::GTourStream Stream(GoTour.FileName, ftGoDraw, faOpen);

      	Stream.SaveHeader(GoTour);

     		if(newcodes)
         	Stream.SaveCodes(GoTour);

         Stream.SavePlayer(plist, GoTour);
      }
      catch(ToolsGUI::ETour& Err)
      {
      	ToolsGUI::Serve(Err, IDE_RegMaster, IDE_save_player);

         return -1;
      }
      catch(...)
      {
      	ToolsGUI::ErrorReport(IDE_RegMaster, IDE_save_player, IDE_NoMemory);

         return -1;
      }
      return rfFields;
   }

   int toggle_reg_status(const int p)
   {
   	using Tours::GPlayerPtr;

      using TourMaster::GoTour;

      if(GoTour.State>=Tours::tsRegClosed) return -1;

   	GPlayerPtr Player = GoTour.GetPlayerPtr(p);

      if(Player->State == Tours::psNew) return -1;

      switch(Player->State)
      {
      	case Tours::psEntered:    Player->State = Tours::psRegistered; break;
         case Tours::psRegistered: Player->State = Tours::psEntered;    break;

         default: Player->State = Tours::psEntered;									// GK 6.3.9  return -1;
      }
      if(save_player(false, GoTour.Sorted[p])<0) return -1;

      return Player->State-1;
   }

   void delete_players(size_t pstart, size_t pend)
   {
   	using std::list;
      using std::vector;

      using Tours::GPlayerPtr;
      using Tours::GTourStream;

      GTourStream TourStream(GoTour.FileName, ftGoDraw, faOpen);

   	try
      {
      	if(pend>GoTour.Players || pstart<=0 || pend<pstart)
      		Throw(ToolsGUI::erBug, IDE_delete_players, IDE_BadDelete);

         // WE TRIM THE FILE FROM THE BACK
         GPlayer Plast = GPlayer(0, GoTour.Rules.Rounds);

         const size_t FileTrim = Plast.Size()*(pend - pstart + 1);

         // PASS 1: MARK THE PLAYERS TO BE DELETED
         for(size_t p=pstart; p<=pend; p++)
         	GoTour.Listed[GoTour.Sorted[p]]->Pid = 0;

         // PASS 2: OVERWRITE THE DELETED PLAYER(S) WITH THE LAST PLAYER(S)
         //			  KEEPING THE LISTS AND THE FILE IN STEP.
         {
            int playersnow = GoTour.Players;

            list<GPlayer>::iterator lfront = GoTour.PlayerList.begin();
   	      list<GPlayer>::iterator lback  = GoTour.PlayerList.end();
            list<GPlayer>::iterator lnow;

            --lback;															// Iterator to last player in PlayerList

            lnow = lback;													// Iterator to current player

            for(size_t p=GoTour.Players; p>=1; p--)
            {
            	GPlayer& Pnow = *lnow;									// Current player

               if(!Pnow.Pid)												// This one to be deleted
               {
               	if(lnow != lback)										// Move last player to deleted player's slot
                  {
                     *lnow = *lback;									// Overwrite player in PlayerList and hence in Listed

                     TourStream.SavePlayer(p, GoTour, Pnow);	// Overwrite player on file
                  }

                  GoTour.PlayerList.erase(lback);   --lback;	// Now delete the last player in the PlayerList

						--playersnow;
               }

               --lnow;
            }

            
            // EPILOGUE
            GoTour.Players = playersnow;						// New player count

            if(!GoTour.Players)
            	GoTour.State = Tours::tsTourStart;

            GoTour.Listed.resize(playersnow + 2);			// NULL+Players+NewPlayer

            GoTour.Listed[playersnow+1] = &GoTour.Dummy;	// New player entry

            TourStream.SaveHeader(GoTour);

            TourStream.Stream->Size =							// Delete last records
            TourStream.Stream->Size - FileTrim;

            GoTour.PopulationCount();

            RegSort(GoTour.SortOrder, GoTour.SortSense);

            // FIND LARGEST .Pid FOR NEXT NEW PLAYER
            GoTour.NewPid = 0;

            for(int p=1; p<=playersnow; p++)									// Search for latest Pid
            	if(GoTour.Listed[p]->Pid > (size_t)GoTour.NewPid)
               	GoTour.NewPid = GoTour.Listed[p]->Pid;

            ++GoTour.NewPid;													   // NewPid > playersnow+1 is possible
         }
      }
      catch(ToolsGUI::ETour& Err)
      {
      	TourStream.~GTourStream();

      	ToolsGUI::Serve(Err, IDE_RegMaster, IDE_delete_players);
      }
      catch(...)
      {
      	TourStream.~GTourStream();

      	ToolsGUI::ErrorReport(IDE_RegMaster, IDE_delete_players, IDE_NoFileWrite);
      }
   }
}
//------------------------------------------------------------------------------

namespace RegMaster								  			// SORTING
{
	using std::vector;
   using std::sort;

	using TourMaster::GoTour;
   
	void RegSort(int sortorder, int sortdir)
   {
   	Tours::SortInterface(&GoTour);	GoTour.StartSorted();

      vector<size_t>::iterator Ss = GoTour.Sorted.begin() + 1;
      vector<size_t>::iterator Se = GoTour.Sorted.end();

      GoTour.SortOrder = sortorder;
      GoTour.SortSense = sortdir;

      if(!GoTour.Players) return;

      switch(sortorder)
      {
      	case Tours::skEntry:		sort(Ss,Se, Tours::OrderEntry());	break;
         case Tours::skTicked:	sort(Ss,Se, Tours::OrderTicked()); 	break;
         case Tours::skName:  	sort(Ss,Se, Tours::OrderName());  	break;
         case Tours::skGrade:		sort(Ss,Se, Tours::OrderGrade());   break;
         case Tours::skClub:		sort(Ss,Se, Tours::OrderClub()); 	break;
         case Tours::skLand:		sort(Ss,Se, Tours::OrderLand()); 	break;
         case Tours::skRounds:	sort(Ss,Se, Tours::OrderRounds()); 	break;
         case Tours::skGroup:		sort(Ss,Se, Tours::OrderGroup()); 	break;
      	case Tours::skNote:		sort(Ss,Se, Tours::OrderNote());	 	break;

         default: return;
      }
   }
}
//------------------------------------------------------------------------------

namespace RegMaster										// CLASS MFindTube
{
	using Tours::GTour;

   MFindTube::MFindTube()
   {
   	for(int f=0; f<fiFields; f++) Lists[f] = 0;

   }

   String MFindTube::PlayerString(Tours::GPlayerPtr P)
   {
   	string items[5], playerstring;

      items[0] = GoTour.Number(P);
      items[1] = GoTour.Name(P);
      items[2] = GoTour.Grade(P);
      items[3] = GoTour.Club(P);
      items[4] = P->Land;

      Tools::pack_tabbed_line(5, items, &playerstring);

      return Tools::s_S(playerstring);
   }

   int MFindTube::Find(MFindTarget targets[fiFields], TStrings* seeked)
   {
   	using std::string;
      using Tools::trim;
      using Tools::is_ascii;
      using Tools::legal_name_syntax;
      using Tools::begins_with;
      using Tools::S_s;
      using TourMaster::GoTour;

      if(seeked->Count != 0)									// Caller must clear seeked before calling
      	return -1;

      bool   rName = targets[fiName].Required;
   	string tName = rName ? trim(S_s(targets[fiName].Target)) : string();
      string tGiven, tMiddle, tFamily;

      if(rName)
      	if(!Tools::parse_extended_family(tName, tGiven, tMiddle, tFamily))
         	return fiName;

      bool   rGrade = targets[fiGrade].Required;
      string sGrade = rGrade ? trim(S_s(targets[fiGrade].Target)): string();

      ints tGrade;

      if(rGrade)
      	if(!parse_grade(sGrade, tGrade))
         	return fiGrade;

      bool rClub = targets[fiClub].Required;
      string tClub = rClub ? trim(S_s(targets[fiClub].Target)) : string();

      if(rClub)
      	if(!legal_name_syntax(tClub))
         	return fiClub;

      bool rLand = targets[fiLand].Required;
      string tLand = rLand ? trim(S_s(targets[fiLand].Target)) : string();

      if(rLand)
      	if(!legal_name_syntax(tLand))
         	return fiLand;

      bool rGroup = targets[fiGroup].Required;
      string sGroup = rGroup ? trim(S_s(targets[fiGroup].Target)): string();

      char iGroup; int tGroup;

      if(rGroup)
      	if(!parse_group(sGroup, iGroup, tGroup))
         	return fiGroup;

      bool rNote = targets[fiNote].Required;
      string tNote = rNote ? trim(S_s(targets[fiNote].Target)) : string();

      if(rNote)
      	if(!is_ascii(tNote))
         	return fiNote;

      bool rMMS = targets[fiMMS].Required;  int tMMS;
      string sMMS = rMMS ? trim(S_s(targets[fiMMS].Target)) : string();

      if(rMMS)
      	if(!Tools::to_int(sMMS.begin(), sMMS.end(), tMMS))
         	return fiMMS;

      bool rRounds = targets[fiRound].Required;
      string tRounds = rRounds ? S_s(targets[fiRound].Target) : string();

      if(rRounds)
         for(size_t r=0; r<tRounds.size(); r++)
            switch(tRounds[r])
            {
               case '0': break;
               case '1': break;
               case '*': break;

               default: return fiRound;
            }

      if(!rName && !rGrade && !rClub && !rLand && !rGroup && !rNote && !rMMS && !rRounds)
      	return rfFields+1;

		for(size_t p=1; p<=GoTour.Players; p++)
      {
      	Tours::GPlayerPtr P = GoTour.Listed[p];

         bool matched = true;

         while(matched)
         {
         	if(rName)
            	if(!begins_with(P->Name, tName))
						{ matched = false;  break; }

            if(rGrade)
            	if(P->Grade != tGrade)
               	{ matched = false;  break; }

            if(rClub)
            	if(!begins_with(GoTour.Club(P), tClub))
						{ matched = false;  break; }

            if(rLand)
            	if(!begins_with(GoTour.Land(P), tLand))
						{ matched = false;  break; }

            if(rGroup)
            	if(sGroup != GoTour.Group(P))
						{ matched = false;  break; }

            if(rNote)
            	if(!begins_with(GoTour.Note(P), tNote))
						{ matched = false;  break; }

            if(rMMS)
            	if(P->MiScore != (ints)(Tours::mc_scale*tMMS))
               	{ matched = false;  break; }

            if(rRounds)
            {
            	vector<bool>Playing(GoTour.Rules.Rounds);  P->RoundsPlayed(Playing);

               for(size_t r=0; r<Playing.size(); r++)
               	if(tRounds[r] != Tools::cStar)							// We DO care
                  {
                  	if(tRounds[r] == Tools::cOne && !Playing[r])		// 1 = playing
                     	{ matched = false;  break; }

                     if(tRounds[r] == Tools::cZero && Playing[r])		// 0 = sleep
                     	{ matched = false;  break; }
                  }
            }

            break;
         }
         if(matched)seeked->Add(PlayerString(P));
      }
      return fiFields;
   }

   MFindTube FindTube;
}
//------------------------------------------------------------------------------

namespace RegMaster										// SEARCHING
{
	using TourMaster::GoTour;

	size_t ExtractPlayerId(String PackedPlayer)
   {
      int TabPos = PackedPlayer.AnsiPos(String(Tools::sTab));	if(TabPos<=1)return 0;

      string pnum = Tools::S_s(PackedPlayer.SubString(1, TabPos-1));

      return (size_t)Tools::to_posint(pnum.begin(), pnum.end());
   }

   int FindPlayerId(size_t pid)
   {
   	for(size_t p=1; p<=GoTour.Players; p++)						// p is the reg row number
      	if(GoTour.Listed[GoTour.Sorted[p]]->Pid == pid)
         	return p;

      return 0;
   }

   int FindPlayerName(string NAME)
   {
   	using Tools::uppercase;

      if(NAME.empty()) return 0;
      
   	for(size_t p=1; p<=GoTour.Players; p++)
      	if(uppercase(GoTour.Listed[GoTour.Sorted[p]]->Name) == NAME)
         	return p;
      return 0;
   }
}
//------------------------------------------------------------------------------

namespace RegMaster										// SCORING
{
	using Tours::mc_scale;

	using TourMaster::GoTour;

   void initial_score(Tours::GPlayer& player)
   {
   	using TourMaster::GoTour;

      if													// Swiss; AllPlayAll
      (
      	GoTour.Rules.Style > Tours::stMcMahon
                            &&
         GoTour.Rules.Style < Tours::stBarLow
      )
      {
      	player.MiScore = 0;

         return;
      }

      if(GoTour.State<Tours::tsRegClosed)		// Unknown bar	GK 6.3.9
      {
      	player.MiScore = player.Grade;

         return;
      }

      const ints MacBar =  GoTour.Rules.MacBar;

   	if(!player.Group.empty())													// SUPERGROUP
      {
      	const char GroupId = player.Group[0];								// 0 - Top, 1-9 Supergroups

         if(isdigit(GroupId))
         {
         	const int iSup = (int)(GroupId - Tools::cZero);			// Integer version

            if(iSup>0)																// A super group
               player.MiScore = MacBar + iSup*GoTour.Rules.SuperDiff;
            else
               player.MiScore = MacBar;                 					// Top McMahon group

         	return;
         }
      }

      if(player.Grade < GoTour.Rules.MinGrade)									// Novice
      {
      	player.MiScore = GoTour.Rules.MinGrade;	return;
      }

      if(player.Grade < MacBar)														// Below the bar
      {
      	player.MiScore = player.Grade;	return;
      }

      // GK 6.3.9	play all rounds above the bar
      player.MiScore = player.PlayingAll()
      					? MacBar
                     : MacBar-Tours::mc_scale;														// Top group
   }
}
//------------------------------------------------------------------------------

namespace RegMaster										// CLOSE REGISTER
{
	void SetMacGroupCodes(void)								// At CloseReg
	{
   	const std::string GroupOne("0:"), GroupNone("  ");

      for(size_t p=1; p<=GoTour.Players; p++)			// Group code 0:  for players above the bar
      {
      	Tours::GPlayerPtr Player = GoTour.Listed[p];

         if
         (
         	GoTour.Rules.Style == Tours::stMcMahon
                               ||
            GoTour.Rules.Style == Tours::stBarLow
         )
         {
            if(Player->Group == GroupOne)						// Clear an old setting
               Player->Group = GroupNone;

            if(Player->MiScore == GoTour.Rules.MacBar)	// New setting
               if(Player->Group == GroupNone)				// Leave other codes alone
                  Player->Group = GroupOne;
         }
         else
         	Player->Group = GroupNone;
      }
   }

	int MissingGradeCount(void)
   {
   	int Count = 0;

      for(size_t p=1; p<=TourMaster::GoTour.Players; p++)
      {
      	int pGrade = (int)GoTour.Listed[p]->Grade;

         if(pGrade == Tours::zero_pro) ++Count;
         if(pGrade == Tours::zero_dan) ++Count;
         if(pGrade == Tours::zero_kyu) ++Count;
      }
      return Count;
   }

   void CloseUnregistered(const int action)
   {
   	// UNREGISTERED PLAYERS ARE AT BACK OF REGISTER
      const first = GoTour.RegCount+1, last = GoTour.Players;

   	if(action==crDelete)											// DELETE
      {
      	delete_players(first, last);	return;
      }

      for(int p=first; p<=last; p++)
      {
      	Tours::GPlayerPtr Player = GoTour.Listed[GoTour.Sorted[p]];

         Player->State = Tours::psRegistered;				// Register the player!

         Player->Games[0].Result = Tours::grSleep;  		// LATE OR KEEP

         if(action==crKeep)
         	for(size_t r=1; r<GoTour.Rules.Rounds; r++)
            	Player->Games[r].Result = Tours::grSleep;	// KEEP
      }
   }
}
//------------------------------------------------------------------------------


namespace RegMaster										// IMPORT
{
   void import_player(GPlayer& playernow)
   {
   	using Tours::gsOne;
      using Tours::grSleep;
      using Tours::grNone;

      // ASSIGN INITIAL McMAHON SCORE - DEPENDS ON GRADE AND GROUP
      initial_score(playernow);

      // SET McMAHON SCORE AND ACCOUNT FOR ROUNDS MISSED
      playernow.McScore = playernow.MiScore;

      for(size_t r=0; r<GoTour.Roundl; r++)
         if(r && r%2==0)
            playernow.McScore += Tours::gsOne;

      // A PLAYER IMPORTED DURING REGISTRATION IS ENTERED/REGISTERED
      // AND playernow.State WAS SET BY THE PARSER

      // A PLAYER IMPORTED AFTER CLOSE OF REG IS ASLEEP
      // AND DOES NOT PLAY THIS ROUND OR ANY PREVIOUS ROUND
      if(GoTour.State > Tours::tsRegClosed)
      {
      	for(size_t r=0; r<GoTour.Roundl; r++)
         	playernow.Games[r].Result = Tours::grSleep;

         playernow.State = Tours::psSleep;
      }
   }
}
//------------------------------------------------------------------------------

