////////////////////////////////////////////////////////////////////////////////
// MODULE  TextSmith:    Production of all Tab-Delimited text files                    G.Kaniuk 05/04/02
//
// Define a class TextFile having a TStringList member which gives us the  filing
// functions LoadFromFile and SaveToFile inherited from TStrings.  The class
// defines two virtual functions HeadLine and TextLine which define how the header
// and individual lines are constructed.  The class method BuildTextList then
// invokes Headline and TextLine to create the complete list which is saved
// to file on destruction.
//
// Specific text files are implemented as classes derived from TextFile,
// where the only implementation needed are the HeadLine and TextLine overrides.

#include <vector>
#include <string>
#include <vcl.h>

#include "TourMaster.h"

#ifndef TextSmithH
#define TextSmithH

//------------------------------------------------------------------------------
namespace TextSmith
{
   using std::string;
   using std::vector;

   using TourMaster::GTourMaster;
   using Tours::GPlayerPtr;

   class TextFile                                     // BASE CLASS FOR ALL TEXT FILES
   {
      public:
         string FileName;                             // Supplied at construction

         TStringList* Lines;                          // Initially empty

      protected:
         GTourMaster* Tour;

      public:
      	TextFile();												// Zeros everything
         TextFile(string& fname, GTourMaster* tour);  // Creates empty Lines
         ~TextFile();                                 // Saves Lines to FileName

         void Add(string line);                       // Add to back of Lines

         virtual void HeadLine();                     // Add any headlines
         virtual void TextLine(int i);                // Add a single line

         virtual void BuildTextList(int items);       // Prepare Lines for filing
   };


   class RankFile : public TextFile                   // TEMPLATE FOR ALL RANKLISTS
   {
      public:
      	size_t Active;

         TStringList* ClubList;
         TStringList* LandList;

         RankFile(string& fname, GTourMaster* tour);
         virtual ~RankFile();

         virtual void RankLine
         (
            bool newgrp, int p, int pos,
            GPlayerPtr P
         );
         void BuildRankList();
   };

   class StdRankFile : public RankFile                // GENERAL PURPOSE LIST & PRINTING
   {
      public:
         StdRankFile(string& fname, GTourMaster* tour);

         virtual void HeadLine();

         virtual void RankLine
         (
            bool newgrp, int p, int pos,
            GPlayerPtr P
         );

         virtual string game_string(Tours::GGame& game);
   };

   class WebRankFile : public StdRankFile                // FOR WEBMASTER
   {
      public:
         WebRankFile(string& fname, GTourMaster* tour);

         string game_string(Tours::GGame& game);

         virtual void HeadLine();

         virtual void RankLine
         (
            bool newgrp, int p, int pos,
            GPlayerPtr P
         );
   };
}
//---------------------------------------------------------------------------

namespace TextSmith
{
   class RegisterPriFile : public TextFile                // REGISTER
   {
      public:
      	TStringList* LandList;
         
         RegisterPriFile(string& fname, GTourMaster* tour);
         ~RegisterPriFile();

         void HeadLine();

         void TextLine(int p);
   };
}
//---------------------------------------------------------------------------

namespace TextSmith
{
	using Tours::GPlayerPtr;
   using Tours::GBoard;

	enum { drDraw, drResult };


   class DrawResPriFile : public TextFile                // DRAW/RESULTS
   {
   	private:
      	int Mode;						// Draw: Mode = 0, Results: Mode = 1

         string Name(GPlayerPtr player);
         string Komi(GBoard board);

      public:
         DrawResPriFile(int mode, string& fname, GTourMaster* tour);

         void HeadLine();

         void TextLine(int p);
   };
}
//---------------------------------------------------------------------------

namespace TextSmith
{
   class PrizePriFile : public TextFile                // PRIZE LIST
   {
      public:
         PrizePriFile(string& fname, GTourMaster* tour);

         void HeadLine();

         void TextLine(int p);
   };
}
//---------------------------------------------------------------------------

namespace TextSmith
{
   class CardsPriFile : public TextFile                // CARDS
   {
   	private:
      	int Pfirst, Plast;

      public:
         CardsPriFile(string& fname, int pstart, int plast, GTourMaster* tour);

         void HeadLine();

         void TextLine(int p);
   };
}
//---------------------------------------------------------------------------
#endif
