////////////////////////////////////////////////////////////////////////////////
// MODULE TourMaster:       All interface to GTour        G.Kaniuk 03/10/02
//
// We define the class GTourMaster derived from GTour.  This implements all
// GoDraw specific data and functionality including:
//    State management
//    Pairing
//    Loading and Saving file, Player and Game result
//
// Any function requring access to the tournament data does so via GTourMaster
// We also declare the one (global) instance GoTour of GTourMaster
//
////////////////////////////////////////////////////////////////////////////////

#include <vector>
#include <string>
#include <list>

#include "Tour.h"

#ifndef TourMasterH
#define TourMasterH

namespace TourMaster                      // CLASS GTourMaster
{
   using std::vector;
   using std::string;
   using std::list;

   using Tours::TourState;
   using Tours::GBoard;
   using Tours::GGame;
   using Tours::GTour;
   using Tours::GPlayer;
   using Tours::GPlayerPtr;

   enum										// GK REGSORT 2015
	{
		regOldSort,							// The register order must be restored if changed

   	regNewSort							// The register has been sorted inside the command
   };

   enum	// ACTIVE COMMANDS PRODUCING A STATE CHANGE
   {
   	cmdNone,   cmdNew,     cmdOpen,
      cmdEnter,  cmdDel,     cmdClose,
      cmdReOpen, cmdRestore, cmdRound,
      cmdDraw,   cmdForced,  cmdEdit,
      cmdSave,   cmdDcancel, cmdEcancel,
      cmdResult, cmdCloseT,
   	cmdActiveCount
   };

   enum	// STATE DEPENDENT COMMANDS - THESE DO NOT CHANGE STATE BUT ARE NOT ALWAYS AVAILABLE
   {
   	cmdFind = cmdActiveCount, cmdRegAll, cmdUnRegAll, cmdFastReg,  // Visible on the main form
   	cmdMacBar, cmdKomBar, 	  cmdNovBar,

      cmdImport,														// File Menu
      cmdCount
   };

   enum                        // SORTED PAIRING FOR DISPLAY
   {
      pkBoardUp, pkBoardDn,    // Board order
      pkKomiUp,  pkKomiDn,     // Handicap, then Komi
      pkNoteUp,  pkNoteDn,     // Comment
      pkSize
   };

   enum
   {
      rkBoardUp, rkBoardDn,   // Board order
      rkBlackUp, rkBlackDn,   // Black result
      rkWhiteUp, rkWhiteDn,   // White result
      rkSize
   };

   enum                       // NAME FORMAT FROM PLAYERS FILE
   {
      nfUnknown,
      nfGivenFamily,          // Given Middle Family
      nfFamilyGiven,          // Family Given Middle
      nfFamily						// Family only
   };

   struct GCodeCount				// Maintain counts of clubs, countries
   {
   	string Code;
      int    Count;

      GCodeCount()
      {
      	Code = string();  Count = 0;
      }

      GCodeCount(string code, int count)
      {
      	Code = code;  Count = count;
      }

      bool operator != (GCodeCount c) { return Code != c.Code; }
      bool operator == (GCodeCount& c) { return Code == c.Code; }
   };

   class GTourMaster : public GTour
   {
      private:
         int InfoID;                               // Status Bar message ID

      public:                                      // Display, not control
         int ClubCount;                            // Total Clubs used
         int LandCount;                            // Total Countries used
         int RegCount;										// Number of registered players
			int PlayNowCount;									// Playing this round
         int PlayNextCount;								// Playing next round
         int JoiningCount;									// Number of players joining the draw
         int MissingCount;									// Missing results
         int PrizeCount;									// Estimated prizes
         int ForcedEdCount;								// Number of players Joining or Leaving current draw

         int SupPop;											// Supergroup population
         int MacPop;											// McMahon group
         int KomiPop;										// Handicap group
         int NovicePop;										// Lowest bar group

         int NewPid;											// Playerid for a new player

         bool RestoreRegStatus;							// RegGrid needs to change regfield code from J/L to R
         int  OldRegOrder, OldRegSense;				// GK REGSORT 2015 Manage re-sorting register after system disturbance

         list <GCodeCount>ClubPops;						// Population each club
         list <GCodeCount>LandPops;						// Population each country

         vector<GBoard>Pair;                       // The draw
         
      public:
         GTourMaster();                            // Default Constructor

         GTourMaster(string FileName);

         void Load(string FileName);
         void Save(string FileName);
         
         bool SaveTournament(void);
         bool SaveHeader(void);
         bool SavePlayer(int p);

         void NewStart(TourState tState);

         void CleanSeed(void);

         void PopulationCount(void);									// Total counts over all players
         bool UpdatePopulationCount(GPlayer& a, GPlayer& b);	// Incremental update
         void SetState(void);												// State change due to Reg/Draw/Result edit

         GPlayerPtr GetPlayerPtr(int p);           // Access sorted list

         bool ActivePlayer(int p);						// At least one result or RegClosed
         bool PlayingThisRound(p);						// False if sleeping

         string Number(GPlayerPtr P);					// Player ID
         string Registered(GPlayerPtr P);          // R or empty
         string Name(GPlayerPtr P);                // Display Given Middle Family form
         string Family(GPlayerPtr P);              // Display Family,Given Middle form
         string ShortName(GPlayerPtr P);      		// G Family
         string ShortNameGrade(GPlayerPtr P);      // G Family(Grade)
         string ShortNameMMS(GPlayerPtr P);        // G Family(MMS)
         string ShortLandClub(GPlayerPtr P);			// CC - Club Name 
         string Grade(GPlayerPtr P);               // Display converted grade
         string Club(GPlayerPtr P);                // Display club name
         string Land(GPlayerPtr P);                // Display country name
         string Rounds(GPlayerPtr P);              // Rounds played
         string Group(GPlayerPtr P);               // Trimmed group
         string Note(GPlayerPtr P);                // Player Note - any ascii
         string Result(GPlayerPtr P);              // Current round result

         string GameNote(GBoard Board);            // Pairing Comment
         string Komi(GBoard Board);                // Stones+Points

         string TourNameYear();							// Name - Year

         void ClearTourInfo()                      // Cleared at start of each action
         { InfoID = 0; }

         bool EnabledCommand(int cmd);					// true if command enabled in current state

         void PostTourInfo(int message)            // Show message on completion
         { if(!InfoID) InfoID = message; }         // of some action

         int TourInfo() { return InfoID; }         // Identify message in TourInfoRES

		private:
      	bool IncPops(GPlayer& Player);
         bool DecPops(GPlayer& Player);
         bool ChangePops(GPlayer& Pa, GPlayer& Pb);

      public:
      	void ChangeRegPops(GPlayer& Pa, GPlayer& Pb);
   };

   extern const char* GroupCodes;                  // CODES for enum Group

   extern const char* RegCode[4];                  // CODE for Registered

   extern GTourMaster GoTour;

   extern const int ActiveCommandMasks[cmdActiveCount];
   extern const int StateMatrix[Tours::tsCount][Tours::tsCount];
   extern const bool EnabledCommands[cmdCount][Tours::tsCount];
}

namespace TourMaster                   // STRING-GTOUR INTERFACE TOOLS
{
   using std::string;
   using std::vector;

   using Tools::ints;

   bool parse_year(string& yrtext);
   bool parse_tourname(string& ntext);

   bool parse_komi(string skomi, int& stones, int& points);
   
   bool is_result(char key);
}
//---------------------------------------------------------------------------
#endif
