//----------------------------------------------------------------------------------------------------------------------
// echamp_sim	Simulate results for a given pairing 										geoff kaniuk	11 Dec 2010
//----------------------------------------------------------------------------------------------------------------------
//
// We are given a pairing as a tab-delimited text file in the form:
//
// #BOARD BLACK(GRADE) WHITE(GRADE) RESULT
//  172 John Smith (3d) - Pete Jones(4d) -
//
// The output file has the result tacked onto the end of whatever is the input so:
//  172 John Smith(3d) - Pete Jones(4d) -   [1-0]
//
// To get the result we use the following algorithm:
//
//
// VERSION1
// 1. Load players' simplified gor file
// 2. Load pairing
// 3. get player's GOR
// 4. Convert GoR to zero shodan scale
// 5. Use pwin to generate the result - no jigos
//
//
// VERSION2
// 1. Load pairing
// 1. Identify the player in EGD
// 2. Get the players GoR
// 3. Convert GoR to zero shodan scale
// 3. Use pwin to generate the result - no jigos
//
// If some players cannot be identified, their GoR is derived from their Grade, and a list of unidentified is produced
// on an output file
//
//----------------------------------------------------------------------------------------------------------------------

#include <math.h>

#include "tokens.h"
#include "egd_error.h"
#include "egd.h"

//----------------------------------------------------------------------------------------------------------------------
// MAGIC
//
namespace EGD
{
	using std::istringstream;

	using std::list;

	using std::cout;
	using std::endl;

	bool extract_grade(const string& grade, int& zgrade)
	{
		string::size_type ibra=grade.find(s_bra), iket=grade.find(s_ket);	// Format is like (33k)

		if(ibra == string::npos || iket == string::npos ) return false;		// Cannot find brackets
		if(iket<ibra+2 )return false;										// Bad grade

		istringstream sgrade(grade.substr(ibra+1, iket-1));					// Peel brackets

		string gclass;	sgrade>>zgrade>>gclass;								// Grade value and class

		switch(gclass[0])													// We want grade value in zero shodan units
		{
			case c_kyu: case c_Kyu: zgrade = -zgrade;   break;
			case c_dan: case c_Dan: zgrade =  zgrade-1; break;
			case c_pro: case c_Pro: zgrade =  zgrade+6; break;

			default: return false;
		}

		return true;
	}

	static const real pwin_u[4] = { 0.057186500, 0.126337000, 0.0, 0.0139085 };
	static const real pwin_v[4] = { 0.002713730, 0.002390260, 0.0, 0.0000040 };
	static const real pwin_K    =   0.204062;

	int simulate_game(GRand* rgen, const real bstrong, const real wstrong, real& pwin, real& psim)	// Generate result from strengths
	{
		real r = bstrong, s = wstrong, q, qq ;

		q = wstrong-bstrong; qq = q*q;

		real lambda=0.0;

		for(int n=0; n<4; n++)
		{
			const real cq = (pwin_u[n] +pwin_v[n]*qq)*q;

			const real nkrs = n*pwin_K *(q>=0 ? r : s);

			lambda +=  cq*exp(nkrs);
		}


		pwin = 0.5*(1 - erf(lambda));					// Probability of win by r in range 0 <= pwin < 1

		psim = g_rand_double(rgen);						// Random in range 0 <= psim <1

		return psim <= pwin ? 1 : 0;
	}
}

namespace EGD
{
	Toursim::Toursim()
	{
		pairinglist = list<Pair>();
	}

	Toursim::~Toursim()
	{
		pairinglist.clear();
	}

	bool Toursim::parse_pair(string& line, real& bstrength, real& wstrength)
	{
		// FORMAT IS TAB DELIMITED BETWEEN PAIRS
		// 172.John.Smith.(3d) ~-~ Pete.Jones.(4d)~-

		std::istringstream pstream(line, std::ios::in);

		if(!pstream.good()) return false;

		// DISECT THE LINE
		int board;

		string bgiven, bfamily, bgrade, wgiven, wfamily, wgrade, sep;

		pstream>>board>>bgiven>>bfamily>>bgrade>>sep>>wgiven>>wfamily>>wgrade;

		// WE WANT GRADES IN CASE OF MISSING IDENTITY
		int brank, wrank, bgor, wgor;
		if(!extract_grade(bgrade, brank)) return false;
		if(!extract_grade(wgrade, wrank)) return false;

		// RATE THE PLAYERS FROM RATING LIST OR JUST ACCEPT GRADE
		if(find_rating(bgiven, bfamily, bgor))
			bstrength = (bgor - 2100)/100.0;
		else
		{
			bstrength = brank;

			EGD::egd_message(e_ublack, board);
		}

		if(find_rating(wgiven, wfamily, wgor))
			wstrength = (wgor - 2100)/100.0;
		else
		{
			wstrength = wrank;

			EGD::egd_message(e_uwhite, board);
		}

		// REBUILD THE LINE
		line.clear();

		std::ostringstream rstream(line, std::ios::out);

		rstream<<board<<tab;
		rstream<<bgiven<<EGD::s_space<<bfamily<<tab<<bgrade;
		rstream<<tab;
		rstream<<wgiven<<EGD::s_space<<wfamily<<tab<<wgrade;

		line = rstream.str();

		return true;
	}

}

namespace EGD
{
	int simulate_european(const char* tour, const char* ratefile,  const char* pairfile, const char* resultfile)
	{
		using std::setprecision;

		using EGD::tab;

		// LOAD THE TOURNAMENT SPECIFIC RATINGS FILE
		Toursim eusim;	if(!eusim.load_tour_ratings(ratefile)) return -1;

		//eusim.show_ratings();

		// OPEN PAIRING FILE
		std::ifstream pstream(pairfile, std::ios::in);

		if(!pstream.good())return EGD::egd_error(e_open, pairfile, -2);

		// SKIP HEADER TO FIRST NON EMPTY LINE NOT STARTING WITH #
		string line; int lcount = 0;

		while(EGD::get_next_line(pstream, lcount, line))
			if(!line.empty())
				if(line[0] != EGD::c_hash)
					break;
		if( lcount<3 || !pstream.good())
		{
			pstream.close(); return EGD::egd_error(EGD::e_heading, pairfile, -4);
		}

		// OPEN RESULT FILE AND PLANT HEADING
		std::ofstream qstream(resultfile, std::ios::out);

		if(qstream.good())
		{
			qstream<<EGD::hash<<EGD::s_title<<pairfile<<endl;
			qstream<<EGD::hash<<tab<<EGD::s_black<<tab<<tab<<EGD::s_white<<tab<<tab;
			qstream<<EGD::s_bstreng<<tab<<EGD::s_wstreng<<tab<<EGD::s_pwin<<tab<<EGD::s_psim<<tab<<EGD::s_result;
			qstream<<endl;
		}
		else
		{
			pstream.close(); return EGD::egd_error(e_open, resultfile, -5);
		}

		// RANDOM NUMBER SEED - DIFFERENT FOR EACH ROUND IN EACH TOURNAMENT
		int seed=0, i=0;

		if(tour)
			while(tour[i]) seed += (int)tour[i++];

		i=0; while(pairfile[i]) seed += (int)pairfile[i++];

		GRand* rgen = g_rand_new_with_seed(seed);

		// PROCESS PAIRING LINE BY LINE
		int report = 0;

		while(pstream.good())
		{
			// EXTRACT PLAYING STRENGTHS
			Appmaths::real bstrong, wstrong, pwin,psim;

			if(!eusim.parse_pair(line, bstrong, wstrong))
			{
				report = egd_error(EGD::e_pair, line.c_str(), -6); 	break;
			}

			// GET RESULT, APPEND TO LINE, AND WRITE TO FILE
			const int bresult = simulate_game(rgen, bstrong, wstrong, pwin, psim);

			if(qstream.good())
			{
				qstream <<line<<tab
						<<std::fixed
						<<setprecision(1)<<bstrong
						<<tab<<setprecision(1)<<wstrong
						<<tab<<setprecision(2)<<pwin
						<<tab<<setprecision(2)<<psim
						<<tab<<EGD::s_bral<<bresult<<EGD::s_hyphen<<1-bresult<<EGD::s_ketl
						<<endl;
			}
			else
			{
				report = egd_error(EGD::e_result, line.c_str(), -7); break;
			}

			// GET NEXT LINE FROM FILE - SKIPPING ANY BLANKS
			while(pstream.good())
			{
				if(!get_next_line(pstream, lcount, line))break;

				if(!line.empty()) break;
			}
		}
		if(!qstream.good())
			if(!report)
			{
				report = egd_error(EGD::e_result, resultfile, -8);

				EGD::show_line(lcount, line);
			}
		qstream.close();
		pstream.close();

		return report;
	}
}
// END echamp_sim
//----------------------------------------------------------------------------------------------------------------------
