////////////////////////////////////////////////////////////////////////////////
// MODULE PrinterAPI: Printer Interface to the Windows API     G.Kaniuk 12/03/02
//

#define STRICT

#include <windows.h>
#include <vcl.h>
#include <algorithm>

#pragma hdrstop

#include "PrinterRES.h"

#include "Maths.h"
#include "ETour.h"

#include "PrinterAPI.h"

extern ToolsGUI::ErrorLog AppLog;

/////////////////////////////////////////////////////DONE///////////////////////
// Weird
// Strange behaviour with SelFont & DelFont originally named SelectFont and
// DeleteFont which  happen to be API macros.  Compile fine but handles seem
// to become invalid
//

namespace PrinterAPI
{
   using ToolsGUI::EType;
   using ToolsGUI::erBug;

   /////////////////////////////////////////////////////////////////////////////
   // Special characters

   const char pr_quarter       = (char)0xbc;
   const char pr_half          = (char)0xbd;
   const char pr_threequarters = (char)0xbe;

   /////////////////////////////////////////////////////////////////////////////
   // Printer Interface

   const static UINT pr_align[] =
   {
      DT_TOP     | DT_LEFT,  DT_TOP     | DT_CENTER,  DT_TOP     | DT_RIGHT,
      DT_VCENTER | DT_LEFT,  DT_VCENTER | DT_CENTER,  DT_VCENTER | DT_RIGHT,
      DT_BOTTOM  | DT_LEFT,  DT_BOTTOM  | DT_CENTER,  DT_BOTTOM  | DT_RIGHT
   };

   static PRINTDLG     pr_config;                          // Initialised at Startup, changed by user
   static String       pr_fontname;                        // Name of an enumerated font

   static bool         pr_ready    = false;                // Set true by CreatePrinter
   static bool         pr_started  = false;                // Managed by Start_ End_ Printer()
   static int          pr_fontsize = 11;                   // Font size in points

   static bool FontExists(HDC DeviceHandle, String FontName);

   /////////////////////////////////////////////////////////////////////////////
   // PrintDlg Errors

   const int pr_dlgerrcount = 13;

   const DWORD pr_dlgerrcode[pr_dlgerrcount] =
   {
      PDERR_NODEFAULTPRN,                                                  // No default printer

      PDERR_LOADDRVFAILURE,   PDERR_GETDEVMODEFAIL, PDERR_NODEVICES,       // Driver problems

      PDERR_PARSEFAILURE,     PDERR_PRINTERNOTFOUND,                       // win.ini problems

      PDERR_SETUPFAILURE,     PDERR_RETDEFFAILURE,    PDERR_INITFAILURE,   // Software
      PDERR_DNDMMISMATCH,     PDERR_CREATEICFAILURE,  PDERR_DEFAULTDIFFERENT,

      PDERR_PRINTERCODES
   };

   void Throw(EType etype, short idproc, short idmsg)
   {
      using ToolsGUI::Forge;

      if(pr_started && pr_config.hDC != NULL)
      {
         DeleteDC(pr_config.hDC); pr_config.hDC = NULL;

         pr_started = false;
      }

      std::string message = Forge(etype, IDP_PrinterAPI, idproc, idmsg);

      if(etype==erBug)::AppLog.Add(message);

      throw ToolsGUI::ETour(etype, message);
   }

   void ThrowDlgErr(short routine, DWORD dlgerr)
   {
      using ToolsGUI::erOperr;

      EType etype;   short cause;

      int err = 0;

      while(err<pr_dlgerrcount)
      {
         if(dlgerr == pr_dlgerrcode[err]) break;

         ++err;
      }

      while(err>=0)
      {
         if(err ==  0 ){ etype = erOperr;  cause = IDP_NoDefault;    break; }
         if(err <=  3 ){ etype = erOperr;  cause = IDP_NoDriver;     break; }
         if(err <=  5 ){ etype = erOperr;  cause = IDP_BadWin;       break; }
         if(err <= 11 ){ etype = erBug;    cause = IDP_PrDlgSoft;    break; }
         if(err == 12 ){ etype = erBug;    cause = IDP_PrDlgUnknown; break; }

                         etype = erBug;    cause = IDP_PrDlgOther;   break;
      }

      Throw(etype, routine, cause);
   }



   /////////////////////////////////////////////////////////////////////////////
   // SelectFontName
   //
   // Called by StartNewFont it sets up the font name only
   //
   // If the supplied Font->Name is empty:
   //    It tries to initialise with Arial;
   //    Then with Times New Roman;
   //    And lastly with the font that comes with pr_config
   //
   // If the supplied Font->Name is not empty:
   //    We assume that the Font Name has been set up from an ini file.
   //    We test that the font exists.

   void SetPrinterFont(String name)
   {
      pr_fontname = name;
   }

   void SelectFontName(String* FontName)
   {
      bool FontChosen;

      String s; int i,j;

      for(i=1; i<=64; i++)
      {
      	j= i+IDP_START;
      	s = LoadStr(j);

         if(s.IsEmpty()) break;
      }

      if(FontName->IsEmpty())                                            // A VIRGIN START
      {
      	String IDPfont = String();

         // First: try Arial
         IDPfont = LoadStr(IDP_Arial);

         FontChosen = FontExists(pr_config.hDC, IDPfont);

         if(FontChosen)
         {
            *FontName = LoadStr(IDP_Arial);

             return;
         }

         // Second: try Times New Roman
         IDPfont = LoadStr(IDP_Times);

         FontChosen = FontExists(pr_config.hDC, IDPfont);

         if(FontChosen)
         {
            *FontName = LoadStr(IDP_Times);

            return;
         }

         // Third: set to font supplied with the pr_config
         {
            TCHAR Name[LF_FACESIZE];  // TCHAR is wide char for UNICODE, otherwise unsigned char

            if(!GetTextFace(pr_config.hDC, LF_FACESIZE, (LPTSTR)Name))
               Throw(erBug, IDP_SelectFontName, IDP_BadTextFace);

            *FontName = String(Name);   return;
         }
      }
      else                                                              // RESTART FROM INI FILE
      {
         FontChosen = FontExists(pr_config.hDC, *FontName);

         if(!FontChosen)
         {
            *FontName = String();                                             // Start from scratch

            SelectFontName(FontName);
         }
      }
   }

   /////////////////////////////////////////////////////////////////////////////
   // GetPrinterFontName
   //
   // We need to retrieve the current device font whenever we wish to change it
   //

   String GetCurrentPrinterFontName()
   {
      TCHAR FaceName[LF_FACESIZE];

      if(GetTextFace(pr_config.hDC, LF_FACESIZE, (LPTSTR)FaceName))
         return String(FaceName);
      else
         return String();
   }

   String GetPrinterFontName()
   {
      return pr_fontname;
   }


   static void GetPrinterFont(LOGFONT* font)
   {
      TCHAR FaceName[LF_FACESIZE];

      TEXTMETRIC tmet;

      String FontName;

      if(!GetTextFace(pr_config.hDC, LF_FACESIZE, (LPTSTR)FaceName))
         Throw(erBug, IDP_GetPrinterFont, IDP_BadTextFace);

      if(!GetTextMetrics(pr_config.hDC, (LPTEXTMETRIC)(&tmet)))
         Throw(erBug, IDP_GetPrinterFont, IDP_BadMetrics);

      ZeroMemory(font, sizeof(LOGFONT));

      for(int i=0; i<LF_FACESIZE; i++)font->lfFaceName[i] = FaceName[i];

      FontName = String(font->lfFaceName);

      font->lfHeight = tmet.tmHeight;
      font->lfWidth  = tmet.tmAveCharWidth;
      font->lfWeight = tmet.tmWeight;

      font->lfItalic    = tmet.tmItalic;
      font->lfUnderline = tmet.tmUnderlined;
      font->lfStrikeOut = tmet.tmStruckOut;

      font->lfCharSet   = tmet.tmCharSet;

      font->lfPitchAndFamily = tmet.tmPitchAndFamily;
   }


   /////////////////////////////////////////////////////////////////////////////
   // CreatePrinter
   //
   // Initialises the following for the default printer at startup:
   //
   // PRINTDLG    pr_config
   // String      pr_fontname
   //
   // It is only ever called once, and does not show a dialogue
   //
   // It calls SelectFontName to set up a suitable font.  The font
   // name parameter comes from an external ini file or is empty
   // in which case pr_fontname is set internaly
   //

   void CreatePrinter(String FontName)
   {
      if(pr_ready) Throw(erBug, IDP_CreatePrinter, IDP_BadInit);

      ZeroMemory(&pr_config, sizeof(PRINTDLG));

      pr_config.lStructSize = sizeof(PRINTDLG);

      pr_config.Flags = PD_RETURNDEFAULT|PD_RETURNDC;

      if(PrintDlg(&pr_config))
      {
         pr_fontname = FontName;

         SelectFontName(&pr_fontname);

         EndPrinter();

         pr_config.Flags &= ~PD_RETURNDEFAULT;   // restored to zero
         pr_config.Flags &= ~PD_RETURNDC;        // restored to zero

         pr_ready = true;
      }
      else
         ThrowDlgErr(IDP_CreatePrinter, CommDlgExtendedError());
   }

   void DeletePrinter()
   {

   }


   /////////////////////////////////////////////////////////////////////////////
   // SetupPrinter
   //
   // Called by user.  Shows dialog box

   bool SetupPrinter()
   {
      if(!pr_ready)Throw(erBug, IDP_SetupPrinter, IDP_BadInit);

      if(pr_config.hDevNames == NULL && pr_config.hDevMode == NULL)
         Throw(erBug, IDP_SetupPrinter, IDP_NoNamesMode);

      pr_config.Flags &= ~PD_RETURNDC;
      pr_config.Flags |= PD_PRINTSETUP;

      if(PrintDlg(&pr_config))
         return true;
      else
      {
         DWORD Err = CommDlgExtendedError();

         if(Err==0) return false;   // Dialogue cancelled

         ThrowDlgErr(IDP_SetupPrinter, Err);
      }
      return false;
   }


   /////////////////////////////////////////////////////////////////////////////
   // StartNewPrinter
   //
   // Called by user.  Shows dialog box which can be cancelled
   //

   bool StartNewPrinter()
   {
      if(!pr_ready)Throw(erBug, IDP_StartNewPrinter, IDP_BadInit);

      if(pr_config.hDevNames == NULL && pr_config.hDevMode == NULL)
         Throw(erBug, IDP_StartNewPrinter, IDP_NoNamesMode);

      pr_config.Flags |= PD_RETURNDC;
      pr_config.Flags |= PD_HIDEPRINTTOFILE;
      pr_config.Flags |= PD_USEDEVMODECOPIES;

      pr_config.nCopies = 1;

      if(PrintDlg(&pr_config))
      {
         SelectFontName(&pr_fontname);      // Maybe changed printer

         return true;
      }
      else
      {
         DWORD Err = CommDlgExtendedError();

         if(Err==0)return false;                   // Dialogue cancelled

         ThrowDlgErr(IDP_StartNewPrinter, Err);
      }
      return false;
   }

   /////////////////////////////////////////////////////////////////////////////
   // StartPrinter
   //
   // Called by Application.  Does not display dialog box.

   void StartPrinter()
   {
      SetLastError(0);

      if(!pr_ready || pr_config.hDevNames == NULL)     // Initialised by CreatePrinter
         Throw(erBug, IDP_StartPrinter, IDP_BadInit);  // or StartNewPrinter

      if(pr_config.hDC != NULL)                        // So missing EndPrinter call
          Throw(erBug, IDP_StartPrinter, IDP_BadIniDC);

      // We need only the Device name from  DEVNAMES

      LPDEVNAMES pDevNames = (LPDEVNAMES)GlobalLock(pr_config.hDevNames);

            if(pDevNames == NULL)Throw(erBug, IDP_StartPrinter, IDP_NoDevnames);

      LPCTSTR Device = (LPCTSTR)&((LPCSTR)pDevNames)[pDevNames->wDeviceOffset];

      GlobalUnlock(pr_config.hDevNames);

            if(GetLastError())Throw(erBug, IDP_StartPrinter, IDP_NoDevnames);


      // We need the whole of DEVMODE if it is non null

      LPDEVMODE pDevMode = NULL;

      if(pr_config.hDevMode != NULL)
      {
         pDevMode = (LPDEVMODE)GlobalLock(pr_config.hDevMode);

               if(pDevMode == NULL)Throw(erBug, IDP_StartPrinter, IDP_NoDevmode);

         GlobalUnlock(pr_config.hDevMode);

               if(GetLastError())Throw(erBug, IDP_StartPrinter, IDP_NoDevmode);
      }


      // Now finaly we get to create the DC

      pr_config.hDC = CreateDC(NULL, Device, NULL, pDevMode);

            if(pr_config.hDC == NULL)Throw(erBug, IDP_StartPrinter, IDP_BadDC);

      SelectFontName(&pr_fontname);      // Maybe changed printer

      pr_started = true;
   }


   /////////////////////////////////////////////////////////////////////////////
   // EndPrinter
   //
   // Termination of print process

   void EndPrinter()
   {
      if(!pr_config.hDC)
         Throw(erBug, IDP_EndPrinter, IDP_BadIniDC);

      if(DeleteDC(pr_config.hDC))
         pr_config.hDC = NULL;
      else
         Throw(erBug, IDP_EndPrinter, IDP_BadDC);

      pr_started = false;
   }

   /////////////////////////////////////////////////////////////////////////////
   // StartDoc and EndDoc
   //

   void StartDoc(char* FileName)
   {
      DOCINFO docinfo;

      ZeroMemory(&docinfo, sizeof(DOCINFO));

      docinfo.cbSize = (int)sizeof(DOCINFO);

      docinfo.lpszDocName = (LPCTSTR)&((LPCSTR)FileName)[0];

      docinfo.lpszOutput = NULL;

      if(::StartDoc(pr_config.hDC, &docinfo) > 0)
         return;
      else
         Throw(erBug, IDP_StartDoc, IDP_BadPrDoc);
   }


   void EndDoc()
   {
      if(::EndDoc(pr_config.hDC) > 0)
         return;
      else
         Throw(erBug, IDP_EndDoc, IDP_BadPrDoc);
   }

   /////////////////////////////////////////////////////////////////////////////
   // StartPage and EndPage
   //

   void StartPage()
   {
      if(::StartPage(pr_config.hDC) > 0)
         return;
      else
         Throw(erBug, IDP_StartPage, IDP_BadPrDoc);
   }

   void EndPage()
   {
      if(::EndPage(pr_config.hDC) > 0)
         return;
      else
         Throw(erBug, IDP_EndPage, IDP_BadPrDoc);
   }


   /////////////////////////////////////////////////////////////////////////////
   // CLASS PrSheet IMPLEMENTATION
   //

   static void SplitStyle
   (
      PrFontStyle style,
      int* weight,
      BYTE* italic, BYTE* uline, BYTE* strike
   )
   {
      *italic = (BYTE)(style & pfItalic);
      *uline  = (BYTE)(style & pfUline);
      *strike = (BYTE)(style & pfStrike);

      *weight = style & pfBold ? FW_BOLD : FW_NORMAL;
   }

   static PrFontStyle PackStyle(int weight, BYTE italic, BYTE uline, BYTE strike)
   {
      PrFontStyle style = pfNormal;

      if(italic) style = (PrFontStyle)(style | pfItalic);
      if(uline)  style = (PrFontStyle)(style | pfUline);
      if(strike) style = (PrFontStyle)(style | pfStrike);

      if(weight>= FW_BOLD) style = (PrFontStyle)(style | pfBold);

      return style;
   }


   __fastcall PrSheet::PrSheet()
   {
      // ATTACH DC TO PrSheet

      hDevice = pr_config.hDC;          // The handle created at application start

      if(hDevice == NULL)Throw(erBug, IDP_PrSheet, IDP_BadDC);

      // SET PAGE METRICS

      Hscale  = GetDeviceCaps(hDevice, LOGPIXELSX);
      Vscale  = GetDeviceCaps(hDevice, LOGPIXELSY);

      Hoffset = GetDeviceCaps(hDevice, PHYSICALOFFSETX);
      Voffset = GetDeviceCaps(hDevice, PHYSICALOFFSETY);

      Harea   = GetDeviceCaps(hDevice, HORZRES);
      Varea   = GetDeviceCaps(hDevice, VERTRES);

      Hsize   = GetDeviceCaps(hDevice, PHYSICALWIDTH);
      Vsize   = GetDeviceCaps(hDevice, PHYSICALHEIGHT);

      // PAGE MARGINS WITHIN THE PRINTABLE PAGE AREA
      {
         using Tools::real;
         using Tools::round;

         real Hmarg = (real)Harea/20,  Vmarg = (real)Varea/30;

         real Htext = Harea - 2*Hmarg, Vtext = Varea - 2*Vmarg;

         Lmargin = round(Hmarg);  Rmargin = round(Hmarg + Htext);
         Tmargin = round(Vmarg);  Bmargin = round(Vmarg + Vtext);

         if(Lmargin < Hoffset)         Lmargin = Hoffset;
         if(Rmargin > Hoffset + Harea) Rmargin = Hoffset + Harea;

         if(Tmargin < Voffset)         Tmargin = Voffset;
         if(Bmargin > Voffset + Varea) Bmargin = Voffset + Varea;
      }
      if(Hoffset + Harea > Hsize || Voffset + Varea > Vsize)
         Throw(ToolsGUI::erOperr, IDP_PrSheet, IDP_BadDriver);


      // CREATE THE FONT NAMED IN pr_fontname with standard font height

      GetPrinterFont(&LogFont);

      int SizeName = pr_fontname.Length(), i;

      for(i=0; i<SizeName; i++)LogFont.lfFaceName[i] = pr_fontname[i+1];
      for(i=SizeName; i<LF_FACESIZE; i++) LogFont.lfFaceName[i] = 0;

      LogFont.lfHeight = (pr_fontsize*Vscale)/72;

      hStdFont = CreateFontIndirect(&LogFont);
                 if(!hStdFont)Throw(erBug, IDP_PrSheet, IDP_BadFont);

      hDefFont = (HFONT)SelectObject(hDevice, (HGDIOBJ)hStdFont);
                 if(hDefFont == NULL)Throw(erBug, IDP_PrSheet, IDP_BadFont);


      // FONT METRICS

      hCurFont = hStdFont;             // Start with Standard font

      SetFontMetrics();                // Height width style

		StdFontWidth  = FontWidth;			// Basic reference
      StdFontHeight = FontHeight;
   }

   __fastcall PrSheet::~PrSheet()
   {
      HFONT hcurfont = (HFONT)GetCurrentObject(hDevice, OBJ_FONT);

      if(hcurfont != hStdFont)      // We need to retrieve the StdFont
      {
         hcurfont = (HFONT)SelectObject(hDevice, (HGDIOBJ)hStdFont);

         DeleteObject((HGDIOBJ)hcurfont);
      }

      // Delete the StdFont

      hStdFont = (HFONT)SelectObject(hDevice, (HGDIOBJ)hDefFont);

      DeleteObject((HGDIOBJ)hStdFont);
   }


   /////////////////////////////////////////////////////////////////////////////
   // FONT MANAGEMENT

   void __fastcall PrSheet::DelFont(HFONT hparked)
   {
      if(hparked == hStdFont || hparked == hDefFont)
         Throw(erBug, IDP_DelFont, IDP_BadFont);

      if(hparked == (HFONT)GetCurrentObject(hDevice, OBJ_FONT))
         if(SelectObject(hDevice, (HGDIOBJ)hStdFont)==NULL)
            Throw(erBug, IDP_DelFont, IDP_BadFont);

      if(!DeleteObject((HGDIOBJ)hparked))
         Throw(erBug, IDP_DelFont, IDP_BadFont);
   }

   void __fastcall PrSheet::SelFont(HFONT hparked)
   {
      if(hparked == (HFONT)GetCurrentObject(hDevice, OBJ_FONT))return;

      if(SelectObject(hDevice, (HGDIOBJ)hparked)==NULL)
         Throw(erBug, IDP_SelFont, IDP_BadFont);

      hCurFont = hparked;

      SetFontMetrics();
   }

   void __fastcall PrSheet::NewFont(int height, int width, PrFontStyle style)
   {
      if(width<1 || width>Harea ||height<1 || height>Varea)
         Throw(erBug, IDP_NewFont, IDP_BadFontPar);

      BYTE italic, underline, strikeout; int weight;

      SplitStyle(style, &weight, &italic, &underline, &strikeout);

      // Use the LOGFONT created at PrSheet::construction.
      // Change only the required parameters

      LogFont.lfWidth     = (LONG)width;
      LogFont.lfHeight    = (LONG)height;

      LogFont.lfWeight    = (LONG)(weight);

      LogFont.lfItalic    = italic;
      LogFont.lfUnderline = underline;
      LogFont.lfStrikeOut = strikeout;


      // Create and select the new font

      HFONT hnewfont = CreateFontIndirect(&LogFont);
                        if(!hnewfont)Throw(erBug, IDP_NewFont, IDP_BadFont);

      HFONT holdfont = (HFONT)SelectObject(hDevice, (HGDIOBJ)hnewfont);


      // Get the actual font obtained

      hCurFont = hnewfont;

      SetFontMetrics();


      // Delete the old font - holdfont != hStdFont because of StartPage!

      if(holdfont != hStdFont)
         if(!DeleteObject((HGDIOBJ)holdfont))
            Throw(erBug, IDP_NewFont, IDP_BadFont);
   }

   void __fastcall PrSheet::SetFontMetrics()
   {
      TEXTMETRIC tmet;

      if(!GetTextMetrics(hDevice, (LPTEXTMETRIC)(&tmet)))
         Throw(erBug, IDP_SetFontMetrics, IDP_BadFont);

      LONG Weight    = tmet.tmWeight;
      BYTE Italic    = tmet.tmItalic;
      BYTE Underline = tmet.tmUnderlined;
      BYTE Strikeout = tmet.tmStruckOut;

      FontWidth  = tmet.tmAveCharWidth;
      FontHeight = tmet.tmHeight;
      FontStyle  = PackStyle(Weight, Italic, Underline, Strikeout);
   }

   void __fastcall PrSheet::ResetStdFontSize(PrFontStyle style)
   {
   	NewFont(StdFontHeight, StdFontWidth, style);
   }

   /////////////////////////////////////////////////////////////////////////////
   // LAYOUT TOOLS

   void __fastcall PrSheet::GetTextArea(string& text, int& width, int& height)
   {
      SIZE textcell;

      width = 0;  height = 0;

      if(!::GetTextExtentPoint32(hDevice, text.c_str(), text.size(), (LPSIZE)(&textcell)))
      	return;

      width = (int)textcell.cx;  height = (int)textcell.cy;
   }

   int __fastcall PrSheet::DrawTextCell(TRect area, PrAlign posn, std::string text)
   {
   	int wtext, htext; GetTextArea(text, wtext,htext);

      int warea = area.right - area.left;

      if(wtext>warea && FontWidth>0)
      {
      	int strlen = text.size();

      	int excess = (wtext-warea)/FontWidth;

         if(strlen - (excess+3) > 3)
         	text = text.substr(0, strlen-excess-3) + string("...");
         else
         	text = text.substr(0, strlen-excess);
      }

      RECT Area;

      Area.top = (LONG)area.Top;     Area.bottom = (LONG)area.Bottom;
      Area.left = (LONG)area.Left;   Area.right  = (LONG)area.Right;

      UINT Format = DT_NOPREFIX | DT_SINGLELINE | pr_align[posn];


      return ::DrawText
      (
         hDevice,
         text.c_str(), text.size(),
         (LPRECT)&Area,
         Format
      );
   }

   static bool terminator(char c)
   {
      return c == '\t' || c == '\0';
   }

   void __fastcall PrSheet::TextLineOut
   (
      int cellheight,            // Height of a text cell
      int row,                   // Row position in dots
      std::vector<int> col,      // Col positions in dots
      std::string text           // Tab starts a new column
   )
   {
      using std::string;

      int fields = col.size()-1, f = 0;  // Number of separate fields

      string::iterator now = text.begin(), end = text.end(), iterm;

      TRect area;

      area.Top = row;   area.Bottom = row + cellheight - 1;

      while(now != end)
      {
         iterm = std::find_if(now, end, terminator);

         string istring = string(now, iterm);

         area.Left = col[f];  area.Right = col[f+1] - 1;

         DrawTextCell(area, paTopLeft,  istring);

         now = iterm;

         if(now != end) { ++f;  ++now; };

         if(f>=fields) break;
      }
   }

   void __fastcall PrSheet::MoveTo(int i, int j)
   {
      (void)::MoveToEx(hDevice, i, j, NULL);
   }

   void __fastcall PrSheet::LineTo(int i, int j)
   {
      (void)::LineTo(hDevice, i, j);
   }

   void __fastcall PrSheet::DrawTick(TRect& area)
   {
   	MoveTo(area.left, (area.top + area.bottom)/2);
      LineTo((area.left + area.right)/2, area.bottom);
      LineTo(area.right, area.top);
   }

   static POINT MakePoint(int i, int j)
   {
      POINT p;

      p.x = i;  p.y = j;

      return p;
   }

   void __fastcall PrSheet::FrameRect(int is, int js, int ie, int je)
   {
      POINT cnr[5];

      cnr[0] = MakePoint(is, js);    cnr[1] = MakePoint(ie, js);
      cnr[2] = MakePoint(ie, je);    cnr[3] = MakePoint(is, je);
      cnr[4] = MakePoint(is, js);

      (void) ::Polyline(hDevice, cnr, 5);
   }

   void __fastcall PrSheet::DrawHalf(int is, int js, int w, int h)
   {
      int OldHeight = FontHeight, OldWidth = FontWidth;

      PrFontStyle OldStyle = FontStyle;

      int hchar = (4*h)/10, wchar = w/2;

      NewFont(hchar, wchar, pfNormal);    // Font for numerator and denominator

      TRect Cell;

      Cell.Left = is; Cell.Right  = is + w - 1;
      Cell.Top  = js; Cell.Bottom = js + h - 1;

      DrawTextCell(Cell, paTopMid, "1");
      DrawTextCell(Cell, paBotMid, "2");

      NewFont(OldHeight, OldWidth, OldStyle);

      DrawTextCell(Cell, paCentre, "-");
   }

   int __fastcall PrSheet::BatchesPerPage
   (
   	const int uFont, const int lFont, const int uSpace,
   	const int headHeight
   )
   {
   	// CURRENT FONT IS SET FOR LINE PRINTING
      const int ActiveHeight = Bmargin - Tmargin;

      const int LinesHeight = ActiveHeight - headHeight;
      					if(LinesHeight < 2*FontHeight) return 0;

      const int LineHeight  = 5*((uFont*FontHeight)/lFont);
      const int SpaceHeight = (uSpace*FontHeight)/lFont;

      // We solve B*LineHeight + (B-1)*SpaceHeight = LinesHeight
      // for the BatchCount B

      return  (SpaceHeight + LinesHeight)/(LineHeight+SpaceHeight);
   }


   /////////////////////////////////////////////////////////////////////////////
   // FontExists
   //
   // Uses EnumerateFontFamilies to test whether a given fontname actually
   // exists for the device context
   //
   // Looks for variable pitch fonts

   static int CALLBACK FontProc                          // Note that the documentation
   (                                                     // has errors in the spec for
      CONST LOGFONT* elfont, CONST TEXTMETRIC* elfmet,   // the parameter types.
      DWORD elftype,
      LPARAM elfpar                                      // These come from wingdi.h
   )
   {
      int* Count = (int*)elfpar;

      ++(*Count);    // We have a font!

      return 0;      // We can stop enumerating
   }


   static bool FontExists(HDC DeviceHandle, String FontName)
   {
      if(FontName.Length() >= LF_FACESIZE)
         Throw(erBug, IDP_FontExists, IDP_BadLength);

      LOGFONT lfont; ZeroMemory(&lfont, sizeof(LOGFONT));

      for(int i=0; i<FontName.Length(); i++)
         lfont.lfFaceName[i] = (TCHAR)FontName[i+1];

      lfont.lfCharSet = ANSI_CHARSET;
      lfont.lfPitchAndFamily = VARIABLE_PITCH | FF_DONTCARE;

      int Count = 0;

      EnumFontFamiliesEx(DeviceHandle,(LPLOGFONT)&lfont,FontProc,(LPARAM)&Count,0);

      return Count>0;
   }

   /////////////////////////////////////////////////////////////////////////////
   // ListFonts
   //
   // Uses EnumerateFontFamilies to create a list of all the printer fonts known
   // to the printer device context pr_config.hDC.
   //
   // Looks for Ansi variable pitch fonts
   //

   static int CALLBACK NextFontProc                      // Note that the documentation
   (                                                     // has errors in the spec for
      CONST LOGFONT* elfont, CONST TEXTMETRIC* elfmet,   // the parameter types.
      DWORD elftype,
      LPARAM elfpar                                      // These come from wingdi.h
   )
   {
      char FaceName[LF_FACESIZE];

      for(int i=0; i<LF_FACESIZE; i++)FaceName[i] = elfont->lfFaceName[i];

      String FontName = String(FaceName);

      TStrings* Fonts = (TStrings*)elfpar;

      Fonts->Add(FontName);

      return Fonts->Count;
   }

   int ListFonts(TStrings* Fonts)  // TStrings* usually supplied by some control
   {
      Fonts->BeginUpdate();

      Fonts->Clear();

      LOGFONT lfont; ZeroMemory(&lfont, sizeof(LOGFONT));

      lfont.lfFaceName[0] = 0;

      lfont.lfCharSet = ANSI_CHARSET;

      lfont.lfPitchAndFamily = 0;               // VARIABLE_PITCH | FF_ROMAN;

      int TotalFonts = EnumFontFamiliesEx(pr_config.hDC,(LPLOGFONT)&lfont,NextFontProc,(LPARAM)Fonts,0);

      Fonts->EndUpdate();

      return TotalFonts;
   }

   void PrTextArea::DrawText(string text, PrSheet& Sheet)
   {
   	Sheet.DrawTextCell(area, align, text);
   }
}

//---------------------------------------------------------------------------







