////////////////////////////////////////////////////////////////////////////////
// MODULE PrinterAPI: Printer Interface to the Windows API     G.Kaniuk 12/03/02
//

#define STRICT

#include <vector>
#include <windows.h>

#include "Strings.h"

#ifndef PrinterAPIH
#define PrinterAPIH

namespace PrinterAPI
{
   /////////////////////////////////////////////////////////////////////////////
   // Special characters

   using std::string;

   extern const char pr_quarter;
   extern const char pr_half;
   extern const char pr_threequarters;


   enum PrAlign
   {
      paTopLeft, paTopMid, paTopRight,
      paMidLeft, paCentre, paMidRight,
      paBotLeft, paBotMid, paBotRight
   };

   enum PrFontStyle
   {
      pfNormal  = 0x00,
      pfBold    = 0x01,     pfItalic = 0x02,
      pfUline   = 0x04,     pfStrike = 0x08
   };


   class PrSheet
   {
      private:
         LOGFONT LogFont;        // FONT MECHANICS
         HFONT hDefFont;         // Default font supplied with printer DC
         HFONT hStdFont;         // Standard font is 11 point Ariel

      public:
      	HDC hDevice;            // Handle to the device context
         HFONT hCurFont;         // The current selected font
         int StdFontWidth;			// Standard font width
         int StdFontHeight;		// Standard font height

         int FontWidth;          // Current font average character width(dots)
         int FontHeight;         // Current font character height including leading (dots)

         PrFontStyle FontStyle;  // Clear, Bold, Italic, Underline, Strikeout

      public:                    // PAGE METRICS
         int Hscale,  Vscale;    // Horizontal and Vertical scales - dots/inch.
         int Hsize,   Vsize;     // Hor, Ver paper sizes(dots)
         int Hoffset, Voffset;   // Hor, Ver offsets(dots).  Start of printable area.
         int Harea,   Varea;     // Printable area dimensions(dots)
         int Tmargin, Bmargin;   // Horizontal margin posns (dots)
         int Lmargin, Rmargin;   // Vertical margin posns(dots)

      public:
         __fastcall PrSheet();   // Creates and selects standard font
         __fastcall ~PrSheet();  // Selects default font and deletes standard

         void __fastcall NewFont(int height, int width, PrFontStyle style);   // Create and select
         void __fastcall SelFont(HFONT hparked);                              // Select a created font
         void __fastcall DelFont(HFONT font);											// Select Std and delete
         void __fastcall ResetStdFontSize(PrFontStyle style);                 // Size of

         void __fastcall TextLineOut
         (
            int cellheight,            // Height of a text cell
            int row,                   // Row top position in dots
            std::vector<int> col,      // Col positions in dots
            std::string text           // Tab starts a new column
         );

         int __fastcall DrawTextCell(TRect area, PrAlign posn,std::string text);
         void __fastcall DrawHalf(int is, int js, int w, int h);

         void __fastcall MoveTo(int i, int j);
         void __fastcall LineTo(int i, int j);
         void __fastcall DrawTick(TRect& area);
         
         void __fastcall FrameRect(int is, int js, int ie, int je);
        
         void __fastcall GetTextArea(string& text, int& width, int& height);

         int __fastcall BatchesPerPage
   		(
   			const int uFont, const int lFont, const int uSpace,
   			const int headHeight
   		);

      private:
         void  __fastcall SetFontMetrics();                                   // Sets Font height/width
   };

   class PrTextArea						// TEXT AREA METRICS
   {
   	public:
      	TRect area;						// Text field boundary

   		PrAlign align;					// Alignment within field

      public:
      	PrTextArea()
         {
         	align = paCentre; area = TRect(0,0,0,0);
         }

         PrTextArea(const int Col, const int Size, const PrAlign Align)
         {
         	align = Align;

            area = TRect(Col,0,Col+Size,0); // Col can be an offset or an absolute value
         }

         PrTextArea operator+ (const PrTextArea g)
         {
         	PrTextArea G;

            G.align = g.align;

            const int spacer = g.area.left;	// We treat g.area.left as an offset
            const int sizer  = g.area.right - g.area.left + 1;

            G.area.left  = area.right + spacer;
            G.area.right = G.area.left + sizer;

            return G;
         }

         PrTextArea displace(const int hstep, const int vstep)
         {
         	PrTextArea displaced = *this;

            displaced.area.top  += vstep;  displaced.area.bottom += vstep;
            displaced.area.left += hstep;  displaced.area.right  += hstep;

				return displaced;
         }

         void DrawText(string text, PrSheet& Sheet);
   };


   void Throw(ToolsGUI::EType etype, short idproc, short idmsg);
   
   void CreatePrinter(String FontName);
   bool SetupPrinter();
   bool StartNewPrinter();
   void DeletePrinter();
   void StartPrinter();
   void EndPrinter();

   void StartDoc(char* FileName);
   void EndDoc();

   void StartPage();
   void EndPage();

   void SelectFontName(String* FontName);
   void SetPrinterFont(String name);

   int ListFonts(TStrings* Fonts);

   String GetCurrentPrinterFontName();
   String GetPrinterFontName();
}

//---------------------------------------------------------------------------
#endif
