////////////////////////////////////////////////////////////////////////////////
//                    (C) MS Instruments plc
////////////////////////////////////////////////////////////////////////////////
//
// MODULE:   Format            		CONVERSION OF BASIC TYPES TO STRING
//
// SYSTEM:   MATH DEVELOPMENT
//
// SPEC:     MSI Maths				  ISSUE: A
//
// AUTHOR:   G Kaniuk              DATE:   20 Jul 2001
//
////////////////////////////////////////////////////////////////////////////////
//
// HISTORY:  New at creation date
//
////////////////////////////////////////////////////////////////////////////////

#include <float.h>
#include <math.h>
#include <ctype.h>
#include<iomanip>
#include <ostream.h>
#include <sstream.h>

#pragma hdrstop

#include "Format.h"

namespace Tools
{
	using std::string;

   enum  // ONLY USED LOCALLY
   {
   	cSpace = ' ', cPlus = '+', cMinus = '-', cPoint = '.', cFloat = 'e',  // Syntax marks
      cDigit = '#', cZero = '0',										  					 // Sequences
      cEOS = 0,                                               					 // End of string
      cQuery = '?', cStar = '*'														  	 // Error digit
   };

   static const char *sSpace = " ", *sExp = "e";
   static const char *sZero = "0",  *sPlus = "+", *sMinus = "-", *sPoint = ".";

   char* p_err[] = { "", "-", "+" }; 													 // Error prefix

   const string bad_pars = "~~~";														 // Parameter error

   // FORMAT ANALYSIS

   bool scan_zero_filled_seq												// 0000E
   (
   	string::iterator* finger, string::iterator fend,
      int* i
   )
   {
   	while(*finger != fend)
      {
      	if(**finger != cZero) return false;

         *i += 1;

         ++(*finger);
      }
      return *i>0;
   }

   bool scan_integer_seq														// ####*
   (
   	string::iterator* finger, string::iterator fend,

      char* term, int* i
   )
   {
		*term = cEOS;

   	while(*finger != fend)
      {
      	if(**finger != cDigit) { *term = **finger;  return *i>0; }

         *i += 1;

         ++(*finger);
      }

      return *i>0;
   }

   bool scan_float_seq                                            // +000E
   (
   	string::iterator* finger, string::iterator fend,

      int* e
   )
   {
   	if(**finger != cPlus) return false;

      ++(*finger);

      return scan_zero_filled_seq(finger, fend, e);
   }

   bool scan_numeric_seq
   (
   	string::iterator* finger, string::iterator fend,

      Tools::Numeric* numtype,			int* i, int* d, int* e
   )
   {
   	char term;

      if(!scan_integer_seq(finger, fend, &term, i)) return false;

      switch(term)
      {
      	case cEOS:
         {
         	*numtype = nuInteger;  return true;
         }

         case cPoint:
         {
            ++(*finger);  if(!scan_integer_seq(finger, fend, &term, d)) return false;

            switch(term)
            {
            	case cEOS:
               {
               	*numtype = nuDecimal;  return true;
               }

               case cFloat:
               {
               	*numtype = nuFloat;

                  ++(*finger);  return scan_float_seq(finger, fend, e);
               }

               default: return false;
            }
			}
         default: return false;
      }
   }

   bool scan_number_seq
   (
   	string::iterator* finger, string::iterator fend,

      Tools::Numeric* numtype,		int* i, int* d, int* e
   )
   {
      if(**finger == cZero)
      {
      	*numtype = nuZero;

      	return scan_zero_filled_seq(finger, fend, i);
      }
      else
         return scan_numeric_seq(finger, fend, numtype, i,d,e);
   }


	bool scan_format
   (
   	string format,
      Prefix* prefix, Tools::Numeric* numtype,
      int* i, int* d, int* e
   )
   {
   	string::iterator finger = format.begin(), fend = format.end();

      switch(*finger)
      {
      	case cSpace: *prefix = pxHidden;   ++finger;  break;

         case cPlus:  *prefix = pxVisible;  ++finger;  break;

         default:		 *prefix = pxNone;
       }

      return scan_number_seq(&finger, fend, numtype, i, d, e);
   }


   // ERROR STRINGS

   string error_string(Prefix pref, int i)
   {
    	return p_err[pref] + string(i, cQuery);
   }

   string error_string(Prefix pref, int i, int d)
   {
   	return
      	error_string(pref,i) + "." + string(d, cQuery);
   }

   string error_string(Prefix pref, int i, int d, int e)
   {
   	return
      	error_string(pref,i,d) + "e" + error_string(pxVisible,e);
   }

   bool conv_error(string out)
   {
   	int pos = out.size()>1 ? 1 : 0;

   	return out[pos] == cQuery;
   }


   string zero_string(Prefix pref, int i, int d, int e)
   {
   	return
      	to_string(pref,i,d, 0.0) + "e+" + string(e,cZero);
   }

   string huge_string(Prefix pref, int i, int d, int e)
   {
   	return
      	to_string(pref,i,d, 1.0) + "e+INF";
   }


   // DIGIT GENERATOR

   char next_digit(int* i)
   {
   	if(*i<0) return cStar;

      int j = *i%10;	*i /= 10;

      return (char)(cZero + j);
   }

   char next_digit(real* r)
   {
   	if(*r<0 || *r>=1) return cStar;

      real s = *r * 10;  int j = (int)s;

      *r = s  - j;

      return (char)(cZero + j);
   }

   // CONVERSION TO STRING
   string to_stringi(const int ivalue)
   {
   	std::ostringstream osink;

      osink<<ivalue;

      return osink.str();
   }

   string to_stringr(const int dprec, const real rvalue)
   {
   	std::ostringstream rsink;

      rsink<<std::fixed;

      if(dprec>=0)rsink<<std::setprecision(dprec);

      rsink<<rvalue;

      return rsink.str();
   }

   string to_string
   (
   	const Prefix pref, const Numeric numtype, const int i,
      const int value
   )
   {
      char cpref = cZero;

      if(numtype>nuInteger || i<=0) return bad_pars;

      switch(pref)
      {
      	case pxHidden:  cpref = value<0 ? cMinus : cSpace;				break;

         case pxVisible: cpref = value<0 ? cMinus : cPlus;				break;

         case pxNone:    if(value<0) return error_string(pref, i);	break;

         default:			 return "?";
      }

		int ival = modulus(value), start = (pref == pxNone ? 0 : 1), j;

      string sval = string(i+start, cSpace);

      int width = sval.size();

      for(j=width-1; j >= start; j--)sval[j] = next_digit(&ival);

      if(ival>0) return error_string(pref,i);


		if(numtype == nuZero)										// Zero filled
      {
      	if(cpref != cZero) sval[0] = cpref;

         return sval;
      }

      int ipref = 0;

      for(j=start; j<width; j++)									// Space fill from left
      {
      	if(sval[j] != cZero)	 break;

         sval[j] = cSpace;

         ipref = j;
      }
      if(ipref==width-1) --ipref;								// All zero

      if(pref != pxNone) sval[ipref] = cpref;       		// Implant prefix

      if(value == 0)      sval[ipref+1] = cZero;         // Show one zero!

      return sval;
   }

   string to_string
   (
   	const Prefix pref,  const int i, const int d,
      const real value
   )
   {
   	using Tools::modulus;
      using Tools::entier;

      if(i<0 || d<0) return bad_pars;

      real rval = modulus(value) + 5.0*pow(10, -(d+1)), dval;   int ival;

      split(rval, ival, dval);   if(value<0) ival = -ival;

      string sival = to_string(pref, nuInteger, i, ival);
      					if(conv_error(sival)) return error_string(pref, i,d);

      string sdval = string(d, cZero);

      for(int j=0; j<d; j++)sdval[j] = next_digit(&dval);

   	return sival + sPoint + sdval;
   }

   string to_string
   (
   	const Prefix pref,	const int i, const int d, const int e,
      const real value
   )
   {
      if(i<0 || d<0 || e<0) return bad_pars;

      if(value==0)
      {
         string sval = to_string(pref, i,d, 0.0);

         string sexp = to_string(pxVisible, nuZero, e, 0);

         return sval + sExp + sexp;
      }

      real rval = modulus(value), aval, kval;										 // Std: r = a*10^k,  1 <= a < 10
                  if(rval>=DBL_MAX)return huge_string(pref,i,d,e);

      int nexp;  real bval = frexp(rval, &nexp);     								 // Bin:   r = b*2^n,  1/2 <= b < 1

      int kexp;  real dexp;	split(nexp*log2, kexp, dexp);					 // Conv:  a = b*10^d,  k = [n log(2)]


      aval = bval*pow(10, dexp);   if(aval<1) { aval *= 10;  kexp -= 1; }   // Require 1<=a<10

      if(i != 1) { aval *= pow(10, i - 1);	  kexp -= i - 1; }			 	 // Shift for i != 1

      kval = aval + 5*pow(10, -(d+1)); 	 											 // Rounding may affect shift

      if(kval>=pow(10, i)){ aval /= 10;  kexp += 1; }

 		string sval = to_string(pref,i,d, sign(value)*aval);                  // DECIMAL PART
                    if(conv_error(sval)) return error_string(pref,i,d,e);

      string sexp = to_string(pxVisible, nuZero, e, kexp);						 // EXPONENT PART
      				  if(conv_error(sexp)) return error_string(pref,i,d,e);

   	return sval + sExp + sexp;
   }

   // UNFORMATTED INTEGER CONVERSION
   string to_string(const int value)
   {
      if(value==0) return string(sZero);

      int v = Tools::modulus(value);

      string sv = string();                     // Start the output string

      while(v>0)                                // The value of v reduces by a factor of 10
         sv.insert(sv.begin(), next_digit(&v)); // each call, so this is guaranteed to terminate

      if(value<0)
         sv.insert(sv.begin(), Tools::cMinus);

      return sv;
   }

   // CONVERSION FROM RAW FORMAT
   string to_string(const string format, const int value)
   {
   	Prefix pref;  Numeric numtype;

   	int i=0, d=0, e=0;

      if(!scan_format(format, &pref, &numtype, &i,  &d, &e)) return format;

      switch(numtype)
      {
      	case nuZero: case nuInteger:
         	return to_string(pref, numtype, i, value);

         case nuDecimal:
         	return to_string(pref, i, d, (real)value);

         case nuFloat:
            return to_string(pref, i, d, e, (real)value);

         default:
         	return "?";
      }
   }


   string to_string(const string format, const real value)
   {
   	Prefix pref;  Numeric numtype;

   	int i=0, d=0, e=0;

      if(!scan_format(format, &pref, &numtype, &i,  &d, &e)) return format;

      switch(numtype)
      {
      	case nuZero: case nuInteger:
         	return to_string(pref, numtype, i, round(value));

         case nuDecimal:
         	return to_string(pref, i, d, value);

         case nuFloat:
            return to_string(pref, i, d, e, value);

         default:
         	return "?";
      }
   }
}

namespace Tools                        // ARITHMETIC SCANNING
{
   int to_posint
   (
      const string::iterator btext,    // Start of digit sequence
      const string::iterator etext     // End of contiguous digit sequence
   )
   {
      int ival = -1;

      if(btext==etext)return ival;

      ival = 0;

      string::iterator stext = btext;

      while(stext != etext)
      {
         ival = 10*ival + (int)(*stext - Tools::cZero);

         ++stext;
      }
      return ival;
   }

   bool to_int
   (
      const string::iterator btext,    // Start of digit sequence
      const string::iterator etext,    // End of contiguous digit sequence

      int& value
   )
   {
      value = 0;	if(btext==etext)return false;

      int sign = 0;

      string::iterator stext = btext;

      while(!sign)
      {
      	if(*stext == cPlus) { sign =  1; ++stext; break; }
         if(*stext == cMinus){ sign = -1; ++stext; break; }

         sign = 1; break;
      }

      string::iterator ptext = stext;

      while(ptext != etext)
      {
      	if(!isdigit(*ptext)) return false;

         ++ptext;
      }

      value = to_posint(stext, etext);

      if(sign<0) value = -value;

      return true;
   }

}

// Format
////////////////////////////////////////////////////////////////////////////////


