//----------------------------------------------------------------------------------------------------------------------
// PMatch	GENERATE PERFECT MATCHINGS ON A COMPLETE GRAPH									geoff kaniuk	2 Nov 2011
//----------------------------------------------------------------------------------------------------------------------
#include <vector>

#include "pmatch.h"

//----------------------------------------------------------------------------------------------------------------------
// CLASS PMatch
//
// An algorithm for generating all the perfect matchings on a complete graph has a simple recursive implementation:
// Assume we have an array m of size N.  We call routine pmatch(m,m) to generate all matchings.
//
// ROUTINE pmatch(tail, m)
// PM1: # set up an index to tail
// 		j = 1.
//
// PM2: # generate new matching
//		swap tail[j], tail[1].
//
// PM3: # match the elements from tail+2 on
//		if size of tail > 4 then call pmatch(tail+2, m),
//		otherwise output m.
//
// PM4: # undo the previous matching
//		swap tail[j], tail[1].
//
// PM5: # get next vertex for swap
//		increment j.
//		if j < size of tail then continue at PM2, otherwise return.
//
// For the 'on-demand' implementation we simulate the stack with vector<int> j and vector<int*> tail.  The integer k
// represents the stack level, so for example, j[k] is the index j that the recursive routine puts on the stack.

namespace Pmatch
{
	PMatch::PMatch(const int v)
	{
		if(v>0 && !(v%2))													// Even number of vertices
		{
			N = v; 	K = N/2;

			tail.resize(K-1);	n.resize(K-1);  j.resize(K-1);	// Matching mechanics

			matched.resize(N);											// Waiting for delivery

			for(k=0; k<K-1; k++)
			{
				tail[k] = &(matched[2*k]);								// Fixed pointers to match

				n[k]	= N - 2*k;											// Size of the tail

				j[k] = 1;													// Index for tail[k]
			}

			started = stopped = false;
		}
		else
			N=0;
	}

	void PMatch::swap(void)
	{
		int* tailk = tail[k];

		const int jk = j[k];

		const int tjk = tailk[jk];

		tailk[jk] = tailk[1];  tailk[1]= tjk;
	}

	bool PMatch::next_matching(vector<int>& match)
	{
		if(match.size() != matched.size())
			return false;

		// MATCHINGS COMPLETED
		if(stopped)												// No more matchings were detected on last call
		{
			started = false;									// Ready for next run
			stopped = false;

			return false;										// Tell caller there are no more matchings
		}

		// FIRST TIME INITIALISATION
		if(!started)
		{
			k = K-2;												// At bottom of stack

			matched = match;									// Caller supplies initialised match

			started = true;
			stopped = false;
		}

		// THE MATCHING WAS FULLY GENERATED ON THE PREVIOUS CALL
		match = matched;										// Fill

		if(k<0)													// No more matchings to deliver
			return stopped = true;


		// PREPARE FOR NEXT MATCHING
		while(k>=0)												// Ascend the tail stack
		{
			swap();												// Undo previous interchange

			j[k] += 1;											// Next vertex in tail[k]

			if(j[k] < n[k]) break;							// More to do this level

			k -= 1;												// Move up a level

			if(k<0) 												// We have just completed the last matching
				return stopped = true;
		}

		while(true)
		{
			swap();												// Interchange vertices j[k] and 1 in tail[k]

			if(k == K-2)										// The matching is ready for next call
				return true;

			k += 1;												// Move down a level in the stack

			j[k] = 1;
		}
	}
}
// END pmatch
//----------------------------------------------------------------------------------------------------------------------
