////////////////////////////////////////////////////////////////////////////////
// MODULE ETour:       Project-Wide Exception support          G.Kaniuk 22/04/02
//
// Each module that throws exceptions creates a routine called
//                      Throw(idType, idProc, idMess)
// This routine has the form:
//
//    ETour Throw(idType, idProc, idMessage)
//   {
//      string TextMess = Forge(idType, idSource, idProc, idMess);
//
//      AppLog.Add(TextMess)     AppLog of class ErrorLog identifies the
//                               Filename globally and is also required in the
//                               catch clause of the generating event

//      throw ETour(idType, TextMess);
//   }
//
////////////////////////////////////////////////////////////////////////////////

#include <vcl.h>
#include <vcl\sysutils.hpp>

#pragma hdrstop

#include "Strings.h"
#include "ETourRES.h"
#include "ETour.h"
//---------------------------------------------------------------------------

#pragma package(smart_init)
//---------------------------------------------------------------------------


namespace ToolsGUI                                 // ETour TOOLS
{
   using std::string;

   static const int max_log_text = 1024;           // We trim any garbage
   static const int max_sho_text =  1024;

   static const string softerr("SOFTWARE ERROR");  // Avoid loading from RES

   string Forge(EType Etype, short idSource, short idProc, short idMess)
   {
      using namespace Tools;

      short idType = IDS_HBug + Etype;

      string mess = sEmpty;

      mess += sTilde + load_str(idType);
      mess += sTilde + load_str(idSource);
      mess += sTilde + load_str(idProc);
      mess += sTilde + load_str(idMess);
      mess += sTilde;

      return mess;
   }

   string Forge(short idSink, short idEvent, String Message)
   {
      using namespace Tools;

      string mess = sEmpty;

      mess += sTilde + load_str(IDS_Detected);
      mess += sTilde + load_str(idSink);
      mess += sTilde + load_str(idEvent);
      mess += sTilde + string(sCrLf);

      return mess + S_s(Message) + sCrLf;
   }

   static string NextItem(string::iterator& is, string::iterator& end)
   {
      using namespace Tools;

      string::iterator ie = is;

      while(ie != end)
      {
         if(*ie == cTilde) break;

         ++ie;
      }
      if(ie==end) return string(0);       // String terminates with cTilde

      string item = string(is,ie);        // Extracted from ~ to ~

      is = ++ie;                          // Skip over ~

      return item;
   }

   static bool Parse(String& Message, string& ErrCode, string& Source, string& Proc, string& Mess)
   {
      using std::string;

      using namespace Tools;

      string message = S_s(Message);

      string errcode=sEmpty, source=sEmpty, proc=sEmpty, mess=sEmpty;

      string::iterator is = message.begin(), end = message.end();

      ErrCode = load_str(IDS_HAlien);

      Source  = source;  Proc = proc; Mess = mess;

      if(*is != cTilde){return false; }

      ++is;

      errcode = NextItem(is, end);  if(errcode.empty()) return false;
      source  = NextItem(is, end);  if(source.empty()) return false;
      proc    = NextItem(is, end);  if(proc.empty()) return false;
      mess    = NextItem(is, end);  if(mess.empty()) return false;

      ErrCode = errcode;
      Source  = source;
      Proc    = proc;
      Mess    = mess;

      return true;
   }
}

namespace ToolsGUI									// MESSAGE BOXES
{
   int ShowMessage(EType etype, string mess) // returns IDOK,IDCANCEL,IDABORT,IDRETRY,IDIGNORE,IDYES,IDNO
   {
      using namespace Tools;

      string title;

      int okhelpbuttons = MB_OK;									// RESEARCH NEEDED FOR  | MB_HELP

      int style;

      switch(etype)
      {
         case erBug:
         {
            style = MB_ICONSTOP | okhelpbuttons;
            title = softerr;                             // For memory low not a good idea to load from RES
         }
         break;

         case erOperr:
         {
            style = MB_ICONEXCLAMATION | okhelpbuttons;
            title = load_str(IDS_HOperr);
         }
         break;

         case erWarn:
         {
            style = MB_ICONEXCLAMATION | MB_OKCANCEL| okhelpbuttons;
            title = load_str(IDS_HWarn);
         }
         break;

         case erInfo:
         {
            style = MB_ICONINFORMATION | okhelpbuttons;
            title = load_str(IDS_HInfo);
         }
         break;

         default:
            style = MB_ICONSTOP | MB_OK;
            title = softerr;                             // For memory low not a good idea to load from RES
      }

      if(mess.size()>max_sho_text) mess.resize(max_sho_text);

      Screen->Cursor = crDefault;

      return Application->MessageBox(mess.c_str(), title.c_str(), style);
   }


   int WarnOverwrite(string filename) // returns IDYES,IDNO
   {
      using Tools::load_str;
      using Tools::sCrLf;

      int style = MB_ICONEXCLAMATION | MB_YESNO;

      string title = load_str(IDS_HFile);

      string mess = filename + load_str(IDS_Exist) + string(sCrLf)
                  + string(sCrLf)
                  + load_str(IDS_Overwrite);

      if(mess.size()>max_sho_text) mess.resize(max_sho_text);

      return Application->MessageBox(mess.c_str(), title.c_str(), style);
   }

   int WarnOK(string mess)
   {
      using Tools::load_str;
      using Tools::sCrLf;

      int style = MB_ICONEXCLAMATION | MB_OK;

      string title = load_str(IDS_HWarn);

      if(mess.size()>max_sho_text) mess.resize(max_sho_text);

      return Application->MessageBox(mess.c_str(), title.c_str(), style);
   }

   int ShowSyntaxError(const string text, const string explain, const string format)
   {
   	using Tools::load_str;
      using Tools::sSqBra;
      using Tools::sSqKet;
      
   	const int style = MB_ICONSTOP | MB_OK;

      string title = load_str(IDS_HSyntax);

      string mess = string(sSqBra)+text+string(sSqKet) + load_str(IDS_BadInput);

      if(!explain.empty()) mess += explain;
      if(!format.empty()) mess += format;

      if(mess.size()>max_sho_text) mess.resize(max_sho_text);

      return Application->MessageBox(mess.c_str(), title.c_str(), style);
   }

   void OperatorError(const int ID_message)
   {
   	ShowMessage(erOperr, Tools::load_str(ID_message));
   }

   bool OperatorWarning(const int ID_message)
   {
   	switch(ShowMessage(erWarn, Tools::load_str(ID_message)))
      {
      	case IDOK: 		return false;				// Continue
         case IDCANCEL:	return true;				// Cancel

      	default: 		return true;				// Low memory
      }
   }

   void OperatorInfo(const int ID_message)
   {
   	ShowMessage(erInfo, Tools::load_str(ID_message));
   }
}

namespace ToolsGUI                           // THE PUBLIC ETOUR INTERFACE
{
   using Tools::s_S;
   using Tools::S_s;
   
   __fastcall ETour::ETour():Exception(NULL)
   {
      Etype   = erUnknown;
      Message = String();
   }

   __fastcall ETour::ETour(EType ErrType,  string MessText):Exception(NULL)
   {
      Etype   = ErrType;
      Message = s_S(MessText);
   }

   int Serve(Exception& Exceptor, short idSink, short idEvent)
   {
      EType ErrType;

      string errcode, source, proc, mess;

      bool TourClass  = String(Exceptor.ClassName()) == LoadStr(IDS_ETour);
      bool Parsed     = false;

      if(TourClass)
         Parsed = Parse((String&)Exceptor.Message, errcode, source, proc, mess);

      if(!TourClass || !Parsed)
      {
         ::ShowMessage(Exceptor.Message);

         string SinkText = Forge(idSink, idEvent, Exceptor.Message);

         AppLog.Add(SinkText);

         Screen->Cursor = crDefault;
         
         return -1;
      }

      ETour& ExcepTour = dynamic_cast<ETour&>(Exceptor);

      int Response = -1;

      switch(ExcepTour.Etype)
      {
      	case erBug:
         {
            Response = ShowMessage(ExcepTour.Etype, mess);

         	string SinkText = Forge(idSink, idEvent, Exceptor.Message);

         	AppLog.Add(SinkText);
         }
         break;

         case erLog:
         {
         	string SinkText = Forge(idSink, idEvent, Exceptor.Message);

         	AppLog.Add(SinkText);
         }
         break;

         default:
         	// 0 no memory; IDOK, IDCANCEL
         	Response = ShowMessage(ExcepTour.Etype, mess);
      }

      return Response;
   }

   int BugReport(short idmodule, short idproc, short idmsg)
   {
      string message = Forge(erBug, idmodule, idproc, idmsg);

      AppLog.Add(message);

      return ShowMessage(erBug, message);  		// 0 no memory; IDOK, IDCANCEL
   }

   int ErrorReport(short idmodule, short idproc, short idmsg)
   {
      string message = Forge(erOperr, idmodule, idproc, idmsg);

      AppLog.Add(message);

      return ShowMessage(erOperr, message);  		// 0 no memory; IDOK, IDCANCEL
   }

}

namespace ToolsGUI                              // ERROR LOG
{
   using std::string;

   static const int MaxErrors = 16;             // Log the First 16 errors from startup

   ErrorLog::ErrorLog()
   {
      Count = 0;  FilePath = Tools::sEmpty;   Stream = 0;
   }

   void ErrorLog::StartUp(string version, string filename)
   {
      try
      {
         string times = S_s(TDateTime().CurrentDateTime().DateTimeString());

         string head = times + Tools::sSpace + version + Tools::sCrLf;

         FilePath = GetCurrentDir() + Tools::sSep + s_S(filename);

         Stream = new TFileStream(FilePath, fmCreate|fmShareExclusive);
            Stream->Write(head.c_str(), head.size());
         delete Stream;  Stream = 0;

         Count = 0;
      }
      catch(...)
      {
         delete Stream;

         throw Exception(IDS_Start);
      }
   }

   void ErrorLog::Add(string message)  // Errors get logged in
   {                                   // generator/detector pairs
      using Tools::sCrLf;

      if(Count >= 2*MaxErrors) return;

      string times = S_s(TDateTime().CurrentDateTime().DateTimeString());

      string text = times + sCrLf + message + sCrLf;

      try
      {
         Word OpenMode = fmOpenReadWrite|fmShareExclusive;

         if(Stream)delete Stream;

         Stream = new TFileStream(FilePath, OpenMode);

            Stream->Seek(0, soFromEnd);
            Stream->Write(text.c_str(), text.size());

         delete Stream;  Stream = 0;

         ++Count;
      }
      catch(...)
      {
         delete Stream; Stream = 0;
      }
   }

   void ErrorLog::ShutDown()
   {
      --Count;

      Add(Tools::load_str(IDS_Shutdown));
   }
}

namespace ToolsGUI                              // ERROR INTERFACE
{
   void ThrowError(EType etype, short idmodule, short idproc, short idmsg)
   {
      string message = Forge(etype, idmodule, idproc, idmsg);

      if(etype==erBug)::AppLog.Add(message);

      throw ETour(etype, message);
   }

   string FormWin32Error(short idSink, short idEvent)
   {
      string errormsg = string();

      try
      {
         RaiseLastWin32Error();
      }
      catch(EWin32Error& Ewin32)
      {
         string fatal = Forge(idSink, idEvent, Ewin32.Message);

         AppLog.Add(fatal);

         errormsg = fatal;
      }
      catch(...)
      {
         errormsg =  Tools::load_str(IDS_FatalError);
      }

      return errormsg;
   }

   void ShowWin32Err(short idSink, short idEvent)
   {
      string win32msg = FormWin32Error(idSink, idEvent);

      ShowMessage(erOperr, win32msg);
   }

}

////////////////////////////////////////////////////////////////////////////////



