////////////////////////////////////////////////////////////////////////////////
// MODULE GridGUI:       Base class for GoDraw's grids        G.Kaniuk 08/06/03
//

#include <vcl.h>
#pragma hdrstop

#include "GraphicsAPI.h"
#include "Strings.h"
#include "GridGUI.h"

namespace ToolsGUI                        // MAGIC numbers
{
   const int defFontSize   =   8;         // Main grid font size - Points
   const int defStrideSize =  14;         // Grid row stride - Points

   const TColor defPaneColor = clWindow;  // Main form pane input

   /////////////////////////////////////////////////////////////////////////////
   // EDIT Termination Strategy
   //
   // EditCell is global sentry which flags an unterminated cell edit.
   // It is checked on clicking a grid cell.
   // If it is not clear, we have a possible unterminated edit and the user must
   // be instructed.  This is achieved via overrides to the
   // RowChanged ColChanged or  GridChanged routines.  These routines issue an
   // operator message if appropriate and clear the sentry if the action
   // is allowed.
   //
   // If now the Edit Cell is clear, we can dismount the old field and mount the new field
   // if it is editable

   GField* EditCell = NULL;

   void ClearEditCell(){ EditCell = NULL; }

   int  EditCellRow()
   {
   	if(!EditCell) return -1;
      
   	return EditCell->RowM;
   }

   bool EditCellIsNot(GField* Field) { return EditCell != Field; }
   bool EditCellIs(GField* Field) { return EditCell == Field; }

   bool EditingCell(){ return EditCell != NULL; }
   
   bool EditingGrid(GGrid* grid)
   {
      if(EditCell == NULL) return false;

     return EditCell->Grid == grid;
   }

   void ResetToEditCell(void)
   {
		if(!EditCell) return;

      GGrid* EditGrid = EditCell->Grid;		if(!EditGrid) return;

      Beep();

      EditGrid->SelectGridCell(EditCell->ColM, EditCell->RowM); // Reselect old cell

      if(EditCell->FieldState()!=fsShowing)                   // Ensure ready for action
         EditGrid->TopRow = EditCell->RowM;

      EditCell->SetFocusEd();                        // It will have lost focus
   }

}

namespace ToolsGUI
{
   __fastcall GEdit::GEdit(GField* field):TEdit(field->Grid->Owner)     //TComponent* owner,
   {
      Visible        = false;
      Parent         = field->Grid->Parent;

      Text           = String();
      TabStop        = false;
      TabOrder       = -1;
      BorderStyle    = bsNone;
      AutoSize       = true;
      AutoSelect     = false;
      Color          = clSilver;

      Field          = field;
   }

   void __fastcall GEdit::KeyDown(Word& Key, TShiftState Shift)
   {
      if(Field)
      {
         Field->Grid->ClearInfo();
         
         Field->FieldKeyDown(Key, Shift);
      }

      if(!Key) return;

      TEdit::KeyDown(Key, Shift);
   }

   void __fastcall GEdit::KeyDownSpecial(TCMWantSpecialKey& Message)
   {
      Message.Result = 1;
   }

   GField* __fastcall GEdit::GetField(){ return  Field; }

   GGrid*  __fastcall GEdit::GetGrid() { return  Field->Grid; }

   void __fastcall GEdit::SelectText()
   {
      SelStart  = 0;
      SelLength = Text.Length();
   }
}

namespace ToolsGUI
{
   __fastcall GField::GField(GGrid* grid)
   {
      Grid        = grid;
      Editor      = NULL;

      State       = fsDormant;
      ColM        = grid->Col;
      RowM        = grid->Row;

      CellText    = String();    // OBSOLETE
   }
   //---------------------------------------------------------------------------

   void __fastcall GField::ShowEd()
   {
      if(Editor)
         Editor->Show();
   }

   void __fastcall GField::HideEd()
   {
      if(Editor)
         Editor->Hide();
   }

   void __fastcall GField::MountEd(TRect& Cell)
   {
      if(Editor)
      {
         Editor->Width  = Cell.Right - Cell.Left;   // Edit Size
         Editor->Height = Cell.Bottom - Cell.Top;
      }
   }

   void __fastcall GField::MoveEd(TRect& Cell)
   {
      if(Editor)
      {
         Editor->Left = Grid->Left + Cell.Left + 2*Grid->GridLineWidth;
         Editor->Top  = Grid->Top  + Cell.Top  + 2*Grid->GridLineWidth;
      }
   }

   void __fastcall GField::ReMountEd(void)
   {
      if(Editor)
      {
         TRect Cell = Grid->CellRect(Grid->Col, Grid->Row);

         Editor->Left   = Grid->Left + Cell.Left + 2*Grid->GridLineWidth;
         Editor->Top    = Grid->Top  + Cell.Top  + 2*Grid->GridLineWidth;

         Editor->Width  = Cell.Right - Cell.Left;
         Editor->Height = Cell.Bottom - Cell.Top;
      }
   }

   void __fastcall GField::SetFocusEd(void)
   {
      if(!Editor)
         Grid->SetFocus();
      else
         if(Editor->Visible)Editor->SetFocus();
   }
   //---------------------------------------------------------------------------

   
   void __fastcall GField::TransferFocus()
   {
      if(Grid->SelectedCellShowing())
         SetFocusEd();
      else
         Grid->SetFocus();
   }

   bool __fastcall GField::Scroll()
   {
      bool Scrolled = !Grid->SelectedCellShowing();

      if(Scrolled)                       // The current cell IS not showing
      {
         if(State == fsShowing)          // It WAS showing
         {
            State = fsMounted;           // KEEP attached ..

            HideEd();                     // .. but switch OFF
         }
      }
      else                               // The Current cell IS showing
      {
         if(State == fsMounted)          // It WAS not showing
            State = fsShowing;           // Switch ON
      }
      return Scrolled;
   }

   void __fastcall GField::Move()
   {
      if(State < fsMounted) return;

      if(Scroll()) return;

      TRect Cell = Grid->CellRect(Grid->Col, Grid->Row);

      MoveEd(Cell);   ShowEd();   SetFocusEd();
   }

   void __fastcall GField::Mount()
   {
      if(State != fsDormant)return;                               // Start with fsDormant to fsMounted is valid

      State = fsMounted;                                          // fsDormant->fsMounted is valid

         HideEd();

         ColM = Grid->Col;   Grid->CurFieldId = ColM;
         RowM = Grid->Row;

         CellText = Grid->Cells[ColM][RowM];                       // Preserve for VK_ESCAPE

         TRect CellArea = Grid->CellRect(ColM, RowM);              // TODO:: MountEd should take A Trect, NOT TRect&

         MountEd(CellArea);                                        // Set Editor dimensions from visible cell

         Load();                                                   // Renders cell string for display in editor

      Move();                                                      // Position and show
   }

   void __fastcall GField::Dismount()                             // END OF EDIT SESSION
   {
      if(State==fsDormant) return;

      Store();

      State = fsDormant;

      EditCell = NULL;

      HideEd();
   }

   void __fastcall GField::HardDismount()                             // END OF EDIT SESSION
   {
      if(State==fsDormant) return;

      State = fsDormant;

      HideEd();

      EditCell = NULL;
   }
   //---------------------------------------------------------------------------

   void __fastcall GField::Load()
   {
      // Copy  [CellText] Grid->Cells[ColM][RowM] to Editor text
   }

   void __fastcall GField::Store()
   {
      // Set Grid->Cells[ColM][RowM] = Editor text;
   }

   bool __fastcall GField::SameValue()
   {
      return false;
   }
   //---------------------------------------------------------------------------

   void __fastcall GField::FieldKeyDown
   (
      Word& Key, Classes::TShiftState ShiftState
   )
   {
      if(State == fsDormant) return;

      switch(Key)
      {
         case VK_UP:
         {
            if(SameValue())
            {
               Grid->NextGridRow(-1);  Grid->MountFieldAfter(this);
            }
            else
            {
               Grid->RowChange(GGrid::rcField);
            }
         }
         break;

         case VK_DOWN:
         {
            if(SameValue())
            {
               Grid->NextGridRow(1);   Grid->MountFieldAfter(this);
            }
            else
            {
               Grid->RowChange(GGrid::rcField);
            }
         }
         break;

         case VK_PRIOR:
         {
            Grid->NextGridPage(-1);    TransferFocus();
         }
         break;

         case VK_NEXT:
         {
            Grid->NextGridPage( 1);   TransferFocus();
         }
         break;

         case VK_TAB:
         {
            Grid->SetFocus();

            Grid->TStringGrid::KeyDown(Key, ShiftState);
         }
         break;
         

         case VK_ESCAPE:
         {
            Dismount();

            Grid->Cells[ColM][RowM] = CellText;  

            Grid->SetFocus();
            
            EditCell = 0;
         }
         break;

         case VK_RETURN:
         {
         	if(!Grid->Parse(Key))
            	return;

            Dismount();

            Grid->SetFocus();
            Grid->PostParse();
         }
         break;

         default:  // IT IS UP TO CALLER TO DECIDE IF FURTHER KEY PROCESSING IS NEEDED
                   // IF NOT THEN CALLER MUST SET Key = 0
                   return;
      }
      Key=0;
   }
}
//---------------------------------------------------------------------------



namespace ToolsGUI                        // GENERIC GRID
{
    __fastcall GGrid::GGrid               // Virgin startup constructor
   (
      TWinControl* parent,                // Usually a GroupBox owned by Form
      TStatusBar* statusbar,              // Operator info

      const int fields                    // Total field count
   )
   : TStringGrid(parent->Owner)
   {
      Visible        = false;
      Parent         = parent;
      StatusBar      = statusbar;
      Color          = clWindow;
      GridLineWidth  = 1;

      Options.Clear();
      Options<<goFixedVertLine<<goFixedHorzLine<<goVertLine<<goHorzLine;
      Options<<goThumbTracking;
      Options<<goTabs;

      Font->Size = defFontSize;

      BackGround = Canvas->Brush->Color;
      Canvas->Brush->Bitmap = NULL;                   // Some descendents need
      Canvas->Brush->Style = bsSolid;                 // to change the background

      ColCount = fields;      FixedCols = 1;
      RowCount = 2;           FixedRows = 1;

      OriginalWindowProc = WindowProc;
      WindowProc         = GridWindowProc;

      OnMouseWheel     = MouseWheel;
      OnMouseWheelUp   = MouseWheelUpDn;
      OnMouseWheelDown = MouseWheelUpDn;

      BodyStride = (defStrideSize*Font->PixelsPerInch) / 72 ;
      hScrollBar = false;

      if(fields>0)
      {
         Fields.resize(fields);
         SortDir.resize(fields);

         for(int f=0; f<fields; f++)
         {
            Fields[f] = 0;

            SortDir[f] = gsNa;
         }
      }
      NoCellSelected();
      CurFieldId = 0;

      SortEnabled         = false;


      SortDn = new Graphics::TBitmap();
      SortUp = new Graphics::TBitmap();

      SortDn->LoadFromResourceName((unsigned)HInstance, "SORTUP");
      SortUp->LoadFromResourceName((unsigned)HInstance, "SORTDN");

      SortUp->Transparent = true;   SortUp->TransparentColor = SortUp->Canvas->Pixels[0][0];
      SortDn->Transparent = true;   SortDn->TransparentColor = SortDn->Canvas->Pixels[0][0];
   }

   __fastcall GGrid::~GGrid()
   {
      WindowProc = OriginalWindowProc;

      delete SortUp;
      delete SortDn;
   }
   //---------------------------------------------------------------------------

   int __fastcall GGrid::BodyRowCount()
   {
      return ClientHeight/BodyStride - 1;                            // Client = Header + Body
   }

   void __fastcall GGrid::Sheave                                     // Expand columns to fit client
   (
      const ColSize colsizes[],                                      // Minimum column widths
      int first,                                                     // Given starting column
      int& last                                                      // From first to last adjusted to fill ClientWidth
   )
   {
      int ScrollWidth = ClientWidth - colsizes[0].minwidth;          // The first col does not scroll

      int TotalColWidth, ExpandColWidth = 0, RigidColWidth = 0;

      last = first;  if(first<1) return;


      // DETERMINE THE SET OF COLUMNS TO EXPAND
      for(int f=first; f<(int)Fields.size(); f++)
      {
         ColSize colsize = colsizes[f];

         if(ExpandColWidth + RigidColWidth + colsize.minwidth > ScrollWidth)
            break;

         last = f;

         if(colsize.expand)
            ExpandColWidth += GridLineWidth + colsize.minwidth;
         else
            RigidColWidth += GridLineWidth + colsize.minwidth;
      }

      TotalColWidth =  ExpandColWidth + RigidColWidth;

      if(TotalColWidth == ScrollWidth)
      {
         for(int f=0; f<last; f++)ColWidths[f] = colsizes[f].minwidth;

         ColWidths[last] = colsizes[last].minwidth - 1;

         return;
      }


      // ESTABLISH THE SCALING
      int nScale, dScale;

      if(ExpandColWidth > 0)                    // Fill available space by expanding
      {                                         // just the desired columns
         nScale = ScrollWidth - RigidColWidth;  // ClientWidth > AllRigidColWidth
         dScale = ExpandColWidth;
      }
      else
      {
         nScale = ScrollWidth;                  // Fill available space by expanding
         dScale = TotalColWidth;                // all columns in proportion
      }

      GRatio Scale(nScale, dScale);


      // NOW EXPAND THE COLUMNS FROM first TO last

      ColWidths[0] = colsizes[0].minwidth;

      int NewTotalWidth = 0;

      for(int f=first; f<=last; f++)
      {
         ColSize colsize = colsizes[f];

         if(!ExpandColWidth)                                // All columns have low priority for expansion
            colsize.expand = 1;                             // Share excess over ALL columns

         if(colsize.expand)                                 // Expand this column
            ColWidths[f] = Scale(colsize.minwidth);
         else
            ColWidths[f] = colsize.minwidth;

         NewTotalWidth += GridLineWidth + ColWidths[f];
      }

      if(NewTotalWidth != ScrollWidth)                      // Truncation error
         ColWidths[last] += ScrollWidth-NewTotalWidth-1;    // A small increase
   }

   void __fastcall GGrid::ReBuild(const ColSize colsizes[])          // Called by OnFormResize()
   {
   	// SCROLLBARS AFFECT CLIENT SIZE
      ScrollBars = ssNone;

      // FIX CLIENT HEIGHT FROM HORIZONTAL SCROLLBAR
      int ScrollWidth = ClientWidth - colsizes[0].minwidth;				// Max possible

      int TotalColWidth = 0;

      for(size_t f=1; f<Fields.size(); f++)                          // Start with minimum column widths
         TotalColWidth += GridLineWidth + colsizes[f].minwidth;

      hScrollBar = TotalColWidth > ScrollWidth;

      if(hScrollBar)ScrollBars = ssHorizontal;                       // Fixes ClientHeight


      // FIX CLIENT WIDTH FROM VERTICAL SCROLLBAR
      Font->Size = defFontSize;                                      // Do not scale this font

      BodyStride = (defStrideSize*Font->PixelsPerInch) / 72 ;        // Convert from points to pixels
      if(RowCount-1 > BodyRowCount())                                  // Add Vertical ScrollBar if reqd
         ScrollBars = hScrollBar ? ssBoth : ssVertical;

      HeadRowHeight = ClientHeight - BodyRowCount()*BodyStride - 1;  // Take up extra in fixed row

      RowHeights[0] = HeadRowHeight;

      for(int j = 1; j<RowCount; j++)                                // All body rows same height
         RowHeights[j] = BodyStride-GridLineWidth;

      // ADJUST COLUMNS TO FILL CLIENT IF CLIENT IS BIG ENOUGH
      while(true)
      {
         int LastCol, EndCol = (int)Fields.size()-1;

         if(TotalColWidth!=ScrollWidth)
            Sheave(colsizes, 1, LastCol);
         else
            break;

         if(LastCol==EndCol)
            break;

         // ASSUME THAT TOTAL COLUMN WIDTH IS LESS THAN 2 CLIENT WIDTHS
         // Adjust the width of the last column to fill the ScrollWidth
         // with a whole number of columns.  All columns from 1 to LastCol
         // have already been adjusted

        ScrollWidth = ClientWidth - ColWidths[0];

         TotalColWidth = 0;

         for(int f=EndCol; f>1; f--)
         {
            int colwidth = f > LastCol ? colsizes[f].minwidth : ColWidths[f];

            if(TotalColWidth + GridLineWidth + colwidth > ScrollWidth)
               break;

            TotalColWidth += GridLineWidth + colwidth;
         }

         if(TotalColWidth != ScrollWidth)
            ColWidths[EndCol] = ScrollWidth-TotalColWidth+colsizes[EndCol].minwidth-1;

         break;
      }

      // IF EDITING IN PROGRESS WE NEED TO RE-ATTACH THE FIELD TO THE CELL
      if(EditCell)                           // We are editing a cell
      if(EditCell->Grid == this)             // The rebuilt grid contains the cell
      {
         SelectGridCell(EditCell->ColM, EditCell->RowM);

         ExposeSelectedCell();

         EditCell->ReMountEd();
         
         EditCell->Move();
      }
   }
   //---------------------------------------------------------------------------


   void __fastcall GGrid::AddRow(int r)
   {
      RowCount = RowCount+1;

      RowHeights[RowCount-1] = BodyStride - GridLineWidth;
   }

   void __fastcall GGrid::AddRows(int rows)
   {
      int OldRowCount = RowCount, r;

      if(rows<=0) return;

      RowCount = OldRowCount + rows;

      for(r=OldRowCount; r<RowCount; r++)
         RowHeights[r] = BodyStride - GridLineWidth;
   }
   //---------------------------------------------------------------------------

   void __fastcall GGrid::ShowInfo(int messid)
   {
      if(!StatusBar) return;

      if(StatusBar->Panels!= NULL)
      {
         StatusBar->Panels->Items[0]->Text = LoadStr(messid);
         StatusBar->Repaint();
      }
   }

   void __fastcall GGrid::ShowInfo(String mess)
   {
     if(StatusBar)
      if(StatusBar->Panels)
         StatusBar->Panels->Items[0]->Text = mess;
   }

   void __fastcall GGrid::ShowInfoStatus(int idinfo, int idstatus)
   {
      if(!StatusBar) return;

      if(StatusBar->Panels!= NULL)
      {
         StatusBar->Panels->Items[0]->Text = LoadStr(idinfo);
         StatusBar->Panels->Items[1]->Text = LoadStr(idstatus);
      }
   }

   void __fastcall GGrid::ClearInfo()
   {
      if(!StatusBar) return;

      if(StatusBar->Panels!= NULL)
         StatusBar->Panels->Items[0]->Text = String();
   }
   //---------------------------------------------------------------------------

   void __fastcall GGrid::NoCellSelected()
   {
      TGridRect NoSelect;

      NoSelect.Top    = -1;   NoSelect.Left  = -1;
      NoSelect.Bottom = -1;   NoSelect.Right = -1;

      Selection = NoSelect;
   }

   void __fastcall GGrid::SelectGridCell(int col, int row)
   {
      if(col<0 || col >= ColCount) { Beep(); return; }
      if(row<0 || row >= RowCount) { Beep(); return; }

      TGridRect GridCell;

      GridCell.Top    = row;   GridCell.Left  = col;
      GridCell.Bottom = row;   GridCell.Right = col;

      Selection = GridCell;
   }
   
   bool __fastcall GGrid::BlockSelected(void)
   {
   	if(Selection.Left != Selection.Right) return true;
      if(Selection.Top  != Selection.Bottom)return true;

      return false;
   }

   void __fastcall GGrid::ExposeSelectedCell()
   {
      if(Col < LeftCol)                      LeftCol = Col>0 ? Col :1;
      if(Col > LeftCol + VisibleColCount - 1)LeftCol = Col - VisibleColCount + 1;

      if(Row < TopRow)                        TopRow = Row;
      if(Row > TopRow + VisibleRowCount - 1)  TopRow = Row - VisibleRowCount + 1;
   }

   bool __fastcall GGrid::SelectedCellShowing()
   {
      if(Row<TopRow  || Row > TopRow + VisibleRowCount - 1) return false;
      if(Col<LeftCol || Col > LeftCol + VisibleColCount - 1)return false;

      return true;
   }

   bool __fastcall GGrid::SelectedFieldScrolled()
   {
      GField* Field = Fields[Col];

      if(!Field) return false;

      switch(Field->State)
      {
         case fsDormant:   return false;  // Inactive
         case fsMounted:   return true;   // Active but not showing
         default:          return false;  // Showing
      }
   }

   void __fastcall GGrid::SelectPopupCell(TPoint PopPoint)
   {
      TPoint GridPt = ScreenToClient(PopPoint);			// Top Left corner

      int Gcol, Grow;

      MouseToCell(GridPt.x, GridPt.y, Gcol, Grow);		// Cell location

      SelectGridCell(Gcol,Grow);
   }

   void __fastcall GGrid::AnalysePopupCell(TPoint PopPoint, int&PopCol, int& PopRow)
   {
      TPoint GridPt = ScreenToClient(PopPoint);				// Top Left corner

      MouseToCell(GridPt.x, GridPt.y, PopCol, PopRow);	// Cell location
   }

   //---------------------------------------------------------------------------

   void __fastcall GGrid::DrawCell                 // HANDLER DrawCell
   (
      int col, int row,
      const Windows::TRect& cell,
      Grids::TGridDrawState state
   )
   {
      using ToolsGUI::DrawAlignedText;
      using ToolsGUI:: jyCentre;

      DrawAlignedText(Canvas->Handle, jyCentre, cell, Cells[col][row]);
   }
   //------------------------------------------------------------------

   void __fastcall GGrid::MountFieldAfter(GField* PrevField)
   {
      if(PrevField)PrevField->Dismount();       // Dismount the Old

      SetFocus();

      if(Fields[Col])                           // The new field is editable
      {
         Fields[Col]->Mount();                  // Mount the New

         EditCell =  Fields[Col];               // Wake up Sentry
      }
   }

   void __fastcall GGrid::ResetToField(GField* PrevField)
   {
		if(!PrevField) return;
         	
      Beep();

      SelectGridCell(PrevField->ColM, PrevField->RowM); // Reselect old cell

      if(PrevField->State!=fsShowing)                   // Ensure ready for action
         PrevField->Grid->TopRow = PrevField->RowM;
      else
         PrevField->SetFocusEd();                        // It will have lost focus
   }

   bool __fastcall GGrid::EditPossible(void)
   {
   	if(!EditCell) return true;									// No other grid is editing

      return EditCell->Grid == this;            			// The current edited cell is owned by the grid
   }

   void __fastcall GGrid::Click()                        // HANDLER Click
   {
      if                                                 // After action on this selection
      (                                                  // we will then deselect
         Selection.Left != Selection.Right
                        ||
         Selection.Top  != Selection.Bottom
      )
      return;

      int PrevRow, PrevCol;

      GGrid* PrevGrid = this;

      GField* PrevField = NULL;

      ClearInfo();

      if(EditCell)
      {
         PrevField = EditCell;      PrevGrid = PrevField->Grid;

         PrevRow = PrevField->RowM;  PrevCol = PrevField->ColM;

         if(PrevGrid != this)
            PrevGrid->GridChange();             // EditCell normally stays set

         while(PrevGrid == this)                // Cell selection in current grid may have changed
         {
            if(PrevField->SameValue())          // OK to hop to new cell
            {
               EditCell = NULL;  break;
            }

            if(PrevRow != Row)                  // Usually not allowed
            {                                   // Beep & issue message
               RowChange(GGrid::rcGrid);

               break;
            }

            if(PrevCol != Col)                  // Usually allowed
                ColChange(GGrid::rcGrid);       // Clears EditCell if allowed

            break;
         }
      }

      if(EditCell)                                       // The change was forbidden
      {
         ResetToField(PrevField);

         if(0<=CurFieldId && (size_t)CurFieldId<Fields.size())
         {
            GField* FieldSel = Fields[CurFieldId];

            if(FieldSel)
               SelectGridCell(FieldSel->ColM, FieldSel->RowM);
            else
               NoCellSelected();
         }
         else
            NoCellSelected();

         return;
      }

      MountFieldAfter(PrevField);
   }

   bool __fastcall GGrid::SelectCell(int col, int row)
   {
      return TStringGrid::SelectCell(col, row);
   }
   //------------------------------------------------------------------

   void __fastcall GGrid::TopLeftChanged()      // Scrolling
   {
      if(Col<0) return;
      if(!Fields[Col]) return;                  // No editor in current column

      Repaint();

      Fields[Col]->Move();
   }

   void __fastcall GGrid::GridWindowProc(TMessage& Message)
   {
      switch(Message.Msg)
      {
      /*
         case WM_KEYDOWN:
         {
            // TODO: HERE FOR DEBUG ONLY
            int W = Message.WParam;
            int L = Message.LParam;
            int R = Message.Result;

            R=L; L=W; W = R;
         }
         break;
      */

         case WM_VSCROLL: case WM_HSCROLL:
         {
            Word ScrollCode = Message.WParamLo;

            if(ScrollCode==SB_ENDSCROLL)
            {
               if(Col<0)break;

               GField* Field = Fields[Col];

               if(Field)
               {
                  Field->SetFocusEd();

                  return;
               }
            }
         }
         break;

         default: break;
      }
      OriginalWindowProc(Message);
   }
   //---------------------------------------------------------------------------

   void __fastcall GGrid::NextGridRow(int dir)
   {
      if(Row<0) return;

      if(!SelectedCellShowing())
      {
         TopRow = Row;  return;        // Fires TopLeftChanged
      }

      switch(dir)
      {
         case -1:
         {
            if(Row>1)
               SelectGridCell(Col,Row-1);
            else
               Beep();
         }
         break;

         case 1:
         {
            if(Row<RowCount-1)
               SelectGridCell(Col,Row+1);
            else
            {
               Beep();  NoDownArrow();
            }
         }
         break;

         default: SelectGridCell(Col,Row);
      }

      if(!SelectedCellShowing())
      {
         int NewTopRow = TopRow + dir;

         if(NewTopRow < 1)           NewTopRow = 1;
         if(NewTopRow > RowCount - 1)NewTopRow = RowCount - 1;

         TopRow = NewTopRow;
      }
   }

   void __fastcall GGrid::NextGridCol(int dir)
   {
      if(Col<0) return;

      if(!SelectedCellShowing())
      {
         int NewLeftCol = LeftCol + dir;

         if(NewLeftCol<1) NewLeftCol = 1;
         if(NewLeftCol>=ColCount) NewLeftCol = ColCount-1;

         LeftCol = NewLeftCol;
      }

      switch(dir)
      {
         case -1:
         {
            if(Col>1)
               SelectGridCell(Col-1,Row);
            else
               Beep();
         }
         break;

         case 1:
         {
            if(Col<ColCount-1)
               SelectGridCell(Col+1,Row);
            else
              Beep();
         }
         break;

         default: SelectGridCell(Col,Row);
      }

      if(!SelectedCellShowing())
      {
         int NewLeftCol = LeftCol + dir;

         if(NewLeftCol<1) NewLeftCol = 1;
         if(NewLeftCol>=ColCount) NewLeftCol = ColCount-1;

         LeftCol = NewLeftCol;
      }
   }

   void __fastcall GGrid::NextGridPage(int dir)
   {
      int BodyRows = BodyRowCount();

      int Page = TopRow/BodyRows +1;

      int TopBdy = (Page-1)*BodyRows + 1;

      if(dir<0)   // Page Up
      {
         if(TopRow == 1)
         {
            Beep();  return;
         }

         if(TopBdy > BodyRows)
            TopRow = TopBdy - BodyRows;
         else
         {
            TopRow = TopBdy;  Beep();
         }
      }

      if(dir>0)   //Page Down
      {
         if(TopBdy < RowCount-BodyRows)
            TopRow = TopBdy + BodyRows;
         else
         {
            TopRow = TopBdy;  Beep();

            NoDownArrow();
         }
      }

      if(Col>=0 && Row>=0)                   // Some Cell selected
      {
         GField* Field = Fields[Col];        // Changes in TopRow may force Editor to Hide

         if(Field)                           // The field is an editor
            if(Field->State>fsDormant)       // The editor is active
               Field->Move();                // So move it
      }
   }


   void __fastcall GGrid::KeyDown                  // HANDLER KeyDown
   (
      Word &Key, Classes::TShiftState Shift
   )
   {
      ClearInfo();

      KeyClick = false; KeyPressed = 0;

      if(SelectedFieldScrolled())         // The grid may have focus, because
      {                                   // there is an active field which
         ExposeSelectedCell();            // has been scrolled out of view.

         Key = 0;                         // Refocus the active field
                                          // but take no action
         return;
      }
      
      if(Shift.Contains(ssShift))                               // MULTI SELECT
      if
      (
         Key==VK_UP   || Key==VK_DOWN
                      ||
         Key==VK_LEFT || Key==VK_RIGHT
      )
      {
         KeyClick = true;  KeyPressed = Key;

         TStringGrid::KeyDown(Key, Shift);

         Key = 0;

         return;
      }

      switch(Key)
      {
         case VK_HOME: case VK_END:
         {
            KeyClick = true;  KeyPressed = Key;

            TStringGrid::KeyDown(Key, Shift);
         }
         break;

         case VK_UP:       NextGridRow(-1);                 break;
         case VK_DOWN:     NextGridRow( 1);                 break;

         case VK_LEFT:     NextGridCol(-1);                 break;
         case VK_RIGHT:    NextGridCol( 1);                 break;

         case VK_PRIOR:    NextGridPage(-1);                break;
         case VK_NEXT:     NextGridPage( 1);                break;

         case VK_ESCAPE:
         {
            if
            (
               Selection.Left != Selection.Right
                              ||
               Selection.Top  != Selection.Bottom
            )
               NoCellSelected();
         }
         break;

         default:       TStringGrid::KeyDown(Key, Shift);   return;
      }
      Key = 0;
   }

   void __fastcall GGrid::MouseDown                // HANDLER MouseDown
   (                                               // Click on Col or Row 0
      TMouseButton Button, TShiftState Shift,
      int i, int j
   )
   {
      TGridCoord CellNow = MouseCoord(i, j);

      int col = CellNow.X, row = CellNow.Y;

      if(row>0 && col>0 || Button != mbLeft)
      {
         TStringGrid::MouseDown(Button,Shift, i,j);

         return;
      }

      if(row == 0)                                   // Click Header Row - to sort
      if(SortEnabled)
      {
         for(int c=0; c<(int)SortDir.size(); c++)    // Clear the previous sort
            if(c != col)SortDir[c] = gsNo;

         SortColumn(col);

         Invalidate();

         return;
      }
   }
   //---------------------------------------------------------------------------

   bool __fastcall GGrid::ColChange(int Control)
   {
      // INVOKED if
      // a. We are TABBING in a field and the text has changed
      // b. We have clicked the grid with EditCell != 0
      //    The text has changed
      //    The previous row is the same, but PrevCol != Col

      switch(Control)
      {
         case rcField: Beep();           return false;   // Not Allowed in TAB
         case rcGrid:  EditCell = NULL;  return true;    // Allowed in click

         default:      EditCell = NULL;
      }
      return true;                                       // Never gets here
   }

   void __fastcall GGrid::RowChange(int Control)
   {
      // INVOKED if
      // a. We are DOWN/UP arrow in a field and the text has changed
      // b. We have clicked the grid with EditCell != 0
      //    The text has changed and PrevRow != Row

      switch(Control)
      {
         case rcField: Beep();   break;   // Not Allowed in Arrow
         case rcGrid:  Beep();   break;   // Not Allowed in click

         default:      EditCell = NULL;
      }
   }

   void __fastcall GGrid::GridChange()
   {
     // INVALID GRID CHANGE MESSAGE BECAUSE INCOMPLETE EDIT
   }

   void __fastcall GGrid::NoDownArrow()
   {
      // INVALID DownArrow MESSAGE BECAUSE NO MORE ROWS WITH DATA
   }

   void __fastcall GGrid::NotShowing()
   {
      // INVALID Escape OR Return MESSAGE BECAUSE EDITOR HIDDEN
   }

   void __fastcall GGrid::SortColumn(int col)
   {
      // The override looks at SortDir[] to decide what to do
      // It does the sort
      // It updates SortDir
      // DrawCell override uses SortDir to render the header

      if(SortDir[col] == gsNa) return;
      
      switch(SortDir[col])
      {
         case gsUp: SortDir[col] = gsDn;    break;
         case gsDn: SortDir[col] = gsUp;    break;

         default:   SortDir[col] = gsUp;
      }
   }

   bool __fastcall GGrid::Parse(const Word Key)
   {
      // Invoked by GGrid on Enter in FieldKeyDown
      // Descendants may invoke from Field.KeyPress
      
      return true;
   }

   void __fastcall GGrid::PostParse(void)
   {
      // Called after the parse in FieldKeyDown and when the grid
      // has regained focus
   }

   void __fastcall GGrid::InvalidateGridRow(int row)
   {
   	if(0<=row && row<RowCount)
      	InvalidateRow(row);
   }

}
//---------------------------------------------------------------------------

#pragma package(smart_init)
 