////////////////////////////////////////////////////////////////////////////////
// MODULE GridGUI:       Base class for GoDraw's grids        G.Kaniuk 08/06/03
//
#include <string>
#include <vector>

#include <grids.hpp>
#include <classes.hpp>
#include <comctrls.hpp>

#ifndef GridDUIH
#define GridDUIH

namespace ToolsGUI                       // MAGIC numbers
{
   extern const int defFontSize;         // Main grid font size - Points
   extern const int defStrideSize;       // Grid row stride - Points

   extern const TColor defPaneColor;     // Main form pane input

   enum                                  // Sort direction
   {
      gsNo =  0,                         // Not sorted
      gsUp =  1,                         // Ascending
      gsDn = -1,                         // Descending
      gsNa = -2                          // Not allowed
   };

   struct GRatio                         // SCALING SUPPORT
   {
      int numerator, denominator;

      GRatio(){ numerator = 1;  denominator = 1; }

      GRatio(const int n, const int d){ numerator = n;  denominator = d; }

      int operator()(int val){ return (numerator*val)/denominator; }

      int Rounding(int val)
      {
         return (2*numerator*val + denominator)/(2*denominator);
      }
   };

   class GField;
   class GGrid;

   extern GField* EditCell;

   void ClearEditCell();
   int  EditCellRow();
   bool EditCellIsNot(GField* Field);
   bool EditCellIs(GField* Field);
   bool EditingCell();
   bool EditingGrid(GGrid* grid);
   void ResetToEditCell(void);
}
//------------------------------------------------------------------------------

namespace ToolsGUI
{
   class GEdit : public TEdit
   {
      private:
         GField* Field;

      protected:
         DYNAMIC void __fastcall KeyDown(Word& Key, TShiftState Shift);

         void __fastcall KeyDownSpecial(TCMWantSpecialKey& Message);

      public:
         __fastcall GEdit(GField* Field);

         virtual GField* __fastcall GetField();
         virtual GGrid*  __fastcall GetGrid();
         virtual void    __fastcall SelectText();

         BEGIN_MESSAGE_MAP
            MESSAGE_HANDLER(CM_WANTSPECIALKEY, TCMWantSpecialKey, KeyDownSpecial)
         END_MESSAGE_MAP(TEdit)
   };
}

namespace ToolsGUI
{
   enum              // FIELD STATES
   {
      fsDormant,     // Initial state on construction
      fsMounted,     // Field attached to cell, not displayed
      fsShowing      // Field visible
   };

   class GGrid;

   class GField
   {
      friend GGrid;

      protected:
         int  State;             // Controller

      public:
         GGrid* Grid;            // Access grid geometry

         GEdit* Editor;          // Pointer to the Cell editor 

         String CellText;        // Preserved on Mount()

         int ColM, RowM;         // Set on Mount() to Grid->cell

      public:
         __fastcall GField(){};
         __fastcall GField(GGrid* grid);

         int __fastcall FieldState() { return State; }
         void __fastcall NewState(int state)
         {
         	if(fsDormant<=state && state <= fsShowing)State = state;
         }


      private:
         void __fastcall TransferFocus();
         
      public:
         virtual void __fastcall MountEd(TRect& Cell);
         virtual void __fastcall MoveEd(TRect& Cell);
         virtual void __fastcall ShowEd(void);
         virtual void __fastcall HideEd(void);
         virtual void __fastcall SetFocusEd(void);
         virtual void __fastcall ReMountEd(void);

         virtual bool __fastcall Scroll();
         virtual void __fastcall Mount();
         virtual void __fastcall Dismount();
         virtual void __fastcall HardDismount();                             // END OF EDIT SESSION

         virtual void __fastcall Move();

         virtual void __fastcall Load();
         virtual void __fastcall Store();
         virtual bool __fastcall SameValue();

      public:
         virtual void __fastcall FieldKeyDown(Word& Key, TShiftState Shift);
   };
}
//------------------------------------------------------------------------------

namespace ToolsGUI                              // CLASS GGRID
{
   using std::string;
   using std::vector;

   class GGrid : public TStringGrid
   {
      friend GField;

      public:
         enum                                   // ROW/COL CHANGE CONTROL
         {
            rcField,
            rcGrid
         };

         struct ColSize                         // CONTROL COLUMN WIDTH ADJUSTMENTS
         {
            int minwidth;                       // Starting size
            int expand;                         // Priority for expansion
         };

      protected:
         Graphics::TBitmap *SortDn, *SortUp;    // PaneBIT.res

      protected:
         bool hScrollBar;                       // True if permanent horizontal scroll bar
         bool SortEnabled;                      // Allows click on header tabs

         int HeadRowHeight;                     // Header row height is bigger than rows
         int BodyStride;                        // Body row height + GridLineWidth

         TStatusBar* StatusBar;                 // Message Display

         vector<GField*>Fields;                 // Non-NULL fields are editable
         vector<int> SortDir;                   // +1 Up, -1 Down, 0 None , -2 Never

         int CurFieldId;                        // Current selected field

         bool KeyClick;                         // True when a Key generates Click()
         Word KeyPressed;                       // The key that caused the Click()

      public:
         TColor BackGround;                     // The default brush colour

      public:
         __fastcall GGrid
         (
            TWinControl* parent,
            TStatusBar* statusbar,
            const int fields
         );

         __fastcall ~GGrid();

         DYNAMIC void __fastcall Click(void);
         
         int  __fastcall BodyRowCount();

         void __fastcall Sheave(const ColSize colsizes[],int first,int& last);
         void __fastcall ReBuild(const ColSize colsizes[]);

         void __fastcall AddRow(int r);
         void __fastcall AddRows(int rows);
         void __fastcall ShowInfo(int messid);
         void __fastcall ShowInfo(String mess);
         void __fastcall ShowInfoStatus(int info, int status);
         void __fastcall ClearInfo();

         void __fastcall NoCellSelected();
         bool __fastcall BlockSelected(void);
         void __fastcall SelectGridCell(int col, int row);
         bool __fastcall SelectedCellShowing();
         bool __fastcall SelectedFieldScrolled();
         void __fastcall SelectPopupCell(TPoint PopPoint);
         void __fastcall AnalysePopupCell(TPoint PopPoint, int&PopCol, int& PopRow);

         void __fastcall ExposeSelectedCell();

         void __fastcall NextGridRow(int dir);
         void __fastcall NextGridCol(int dir);
         void __fastcall NextGridPage(int dir);

         void __fastcall SetColor(TColor color){ Color = color; }

         void __fastcall MountFieldAfter(GField* PrevField);
         void __fastcall ResetToField(GField* PrevField);

         void __fastcall InvalidateGridRow(int row);

         bool __fastcall EditPossible(void);

         GField* GetCurField(void){ return Fields[CurFieldId]; }

      protected:                                // EVENT HANDLERS
         virtual void __fastcall DrawCell
         (
            int col, int row,
            const Windows::TRect& Rect,
            TGridDrawState State
         );

         virtual bool __fastcall SelectCell(int col, int row);

         DYNAMIC void __fastcall TopLeftChanged(void);

         DYNAMIC void __fastcall KeyDown(Word &Key, Classes::TShiftState Shift);
         DYNAMIC void __fastcall MouseDown(TMouseButton Button, TShiftState Shift, int X, int Y);

         Controls::TWndMethod OriginalWindowProc;

         virtual void __fastcall GridWindowProc(TMessage& Message);

         virtual void __fastcall RowChange(int Control);
         virtual bool __fastcall ColChange(int Control);
         
         virtual void __fastcall GridChange(void);
         virtual void __fastcall NoDownArrow(void);
         virtual void __fastcall NotShowing(void);

         virtual bool __fastcall Parse(const Word Key);
         virtual void __fastcall PostParse(void);
         virtual void __fastcall SortColumn(int col);

         void __fastcall MouseWheelUpDn(TObject *Sender,TShiftState Shift, const TPoint &MousePos, bool &Handled){ Handled = true; }
         void __fastcall MouseWheel(TObject *Sender,TShiftState Shift,int WheelDelta, const TPoint &MousePos, bool &Handled){ Handled = true; }

   };
}
//---------------------------------------------------------------------------
#endif
