////////////////////////////////////////////////////////////////////////////////
// MODULE LoadOldFile:  Loads DOS godrawd files into GTour      G.Kaniuk 16/06/02
//                      The two purposes of this are:
//                      1. Produce file for Euro ratings
//                      2. Produce Tab-Delimited file for Web
//
// Once GTour has been set up by LoadOldTour, it is in the same state as
// as if it had been loaded from a .gdt file

#include <algorithm>
#include <ctype.h>
#include <vcl.h>
#pragma hdrstop

#include "TourRES.h"
#include "ETour.h"
#include "Strings.h"

#include "Tour.h"
#include "TourStream.h"
#include "LoadOldFile.h"

#pragma package(smart_init)

namespace OldTourFile                                        // STRUCTURE CONSTANTS AND SIMPLE TOOLS
{
   using namespace ToolsGUI;
   using namespace Tours;
   
   using Tools::intsi;
   using Tools::ints;

   static const int MaxPlayers = 512;                        // LIMITS
   static const int MaxBoards  = 256;
   static const int MaxRounds  =  12;

                                                             // FIELD SIZES
   static const int SizeHead  = 191;			                // The entire header
   static const int SizePline =  74;                         // Player details
   static const int SizeGline =  86;                         // Max Game details list
   static const int SizeTerm  =   2;                         // Line terminator is CR+LF
   static const int SizeGame  =   6;                         // One game detail

   static const int SizeRoom  =   4;                         // Boards in room
   static const int SizeFname =  12;                         // DOS root filename
   static const int SizePname =  22;                         // Player name field
   static const int SizeArea  =  13;                         // Country or club
   static const int SizeClub  =   3;                         // Three char club code
   static const int SizeLand  =   2;                         // Two char country code

   static const int SizeNote  =  12;                         // Free field

                                                             // HEADER FIELD POSITIONS
   static const int PosFilename = 0;                         // Embedded tournament file name
   static const int PosState    = 16;                        // Current State
   static const int PosRound    = 20;                        // Current Round
   static const int PosRounds   = 25;                        // Number of rounds
   static const int PosTie1     = 29;                        // First tie break
   static const int PosTie2     = 33;                        // Second tie break

   static const int PosBar      = 37;                        // MacMahon Bar
   static const int PosBlack    = 46;                        // First named
   static const int PosWhite    = 53;                        // Second named
   static const int PosBarSeg   = 88;                        // Segregation above bar
   static const int PosBoards   = 91;                        // Boards per room
   static const int PosPlayers  = 186;                       // Number of players

                                                             // PLAYER FIELD POSITIONS
   static const int PosMMS      = 0;                         // Start Entr MMS field(1 char)
   static const int PosName     = 2;                         // Start of player name field
   static const int PosGrade    = PosName + SizePname + 1;   // Name_Grade
   static const int PosClub     = PosGrade +4;               // Grade_Club
   static const int PosLand     = PosClub + SizeArea + 1;    // Club_Country
   static const int PosNote     = PosLand + SizeArea + 1;    // Admin

                                                             // GAME FIELD POSITIONS
   static const int PosGame     = 1;                         // Start of first game on line
   static const int PosOrder    = 3;                         // Board order code
   static const int PosDir      = 4;                         // Direction of draw
   static const int PosRes      = 5;                         // Result code

   static const int EmptyRoom   = 999;
                                                             // ENCODING
   static const char SleepCode  = 'N'; // Why not char('N')?
   static const char YesCode    = char('Y');

   static const char KyuCode = char('K');                   // Conversions between strings
   static const char kyuCode = char('k');
   static const char DanCode = char('D');                   // & zero-shodan representation
   static const char danCode = char('d');

   static const ints  MinGrade = -60;                         // Godraw minimum is 30k
   static const ints  MaxGrade = 16;                          // Godraw maximum is 9D
   static const ints  MoScore  = 93;   	                    // Initial MacMahon Score  ranges from -60 to 44 = 20 + 12*2

   static const intsi First    = -1;
   static const intsi Second   =  1;
   static const ints  Zero     = (ints)Tools::cZero;

   static const Tours::tie_t TieBreak[6] =                                    // Index from old file
   {
      Tours::tbWins,
      tbSoS, tbSoDoS,
      tbCuss, tbCusp,
      tbNone
   };

   static const Tours::TourState StateCode[13] =
   {
      tsVirgin,     tsTourStart, tsRegNow,    tsRegClosed,
      tsDrawDone,   tsDrawDone,   tsResultNow,
      tsRoundDone,  tsTourDone,
      tsEditForced, tsEditNow,   tsForcedPair,
      tsPastRound
   };

   static const String ErGoDraw = "GoDraw";
   static const String ErCorrupt = "GoDraw File Corrupt";
   static const String ErHandle = "Bad Handle";
   static const String ErLength = "Bad File Length";
   static const String ErHeader = "Bad File Header";
   static const String ErFileName = "Bad File Name";

   static int GetInt(Byte* Mem)                           // Convert a 3 character integer ddd to int
   {
      try
      {
         return String((char*)Mem, 3).ToInt();
      }
      catch(...)
      {
         Tours::Throw(ToolsGUI::erOperr, IDT_GetInt, IDT_Corrupt);
      }
      return 0;
   }

   static ints GradeToInts(Byte* Grade)
   {
      Byte* pGrade = Grade;

      ints Gval = (ints)(*pGrade) - Zero;  ++pGrade;

      if(isdigit(*pGrade))
      { Gval = 10*Gval + (ints)(*pGrade) - Zero;  ++pGrade; }

      switch(*pGrade)
      {
         case KyuCode: case kyuCode: Gval = -Gval;    break;

         case DanCode: case danCode: Gval =  Gval-1;  break;

         default: Throw(erOperr, IDT_GradeToInt, IDT_Corrupt);
      }

      if(Gval<MinGrade || Gval>MaxGrade)
         Throw(erOperr, IDT_GradeToInt, IDT_Corrupt);

      return Gval*Tours::mc_scale;
   }

   static intsi DecodeResult(char ResultCode)
   {
      switch(ResultCode)
      {
         case '1': return Tours::grWin;                // Worth mc_scale
         case '0': return Tours::grLose;               // Worth 0
         case 'j': return Tours::grJigo;               // Worth mc_scale/2

         case 'b': return Tours::grFwin;               // Defined by GoDraw
         case 'd': return Tours::grDlos;
         case 'v': return Tours::grVoid;

         case 'N': return Tours::grSleep;              // Worth 1 potentially

         case 'Y': return Tours::grNone;               // Still playing, no result

         default:  return Tours::grLose;               // P,Y or garbage
      }
   }

   static intsi DecodeOrder(char OrderCode)
   {
      switch(OrderCode)
      {
         case 'F': return Tours::gcBlack;        // First named player
         case 'S': return Tours::gcWhite;        // Second named player

         default:  return Tours::gcNone;         // Sleeping
      }
   }

   static intsi DecodeParity(char ParityCode)
   {
      switch(ParityCode)
      {
         case '-': return Tours::gdDown;
         case '+': return Tours::gdUp;
         case '.': return Tours::gdEven;

         dfault:
            Tours::Throw(erOperr, IDT_DecodeParity, IDT_Corrupt);
      }
      return gdEven;
   }

}
////////////////////////////////////////////////////////////////////////////////

namespace OldTourFile                                        // WORKSHOP
{
   using namespace Tours;

   using Tools::s_S;
   using Tools::S_s;

   static string ForgeFileName(string FullName, GTour& Tour)
   {
      size_t i = FullName.find(Tools::sDot);

      int j, k;

      if(i<FullName.size()) FullName[i] = Tools::cUline;			// C:\...\oldname.dda -> C:\...\oldname_dda.gdt

      for(j=i; j>=0; j--)
      	if(FullName[j] == Tools::cSep)
         {
         	++j;

         	break;
         }

      if(i-j > GTour::SizeName)											// TourName is fixed with trailing spaces
      	i = j + GTour::SizeName;

      for(k=j; k<(int)i; k++)												// Extract the root to make TourName = oldname
      	Tour.TourName[k-j] = FullName[k];

      for(k=i; k<GTour::SizeName; k++)
      	Tour.TourName[k-j] = Tools::cSpace;

      return FullName + Tours::TourExt;
   }

   void LoadOldHeader(GTourStream& TourFile, GTour& Tour)
   {
      using Tools::s_S;

      Tour.Rules.Default();

      Tour.SpareOne = Tour.SpareTwo = 0;

      if(TourFile.Stream->Size < SizeHead)
         Throw(erOperr, IDT_LoadOldHeader, IDT_Corrupt);

      // READ THE HEADER
      Byte Header[SizeHead];

      if(TourFile.Stream->Read(Header, SizeHead) != SizeHead)
         Throw(erBug, IDT_LoadOldHeader, IDT_BadReadSize);


      // VERIFY EMBEDDED FILENAME
      String Fname = ExtractFileName(s_S(TourFile.FileName));
      String Hname = String((char*)Header, SizeFname).TrimRight();

      if(Fname.LowerCase()!= Hname.LowerCase())
         Throw(erOperr, IDT_LoadOldHeader, IDT_Corrupt);

      // CURRENT TOURNAMENT STATE
      Tour.State = StateCode[GetInt(Header + PosState)];

      // EXTRACT Rounds, Players, and scaled MacMahon Bar
      Tour.Rules.Rounds  = (size_t)GetInt(Header + PosRounds);
      Tour.Round         = (size_t)GetInt(Header + PosRound);
      Tour.Roundl        = Tour.Round;

      if(Tour.Round>0 && Tour.State!=tsRoundDone && Tour.State!=tsTourDone)
         Tour.Roundl = Tour.Round - 1;

      Tour.Players = (size_t)GetInt(Header + PosPlayers);

      int OldTie1 = GetInt(Header + PosTie1);
      int OldTie2 = GetInt(Header + PosTie2);

      if(0<=OldTie1 && OldTie1<=6)
         Tour.Rules.TieOne = TieBreak[OldTie1];
      else
         Throw(erOperr, IDT_LoadOldHeader, IDT_Corrupt);

      if(0<=OldTie2 && OldTie2<=6)
         Tour.Rules.TieTwo = TieBreak[OldTie2];
      else
         Throw(erOperr, IDT_LoadOldHeader, IDT_Corrupt);

      int GradeBar = GetInt(Header + PosBar);

      Tour.Rules.MacBar = GradeBar>0 ? mc_scale*(GradeBar-1) : mc_scale*GradeBar;

      Tour.Rules.BarSeg = *(Header + PosBarSeg)==YesCode;

      Tour.Rules.KomiDiff = -mc_scale;	// Specified in scaled units of stones
      Tour.Rules.KomiRate = 0;      	// Units are points per stone scaled;

      Tour.Rules.SupBar = Tours::sup_bar + mc_scale*(Tour.Rules.Rounds + 1);

      Tour.Rules.Prizes = Tour.Rules.Rounds*mc_scale;

      Tour.Rooms = 0;

      Byte* PosRoom = Header + PosBoards + 1;

      for(int i=0; i<19; i++)
      {
         Tour.Boards[i] = GetInt(PosRoom);

         if(Tour.Boards[i] != EmptyRoom)
         {
            ++Tour.Rooms;  PosRoom += SizeRoom;
         }
         else
         {
            Tour.Boards[i] = 0;

            break;
         }
      }

      Tour.Rules.Checked = true;
   }

   void LoadPlayerGames(TStream* Stream, GRules& Rules, Byte PlayerLine[], Byte GamesLine[])
   {
      Byte TermLine[SizeTerm];                               // Two line record separated by blank line

      int SizeHline = (SizeGame+1)*Rules.Rounds + SizeTerm;  // Actual usage of GamesLine

      if(SizeHline<=0 || SizeHline>SizeGline)
         Throw(erOperr, IDT_LoadPlayerGames, IDT_Corrupt);

      if(Stream->Read(PlayerLine, SizePline) != SizePline)Throw(erBug, IDT_LoadPlayerGames, IDT_BadReadSize);
      if(Stream->Read(GamesLine,  SizeHline) != SizeHline)Throw(erBug, IDT_LoadPlayerGames, IDT_BadReadSize);
      if(Stream->Read(TermLine,   SizeTerm)  != SizeTerm) Throw(erBug, IDT_LoadPlayerGames, IDT_BadReadSize);
   }

   static void FormPlayer(Byte PlayerLine[], GPlayer& P, GTour& Tour)
   {
      using std::string;

      using Tools::cSpace;
      using Tools::trim_right;
      using Tools::slice;
      using Tools::intsi;

      char* Ns = PlayerLine+PosName;
      char* Cs = PlayerLine+PosClub;
      char* Ls = PlayerLine+PosLand;
      char* Ms = PlayerLine+PosNote;

      string Name = trim_right(string(Ns, Ns+SizePname)),Given,Middle,Family;
      string Space(Tools::sSpace);
      
      if(Tools::parse_name(Name,Given,Middle,Family))
      	P.Name = Family + string(Tools::sComma) + Given + Space + Middle;
      else
      	P.Name = Name;

      string Club = trim_right(string(Cs, Cs+SizeArea));
      string Land = trim_right(string(Ls, Ls+SizeArea));
      string Note = trim_right(string(Ms, Ms+SizeNote));

      P.Note = Note;

      Tour.Clubs.encode(Club, P.Club);                   // Ignore any Query return
      Tour.Lands.encode(Land, P.Land );                  // for legacy files

      P.Group   = string(2, cSpace);                     // Group = ID+property - concept does not exist before V6

      P.Grade   = GradeToInts(PlayerLine+PosGrade);      // Grade in Zero Shodan units

      P.MiScore = 2*((ints)(PlayerLine[0]) - MoScore);   // Entry MacMahon score coded in first char
                                                         // Factor 2 for mc_scale = 4

      P.McScore = P.MiScore;                             // Initialisation
      P.Cbal    = 0;
      P.Mbal    = 0;
      P.Mout    = 0;

      switch(Tour.State)                                 // INCOMPLETE STATE INIT!!!
      {
      	case Tours::tsRegNow: 		P.State = Tours::psEntered;
         break;

         case Tours::tsRegClosed:	P.State = Tours::psRegistered;
         break;

         default: 						P.State = Tours::psSleep;
      }
   }

   static void FormGames(Byte GamesLine[], GPlayer& P, int& MaxBoard)
   {
      using std::vector;

      Byte* fGame = GamesLine + PosGame;                 // Source

      vector<GGame>::iterator rGame = P.Games.begin();
      vector<GGame>::iterator eGame = P.Games.end();

      int Sleeping = 0;                                  // Gain a point when even

      while(rGame != eGame)
      {
         rGame->Board  = GetInt(fGame);                     // Board number
         rGame->Colour = DecodeOrder(*(fGame+PosOrder));    // Board order
         rGame->Parity = DecodeParity(*(fGame+PosDir));
         rGame->Result = DecodeResult(*(fGame+PosRes));     // Result is scaled*2


         if(rGame->Board > (size_t)MaxBoard)                // Monitor the maximum board used
            MaxBoard = rGame->Board;

         if(rGame->Board > 0)                               // Actually playing
         {
            P.Cbal += rGame->Colour;
            P.Mbal += rGame->Parity;
            P.Mout += Tools::modulus(rGame->Parity);
         }

         switch(rGame->Result)
         {
            case grWin: case grLose: case grJigo:
            {
               P.McScore += Score(rGame->Result);
            }
            break;

            case grFwin:
            {
               P.McScore += gsOne;
            }
            break;

            case grSleep:
            {
               ++Sleeping;

               if(Sleeping%2 == 0)P.McScore += gsOne; // Get whole point every two rounds missed
            }
            break;

            case grNone:                              // No result yet
            {
               // P.State = psPlaying;               // For all rounds beyond current round
            }
            break;

            default:
            {
               P.McScore  += gsZero;
            }
         }

         ++rGame;  fGame += SizeGame+1;
      }
   }
}
////////////////////////////////////////////////////////////////////////////////

namespace OldTourFile                                  // INTERFACE
{
   using std::vector;

   void LoadOldTour
   (
      GTourStream& TourFile,                          // Already opened file
      GTour& Tour
   )
   {
      Byte PlayerLine[SizePline];                     // Register
      Byte GamesLine[SizeGline];                      // Results

      LoadOldHeader(TourFile, Tour);                  // In particular, get rounds and players

      Tour.ReBuild(Tour.Players);                     // Destroys, then recreates lists

      std::list<GPlayer>::iterator sPlayer = Tour.PlayerList.begin();
      std::list<GPlayer>::iterator aPlayer;


      // FIRST PASS - load player details and games with board numbers
      int MaxBoard = 0;

      aPlayer = sPlayer;      

      for(size_t p=1; p<=Tour.Players; p++)
      {
         LoadPlayerGames(TourFile.Stream, Tour.Rules, PlayerLine, GamesLine);

         GPlayer& Player = *aPlayer;

         Player.Pid = p;

         FormPlayer(PlayerLine, Player, Tour);

         FormGames(GamesLine, Player, MaxBoard);

         switch(Player.Games[Tour.Roundl-1].Result)
         {
         	case Tours::grSleep: Player.State = Tours::psSleep; 	break;
            case Tours::grNone:	Player.State = Tours::psPlaying; break;

            default:       		Player.State = Tours::psPlaying;
         }

         ++aPlayer;
      }


      // SECOND PASS - attach paired players to their boards.
      //             - identify opponents, and calculate handicaps
      vector <GBoard> Pair(MaxBoard+1);

      for(size_t r=0; r<Tour.Rules.Rounds; r++)
      {
         for(int b=0; b<=MaxBoard; b++)
            Pair[b] = GBoard();
            
         aPlayer = sPlayer;

         for(size_t p=1; p<=Tour.Players; p++)
         {
            GPlayer& Player = *aPlayer;

            if(Player.Pid>0)                                        // Played at least one Game
            {
               GGame& Game = Player.Games[r];

               GBoard& PairNow = Pair[Game.Board];

               if(Game.Colour == gcBlack)PairNow.Black = Player.Pid;
               if(Game.Colour == gcWhite)PairNow.White = Player.Pid;

               if(PairNow.Black>0 && PairNow.White>0)                // We now know the pair
               {
                  // FIRST - identify opponents by entry number
                  GPlayerPtr B = Tour.Listed[PairNow.Black];
                  GPlayerPtr W = Tour.Listed[PairNow.White];

                  GGame& Bg = B->Games[r];
                  GGame& Wg = W->Games[r];

                  Bg.Oppid = PairNow.White;
                  Wg.Oppid = PairNow.Black;


                  // AND NOW - calculate the handicaps
                  int MacBar = Tour.Rules.MacBar;

                //  int Bmac = B->McScore  - Score(Bg.Result);         // MacMahon scores
                //  int Wmac = W->McScore  - Score(Wg.Result);         // at beginning of round

                  ints Bmac = B->McMahonScore(r);							// MacMahon scores
                  ints Wmac = W->McMahonScore(r);  						// at beginning of round

                  int McDiff = Tools::modulus(Bmac-Wmac);            // McMahon score difference

                  int Stones = 0, Lstone = 9*Tours::mc_scale;

                  if(McDiff > Tours::mc_scale)
                  	Stones = McDiff - Tours::mc_scale;					// MMdiff - 1

                  if(Stones > Lstone)
                  	Stones = Lstone;

                  if(B->Grade>MacBar || W->Grade>MacBar) Stones = 0; // No handicap above the bar

                  if(Stones > 0)
                  {
                     if(Bmac < Wmac) { Bg.Stones = Stones;  Wg.Stones = -Stones; }
                     if(Wmac < Bmac) { Wg.Stones = Stones;  Bg.Stones = -Stones; }
                  }
               }
            }
            ++aPlayer;
         }// while
      }//for

      Tour.FileName = ForgeFileName(TourFile.FileName, Tour);
   }

}
////////////////////////////////////////////////////////////////////////////////








