////////////////////////////////////////////////////////////////////////////////
// MODULE Sort:       Implement general sorting for GTour      G.Kaniuk 7/08/02
////////////////////////////////////////////////////////////////////////////////
// RELEASE on 20/09/02
////////////////////////////////////////////////////////////////////////////////

#pragma hdrstop

#include "ETourRes.h"
#include "TourRes.h"
#include "Strings.h"
#include "Sort.h"
#include "Tour.h"
////////////////////////////////////////////////////////////////////////////////

namespace Tours                                    // MANAGEMENT FOR ALL SORTING
{
   static GTour* Tour = NULL;                      // Protected pointer to GTour

   void SortInterface(GTour* ThisTour)
   {
      Tour = ThisTour;
   }
}
////////////////////////////////////////////////////////////////////////////////

namespace Tours                                   // REGISTER SORT COMPARE FUNCTIONS
{
   bool OrderEntry::operator()(const size_t p,  const size_t q)
   {
   	if(Tour->SortSense>0) return Tour->Listed[p]->Pid < Tour->Listed[q]->Pid;
      if(Tour->SortSense<0) return Tour->Listed[p]->Pid > Tour->Listed[q]->Pid;

      return p<q;
   }

   bool OrderTicked::operator()(const size_t p,  const size_t q)
   {
      GPlayerPtr P = Tour->Listed[p], Q = Tour->Listed[q];

      int Tp = P->State >= Tours::psRegistered ? 0 : 1;
      int Tq = Q->State >= Tours::psRegistered ? 0 : 1;

      if(Tour->State > Tours::tsRegClosed)
      {
         if(P->Joining(Tour->Round)) Tp = 2;
         if(P->Leaving(Tour->Round)) Tp = 3;

         if(Q->Joining(Tour->Round)) Tq = 2;
         if(Q->Leaving(Tour->Round)) Tq = 3;
      }

      if(Tp == Tq) return OrderName()(p,q);

      if(Tour->SortSense>0) return Tp < Tq;
      if(Tour->SortSense<0) return Tq < Tp;

      return p<q;
	}

   bool OrderName::operator()(const size_t p,  const size_t q)
   {
      using Tools::parse_family;
      using std::string;

      GPlayerPtr P = Tour->Listed[p], Q = Tour->Listed[q];

      string pfirst,pmid,plast;  parse_family(P->Name, pfirst, pmid, plast);
      string qfirst,qmid,qlast;  parse_family(Q->Name, qfirst, qmid, qlast);

      if(plast == qlast)
      	if(pfirst == qfirst)
         {
         	if(Tour->SortSense>0)return pmid < qmid;
            if(Tour->SortSense<0)return pmid > qmid;
            return p<q;
         }
         else
         {
         	if(Tour->SortSense>0)return pfirst < qfirst;
            if(Tour->SortSense<0)return pfirst > qfirst;
            return p<q;
         }
      else
      {
      	if(Tour->SortSense>0)return plast < qlast;
         if(Tour->SortSense<0)return plast > qlast;
         return p<q;
      }
   }

   bool OrderGrade::operator()(const size_t p,  const size_t q)
   {
      GPlayerPtr P = Tour->Listed[p], Q = Tour->Listed[q];

      if(P->Grade == Q->Grade)
         return OrderName()(p,q);
      else
      {
         if(Tour->SortSense>0)return P->Grade > Q->Grade;
         if(Tour->SortSense<0)return P->Grade < Q->Grade;
         return p<q;
      }
   }

   bool OrderClub::operator()(const size_t p,  const size_t q)
   {
      GPlayerPtr P = Tour->Listed[p], Q = Tour->Listed[q];

      string Pclub = Tour->ClubName(P->Club), Qclub = Tour->ClubName(Q->Club);

      if(Pclub == Qclub)
         return OrderName()(p,q);
      else
      {
         if(Tour->SortSense>0)return Pclub < Qclub;
         if(Tour->SortSense<0)return Pclub > Qclub;

         return p<q;
      }
   }

   bool OrderLand::operator()(const size_t p,  const size_t q)
   {
      GPlayerPtr P = Tour->Listed[p], Q = Tour->Listed[q];

      string Pland = Tour->LandName(P->Land), Qland = Tour->LandName(Q->Land);

      if(Pland == Qland)
         return OrderName()(p,q);
      else
      {
         if(Tour->SortSense>0)return Pland < Qland;
         if(Tour->SortSense<0)return Pland > Qland;

         return p<q;
      }
   }

   bool OrderRounds::operator()(const size_t p,  const size_t q)
   {
      GPlayerPtr P = Tour->Listed[p], Q = Tour->Listed[q];

      int Prounds = P->BinaryRounds(), Qrounds = Q->BinaryRounds();

      if(Prounds == Qrounds)
         return OrderName()(p,q);
      else
      {
         if(Tour->SortSense>0)return Prounds > Qrounds;
         if(Tour->SortSense<0)return Prounds < Qrounds;

         return p<q;
      }
   }

   bool OrderGroup::operator()(const size_t p,  const size_t q)
   {
      GPlayerPtr P = Tour->Listed[p], Q = Tour->Listed[q];

      int oP = P->GroupOrdinal(), oQ = Q->GroupOrdinal();

      if(oP == oQ)
         return OrderName()(p,q);
      else
      {
         if(Tour->SortSense>0)return oP > oQ;
         if(Tour->SortSense<0)return oP < oQ;

         return p<q;
      }
   }

   bool OrderNote::operator()(const size_t p,  const size_t q)
   {
   	GPlayerPtr P = Tour->Listed[p], Q = Tour->Listed[q];

      if(P->Note == Q->Note)
      	return OrderName()(p,q);
      else
      {
   		if(Tour->SortSense>0)return P->Note < Q->Note;
         if(Tour->SortSense<0)return P->Note > Q->Note;
         return p<q;
      }
   }
}
////////////////////////////////////////////////////////////////////////////////

namespace Tours                                   // DRAW AND RANK SUPPORT
{
	struct OrderActive : binary_function<size_t, size_t, bool>
   {
   	bool operator()(const size_t p,  const size_t q);
   };

   bool OrderActive::operator()(const size_t p,  const size_t q)
   {
      return
         Tour->Listed[p]->Active()
                  &&
         !Tour->Listed[q]->Active();
   }
   //---------------------------------------------------------------------------

   struct OrderAwake : binary_function<size_t, size_t, bool>
   {
      bool operator()(const size_t p,  const size_t q);
   };

	bool OrderAwake::operator()(const size_t p,  const size_t q)
   {
   	int round = Tour->Round>0 ? Tour->Round-1 : 0;

      bool pAwake = Tour->Listed[p]->Games[round].Result != grSleep;
      bool qAwake = Tour->Listed[q]->Games[round].Result != grSleep;

      return pAwake && !qAwake;
   }
   //---------------------------------------------------------------------------

   struct OrderMacMahon : binary_function<size_t, size_t, bool>
   {
      bool operator()(const size_t p,  const size_t q);
   };

   bool OrderMacMahon::operator()(const size_t p,  const size_t q)
   {
      GPlayerPtr P = Tour->Listed[p], Q = Tour->Listed[q];

      return P->McScore > Q->McScore;
   }
   //---------------------------------------------------------------------------

   struct OrderRank : binary_function<size_t, size_t, bool>
   {
      bool operator()(const size_t p,  const size_t q);
   };

   bool OrderRank::operator()(const size_t p,  const size_t q)
   {
      GPlayerPtr P = Tour->Listed[p], Q = Tour->Listed[q];

      ints pOne = P->TieOne, pTwo = P->TieTwo;
      ints qOne = Q->TieOne, qTwo = Q->TieTwo;

      if(pOne == qOne)
         if(pTwo == qTwo)
            return OrderGrade()(p,q);
         else
            return pTwo>qTwo;
      else
         return pOne>qOne;
   }
   //---------------------------------------------------------------------------

   struct OrderPrize : binary_function<size_t, size_t, bool>
   {
      bool operator()(const size_t p,  const size_t q);
   };

   bool OrderPrize::operator()(const size_t p,  const size_t q)
   {
      GPlayerPtr P = Tour->Listed[p], Q = Tour->Listed[q];

      size_t Pwins = P->Wins(), Qwins = Q->Wins();

      if(Pwins > Qwins) return true;		 // Order by wins

      if(Pwins == Qwins)
      	return P->Rank < Q->Rank;			// Order by higher rank
      else                                
      	return false;
   }
}
////////////////////////////////////////////////////////////////////////////////

namespace Tours                                   // SORT ROUTINES
{
   size_t ActiveSort(GTour* ThisTour)
   {
      SortInterface(ThisTour);      Tour->StartSorted();

      Tour->SortOrder = -1;			Tour->SortSense = 1;

      if(!Tour->Players) return 0;

      vector<size_t>::iterator As = Tour->Sorted.begin() + 1;
      vector<size_t>::iterator Ae = Tour->Sorted.end();

      sort(As,Ae, OrderActive());      // Inactives to back

      size_t ActivePlayers = Tour->Players;

      while(!Tour->Listed[Tour->Sorted[ActivePlayers]]->Active())
      {
         --ActivePlayers;

         if(!ActivePlayers) break;
      }
      return ActivePlayers;
   }
}


namespace Tours                                   					// RANK SORT
{
   size_t RankSort(GTour* ThisTour)
   {
      using std::sort;

      // DEEP SLEEPERS TO BACK
      size_t ActivePlayers = ActiveSort(ThisTour);

      if(!ActivePlayers) return 0;

      // UP TO DATE MACMAHON SCORES AND TIE BREAKS
      for(size_t p = 1; p <= Tour->Players; p++)
      {
      	GPlayerPtr Player = Tour->Listed[p];
         
      	Player->McScore = Player->McMahonScore(Tour->Round);
      }
      Tour->SetTieBreaks();


      // McMAHON SORT
      vector<size_t>::iterator Mb = Tour->Sorted.begin();

      vector<size_t>::iterator Me = Mb + 1 + ActivePlayers;    // Null + Active
      vector<size_t>::iterator Ms = Mb + 1;                    // First player is null
      vector<size_t>::iterator Mm = Ms;

      sort(Ms, Me, OrderMacMahon());

      // SORT EACH McMAHON GROUP IN RANK ORDER
      Mm = Ms;  Tools::intsi McScore = Tour->Listed[*Ms]->McScore;

      while(true)                                              // Rank sort each Active McMahon group
      {
         if(Mm == Me || Tour->Listed[*Mm]->McScore != McScore) // We have completed a McMahon group
         {
            sort(Ms, Mm, OrderRank());

            if(Mm != Me)
            {
               Ms = Mm;
               McScore = Tour->Listed[*Ms]->McScore;
            }
            else
               break;
         }
         ++Mm;
      }

      // ASSIGN RANK TO EACH PLAYER
      for(size_t rank=1; rank<=Tour->Players; rank++)
         Tour->Listed[Tour->Sorted[rank]]->Rank = rank;

      Tour->SortOrder = Tours::skRank;	Tour->SortSense = 1;

      return ActivePlayers;
   }

   size_t PrizeSort(GTour* ThisTour)
   {
		size_t Active = RankSort(ThisTour);

      if(!Active) return 0;

      vector<size_t>::iterator Ps = Tour->Sorted.begin() + 1;
      vector<size_t>::iterator Pe = Tour->Sorted.begin() + 1 + Active;

      sort(Ps, Pe, OrderPrize());

      Tour->SortOrder = -1;	Tour->SortSense = 1;
      
      return Active;
   }
}
// END Sort
////////////////////////////////////////////////////////////////////////////////


