////////////////////////////////////////////////////////////////////////////////
// MODULE Tour:       Implementation of class GTour            G.Kaniuk 28/05/02
//

#include <vcl.h>
#include <algorithm>
#pragma hdrstop

#include "Strings.h"
#include "Format.h"

#include "ETourRES.h"
#include "TourRES.h"

#include "ETour.h"

#include "Sort.h"
#include "Tour.h"
////////////////////////////////////////////////////////////////////////////////

namespace Tours                              // ERROR MANAGER
{
   void Throw(EType etype, short idproc, short idmsg)
   {
      ToolsGUI::ThrowError(etype, IDT_Tour, idproc, idmsg);
   }
}
////////////////////////////////////////////////////////////////////////////////

namespace Tours                              // MAGIC constants
{
   using Tools::uiBounds;

   const uiBounds MacPop[max_rounds] =          // McMahon bar population
   {
      uiBounds( 2, 4), uiBounds( 3, 6), uiBounds( 4, 8), uiBounds( 5,10),
      uiBounds( 6,12), uiBounds( 7,15), uiBounds( 8,18), uiBounds( 9,22),
      uiBounds(10,26), uiBounds(11,30), uiBounds(12,35), uiBounds(13,40)
   };

   const Tools::uiBounds SupPop[max_rounds] =  // Supergroup population
   {
      uiBounds( 2, 4), uiBounds( 3, 6), uiBounds( 4, 8), uiBounds( 5,10),
      uiBounds( 6,12), uiBounds( 7,15), uiBounds( 8,18), uiBounds( 9,22),
      uiBounds(10,26), uiBounds(11,30), uiBounds(12,35), uiBounds(13,40)
   };

   const size_t player_text_size = size_name + sizeof(Tools::cTab) + size_note;
   const size_t player_code_size = size_club + size_land + size_group;

   const ints mac_scale = grWin;                    // Convert Zero Shodan to MacMahon scale

   const string NoClub   = "NOC=No Club";           // Standard noclub	can play ANY club
   const string FullClub = "ZZZ=ZZZ";               // ClubList is full

   const string NoLand   = "XX=No Land";            // Standard no land can play ANY land
   const string FullLand = "ZZ=ZZ";                 // LandList is full

   const string GroupCodes = "*~-+: ";              // CODES for enum Group


}
////////////////////////////////////////////////////////////////////////////////

namespace Tours                              // CLASS GRules
{
   void GRules::Default()							// USED ONLY FOR OLD GODRAW
   {
      Rounds = def_rounds;

      TieOne = tbSoS;           TieTwo = tbNone;

      Style = stMcMahon;         RuleSet = rtBGA;

      BarSeg   = false;

      MacBar   = def_macbar;		SupBar    = def_supbar;
      MinGrade = def_mingrade;   KomiGroup = def_komigrp;
      KomiDiff = def_komidiff;   KomiRate  = def_komirate;

      Unresolved = gsHalf;
      SuperDiff  = mc_scale;

      Prizes     = def_rounds*mc_scale;
   }

   GRules::GRules()                             // CONSTRUCTOR
   {
      Default();

      Checked = false;
   }

   size_t GRules::Size()
   {
      using Tools::intsi;

      using Tours::tie_t;
      using Tours::style_t;
      using Tours::rules_t;
      using Tours::pair_t;

      size_t rsize = 0;

      rsize += sizeof(bool);           // Checked
      rsize += sizeof(size_t);         // Rounds
      rsize += sizeof(tie_t)*2;        // TieBreaks
      rsize += sizeof(style_t);        // Style
      rsize += sizeof(rules_t);        // Rule Set
      rsize += sizeof(ints);           // Prizes
      
      rsize += sizeof(bool);           // Bar segregation
      rsize += sizeof(ints)*3;         // Bars: MacBar+SupBar+MinGrade
      rsize += sizeof(ints)*3;         // Komi: Group+Diff+Rate
      rsize += sizeof(ints)*2;			// Unresolved, Superdiff

      return rsize;
   }

}
////////////////////////////////////////////////////////////////////////////////

namespace Tours                              // CLASS GGame
{
   GGame::GGame()
   {
      Board   = Oppid   = 0;
      Stones  = Komi 	= 0;

      Colour = gcNone;    Parity = gdEven;     Result = grNone;
   };

   size_t GGame::Size()
   {
      size_t gamesize  = 2*sizeof(size_t); 	// Board, Opponent
             gamesize += sizeof(ints);   		// Komi
             gamesize += 4*sizeof(intsi);  	// Stones, Colour, Parity, Result

      return gamesize;
   }

   ints Score(intsi Result)
   {
      switch(Result)
      {
         case grLose:      return gsZero;
         case grJigo:      return gsHalf;
         case grWin:       return gsOne;

         case grUsolv:     return gsHalf;    // Unresolved

         case grDlos:      return gsZero;    // Player fail to show
         case grFwin:      return gsOne;     // Opp fail to show
         case grVoid:      return gsZero;    // Both fail to show

         case grSleep:     return gsZero;    // Not playing

         default:          return gsZero;    // Not yet known
      }
   }

   GGame& GGame::operator=(const GGame& G)
   {
   	Board = G.Board;		Oppid = G.Oppid;
      Komi = G.Komi;			Stones = G.Stones;
      Colour = G.Colour;	Parity = G.Parity;		Result = G.Result;

      return *this;
   }

}
////////////////////////////////////////////////////////////////////////////////

namespace Tours                              // CLASS GPlayer
{
   using std::vector;
   using Tools::sEmpty;
   using Tools::ints;
   using Tools::intsi;

   GPlayer::GPlayer()  									// DEFAULT
   {
   	State = psNew;                            // Becomes psEntered via keyboard

      Grade = MiScore = McScore = 0;            // Set on entry
      Cbal  = Mbal    = Mout    = 0;            // Pairing

      TieOne = TieTwo = 0;                      // TieBreaks

      Rank   = 0;

      Games = vector<GGame>(0);            		// Empty Games

      Pid = 0;

      Spare[0] = gpAlways;                      // size_spare is 4
      Spare[1] = gpPrefer;
      Spare[2] = gpAvoid;
      Spare[3] = gpNever;

      Name  = string();
      Club  = string();
      Land  = string();
      Group = string();
      Note  = string();
   }

   GPlayer::GPlayer(size_t id, size_t rounds)   // DUMMY
   {
      State = psNew;                            // Becomes psEntered via keyboard

      Grade = MiScore = McScore = 0;            // Set on entry
      Cbal  = Mbal    = Mout    = 0;            // Pairing

      TieOne = TieTwo = 0;                      // TieBreaks

      Rank  = 0;

      Games = vector<GGame>(rounds);            // Non-empty Games

      Pid = id;

      Spare[0] = gpAlways;                      // size_spare is 4
      Spare[1] = gpPrefer;
      Spare[2] = gpAvoid;
      Spare[3] = gpNever;

      if(rounds>0)
      {
         using Tools::cQuery;

         Name  = string(size_name, cQuery);
         Club  = string(size_club, cQuery);
         Land  = string(size_land, cQuery);
         Group = string(size_group, cQuery);
         Note  = string(size_note, cQuery);
      }
   }

   GPlayer::~GPlayer()
   {
      int m = 0;

      --m;
   }

   size_t GPlayer::Size()
   {
      size_t playersize  = sizeof(PlayerState);

      playersize += sizeof(size_t);          // Id
      playersize += player_text_size;        // Name + note
      playersize += player_code_size;        // Club+country+group
      playersize += 3*sizeof(ints);          // Grade+Mi+Mc
      playersize += 3*sizeof(intsi);         // Cbal+Mbal+Mout

      playersize += Games.size()*GGame().Size();

      return playersize;
   }

   bool GPlayer::Active()                    // Played at least one game
   {
      for(size_t r=0; r<Games.size(); r++)
         if(Games[r].Result != grSleep)
            return true;

      return false;
   }

   bool GPlayer::PlayingAll()                // GK 2015 6.3.9	Played all rounds
   {
      for(size_t r=0; r<Games.size(); r++)
         if(Games[r].Result == grSleep)
            return false;

      return true;
   }

   bool GPlayer::Awake(int Round)				// Playing THIS round
   {
   	if(Round<1) return false;

   	return Games[Round-1].Result != grSleep;
   }

   bool GPlayer::Free()						 		//  Waiting to be paired
   {
   	return State == psRegistered || State == psFinished;
   }

   void GPlayer::SetFree(int Round)				// Set player State to Free  for draw
   {
   	if(Round<1)
      	State = psRegistered;
      else
      	State = psFinished;
   }

   void GPlayer::SetBusy(void)					// Set player state to busy for draw
   {
   	State = psChosen;
   }

   bool GPlayer::Sleeping(int Round)			// Sound asleep
   {
   	const bool Sleep  = State == psSleep;						// Was not in draw
   	const bool NoPlay = Games[Round-1].Result == grSleep;  // Was grSleep - Editing register after draw

      return Sleep && NoPlay;
   }

   bool GPlayer::Joining(int Round)		  		// Waiting to join draw
   {
   	const bool Asleep = State == psSleep;						// Was not in draw
   	const bool WakeUp = Games[Round-1].Result == grNone;  // Was grSleep - Editing register after draw

      return Asleep && WakeUp;
   }

   bool GPlayer::Leaving(int Round)		  		// Waiting to leave draw
   {
   	const bool Awake = State >= psPlaying;						// Was in draw
      const bool Doze  = Games[Round-1].Result == grSleep;	// Was grNone or a result - Editing register after draw

      return Awake && Doze;
   }

   bool GPlayer::MissingResult(int Round)										// CALLED ONLY WHEN
   {                                                                 // TOURNAMENT STATE >= DrawDone
   	const bool Awake = State != psSleep;									// Is in draw
      const bool NoRes = Games[Round-1].Result == grNone;				// No result yet

      return Awake && NoRes;
   }

   bool GPlayer::Playing(int Round)												// Definitely in draw
   {
   	const bool Awake = State != psSleep;
      const bool Play = Games[Round-1].Result != grSleep;

      return Awake && Play;
   }


   bool GPlayer::Unpaired(int Round)			// Playing but no opponent
   {
   	const bool Playing = State==psRegistered || State==psFinished;
      const bool NoOppnt = !Games[Round-1].Oppid;

      return Playing && NoOppnt;
   }

   size_t GPlayer::Wins()
   {
      size_t wins = 0, win;

      for(size_t r=0; r<Games.size(); r++)
      {
         switch(Games[r].Result)
         {
            case grJigo: win = gsHalf; break;
            case grWin:  win = gsOne;  break;
            case grFwin: win = gsOne;  break;
            
            default:     win = 0;
         }
         wins += win;
      }
      return wins;
   }

   ints GPlayer::McMahonScore(const int Round)
   {
   	if(Round< 0) return 0;								// Round may be zero - gives score at
      															// beginning of round 1
   	ints McScore = MiScore;

      int SleepCount = 0;

      for(int r=0; r<Round; r++)
      {
         const intsi Result = Games[r].Result;

         McScore += Tours::Score(Result);				// Score at end of this round

         if(Result == Tours::grSleep) ++SleepCount;
      }

      McScore += (SleepCount/2)*Tours::mc_scale;	// Sleeper points

      return McScore;
	}

   ints GPlayer::McMahonScore(const vector<bool>playing)
   {
   	size_t nRounds = Games.size();

   	if(playing.size() != nRounds) return min_grade;

   	ints McScore = MiScore;

      int SleepCount = 0;

      for(size_t r=0; r<nRounds; r++)
      {
         const intsi Result = playing[r] ? Games[r].Result : 0;

         McScore += Tours::Score(Result);				// Score at end of this round

         if(!playing[r]) ++SleepCount;
      }

      McScore += (SleepCount/2)*Tours::mc_scale;	// Sleeper points

      return McScore;
	}

   int GPlayer::BinaryRounds()
   {
   	int base = 1, binrounds = 0;

      for(int r=Games.size()-1; r>=0; r--)
      {
      	if(Games[r].Result != grSleep)
         	binrounds += base;

         base <<= 1;
      }

      return binrounds;
   }

   void GPlayer::RoundsPlayed(vector<bool>& pattern)
   {
   	if(pattern.size() != Games.size()) return;

      for(size_t r=0; r<Games.size(); r++)
      	pattern[r] = Games[r].Result != grSleep;
   }

   int GPlayer::GroupOrdinal()
   {
   	using Tools::cZero;
   	using Tools::cNine;
      using Tools::cGlyA;


   	if(Group.empty()) return gsMax;

      if(Group.size() != 2) return gsInf;

      int cmd = -1;

      switch (Group[1])
      {
      	case gpAlways: cmd = 0; break;
      	case gpNever:  cmd = 1; break;
      	case gpAvoid:  cmd = 2; break;
         case gpPrefer: cmd = 3; break;

         default:  cmd = 4;
      }

      int groupid = 0;

      char cGroup = Group[0];

      if(isdigit(cGroup))
         groupid =  cNine - cGroup + cZero;
      else
      	groupid = cGroup - cGlyA + gsTopGroups;

      return gsBase*groupid + cmd;
   }

   GPlayer& GPlayer::operator=(const GPlayer& P)
   {
   	State = P.State;			Pid = P.Pid;

      Grade = P.Grade;

      MiScore = P.MiScore;		McScore = P.McScore;
      Cbal = P.Cbal;				Mbal = P.Mbal;				Mout = P.Mout;
      TieOne = P.TieOne;		TieTwo = P.TieTwo;      Rank = P.Rank;
      Name  = P.Name;			Club  = P.Club;			Land = P.Land;
      Group = P.Group;			Note  = P.Note;

      Games.resize(P.Games.size());

      for(size_t r=0; r<Games.size(); r++)Games[r] = P.Games[r];

      for(size_t i=0; i<4; i++)Spare[i] = P.Spare[i];

      return *this;
   }
}
////////////////////////////////////////////////////////////////////////////////


namespace Tours                              // CLASS GTour
{
   using std::list;

   GTour::~GTour()
   {
      // PlayerList is cleared by the list destructor.
      // Sorted is cleared by the vector destructor.
      // Listed is cleared by the vector destructor.
      // The pointers in Listed point to PlayerList items
   }

   GTour::GTour(std::string DefaultFileName)
   {
      FileName = DefaultFileName;                        // Set on load tour

      SpareOne = 0;    SpareTwo = 0;
      Players  = 1;    Rooms    = 0;
      Round    = 0;    Roundl   = 0;

      for(size_t r=0; r<size_rooms; r++)Boards[r] = 0;
      for(size_t p=0; p<size_param; p++)PairParam[p] = 0;

      PlayerList.push_back(GPlayer(1, def_rounds));      // First player is a dummy

      Listed = vector<GPlayerPtr>(2);

      Listed[0] = NULL;												// Player indexing starts at 1
      Listed[1] =  &(*PlayerList.begin());					// Dummy - new player

      Sorted = vector<size_t>(2);                        // Random access to Players
      Sorted[0] = 0;	Sorted[1] = 1;
      SortOrder = SortSense = 0;									// GK REGSORT 2015

      State = tsVirgin;

      Clubs.codelist.push_back(NoClub);						// GK 6.4.0
      Clubs.codelist.push_back(FullClub);

      Lands.codelist.push_back(NoLand);
      Lands.codelist.push_back(FullLand);
   }

   GTour::GTour()
   {
      Clear();

      Clubs = SCodeList(3);
      Lands = SCodeList(2);
   }

   void GTour::Clear()
   {
   	TourName	= string(SizeName, Tools::cSpace);			 // Fixed name size
      FileName = string();                                // Set on load tour
      Year     = base_year;

      SpareOne = 0;    SpareTwo = 0;
      Players  = 0;    Rooms    = 0;
      Round    = 0;    Roundl   = 0;

      SortOrder = skEntry;
      SortSense = 1;

      for(size_t r=0; r<size_rooms; r++)Boards[r] = 0;
      for(size_t p=0; p<size_param; p++)PairParam[p] = 0;

      State = tsVirgin;
   }

   void GTour::StartSorted()
   {
   	Sorted.resize(Players+1);

      for(size_t i = 0; i<=Players; i++)
         Sorted[i] = i;
   }

   void GTour::ReBuild(size_t players)
   {
      using std::list;

      size_t rounds = Rules.Rounds;

      Players = players;
      
      try
      {
         PlayerList.clear();  Listed.clear();  Sorted.clear();

         Clubs.codelist.clear();    Lands.codelist.clear();

         Listed.resize(players+2);		// We have a starting NULL & a terminating dummy

			StartSorted();

         for(size_t p=0; p<players; p++)
         	PlayerList.push_back(GPlayer(p+1, rounds));

         Dummy = GPlayer(players+1, rounds);
      }
      catch(...)
      {
         Throw(ToolsGUI::erWarn, IDT_ReBuild, IDS_NoMemory);
      }

      list<GPlayer>::iterator nPlayer = PlayerList.begin();

      for(size_t p=1; p<=players; p++)
      {
         Listed[p] = &(*nPlayer);

         ++nPlayer;
      }

      Listed[0]         = NULL;       // First player on list not used
      Listed[players+1] = &Dummy;     // New player entry

      Clubs.codelist.push_back(NoClub);						// GK 6.4.0
      Clubs.codelist.push_back(FullClub);

      Lands.codelist.push_back(NoLand);
      Lands.codelist.push_back(FullLand);
   }
}
////////////////////////////////////////////////////////////////////////////////

namespace Tours                                    // TIE BREAK CALCULATIONS
{
   using ToolsGUI::erBug;
   
   int GTour::PointsValue                          // Sum of scores per round
   (
      bool sleepgain,                              // Sleepers get credit
      size_t rnd,                                  // Sum to round inclusive
      const GPlayerPtr p                           // The player
   )
   {
      int Points = 0, Sleep  = 0;

      intsi  Result;

      for(size_t r=0; r<=rnd; r++)
      {
         Result = p->Games[r].Result;

         if(Result==grSleep)
            ++Sleep;
         else
            Points += Score(Result);
      }
      Sleep /= 2;                                     // Gain 1 point for each 2 rounds slept

      Sleep = sleepgain ? Sleep*mc_scale : 0;

      return Points +  Sleep;                         // Points already scaled
   }

   int GTour::WinsValue(const GPlayerPtr p)           // Sum of actual wins
   {                                                  // includes any jigos
      int Wins = 0, Win;

      for(size_t r=0; r<Round; r++)
      {
         switch(p->Games[r].Result)
         {
            case grJigo: Win = gsHalf; break;
            case grWin:  Win = gsOne;  break;
            case grFwin: Win = gsOne;  break;
            
            default:     Win = 0;
         }
         Wins += Win;
      }
      return Wins;
   }

   int GTour::HomeGamesValue(const GPlayerPtr p)		// Games against same club
   {
   	using std::string;

   	int HomeGames = 0;

      string pClub = p->Club, pLand = p->Land;

      for(size_t r=0; r<Round; r++)
      {
         size_t oppid = p->Games[r].Oppid;

         if(oppid>0)
         {
            const GPlayerPtr opp =  Listed[oppid];

         	if(pLand == opp->Land)
         		if(pClub == opp->Club)
               	++HomeGames;
         }
      }
      return HomeGames;
   }

   int GTour::CuspValue(const GPlayerPtr p)           // Sum of points per round
   {
      int Cusp = 0;

      for(size_t r=0; r<Round; r++)
         Cusp += PointsValue(false, r, p);

      return Cusp;
   }

   int GTour::CussValue(const GPlayerPtr p)           // Sum of McMahon scores per round
   {
      int Cuss = 0;

      for(size_t r=0; r<Round; r++)
         Cuss += (int)p->McMahonScore(r+1);           // Already deals with sleepers

      return Cuss;
   }

   int GTour::SoSValue(const GPlayerPtr p)            // Sum opponents McMahon scores
   {
      int SoS = 0;

      for(size_t r=0; r<Round; r++)
      {
         size_t oppid = p->Games[r].Oppid;

         SoS += oppid>0
              ? (int)Listed[oppid]->McScore
              : (int)p->MiScore;
      }
      return SoS;
   }

   int GTour::SoDoSValue(const GPlayerPtr p)          // Sum defeated opponents McMahon scores
   {
      int SoDoS = 0;

      for(size_t r=0; r<Round; r++)
      {
         size_t oppid = p->Games[r].Oppid;

         switch(p->Games[r].Result)
         {
            case grWin:    SoDoS += (int)Listed[oppid]->McScore;    break;
            case grJigo:   SoDoS += (int)Listed[oppid]->McScore/2;  break;
            case grFwin:   SoDoS += (int)p->MiScore;                break;

            default:       SoDoS += 0;
         }
      }
      return SoDoS;
   }


   int GTour::SoSoSValue(const GPlayerPtr p)          // Sum  opponents SoS values
   {
   	const int OwnSoS = SoSValue(p);

      int SoSoS = 0;

      for(size_t r=0; r<Round; r++)
      {
         size_t oppid = p->Games[r].Oppid;

         SoSoS += oppid>0 ? SoSValue(Listed[oppid]) : OwnSoS;
      }
      return SoSoS;
   }

   int GTour::TieValue(tie_t TieBreak, size_t p)
   {
      if(p<1 || p>Players)
         Throw(ToolsGUI::erBug, IDT_TieValue, IDT_BadIndex);
         
      switch(TieBreak)
      {
         case tbWins:   return WinsValue(Listed[p]);
         case tbCuss:   return CussValue(Listed[p]);
         case tbCusp:   return CuspValue(Listed[p]);

         case tbSoS:    return SoSValue(Listed[p]);
         case tbSoDoS:  return SoDoSValue(Listed[p]);
         case tbSoSoS:  return SoSoSValue(Listed[p]);

         default:       return 0;
      }
   }

   void GTour::SetTieBreaks(void)
   {
   	for(size_t p=1; p<=Players; p++)
      {
         Listed[p]->TieOne = (ints)TieValue(Rules.TieOne, p);
         Listed[p]->TieTwo = (ints)TieValue(Rules.TieTwo, p);
      }
   }
}
////////////////////////////////////////////////////////////////////////////////

namespace Tours                     // TOOLS
{
	void GTour::AddPlayer(GPlayer& player)
   {
      // ADD NEW PLAYER TO BACK OF PlayerList
      PlayerList.push_back(player);

      ++Players;

      // ADD PTR TO NEW PLAYER TO BACK OF Listed - Players bracketed by Null and Dummy
      Listed.resize(Players+2);

      Listed[Players]   = &PlayerList.back();
      Listed[Players+1] = &Dummy;

      // MAINTAIN SORTED
      Sorted.resize(Players+1);
      Sorted[Players] = Players;
   }

   bool GTour::PastRound(void)
   {
   	return Round < Roundl;
   }

   
   void GTour::PlayNowNext(GPlayer& P, bool& Now , bool& Next)
   {
   	intsi ResultNow = grSleep, ResultNext = grSleep;

      Now = Next = false;

      if(Round==0 && P.State==psRegistered)      																	  		// Registering
      {
      	ResultNow  = P.Games[0].Result;
         ResultNext = Rules.Rounds>1 ? P.Games[1].Result : grSleep;
      }

      if(1 <= Round && Round < Rules.Rounds)											// Round is a count
      {
      	ResultNow  = P.Games[Round-1].Result;
         ResultNext = P.Games[Round].Result;
      }

      if(Round==Rules.Rounds)																// Last round
      {
      	ResultNow  = P.Games[Round-1].Result;
         ResultNext = grSleep;
      }

      Now  = ResultNow  != grSleep;
      Next = ResultNext != grSleep;
   }
}
////////////////////////////////////////////////////////////////////////////////

namespace Tours                     // CONVERSION TO STRING
{
   using Tools::to_string;	using Tools::cFractional;
   using Tools::sSpace;  using Tools::sMinus;

   static const string grade_fmt =   "##";
   static const string mms_fmt   = " ###";
   static const string wins_fmt  =   "##";
   static const string num_fmt   = "0000";                   // Player Id

   char result_char(intsi Result)
   {
      switch(Result)
      {
         case grLose:      return Tools::cMinus;
         case grJigo:      return Tools::cEqual;
         case grWin:       return Tools::cPlus;

         case grUsolv:     return Tools::cUsolv;  // Unresolved

         case grDlos:      return Tools::cDlos;   // Player fail to show
         case grFwin:      return Tools::cFwin;   // Opp fail to show

         case grVoid:      return Tools::cVoid;   // Both fail to show

         case grSleep:     return Tools::cNone;  // Not playing

         default:          return Tools::cQuery;  // Unknown result
      }
   }

   string grade_string(ints grade)
   {
      string sgrade;

      if(grade < 0)
      {
         sgrade = to_string(grade_fmt,(-grade)/mc_scale);
         sgrade += string(1, Tools::cKyu);

         return sgrade;
      }

      if(grade == zero_kyu)return string(" 0k");
      if(grade == zero_dan)return string(" 0d");
   	if(grade == zero_pro)return string(" 0p");

      if(grade >= pro_grade)
      {
         sgrade = to_string(grade_fmt, (grade - pro_grade + mc_scale)/mc_scale);
         sgrade += string(1, Tools::cPro);

         return sgrade;
      }

      if(grade >= 0)
      {
         sgrade = to_string(grade_fmt, (grade + mc_scale)/mc_scale);
         sgrade += string(1, Tools::cDan);

         return sgrade;
      }
      return Tools::sQuery;
   }

   void mms_parts(const ints mms, int& smms, int& imms, int& fmms)
   {
   	const int mmsi  = (int)mms;
      const int pmms  = Tools::modulus(mmsi);

      smms = Tools::sign(mmsi);
      imms = pmms/mc_scale;
      fmms = pmms%mc_scale;
   }
   string mms_string(ints mms)
   {
      int sigm,imms,fmms; mms_parts(mms, sigm,imms,fmms);
      string sfmms = (0<fmms && fmms<4)
                   ? string(1, cFractional[fmms])
                   : string(sSpace);

      return to_string(mms_fmt, sigm*imms) + sfmms;
   }

   string wins_string(ints wins)
   {
      int iwins = wins/mc_scale;
      int fwins = wins%mc_scale;

      string sfwins = (fwins==gsHalf)
                    ? string(1,Tools::cHalf)
                    : string(Tools::sSpace);

      string siwins = to_string(wins_fmt, iwins);

      return siwins + sfwins;
   }

   string grade_ustring(ints grade)
   {
   	string sgrade;

   	if(grade < 0)
      {
         sgrade = to_string((-grade)/mc_scale);
         sgrade += string(1, Tools::cKyu);

         return sgrade;
      }

   	if(grade == zero_kyu)return string("0k");
      if(grade == zero_dan)return string("0d");
   	if(grade == zero_pro)return string("0p");

      if(grade >= pro_grade)
      {
         sgrade = to_string((grade - pro_grade + mc_scale)/mc_scale);
         sgrade += string(1, Tools::cPro);

         return sgrade;
      }

      if(grade >= 0)
      {
         sgrade = to_string((grade + mc_scale)/mc_scale);
         sgrade += string(1, Tools::cDan);

         return sgrade;
      }
      return Tools::sQuery;
   }

   string mms_ustring(ints mms)
   {
      int sigm,imms,fmms; 	mms_parts(mms, sigm,imms,fmms);

      string sfmms = (0<fmms && fmms<4)
                   ? string(1, Tools::cFractional[fmms])
                   : string();

      string smms = to_string(imms) + sfmms;

      return sigm>=0 ? smms : string(sMinus) + smms;
   }

   string wins_ustring(ints wins)
   {
      int iwins = wins/mc_scale;
      int fwins = wins%mc_scale;

      string sfwins = (fwins==gsHalf)
                    ? string(1,Tools::cHalf)
                    : string();

      string siwins = to_string(iwins);

      return siwins + sfwins;
   }

   string pid_string(size_t pid)
   {
   	if(pid > max_pid)
      	return string(4, Tools::cQuery);

      return Tools::to_string(num_fmt, (int)pid);
   }
}
////////////////////////////////////////////////////////////////////////////////

namespace Tours                     // AREA LIST MANAGEMENT
{
   using std::list;
   using std::string;

   using SCodeList::qyNewName;
   using SCodeList::qyKnownName;
   using SCodeList::qyFull;
   
   string GTour::ClubCode(string clubname)                      // Unique code from name
   {
      string clubcode;

      switch(Clubs.encode(clubname, clubcode))
      {
         case qyNewName:   return clubcode;
         case qyKnownName: return clubcode;

         case qyFull:      return sEmpty;

         default:          Throw(ToolsGUI::erBug, IDT_ClubCode, IDT_BadList);
      }
      return sEmpty;
   }

   string GTour::ClubName(string clubcode)                      // Name from given code
   {
      string clubname = sEmpty;

      if(Clubs.decode(clubcode, clubname))
         return clubname;
      else
         Throw(ToolsGUI::erBug, IDT_ClubName, IDT_BadList);

      return sEmpty;
   }

   string GTour::LandCode(string landname)                     // Unique code from name
   {
      string landcode;

      switch(Lands.encode(landname, landcode))
      {
         case qyNewName:   return landcode;
         case qyKnownName: return landcode;

         case qyFull:      return sEmpty;

         default:          Throw(ToolsGUI::erBug, IDT_LandCode, IDT_BadList);
      }
      return sEmpty;
    }

   string GTour::LandName(string landcode)                      // Name from given code
   {
      string landname;

      if(Lands.decode(landcode, landname))
         return landname;
      else
         Throw(ToolsGUI::erBug, IDT_LandName, IDT_BadList);

      return sEmpty;
   }

   void UsedCodes(GTour& Tour, list<string>& Items)               // LIST OF ACTUAL USED CODES
   {
      if(!Items.empty()) return;                                  // Start with an empty list

      list<GPlayer>::iterator iPlayer = Tour.PlayerList.begin();

      for(size_t p=1; p<=Tour.Players; p++)
      {
         string ClubItem = Tour.Clubs.code_item(iPlayer->Club);   // Get the club code and name

         if(ClubItem.empty())                                     // The Clubs list is corrupt
            Throw(ToolsGUI::erBug, IDT_ClubName, IDT_BadList);

         if(find(Items.begin(), Items.end(), ClubItem) == Items.end())  // So ClubItem NOT on UsedList
            Items.push_back(ClubItem);                                  // Add it!


         string LandItem = Tour.Lands.code_item(iPlayer->Land);   // Get the club code and name

         if(LandItem.empty())                                     // The Clubs list is corrupt
            Throw(ToolsGUI::erBug, IDT_LandName, IDT_BadList);

         if(find(Items.begin(), Items.end(), LandItem) == Items.end())  // So LandItem not on UsedList
            Items.push_back(LandItem);                                  // Add it!

         ++iPlayer;
      }
   }

   void ImportedCodes				// PRODUCE LIST OF ACTUAL USED CODES
   (
      GTour& Tour,					// Existing codes and players

   	const int PlayerCount,		// Imported player count
      GPlayer* Players, 			// Imported players indexed from 0 to Players-1

      list<string>& Items
   )
   {
      if(!Items.empty()) return;                                  		// Start with an empty list

      // GATHER EXISTING USED CODES TO SEED Items
      UsedCodes(Tour, Items);

      // ADD ANY NEW FROM THE IMPORTED PLAYERS
      GPlayer* iPlayer = Players;

      for(int p=0; p<PlayerCount; p++)
      {
         string ClubItem = Tour.Clubs.code_item(iPlayer->Club);   				// Get the club code and name

         if(ClubItem.empty())                                     		// The Clubs list is corrupt
            Throw(ToolsGUI::erBug, IDT_ClubName, IDT_BadList);

         if(find(Items.begin(), Items.end(), ClubItem) == Items.end())  // So ClubItem NOT on UsedList
            Items.push_back(ClubItem);                                  // Add it!

         string LandItem = Tour.Lands.code_item(iPlayer->Land);   				// Get the club code and name

         if(LandItem.empty())                                     		// The Clubs list is corrupt
            Throw(ToolsGUI::erBug, IDT_LandName, IDT_BadList);

         if(find(Items.begin(), Items.end(), LandItem) == Items.end())  // So LandItem not on UsedList
            Items.push_back(LandItem);                                  // Add it!

         ++iPlayer;
      }
   }

   CodeItem IdentifyCodeItem(string& Item)
   {
      if(Item.size() <(size_t)( min(size_land, size_club)))
         return ciCorrupt;
      
      int i = Item.find(Tools::sEqual, 0);

      if(i==(int)Item.size())
         return ciCorrupt;

      if(i==size_club)
         return ciClub;

      if(i==size_land)
         return ciLand;

      return ciCorrupt;
   }
}
////////////////////////////////////////////////////////////////////////////////




