////////////////////////////////////////////////////////////////////////////////
// MODULE Tour:       Class GTour                              G.Kaniuk 28/05/02
//
// Provides the base class for any project requiring access to a Tournament file.
// So this class defines the structure of the model for a tournament, but leaves
// the functions to the derived classes.
//
// Actual file services such as reading/writing are provided by GtourStream
//
// All quantities referring to scores or grades of any kind are expressed in
// units of 4 * zero shodan init

#include <string>
#include <list>
#include <vector>

#include "Maths.h"
#include "Strings.h"
#include "ETour.h"

#ifndef TourH
#define TourH
////////////////////////////////////////////////////////////////////////////////

namespace Tours                              // MAGIC constants
{
	using std::string;
   
   using Tools::intsi;
   using Tools::ints;
   
   using ToolsGUI::EType;

   enum                                         // SIZES
   {
      mc_scale   = 4,                           // Convert Zero Shodan to McMahon scale

      size_spare = 4,                           // Field sizes
      size_rooms = 16,
      size_name  = 47,
      size_note  = 16,
      size_club  = 3,
      size_land  = 2,
      size_group = 2,
      size_param = 256
   };

   enum                                         // DEFAULTS
   {
      base_year  = 2000,

      def_rounds =  3,

      def_supbar   = 4*mc_scale,						// 5d supergroup
      def_macbar   = 3*mc_scale,						// 4d bar
      
      def_mingrade = -30*mc_scale,              // 30k
      def_komigrp  = -10*mc_scale,              // 10k
      def_komidiff =    -mc_scale,              // MacMahon score difference - 1
      def_komirate =  15*mc_scale               // 15 points per stone
   };

   enum                                         // LIMITS
   {
   	max_pid      = 9999,								// 4 digit player id
      max_stones   = 36,                        // Represent by 0-9 then A-Z
      max_rounds   = 12,
      max_players  = 1024,                      // Try to avoid a limit here

      min_grade    = -99*mc_scale,					// Minimum possible entry grade
      pro_grade    = 7*mc_scale,                // 8d = 1p
      sup_bar      = pro_grade + 8*mc_scale, 	// Supergroup bar at 9p
      
      zero_pro		 = sup_bar + mc_scale,			// Unknown pro grade
      zero_dan		 = sup_bar + 2*mc_scale,      // Unknown dan grade
      zero_kyu		 = sup_bar + 3*mc_scale,      // Unknown kyu grade

      last_limit
   };

   enum                                                  // COLOUR VALUES
   {
      gcBlack = -1,   gcNone = 0,   gcWhite = 1
   };

   enum                                                  // PARITY VALUES
   {
      gdDown  = -1,   gdEven = 0,    gdUp    = 1
   };

   enum                                                  // SCORE VALUES
   {
      gsZero =  0,
      gsHalf = mc_scale/2,
      gsOne  = mc_scale
   };

   enum                                                  // GROUP CODES
   {
      gpAlways = '*', gpNever  = '~',
      gpAvoid  = '-', gpPrefer = '+',
      gpTop    = ':'
   };

   enum                                                  // RESULT VALUES
   {
      grLose = gsZero,  grJigo = gsHalf,  grWin = gsOne,

      grUsolv,

      grDlos,  grFwin,  grVoid,

      grSleep,

      grNone
   };

   extern const Tools::uiBounds MacPop[max_rounds];      // McMahon bar table
   extern const Tools::uiBounds SupPop[max_rounds];      // Supergroup table

   extern const size_t player_text_size;                 // Packed name, note
   extern const size_t player_code_size;                 // Packed club,land, group

   extern const string NoClub, FullClub;           		// Keywords for regions
   extern const string NoLand, FullLand;

   extern const string GroupCodes;               			// Chars as in enum Group

   void Throw(EType etype, short idproc, short idmsg);

   ints Score(intsi Result);                            // Convert Result to Score
}
/////////////////////////////////////////////////////////////////////////////


namespace Tours                                 // CLASS GRules
{
   enum tie_t                                   // Tie Breaks
   {
      tbNone,
      tbCuss, tbCusp,
      tbSoS,  tbSoSoS, tbSoDoS,
      tbWins,
      tbSize
   };

   enum style_t                                 // Tournament Styles
   {
      stMcMahon,
      stSwiss, stSwissHcap,
      stAllPlay, stAllPlayHcap,
      stBarLow,
      stCount
   };

   enum rules_t                                 // Association Rules
   {
      rtBGA, rtEGF,
      rtCount
   };

   enum pair_t                                  // Priority in pairing
   {
      ptUneven, ptArea, ptColour
   };

   struct GRules
   {
      bool  Checked;                         // Settings confirmed

      size_t Rounds;                         // Def is zero to force setting

      tie_t TieOne, TieTwo;                  // tbSoS, tbNone is default

      style_t Style;                         // MacMahon default

      rules_t RuleSet;                       // BGA default

      bool BarSeg;                           // Default is no bias above bar

      ints MacBar;                           // MacMahon Bar
      ints MacDepth;									// Bar depth
      ints SupBar;                           // Supergroup Bar
      ints MinGrade;                         // All players lower than this have same initial mms

      ints KomiGroup;                        // Below this play handicap, not same club
      ints KomiDiff;                         // Handicap = MMS diff - KomiDiff - scaled
      ints KomiRate;                         // Points per stone in high handicap - scaled
      ints Prizes;                           // Prize threshold - scaled
      
      ints Unresolved;								// Unresolved game score - scaled
      ints SuperDiff;								// SuperGroup displacement - scaled

      GRules();
      void Default();
      size_t Size();
   };
}
////////////////////////////////////////////////////////////////////////////////


namespace Tours                                 // CLASS GPlayer and GGame
{
   using std::string;
   using std::vector;

   using Tools::ints;
   
   struct GBoard                                // A pairing is an array of GBoards
   {
      int Black, White;

      GBoard(){ Black = White = 0; }

      GBoard(int black, int white)
      { Black = black;  White = white; }
   };

   struct GGame
   {
      size_t Board;                            // The board no for this game
      size_t Oppid;                            // Opponent number - index to Listed

      ints  Komi;                              // Komi = extra stones*points per stone - THIS IS SCALED
      intsi Stones;                            // Algebraic Handicap - SCALED stones

      intsi Colour;                            // Black Sleep White
      intsi Parity;                            // Down Even Up

      intsi Result;                            // Lose,Jigo,Win,Unresolved,Dloss,Fwin,Void,Sleep, NoResult

      GGame();

      GGame& operator=(const GGame& G);
      
      size_t Size();                          // File record size
   };



   enum PlayerState
   {
      psNew,                                    // Virgin creation
      psEntered,                                // All details valid
      psRegistered,                             // Ticked
      psSleep,                                  // Not playing this round
      psChosen,											// Temporary state during pairing
      psPlaying,                                // In draw, waiting for result
      psFinished                                // Result entered
   };

   enum														// Groups
   {
   	gsTopGroups=10,									// 9-0
      gsMidGroups=26,                           // A-Z
      gsNoGroup=1,										// Empty
      gsBase=41,											// First prime beyond 26+10+1
      gsCmd=5                                   // *~-+:
   };

   enum
   {
   	gsMax = gsBase*(gsTopGroups + gsMidGroups + gsNoGroup) + gsCmd,
      gsInf
   };


   class GPlayer
   {
      public:
         Byte Spare[size_spare];                // Future

         PlayerState State;                     // Dynamic

         size_t Pid;                             // Publication number allocated at entry
                                                // Never use as an index to any container

         string Name;                           // Given Initials Family
         string Club;                           // Coded club
         string Land;                           // Coded country
         string Note;                           // Ascii
         string Group;                          // Group ID + Code

         ints  Grade;                           // Entry grade in zero shodan units
         ints  MiScore;                         // Initial MacMahon score
         ints  McScore;                         // Current MacMahon score
         ints  Rank, TieOne, TieTwo;            // Rank and Tie breaks

         intsi Cbal;                            // White games - Black games
         intsi Mbal;                            // Up games - Down games
         intsi Mout;                            // Uneven game count

         intsi Class; 									// Player's class - one of mcThing
         intsi UpLose;									// Games played up and lost
         intsi DnWin;									// Games played down and won

         ints  McGroup;									// Index to player's McMahon group
         ints  Pool;										// Number of free opponents in own group
         ints  NearPool;								// Opponents in nearest neighbours
         ints  FarPool;									// Opponents in first group which is not a neighbour

         vector<GGame>Games;               		// Each round has a GGame

      public:
      	GPlayer();										// Zero id , empty rounds
         GPlayer(size_t id, size_t rounds);     // Zero rounds is ghost
        ~GPlayer();                             // Deleted from heap

         GPlayer& operator=(const GPlayer& P);	// Assignment resizes Games

         size_t Size();                         // Size of the filed record
         size_t Wins();                         // Total actual wins to current round
         ints McMahonScore(const int Round);		// Up to date score
         ints McMahonScore								// Calculated while editing
         (
         	const vector<bool>playing
         );

         bool Active();                         // Played at least one game
         bool Free();								   // Waiting to be paired
         void SetFree(int Round);					// Set player State to Free
         void SetBusy(void);							// Complement SetFree

         bool Awake(int Round);						// Playing this round

         bool Sleeping(int Round);					// Sound asleep
         bool Joining(int Round);		  			// Waiting to join draw
   		bool Leaving(int Round);		  			// Waiting to leave draw
   		bool Playing(int Round);					// Definitely in draw
         bool PlayingAll(void);						// Playing all rounds

         bool Unpaired(int Round);					// Playing, but no opp
         bool MissingResult(int Round);			// Playing, but no result

         void RoundsPlayed(vector<bool>& play); // Pattern of rounds played
         int BinaryRounds();							// Integer from pattern of rounds played
         int GroupOrdinal();							// Integer from 10*groupid + cmd

         string AreaCode(){ return Land+Club; }	// For Sorting;

         bool operator != (const GPlayer& P){ return Pid != P.Pid; }
         bool operator == (const GPlayer& P){ return Pid == P.Pid; }
   };

   typedef GPlayer* GPlayerPtr;                 // Pointer to GPlayer

   ints Score(intsi Result);                    // Convert result to score in MMS units
}
///////////////////////////////////////////////////////////////////////////////

namespace Tours                                 // BASE CLASS GTour
{
   /////////////////////////////////////////////////////////////////////////////
   //
   // Player data structures GPlayer are held in the PlayerList.
   //
   // Listed[i] gives a GPlayerPtr in the list.
   // We define GPlayer.Pid st Listed[i]->Pid = i.
   //
   // For  any particular ordering, Sorted[order] is an index to Listed.
   //
   // The vector iSorted maintains the inversion of Sorted.
   // So order = iSorted[index] where index = Sorted[order].
   //
   /////////////////////////////////////////////////////////////////////////////

   using std::string;
   using std::vector;
   using std::list;

   using Tools::SCodeList;
   
   enum TourState
   {
      tsVirgin,      // No tournament selected
      tsTourStart,   // New tournament created
      tsRegNow,      // Registration in progress
      tsRegClosed,   // Registration closed
      tsDrawDone,    // Draw completed, not yet printed
      tsResultNow,   // Entering results
      tsRoundDone,   // Round completed
      tsTourDone,    // Tournament completed
      tsEditForced,  // Waiting to edit draw
      tsPartDraw,		// Partial draw needs editing
      tsEditNow,     // Editing in progress
      tsForcedPair,  // Manual draw
      tsPastRound,   // Earlier round loaded
      tsCount        // Number of tournament states
   };

   enum       			// USER AND INTERNAL SORT KEYS
   {
      skEntry,       // PlayerNumber
      skTicked,		// Registered>Name
      skName,        // Surname>Firstname>MiddleName
      skGrade,       // Grade>Name
      skClub,        // Club>Name
      skLand,        // Country>Name
      skRounds,		// Binary pattern>Name
      skGroup,       // Group>Grade>Name
      skNote,        // Alphanumeric>Name
      skPlay,        // PlayNow>Name
      skSleep,       // SleepNow>Name

      skCuss,        // TIE BREAK SORT
      skCusp,
      skWins,
      skSoS,
      skSoSoS,
      skSoDoS,

      skMcMahon,     // INTERNAL SORT
      skMcBar,
      skRank
   };

   enum CodeItem
   {
      ciClub,
      ciLand,
      ciCorrupt
   };

   class GTour
   {
      public:
      	enum { SizeName = 24 };						// Tournament name size

         string FileName;                       // Current file loaded

         GRules Rules;                          // All the tournament parameters

         Byte PairParam[size_param];				// Raw space for user defined pairing parameters

         TourState State;                       // The current tournament state
         
         int SortOrder;                     		// Current ordering rule
         int SortSense;                     		// Control sort direction

         size_t Round;                          // The current round  - count
         size_t Roundl;                         // The highest round drawn - count

         size_t Boards[size_rooms];             // Boards per room
         size_t Rooms;                          // Number of rooms

         size_t SpareOne;                       // Was TourID
         size_t SpareTwo;                       // Was YearID

         size_t Players;                        // The current number of players

         list<GPlayer>PlayerList;               // Players appended in entry order
         vector<GPlayerPtr>Listed;              // Random access the playerlist

         vector<size_t>Sorted;                  // Sorted[order] is index to Listed

         SCodeList Clubs;                       // Listing of club codes & names
         SCodeList Lands;                       // Listing of country codes & names

         string TourName;								// The fixed length tournament name
         int    Year;									// Gets set from system clock

         GPlayer Dummy;									// The new player

      public:                                         // CONSTRUCTION/DESTRUCTION
         GTour();
         GTour(std::string DefaultFileName);          // Basic standard 3 rounder
         ~GTour();                                    // Delete each player in list

         void GTour::Clear();									// Cleares atoms and fixed arrays
         void ReBuild(size_t players);                // Clear, recreate lists

      public:                                         // CODE LISTS
         string ClubCode(string clubname);
         string ClubName(string clubcode);
         string LandCode(string landname);
         string LandName(string landcode);


      public:                                         // SORTING KEYS
         int PointsValue
         (
            bool sleepgain, size_t rnd,
            const GPlayerPtr p
         );

         int WinsValue(const GPlayerPtr p);

         int HomeGamesValue(const GPlayerPtr p);

         int CuspValue(const GPlayerPtr p);
         int CussValue(const GPlayerPtr p);

         int SoSValue(const GPlayerPtr p);
         int SoDoSValue(const GPlayerPtr p);
         int SoSoSValue(const GPlayerPtr p);

         int TieValue(tie_t TieBreak, size_t p);      // Tie break calculation

         void SetTieBreaks(void);
         
         void StartSorted();                          // Initialise Sorted to Listed order

	public:															// TOOLS
   		void AddPlayer(GPlayer& P);						// Add player to lists

         void PlayNowNext										// Player's presence in this round, next round
         (
         	GPlayer& P, bool& Now , bool& Next
         );

         bool PastRound(void);								// Detect past round during editing
   };

   void UsedCodes(GTour& Tour, list<string>& Items);  // List of codes actually used

   void ImportedCodes											// EXTEND LIST OF ACTUAL USED CODES
   (
   	GTour& Tour,

   	const int PlayerCount,									// Imported player count
      GPlayer* Players, 										// Imported players indexed from 0 to Players-1

      list<string>& Items										// The extended used codes
   );

   CodeItem IdentifyCodeItem(string& Item);           // Analyse for Land or Club code
}
////////////////////////////////////////////////////////////////////////////////

namespace Tours                                       // STRING CONVERSION
{
   char result_char(intsi Result);

   string grade_string(ints grade);                   // FORMATTED
   string mms_string(ints mms);
   string wins_string(ints wins);
   string pid_string(size_t pid);

   string grade_ustring(ints grade);                  // UNFORMATTED
   string mms_ustring(ints mms);
   string wins_ustring(ints wins);
}
////////////////////////////////////////////////////////////////////////////////
#endif
