////////////////////////////////////////////////////////////////////////////////
// MODULE TourStream:  Implementation of GTourStream           G.Kaniuk 8/06/02
//                     Provides all file services for GTour

#include <vcl.h>
#include <sysutils.hpp>
#include <ctype.h>
#pragma hdrstop

#include "ETourRES.h"
#include "TourRES.h"

#include "Strings.h"
#include "ByteStream.h"
#include "Tourstream.h"

#include "LoadOldFile.h"

namespace Tours                                    // STANDARD FILES, IDENTIFICATION
{
   using std::string;

   const string TourTit   = "TourFile";            // File Stamp Title
   const string TourVer   = "1.0.4";               // This Software version
   const string TourExt   = ".gdt";                // Tournament file - binary
   const string BakExt    = ".gdt.bak";            // Backup file - binary
   const string TourNow   = "CurrentTour";         // Default filename


   file_t IdentifyFile(string FileName)                           // GoDraw or godraw file
   {
      using Tools::cZero;
      using Tools::cNine;
      using Tools::s_S;
      using Tools::S_s;

      string Ext = Tools::lowercase(S_s(ExtractFileExt(s_S(FileName))));

      if(Ext == TourExt) return ftGoDraw;


      int ExtLen = Ext.size();                                  // Includes .

      if(ExtLen<3 || ExtLen>4) return ftUnknown;


      char ExtChar;

      ExtChar = Ext[1]; if(ExtChar != cZero && ExtChar != cNine)
                           return ftUnknown;

      ExtChar = Ext[2]; if(!isdigit(ExtChar))
                           return ftUnknown;

      if(ExtLen==4)     if(!isalpha(Ext[3]))
                           return ftUnknown;
      return ftGodraw;
   }
}

namespace Tours                                 // CLASS GTourStream
{
   using ToolsGUI::ETour;
   using ToolsGUI::EType;
   using ToolsGUI::erOperr;
   using ToolsGUI::erBug;
   using ToolsGUI::erWarn;
   
   __fastcall GTourStream::~GTourStream()
   {
      if(Stream->Size != OpenSize)                 // The file size has changed
         SaveFileSize();

      delete Stream;
   }

   __fastcall GTourStream::GTourStream
   (
      string Fname,
      file_t Ftype,
      filact_t Action
   )
   {
      using Tools::s_S;

      StartPos  = 0;
      SizePos   = StartPos + TourTit.size() + TourVer.size();  // Store file size

      HeaderPos = SizePos + sizeof(int);                       // Tournament header

      HeadSize  = GRules().Size();										// Write all Rules

      HeadSize += size_param;												// Raw space for pairing parameters

      HeadSize += sizeof(TourState);                           // TourState
      HeadSize += 2*sizeof(int);                               // Round, Roundl

      HeadSize += (size_rooms+1)*sizeof(int);                  // Rooms + Boards per room

      HeadSize += 2*sizeof(int);                               // SpareOne, SpareTwo

      HeadSize += GTour::SizeName;										// Tournament Name
      HeadSize += sizeof(int);											// Year

      HeadSize += sizeof(int);                                 // Players

      CodesPos  = HeaderPos + HeadSize;								// Codes section

      PlayerPos = CodesPos + CodeFileSize;                     // First player here

      FileName = Fname;

      if(Action==faCreate)
      {
         Stream = new TFileStream(s_S(Fname), fmCreate|fmShareExclusive);

         OpenSize = 0;

         SaveFileStamp();           // Write the TourStream software version

         SaveEmptyCodes();          // Fill country/club code space with 0

         return;
      }

      Word FileMode = fmShareExclusive;

      switch(Ftype)
      {
         case ftGoDraw: FileMode |= fmOpenReadWrite; break;

         case ftGodraw: FileMode |= fmOpenRead;      break;

         default:  Throw(erOperr,IDT_GTourStream,IDT_NoRecog);
      }

      try
      {
         Stream = new TFileStream(s_S(Fname), FileMode);
      }
      catch(...)
      {
         OpenSize = Stream->Size;

         Throw(erOperr, IDT_GTourStream, IDT_NoOpen);
      }

      if(Ftype==ftGoDraw)VerifyFileStamp();

      OpenSize = Stream->Size;
   }

   void GTourStream::StreamThrow(EType etype, short idproc, short idmsg)
   {
      string message = Forge(etype, IDT_Stream, idproc, idmsg);

      if(etype==erBug)::AppLog.Add(message);

      throw ETour(etype, message);
   }

}

namespace Tours                                    // FILESTAMP TOOLS
{
   using ToolsGUI::erBug;

   void GTourStream::SaveFileSize()
   {
      int FileSize = Stream->Size;

      if(Stream->Seek(SizePos,soFromBeginning) != SizePos)
         StreamThrow(erBug, IDT_SaveFileSize, IDT_BadPosition);

      if(Stream->Write(&FileSize, sizeof(int)) != sizeof(int))
         StreamThrow(erBug, IDT_SaveFileSize, IDT_BadWriteSize);
   }

   void GTourStream::SaveFileStamp()                              // FIRST ITEM
   {
      if(Stream->Seek(StartPos, soFromBeginning) != StartPos)
         StreamThrow(erBug, IDT_SaveFileStamp, IDT_BadPosition);

      int StampSize =
          TourTit.size() + TourVer.size() + sizeof(int);

      ByteStream FileStamp(StampSize);

      FileStamp.put_string(TourTit + TourVer);
      FileStamp.put_value(OpenSize);

      if(Stream->Write(FileStamp.bytes(), StampSize) != StampSize)
         StreamThrow(erBug, IDT_SaveFileStamp, IDT_BadWriteSize);
   }

   bool GTourStream::VerifyVersion(std::string version)  //OVERRIDABLE
   {
      if(version==TourVer) return true;

      return false;
   }

   void GTourStream::VerifyFileStamp()
   {
      if(Stream->Seek(StartPos, soFromBeginning) != StartPos)
         StreamThrow(erBug, IDT_LoadFileStamp, IDT_BadPosition);

      int StampSize =
          TourTit.size() + TourVer.size() + sizeof(int);

      ByteStream FileStamp(StampSize);

      if(Stream->Read(FileStamp.bytes(), StampSize) != StampSize)
         StreamThrow(erBug, IDT_LoadFileStamp, IDT_BadReadSize);

      string rTitle(TourTit), rVersion(TourVer);

      FileStamp.get_string(rTitle);
      FileStamp.get_string(rVersion);

      if(rTitle != TourTit)
         StreamThrow(erOperr, IDT_LoadFileStamp, IDT_Corrupt);

      if(!VerifyVersion(rVersion))
         StreamThrow(erOperr, IDT_LoadFileStamp, IDT_OldFile);

      int StreamSize;      FileStamp.get_value(StreamSize);

      if(StreamSize!= Stream->Size)
         StreamThrow(erOperr, IDT_LoadFileStamp, IDT_Tamper);
   }
}

namespace Tours                                    // CODE TOOLS
{
   using std::string;
   using std::list;

   using ToolsGUI::erBug;

   using Tools::s_S;

   void GTourStream::SaveEmptyCodes()                                               // SECOND ITEM
   {
      try
      {
         if(Stream->Seek(CodesPos, soFromBeginning) != CodesPos)
            StreamThrow(erBug, IDT_SaveCodes, IDT_BadPosition);

         string Codes = string(CodeFileSize, Tools::cNull);           // Fill with White space

         if(Stream->Write(Codes.c_str(), CodeFileSize) != CodeFileSize)
            StreamThrow(erBug, IDT_SaveCodes, IDT_BadWriteSize);
      }
      catch(...)
      {
         StreamThrow(erWarn, IDT_SaveCodes, IDS_NoMemory);
      }
   }

   void GTourStream::SaveCodes(GTour& Tour)                       // SAVE CodeLists to file
   {
      list<string>MixedCodeItems;                                 // Each item is a string "CODE=name"
                                                                  // Where CODE is 2 or 3 characters
      Tours::UsedCodes(Tour, MixedCodeItems);

      list<string>::iterator iItem = MixedCodeItems.begin();
      list<string>::iterator eItem = MixedCodeItems.end();

      if(Stream->Seek(CodesPos, soFromBeginning) != CodesPos)
         StreamThrow(erBug, IDT_SaveCodes, IDT_BadPosition);

      int CodesLeft = CodeFileSize;                               // Ensure we stay within allocated file space

      while(iItem != eItem)
      {
         string Item  = string(1, Tools::cNull) + *iItem;         // Prefix size to string
         size_t iSize = Item.size();                              // Includes prefix

         if(iSize > Tools::max_item_size)
            StreamThrow(erBug, IDT_SaveCodes, IDT_BadByteSize);

         Item[0] = (char)iSize;                                   // First char of Item specifies length

         if((int)iSize>CodesLeft)                                      // No one else can detect this condition!
            StreamThrow(erOperr, IDT_SaveCodes, IDT_CodesFull);

         if(Stream->Write(Item.c_str(), iSize) != (int)iSize)
            StreamThrow(erBug, IDT_SaveCodes, IDT_BadWriteSize);

         ++iItem;  CodesLeft -= iSize;
      }

      if(CodesLeft>0)
      {
         string Pad(CodesLeft, Tools::cNull);

         if(Stream->Write(Pad.c_str(), CodesLeft) != CodesLeft)
            StreamThrow(erBug, IDT_SaveCodes, IDT_BadWriteSize);
      }

      if(Stream->Position != PlayerPos)
         StreamThrow(erBug, IDT_SaveCodes, IDT_BadPosition);
   }

   void GTourStream::SaveImportedCodes										// SAVE CodeLists to file
   (
   	GTour& Tour, const int PlayerCount, GPlayer* Players
   )
   {
   	using Tours::UsedCodes;

      list<string>MixedCodeItems;                                 // Each item is a string "CODE=name"
                                                                  // Where CODE is 2 or 3 characters
      ImportedCodes(Tour, PlayerCount,Players, MixedCodeItems);

      list<string>::iterator iItem = MixedCodeItems.begin();
      list<string>::iterator eItem = MixedCodeItems.end();

      if(Stream->Seek(CodesPos, soFromBeginning) != CodesPos)
         StreamThrow(erBug, IDT_SaveCodes, IDT_BadPosition);

      int CodesLeft = CodeFileSize;            							// Ensure we stay within allocated file space

      while(iItem != eItem)
      {
         string Item  = string(1, Tools::cNull) + *iItem;         // Prefix size to string
         size_t iSize = Item.size();                              // Includes prefix

         if(iSize > Tools::max_item_size)
            StreamThrow(erBug, IDT_SaveCodes, IDT_BadByteSize);

         Item[0] = (char)iSize;                                   // First char of Item specifies length

         if((int)iSize>CodesLeft)                                 // No one else can detect this condition!
            StreamThrow(erOperr, IDT_SaveCodes, IDT_CodesFull);

         if(Stream->Write(Item.c_str(), iSize) != (int)iSize)
            StreamThrow(erBug, IDT_SaveCodes, IDT_BadWriteSize);

         ++iItem;  CodesLeft -= iSize;
      }

      if(CodesLeft>0)
      {
         string Pad(CodesLeft, Tools::cNull);

         if(Stream->Write(Pad.c_str(), CodesLeft) != CodesLeft)
            StreamThrow(erBug, IDT_SaveCodes, IDT_BadWriteSize);
      }

      if(Stream->Position != PlayerPos)
         StreamThrow(erBug, IDT_SaveCodes, IDT_BadPosition);
   }

   void GTourStream::LoadCodes(GTour& Tour)
   {
      using Tools::cNull;
      using Tours::ByteStream;

      if(Stream->Seek(CodesPos, soFromBeginning) != CodesPos)
         StreamThrow(erBug, IDT_LoadCodes, IDT_BadPosition);

      ByteStream CodeStream(CodeFileSize);

      if(Stream->Read(CodeStream.bytes(), CodeFileSize) != CodeFileSize)   // Read size of item
         StreamThrow(erBug, IDT_LoadCodes, IDT_BadReadSize);

      if(Stream->Position != PlayerPos)
         StreamThrow(erBug, IDT_LoadCodes, IDT_BadPosition);

      byte bSize = 0;

      while(true)
      {
         CodeStream.get_value(bSize);              // The entire record = prefix + item

         if(bSize == 0)return;                     // We are done as codes initialised to 0

         string Item = string(bSize-1, cNull);     // We know the item size

         CodeStream.get_string(Item);              // Either club or country

         switch(IdentifyCodeItem(Item))
         {
            case ciClub:
               Tour.Clubs.codelist.push_back(Item);
               break;

            case ciLand:
               Tour.Lands.codelist.push_back(Item);
               break;

            default:
               StreamThrow(erBug, IDT_LoadCodes, IDT_BadList);
         }
      }
   }
}


namespace Tours                                    // HEADER
{
   using Tools::s_S;

   using ToolsGUI::erBug;

   void GTourStream::SaveHeader(GTour& Tour)
   {
      if(Stream->Seek(HeaderPos, soFromBeginning) != HeaderPos)
         StreamThrow(erBug, IDT_SaveHeader, IDT_BadPosition);

      ByteStream Header(HeadSize);

      Header.put_GRules(Tour.Rules);
      Header.put_bytes(size_param, Tour.PairParam);
      Header.put_value(Tour.State);
      Header.put_value(Tour.Round);
      Header.put_value(Tour.Roundl);

      Header.put_value(Tour.Rooms);

      for(size_t Room=0; Room<size_rooms; Room++)
         Header.put_value(Tour.Boards[Room]);

      Header.put_value(Tour.SpareOne);		// SpareOneTwo for safety
      Header.put_value(Tour.SpareTwo);

      Header.put_string(Tour.TourName);
      Header.put_value(Tour.Year);

      Header.put_value(Tour.Players);

      if(Stream->Write(Header.bytes(), HeadSize) != HeadSize)
         StreamThrow(erBug, IDT_SaveHeader, IDT_BadWriteSize);

      if(Stream->Position != CodesPos)
         StreamThrow(erBug, IDT_SaveHeader, IDT_BadPosition);
   }

   void GTourStream::LoadHeader(GTour& Tour)
   {
      if(Stream->Seek(HeaderPos, soFromBeginning) != HeaderPos)
         StreamThrow(erBug, IDT_LoadHeader, IDT_BadPosition);

      ByteStream Header(HeadSize);

      if(Stream->Read(Header.bytes(), HeadSize) != HeadSize)
         StreamThrow(erBug, IDT_LoadHeader, IDT_BadWriteSize);

      if(Stream->Position != CodesPos)
         StreamThrow(erBug, IDT_LoadHeader, IDT_BadPosition);

      Header.get_GRules(Tour.Rules);
      Header.get_bytes(size_param, Tour.PairParam);
      Header.get_value(Tour.State);
      Header.get_value(Tour.Round);
      Header.get_value(Tour.Roundl);

      Header.get_value(Tour.Rooms);

      for(size_t Room=0; Room<size_rooms; Room++)
         Header.get_value(Tour.Boards[Room]);

      Header.get_value(Tour.SpareOne);
      Header.get_value(Tour.SpareTwo);

      Header.get_string(Tour.TourName);
      Header.get_value(Tour.Year);

      Header.get_value(Tour.Players);
   }
}


namespace Tours                                    // PLAYER
{
	using Tools::truncate_field;
   using ToolsGUI::erBug;

   void GTourStream::SavePlayer(GPlayer& Player)
   {
   	truncate_field(size_name, "XXX", Player.Name);		// GK 6.5.1
      truncate_field(size_note, "YYY", Player.Note);

   	const size_t psize(Player.Size());
      ByteStream PlayerStream(psize);
      PlayerStream.put_GPlayer(Player);

      if(Stream->Write(PlayerStream.bytes(), Player.Size()) != (int)Player.Size())
         StreamThrow(erBug, IDT_SavePlayer, IDT_BadWriteSize);
   }

   void GTourStream::AppendPlayer(GPlayer& Player)
   {
   	if(Stream->Seek(0, soFromEnd) != Stream->Size)
         StreamThrow(erBug, IDT_SavePlayer, IDT_BadPosition);

      SavePlayer(Player);
   }

   void GTourStream::LoadPlayer(GPlayer& Player)
   {
   	size_t PlayerSize = Player.Size();

      ByteStream PlayerStream(PlayerSize);

      if(Stream->Read(PlayerStream.bytes(), PlayerSize) != (int)PlayerSize)
         StreamThrow(erBug, IDT_LoadPlayer, IDT_BadWriteSize);

      PlayerStream.get_GPlayer(Player);
   }

   void GTourStream::SavePlayer(size_t p, GTour& Tour)
   {
      if(p>Tour.Players)
         StreamThrow(erBug, IDT_SavePlayer, IDT_BadIndex);

      GPlayerPtr Player = Tour.Listed[p];

      truncate_field(size_name, "XXX", Player->Name);
      truncate_field(size_note, "YYY", Player->Note);

      ByteStream PlayerStream(Player->Size());

      PlayerStream.put_GPlayerPtr(Player);

      int PlayPos = PlayerPos + (p-1)*Player->Size();

      if(Stream->Seek(PlayPos, soFromBeginning) != PlayPos)
         StreamThrow(erBug, IDT_SavePlayer, IDT_BadPosition);

      if(Stream->Write(PlayerStream.bytes(), Player->Size()) != (int)Player->Size())
         StreamThrow(erBug, IDT_SavePlayer, IDT_BadWriteSize);
   }

   void GTourStream::LoadPlayer(size_t p, GTour& Tour)
   {
      if(p>Tour.Players)
         StreamThrow(erBug, IDT_LoadPlayer, IDT_BadIndex);

      GPlayerPtr Player = Tour.Listed[p];

      ByteStream PlayerStream(Player->Size());

      int PlayPos = PlayerPos + (p-1)*Player->Size();

      if(Stream->Seek(PlayPos, soFromBeginning) != PlayPos)
         StreamThrow(erBug, IDT_LoadPlayer, IDT_BadPosition);

      if(Stream->Read(PlayerStream.bytes(), Player->Size()) != (int)Player->Size())
         StreamThrow(erBug, IDT_LoadPlayer, IDT_BadWriteSize);

      PlayerStream.get_GPlayerPtr(Player);
   }

   void GTourStream::SavePlayer(size_t p, GTour& Tour, GPlayer& Player)
   {
      if(p>Tour.Players)
         StreamThrow(erBug, IDT_SavePlayer, IDT_BadIndex);

      truncate_field(size_name, "XXX", Player.Name);		// GK 6.5.1
      truncate_field(size_note, "YYY", Player.Note);
      
      ByteStream PlayerStream(Player.Size());

      PlayerStream.put_GPlayer(Player);

      int PlayPos = PlayerPos + (p-1)*Player.Size();

      if(Stream->Seek(PlayPos, soFromBeginning) != PlayPos)
         StreamThrow(erBug, IDT_SavePlayer, IDT_BadPosition);

      if(Stream->Write(PlayerStream.bytes(), Player.Size()) != (int)Player.Size())
         StreamThrow(erBug, IDT_SavePlayer, IDT_BadWriteSize);
   }

   void GTourStream::LoadPlayer(size_t p, GTour& Tour, GPlayer& Player)
   {
      if(p>Tour.Players)
         StreamThrow(erBug, IDT_LoadPlayer, IDT_BadIndex);

      size_t PlayerSize = Player.Size();

      ByteStream PlayerStream(PlayerSize);

      int PlayPos = PlayerPos + (p-1)*PlayerSize;

      if(Stream->Seek(PlayPos, soFromBeginning) != PlayPos)
         StreamThrow(erBug, IDT_LoadPlayer, IDT_BadPosition);

      if(Stream->Read(PlayerStream.bytes(), PlayerSize) != (int)PlayerSize)
         StreamThrow(erBug, IDT_LoadPlayer, IDT_BadWriteSize);

      PlayerStream.get_GPlayer(Player);
   }
}


namespace Tours                           // LOAD GTour FROM FILE
{
   void GTourStream::LoadTour(GTour& Tour)
   {
      using std::list;

      LoadHeader(Tour);

      Tour.ReBuild(Tour.Players);               // Destroys, then recreates lists

      LoadCodes(Tour);                          // Code Lists for actually used codes

      for(size_t p=1; p<=Tour.Players; p++)     // Player 0 is null
         LoadPlayer(*Tour.Listed[p]);

      // If the tournament state is such that it is legitimate
      // to add players, then must we set up adummy at the end
      // of the list?

      Tour.FileName = FileName;
   }
}

namespace Tours                           // TOURNAMENT FILE INTERFACE
{
   using std::string;
   using std::list;

   using ToolsGUI::erBug;
   using ToolsGUI::erOperr;

   void Save(GTour& Tour)
   {
      GTourStream TourStream(Tour.FileName, ftGoDraw, faCreate);  // Create or open if exists

      TourStream.SaveHeader(Tour);

      TourStream.SaveCodes(Tour);

      for(size_t p=1; p<=Tour.Players; p++)     // Player 0 is null
         TourStream.SavePlayer(*Tour.Listed[p]);
   }


   void Load(std::string FileName, GTour& Tour)   // FILENAME SUPPLIED BY USER OR INI FILE
   {
      file_t FileType = IdentifyFile(FileName);

      if(FileType==ftUnknown)
         Throw(erOperr,IDT_Load, IDT_BadFileExt);

      GTourStream TourStream(FileName, FileType, faOpen);

      switch(FileType)
      {
         case ftGodraw:
            OldTourFile::LoadOldTour(TourStream, Tour);
            Save(Tour);
            break;

         case ftGoDraw:
            TourStream.LoadTour(Tour);
            break;

         default:
            TourStream.StreamThrow(erBug,IDT_Load, IDT_BadFileExt);
      }
   }
}
// END TourStream  
////////////////////////////////////////////////////////////////////////////////


